@@ -29,6 +29,7 @@
  */
 
 #include <assert.h>
+#include <fcntl.h>
 #include <netdb.h>
 #include <poll.h>
 #include <stdlib.h>
@@ -49,6 +50,7 @@
 #include <rdma/fi_collective.h>
 
 #include <shared.h>
+#include <hmem.h>
 
 struct fi_info *fi_pep, *fi, *hints;
 struct fid_fabric *fabric;
@@ -143,7 +145,6 @@ struct test_size_param test_size[] = {
 
 const unsigned int test_cnt = (sizeof test_size / sizeof test_size[0]);
 
-#define INTEG_SEED 7
 static const char integ_alphabet[] = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
 static const int integ_alphabet_length = (sizeof(integ_alphabet)/sizeof(*integ_alphabet)) - 1;
 
@@ -358,6 +359,48 @@ void ft_free_bit_combo(uint64_t *combo)
 	free(combo);
 }
 
+static int ft_reg_mr(void *buf, size_t size, uint64_t access,
+		     uint64_t key, struct fid_mr **mr, void **desc)
+{
+	struct fi_mr_attr attr = {0};
+	struct iovec iov = {0};
+	int ret;
+
+	if (((!(fi->domain_attr->mr_mode & FI_MR_LOCAL) &&
+	      !(opts.options & FT_OPT_USE_DEVICE)) ||
+	     (!(fi->domain_attr->mr_mode & FI_MR_HMEM) &&
+	      opts.options & FT_OPT_USE_DEVICE)) &&
+	    !(fi->caps & (FI_RMA | FI_ATOMIC)))
+		return 0;
+
+	iov.iov_base = buf;
+	iov.iov_len = size;
+	attr.mr_iov = &iov;
+	attr.iov_count = 1;
+	attr.access = access;
+	attr.offset = 0;
+	attr.requested_key = key;
+	attr.context = NULL;
+	attr.iface = opts.iface;
+
+	switch (opts.iface) {
+	case FI_HMEM_ZE:
+		attr.device.ze = opts.device;
+		break;
+	default:
+		break;
+	}
+
+	ret = fi_mr_regattr(domain, &attr, 0, mr);
+	if (ret)
+		return ret;
+
+	if (desc)
+		*desc = fi_mr_desc(*mr);
+
+	return FI_SUCCESS;
+}
+
 static int ft_alloc_ctx_array(struct ft_context **mr_array, char ***mr_bufs,
 			      char *default_buf, size_t mr_size,
 			      uint64_t start_key)
@@ -380,24 +423,22 @@ static int ft_alloc_ctx_array(struct ft_context **mr_array, char ***mr_bufs,
 		context = &(*mr_array)[i];
 		if (!(opts.options & FT_OPT_ALLOC_MULT_MR)) {
 			context->buf = default_buf + mr_size * i;
+			context->mr = mr;
+			context->desc = mr_desc;
 			continue;
 		}
-		(*mr_bufs)[i] = calloc(1, mr_size);
-		context->buf = (*mr_bufs)[i];
-    		if (((fi->domain_attr->mr_mode & FI_MR_LOCAL) ||
-		     (fi->caps & (FI_RMA | FI_ATOMIC)))) {
-			ret = fi_mr_reg(domain, context->buf,
-					mr_size, access, 0,
-					start_key + i, 0,
-					&context->mr, NULL);
-			if (ret)
-				return ret;
+		ret = ft_hmem_alloc(opts.iface, opts.device,
+				    (void **) &((*mr_bufs)[i]), mr_size);
+		if (ret)
+			return ret;
 
-			context->desc = fi_mr_desc(context->mr);
-		} else {
-			context->mr =  NULL;
-			context->desc = NULL;
-		}
+		context->buf = (*mr_bufs)[i];
+
+		ret = ft_reg_mr(context->buf, mr_size, access,
+				start_key + i, &context->mr,
+				&context->desc);
+		if (ret)
+			return ret;
 	}
 
 	return 0;
@@ -435,11 +476,11 @@ static int ft_alloc_msgs(void)
 		ft_set_tx_rx_sizes(&tx_size, &rx_size);
 		tx_mr_size = 0;
 		rx_mr_size = 0;
-		buf_size = MAX(tx_size, FT_MAX_CTRL_MSG) * opts.window_size + 
+		buf_size = MAX(tx_size, FT_MAX_CTRL_MSG) * opts.window_size +
 			   MAX(rx_size, FT_MAX_CTRL_MSG) * opts.window_size;
 	}
 
-	if (opts.options & FT_OPT_ALIGN) {
+	if (opts.options & FT_OPT_ALIGN && !(opts.options & FT_OPT_USE_DEVICE)) {
 		alignment = sysconf(_SC_PAGESIZE);
 		if (alignment < 0)
 			return -errno;
@@ -452,13 +493,13 @@ static int ft_alloc_msgs(void)
 			return ret;
 		}
 	} else {
-		buf = malloc(buf_size);
-		if (!buf) {
-			perror("malloc");
-			return -FI_ENOMEM;
-		}
+		ret = ft_hmem_alloc(opts.iface, opts.device, (void **) &buf, buf_size);
+		if (ret)
+			return ret;
 	}
-	memset(buf, 0, buf_size);
+	ret = ft_hmem_memset(opts.iface, opts.device, (void *) buf, 0, buf_size);
+	if (ret)
+		return ret;
 	rx_buf = buf;
 
 	if (opts.options & FT_OPT_ALLOC_MULT_MR)
@@ -468,16 +509,12 @@ static int ft_alloc_msgs(void)
 
 	remote_cq_data = ft_init_cq_data(fi);
 
-	if (!ft_mr_alloc_func && !ft_check_opts(FT_OPT_SKIP_REG_MR) &&
-	    ((fi->domain_attr->mr_mode & FI_MR_LOCAL) ||
-	     (fi->caps & (FI_RMA | FI_ATOMIC)))) {
-		ret = fi_mr_reg(domain, buf, buf_size, ft_info_to_mr_access(fi),
-				0, FT_MR_KEY, 0, &mr, NULL);
-		if (ret) {
-			FT_PRINTERR("fi_mr_reg", ret);
+	mr = &no_mr;
+	if (!ft_mr_alloc_func && !ft_check_opts(FT_OPT_SKIP_REG_MR)) {
+		ret = ft_reg_mr(buf, buf_size, ft_info_to_mr_access(fi),
+				FT_MR_KEY, &mr, &mr_desc);
+		if (ret)
 			return ret;
-		}
-		mr_desc = ft_check_mr_local_flag(fi) ? fi_mr_desc(mr) : NULL;
 	} else {
 		if (ft_mr_alloc_func) {
 			assert(!ft_check_opts(FT_OPT_SKIP_REG_MR));
@@ -485,7 +522,6 @@ static int ft_alloc_msgs(void)
 			if (ret)
 				return ret;
 		}
-		mr = &no_mr;
 	}
 
 	ret = ft_alloc_ctx_array(&tx_ctx_arr, &tx_mr_bufs, tx_buf,
@@ -523,6 +559,14 @@ int ft_open_fabric_res(void)
 		return ret;
 	}
 
+	if (opts.options & FT_OPT_DOMAIN_EQ) {
+		ret = fi_domain_bind(domain, &eq->fid, 0);
+		if (ret) {
+			FT_PRINTERR("fi_domain_bind", ret);
+			return ret;
+		}
+	}
+
 	return 0;
 }
 
@@ -640,18 +684,43 @@ int ft_alloc_active_res(struct fi_info *fi)
 	return 0;
 }
 
-static void ft_init(void)
+static int ft_init(void)
 {
 	tx_seq = 0;
 	rx_seq = 0;
 	tx_cq_cntr = 0;
 	rx_cq_cntr = 0;
+
+	//If using device memory for transfers, require OOB address
+	//exchange because extra steps are involved when passing
+	//device buffers into fi_av_insert
+	if (opts.options & FT_OPT_ENABLE_HMEM)
+		opts.options |= FT_OPT_OOB_ADDR_EXCH;
+
+	return ft_hmem_init(opts.iface);
+}
+
+int ft_sock_setup(int sock)
+{
+	int ret, op;
+
+	op = 1;
+	ret = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
+			  (void *) &op, sizeof(op));
+	if (ret)
+		return ret;
+
+	ret = ft_fd_nonblock(sock);
+	if (ret)
+		return ret;
+
+	return 0;
 }
 
 int ft_init_oob(void)
 {
-	int ret, op, err;
 	struct addrinfo *ai = NULL;
+	int ret;
 
 	if (!(opts.options & FT_OPT_OOB_CTRL) || oob_sock != -1)
 		return 0;
@@ -673,7 +742,6 @@ int ft_init_oob(void)
 
 		close(listen_sock);
 	} else {
-
 		ret = getaddrinfo(opts.dst_addr, opts.oob_port, NULL, &ai);
 		if (ret) {
 			perror("getaddrinfo");
@@ -696,11 +764,7 @@ int ft_init_oob(void)
 		sleep(1);
 	}
 
-	op = 1;
-	err = setsockopt(oob_sock, IPPROTO_TCP, TCP_NODELAY,
-			 (void *) &op, sizeof(op));
-	if (err)
-		perror("setsockopt"); /* non-fatal error */
+	ret = ft_sock_setup(oob_sock);
 
 free:
 	if (ai)
@@ -734,6 +798,11 @@ int ft_getinfo(struct fi_info *hints, struct fi_info **info)
 	if (!hints->ep_attr->type)
 		hints->ep_attr->type = FI_EP_RDM;
 
+	if (opts.options & FT_OPT_ENABLE_HMEM) {
+		hints->caps |= FI_HMEM;
+		hints->domain_attr->mr_mode |= FI_MR_HMEM;
+	}
+
 	ret = fi_getinfo(FT_FIVERSION, node, service, flags, hints, info);
 	if (ret) {
 		FT_PRINTERR("fi_getinfo", ret);
@@ -766,7 +835,10 @@ int ft_start_server(void)
 {
 	int ret;
 
-	ft_init();
+	ret = ft_init();
+	if (ret)
+		return ret;
+
 	ret = ft_init_oob();
 	if (ret)
 		return ret;
@@ -886,6 +958,14 @@ int ft_server_connect(void)
 		goto err;
 	}
 
+	if (opts.options & FT_OPT_DOMAIN_EQ) {
+		ret = fi_domain_bind(domain, &eq->fid, 0);
+		if (ret) {
+			FT_PRINTERR("fi_domain_bind", ret);
+			return ret;
+		}
+	}
+
 	ret = ft_alloc_active_res(fi);
 	if (ret)
 		goto err;
@@ -898,6 +978,9 @@ int ft_server_connect(void)
 	if (ret)
 		goto err;
 
+	if (ft_check_opts(FT_OPT_FORK_CHILD))
+		ft_fork_child();
+
 	return 0;
 
 err:
@@ -928,7 +1011,10 @@ int ft_client_connect(void)
 {
 	int ret;
 
-	ft_init();
+	ret = ft_init();
+	if (ret)
+		return ret;
+
 	ret = ft_init_oob();
 	if (ret)
 		return ret;
@@ -953,6 +1039,9 @@ int ft_client_connect(void)
 	if (ret)
 		return ret;
 
+	if (ft_check_opts(FT_OPT_FORK_CHILD))
+		ft_fork_child();
+
 	return 0;
 }
 
@@ -960,7 +1049,10 @@ int ft_init_fabric(void)
 {
 	int ret;
 
-	ft_init();
+	ret = ft_init();
+	if (ret)
+		return ret;
+
 	ret = ft_init_oob();
 	if (ret)
 		return ret;
@@ -985,6 +1077,9 @@ int ft_init_fabric(void)
 	if (ret)
 		return ret;
 
+	if (ft_check_opts(FT_OPT_FORK_CHILD))
+		ft_fork_child();
+
 	return 0;
 }
 
@@ -1019,8 +1114,8 @@ int ft_enable_ep(struct fid_ep *ep, struct fid_eq *eq, struct fid_av *av,
 	uint64_t flags;
 	int ret;
 
-	if (fi->ep_attr->type == FI_EP_MSG || fi->caps & FI_MULTICAST ||
-	    fi->caps & FI_COLLECTIVE)
+	if ((fi->ep_attr->type == FI_EP_MSG || fi->caps & FI_MULTICAST ||
+	    fi->caps & FI_COLLECTIVE) && !(opts.options & FT_OPT_DOMAIN_EQ))
 		FT_EP_BIND(ep, eq, 0);
 
 	FT_EP_BIND(ep, av, 0);
@@ -1209,7 +1304,7 @@ int ft_init_av_dst_addr(struct fid_av *av_ptr, struct fid_ep *ep_ptr,
 		if (ret)
 			return ret;
 	} else {
-		ret = (int) ft_rx(ep, FT_MAX_CTRL_MSG);
+		ret = ft_get_rx_comp(rx_seq);
 		if (ret)
 			return ret;
 
@@ -1221,6 +1316,10 @@ int ft_init_av_dst_addr(struct fid_av *av_ptr, struct fid_ep *ep_ptr,
 		if (ret)
 			return ret;
 
+		ret = ft_post_rx(ep, rx_size, &rx_ctx);
+		if (ret)
+			return ret;
+
 		if (fi->domain_attr->av_type == FI_AV_TABLE)
 			*remote_addr = 0;
 
@@ -1447,21 +1546,21 @@ int ft_exchange_keys(struct fi_rma_iov *peer_iov)
 
 static void ft_cleanup_mr_array(struct ft_context *ctx_arr, char **mr_bufs)
 {
-	int i;
+	int i, ret;
 
 	if (!mr_bufs)
 		return;
 
 	for (i = 0; i < opts.window_size; i++) {
 		FT_CLOSE_FID(ctx_arr[i].mr);
-		free(mr_bufs[i]);
+		ret = ft_hmem_free(opts.iface, mr_bufs[i]);
+		if (ret)
+			FT_PRINTERR("ft_hmem_free", ret);
 	}
 }
 
 static void ft_close_fids(void)
 {
-	if (mr != &no_mr)
-		FT_CLOSE_FID(mr);
 	FT_CLOSE_FID(mc);
 	FT_CLOSE_FID(alias_ep);
 	FT_CLOSE_FID(ep);
@@ -1475,15 +1574,19 @@ static void ft_close_fids(void)
 	FT_CLOSE_FID(rxcntr);
 	FT_CLOSE_FID(txcntr);
 	FT_CLOSE_FID(pollset);
+	if (mr != &no_mr)
+		FT_CLOSE_FID(mr);
 	FT_CLOSE_FID(av);
-	FT_CLOSE_FID(eq);
 	FT_CLOSE_FID(domain);
+	FT_CLOSE_FID(eq);
 	FT_CLOSE_FID(waitset);
 	FT_CLOSE_FID(fabric);
 }
 
 void ft_free_res(void)
 {
+	int ret;
+
 	ft_cleanup_mr_array(tx_ctx_arr, tx_mr_bufs);
 	ft_cleanup_mr_array(rx_ctx_arr, rx_mr_bufs);
 
@@ -1495,7 +1598,9 @@ void ft_free_res(void)
 	ft_close_fids();
 
 	if (buf) {
-		free(buf);
+		ret = ft_hmem_free(opts.iface, buf);
+		if (ret)
+			FT_PRINTERR("ft_hmem_free", ret);
 		buf = rx_buf = tx_buf = NULL;
 		buf_size = rx_size = tx_size = tx_mr_size = rx_mr_size = 0;
 	}
@@ -1511,6 +1616,10 @@ void ft_free_res(void)
 		fi_freeinfo(hints);
 		hints = NULL;
 	}
+
+	ret = ft_hmem_cleanup(opts.iface);
+	if (ret)
+		FT_PRINTERR("ft_hmem_cleanup", ret);
 }
 
 static int dupaddr(void **dst_addr, size_t *dst_addrlen,
@@ -1685,6 +1794,14 @@ void init_test(struct ft_opts *opts, char *test_name, size_t test_name_len)
 		opts->iterations = size_to_count(opts->transfer_size);
 }
 
+static void ft_force_progress(void)
+{
+	if (txcq)
+		fi_cq_read(txcq, NULL, 0);
+	if (rxcq)
+		fi_cq_read(rxcq, NULL, 0);
+}
+
 static int ft_progress(struct fid_cq *cq, uint64_t total, uint64_t *cq_cntr)
 {
 	struct fi_cq_err_entry comp;
@@ -2003,8 +2120,8 @@ ssize_t ft_post_rx_buf(struct fid_ep *ep, size_t size, void *ctx,
 	if (hints->caps & FI_TAGGED) {
 		op_tag = op_tag ? op_tag : rx_seq;
 		FT_POST(fi_trecv, ft_progress, rxcq, rx_seq, &rx_cq_cntr,
-			"receive", ep, op_buf, size, op_mr_desc, 0, op_tag,
-			0, ctx);
+			"receive", ep, op_buf, size, op_mr_desc,
+			remote_fi_addr, op_tag, 0, ctx);
 	} else {
 		FT_POST(fi_recv, ft_progress, rxcq, rx_seq, &rx_cq_cntr,
 			"receive", ep, op_buf, size, op_mr_desc, 0, ctx);
@@ -2523,6 +2640,21 @@ int ft_fork_and_pair(void)
 	return 0;
 }
 
+int ft_fork_child(void)
+{
+	ft_child_pid = fork();
+	if (ft_child_pid < 0) {
+		FT_PRINTERR("fork", ft_child_pid);
+		return -errno;
+	}
+
+	if (ft_child_pid == 0) {
+		exit(0);
+	}
+
+	return 0;
+}
+
 int ft_wait_child(void)
 {
 	int ret;
@@ -2554,7 +2686,6 @@ int ft_finalize_ep(struct fid_ep *ep)
 	int ret;
 	struct fi_context ctx;
 
-	strcpy(tx_buf + ft_tx_prefix_size(), "fin");
 	iov.iov_base = tx_buf;
 	iov.iov_len = 4 + ft_tx_prefix_size();
 
@@ -2572,7 +2703,7 @@ int ft_finalize_ep(struct fid_ep *ep)
 
 		FT_POST(fi_tsendmsg, ft_progress, txcq, tx_seq,
 			&tx_cq_cntr, "tsendmsg", ep, &tmsg,
-			FI_INJECT | FI_TRANSMIT_COMPLETE);
+			FI_TRANSMIT_COMPLETE);
 	} else {
 		struct fi_msg msg;
 
@@ -2585,7 +2716,7 @@ int ft_finalize_ep(struct fid_ep *ep)
 
 		FT_POST(fi_sendmsg, ft_progress, txcq, tx_seq,
 			&tx_cq_cntr, "sendmsg", ep, &msg,
-			FI_INJECT | FI_TRANSMIT_COMPLETE);
+			FI_TRANSMIT_COMPLETE);
 	}
 
 	ret = ft_get_tx_comp(tx_seq);
@@ -2706,6 +2837,7 @@ void ft_addr_usage()
 			"over the, optional, port");
 	FT_PRINT_OPTS_USAGE("-C <number>", "number of connections to accept before "
 			"cleaning up a server");
+	FT_PRINT_OPTS_USAGE("-F <addr_format>", "Address format (default:FI_FORMAT_UNSPEC)");
 }
 
 void ft_usage(char *name, char *desc)
@@ -2733,7 +2865,17 @@ void ft_usage(char *name, char *desc)
 	FT_PRINT_OPTS_USAGE("", "fi_resmgmt_test");
 	FT_PRINT_OPTS_USAGE("", "fi_inj_complete");
 	FT_PRINT_OPTS_USAGE("", "fi_bw");
+	FT_PRINT_OPTS_USAGE("-U", "run fabtests with FI_DELIVERY_COMPLETE set");
+	FT_PRINT_OPTS_USAGE("", "Only the following tests support this option for now:");
+	FT_PRINT_OPTS_USAGE("", "fi_bw");
+	FT_PRINT_OPTS_USAGE("", "fi_rdm");
+	FT_PRINT_OPTS_USAGE("", "fi_rdm_atomic");
+	FT_PRINT_OPTS_USAGE("", "fi_rdm_pingpong");
+	FT_PRINT_OPTS_USAGE("", "fi_rdm_tagged_bw");
+	FT_PRINT_OPTS_USAGE("", "fi_rdm_tagged_pingpong");
+	FT_PRINT_OPTS_USAGE("", "fi_rma_bw");
 	FT_PRINT_OPTS_USAGE("-M <mode>", "Disable mode bit from test");
+	FT_PRINT_OPTS_USAGE("-K", "fork a child process after initializing endpoint");
 	FT_PRINT_OPTS_USAGE("", "mr_local");
 	FT_PRINT_OPTS_USAGE("-a <address vector name>", "name of address vector");
 	FT_PRINT_OPTS_USAGE("-h", "display this help output");
@@ -2757,6 +2899,10 @@ void ft_mcusage(char *name, char *desc)
 	FT_PRINT_OPTS_USAGE("-p <provider>", "specific provider name eg sockets, verbs");
 	FT_PRINT_OPTS_USAGE("-d <domain>", "domain name");
 	FT_PRINT_OPTS_USAGE("-p <provider>", "specific provider name eg sockets, verbs");
+	FT_PRINT_OPTS_USAGE("-D <device_iface>", "Specify device interface: eg ze (default: None). "
+			     "Automatically enables FI_HMEM (-H)");
+	FT_PRINT_OPTS_USAGE("-i <device_id>", "Specify which device to use (default: 0)");
+	FT_PRINT_OPTS_USAGE("-H", "Enable provider FI_HMEM support");
 	FT_PRINT_OPTS_USAGE("-h", "display this help output");
 
 	return;
@@ -2766,10 +2912,13 @@ void ft_csusage(char *name, char *desc)
 {
 	ft_usage(name, desc);
 	FT_PRINT_OPTS_USAGE("-I <number>", "number of iterations");
+	FT_PRINT_OPTS_USAGE("-Q", "bind EQ to domain (vs. endpoint)");
 	FT_PRINT_OPTS_USAGE("-w <number>", "number of warmup iterations");
 	FT_PRINT_OPTS_USAGE("-S <size>", "specific transfer size or 'all'");
 	FT_PRINT_OPTS_USAGE("-l", "align transmit and receive buffers to page size");
 	FT_PRINT_OPTS_USAGE("-m", "machine readable output");
+	FT_PRINT_OPTS_USAGE("-D <device_iface>", "Specify device interface: eg cuda, ze(default: None). "
+			     "Automatically enables FI_HMEM (-H)");
 	FT_PRINT_OPTS_USAGE("-t <type>", "completion type [queue, counter]");
 	FT_PRINT_OPTS_USAGE("-c <method>", "completion method [spin, sread, fd, yield]");
 	FT_PRINT_OPTS_USAGE("-h", "display this help output");
@@ -2823,6 +2972,24 @@ void ft_parseinfo(int op, char *optarg, struct fi_info *hints,
 		if (!strncasecmp("mr_local", optarg, 8))
 			opts->mr_mode &= ~FI_MR_LOCAL;
 		break;
+	case 'D':
+		if (!strncasecmp("ze", optarg, 2))
+			opts->iface = FI_HMEM_ZE;
+		else if (!strncasecmp("cuda", optarg, 4))
+			opts->iface = FI_HMEM_CUDA;
+		else
+			printf("Unsupported interface\n");
+		opts->options |= FT_OPT_ENABLE_HMEM | FT_OPT_USE_DEVICE;
+		break;
+	case 'i':
+		opts->device = atoi(optarg);
+		break;
+	case 'H':
+		opts->options |= FT_OPT_ENABLE_HMEM;
+		break;
+	case 'K':
+		opts->options |= FT_OPT_FORK_CHILD;
+		break;
 	default:
 		/* let getopt handle unknown opts*/
 		break;
@@ -2852,6 +3019,16 @@ void ft_parse_addr_opts(int op, char *optarg, struct ft_opts *opts)
 		else
 			opts->oob_port = default_oob_port;
 		break;
+	case 'F':
+		if (!strncasecmp("fi_sockaddr_in6", optarg, 15))
+			opts->address_format = FI_SOCKADDR_IN6;
+		else if (!strncasecmp("fi_sockaddr_in", optarg, 14))
+			opts->address_format = FI_SOCKADDR_IN;
+		else if (!strncasecmp("fi_sockaddr_ib", optarg, 14))
+			opts->address_format = FI_SOCKADDR_IB;
+		else if (!strncasecmp("fi_sockaddr", optarg, 11)) /* keep me last */
+			opts->address_format = FI_SOCKADDR;
+		break;
 	case 'C':
 		opts->options |= FT_OPT_SERVER_PERSIST;
 		opts->num_connections = atoi(optarg);
@@ -2870,12 +3047,15 @@ void ft_parsecsopts(int op, char *optarg, struct ft_opts *opts)
 		opts->options |= FT_OPT_ITER;
 		opts->iterations = atoi(optarg);
 		break;
+	case 'Q':
+		opts->options |= FT_OPT_DOMAIN_EQ;
+		break;
 	case 'S':
 		if (!strncasecmp("all", optarg, 3)) {
 			opts->sizes_enabled = FT_ENABLE_ALL;
 		} else {
 			opts->options |= FT_OPT_SIZE;
-			opts->transfer_size = atoi(optarg);
+			opts->transfer_size = atol(optarg);
 		}
 		break;
 	case 'm':
@@ -2941,44 +3121,40 @@ int ft_parse_rma_opts(int op, char *optarg, struct fi_info *hints,
 	return 0;
 }
 
-void ft_fill_buf(void *buf, int size)
+void ft_fill_buf(void *buf, size_t size)
 {
 	char *msg_buf;
-	int msg_index;
-	static unsigned int iter = 0;
-	int i;
+	int msg_index = 0;
+	size_t i;
 
-	msg_index = ((iter++)*INTEG_SEED) % integ_alphabet_length;
-	msg_buf = (char *)buf;
+	msg_buf = (char *) buf;
 	for (i = 0; i < size; i++) {
-		msg_buf[i] = integ_alphabet[msg_index++];
-		if (msg_index >= integ_alphabet_length)
+		msg_buf[i] = integ_alphabet[msg_index];
+		if (++msg_index >= integ_alphabet_length)
 			msg_index = 0;
 	}
 }
 
-int ft_check_buf(void *buf, int size)
+int ft_check_buf(void *buf, size_t size)
 {
 	char *recv_data;
 	char c;
-	static unsigned int iter = 0;
-	int msg_index;
-	int i;
+	int msg_index = 0;
+	size_t i;
 
-	msg_index = ((iter++)*INTEG_SEED) % integ_alphabet_length;
 	recv_data = (char *)buf;
 
 	for (i = 0; i < size; i++) {
-		c = integ_alphabet[msg_index++];
-		if (msg_index >= integ_alphabet_length)
+		c = integ_alphabet[msg_index];
+		if (++msg_index >= integ_alphabet_length)
 			msg_index = 0;
 		if (c != recv_data[i])
 			break;
 	}
 	if (i != size) {
-		printf("Error at iteration=%d size=%d byte=%d\n",
-			iter, size, i);
-		return 1;
+		printf("Data check error (%c!=%c) at byte %zu for "
+		       "buffer size %zu\n", c, recv_data[i], i, size);
+		return -FI_EIO;
 	}
 
 	return 0;
@@ -3119,17 +3295,14 @@ int ft_sock_connect(char *node, char *service)
 		goto free;
 	}
 
-	ret = 1;
-	ret = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (void *) &ret, sizeof(ret));
-	if (ret)
-		perror("setsockopt");
-
 	ret = connect(sock, ai->ai_addr, ai->ai_addrlen);
 	if (ret) {
 		perror("connect");
 		close(sock);
+		goto free;
 	}
 
+	ret = ft_sock_setup(sock);
 free:
 	freeaddrinfo(ai);
 	return ret;
@@ -3137,7 +3310,7 @@ free:
 
 int ft_sock_accept()
 {
-	int ret, op;
+	int ret;
 
 	sock = accept(listen_sock, NULL, 0);
         if (sock < 0) {
@@ -3146,48 +3319,51 @@ int ft_sock_accept()
 		return ret;
 	}
 
-	op = 1;
-	ret = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
-			  (void *) &op, sizeof(op));
-	if (ret)
-		perror("setsockopt");
-
-	return 0;
+	ret = ft_sock_setup(sock);
+	return ret;
 }
 
 int ft_sock_send(int fd, void *msg, size_t len)
 {
-	int ret;
+	size_t sent;
+	ssize_t ret, err = 0;
 
-	ret = send(fd, msg, len, 0);
-	if (ret == len) {
-		return 0;
-	} else if (ret < 0) {
-		perror("send");
-		return -errno;
-	} else {
-		perror("send aborted");
-		return -FI_ECONNABORTED;
+	for (sent = 0; sent < len; ) {
+		ret = send(fd, ((char *) msg) + sent, len - sent, 0);
+		if (ret > 0) {
+			sent += ret;
+		} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
+			ft_force_progress();
+		} else {
+			err = -errno;
+			break;
+		}
 	}
+
+	return err ? err: 0;
 }
 
 int ft_sock_recv(int fd, void *msg, size_t len)
 {
-	int ret;
+	size_t rcvd;
+	ssize_t ret, err = 0;
 
-	ret = recv(fd, msg, len, MSG_WAITALL);
-	if (ret == len) {
-		return 0;
-	} else if (ret == 0) {
-		return -FI_ENOTCONN;
-	} else if (ret < 0) {
-		FT_PRINTERR("ft_sock_recv", -errno);
-		perror("recv");
-		return -errno;
-	} else {
-		perror("recv aborted");
-		return -FI_ECONNABORTED;
+	for (rcvd = 0; rcvd < len; ) {
+		ret = recv(fd, ((char *) msg) + rcvd, len - rcvd, 0);
+		if (ret > 0) {
+			rcvd += ret;
+		} else if (ret == 0) {
+			err = -FI_ENOTCONN;
+			break;
+		} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
+			ft_force_progress();
+		} else {
+			err = -errno;
+			break;
+		}
 	}
+
+	return err ? err: 0;
 }
 
 int ft_sock_sync(int value)
