@@ -100,11 +100,7 @@ const struct fi_domain_attr efa_domain_attr = {
 	.control_progress	= FI_PROGRESS_AUTO,
 	.data_progress		= FI_PROGRESS_AUTO,
 	.resource_mgmt		= FI_RM_DISABLED,
-#ifdef HAVE_LIBCUDA
-	.mr_mode		= OFI_MR_BASIC_MAP | FI_MR_LOCAL | FI_MR_BASIC | FI_MR_HMEM,
-#else
 	.mr_mode		= OFI_MR_BASIC_MAP | FI_MR_LOCAL | FI_MR_BASIC,
-#endif
 	.mr_key_size		= sizeof_field(struct ibv_sge, lkey),
 	.cq_data_size		= 0,
 	.tx_ctx_cnt		= 1024,
@@ -427,9 +423,9 @@ static int efa_alloc_fid_nic(struct fi_info *fi, struct efa_context *ctx,
 
 	efa_addr_to_str(src_addr, link_attr->address);
 
-	link_attr->mtu = port_attr->max_msg_sz;
-
-	link_attr->speed = 0;
+	link_attr->mtu = port_attr->max_msg_sz - rxr_pkt_max_header_size();
+	link_attr->speed = ofi_vrb_speed(port_attr->active_speed,
+	                                 port_attr->active_width);
 
 	switch (port_attr->state) {
 	case IBV_PORT_DOWN:
@@ -469,6 +465,55 @@ err_free_nic:
 	return ret;
 }
 
+#if HAVE_LIBCUDA
+/*
+ * efa_get_gdr_support() check if GPUDirect RDMA is supported by
+ * reading from sysfs file "class/infiniband/<device_name>/gdr"
+ * and set content of gdr_support accordingly.
+ *
+ * Return value:
+ *   return 1 if sysfs file exist and has 1 in it.
+ *   return 0 if sysfs file does not exist or has 0 in it.
+ *   return a negatie value if error happened.
+ */
+static int efa_get_gdr_support(char *device_name)
+{
+	static const int MAX_GDR_SUPPORT_STRLEN = 8;
+	char *gdr_path = NULL;
+	char gdr_support_str[MAX_GDR_SUPPORT_STRLEN];
+	int ret, read_len;
+
+	ret = asprintf(&gdr_path, "class/infiniband/%s/device/gdr", device_name);
+	if (ret < 0) {
+		EFA_INFO_ERRNO(FI_LOG_FABRIC, "asprintf to build sysfs file name failed", ret);
+		goto out;
+	}
+
+	ret = fi_read_file(get_sysfs_path(), gdr_path,
+			   gdr_support_str, MAX_GDR_SUPPORT_STRLEN);
+	if (ret < 0) {
+		if (errno == ENOENT) {
+			/* sysfs file does not exist, gdr is not supported */
+			ret = 0;
+		}
+
+		goto out;
+	}
+
+	if (ret == 0) {
+		EFA_WARN(FI_LOG_FABRIC, "Sysfs file %s is empty\n", gdr_path);
+		ret = -FI_EINVAL;
+		goto out;
+	}
+
+	read_len = MIN(ret, MAX_GDR_SUPPORT_STRLEN);
+	ret = (0 == strncmp(gdr_support_str, "1", read_len));
+out:
+	free(gdr_path);
+	return ret;
+}
+#endif
+
 static int efa_get_device_attrs(struct efa_context *ctx, struct fi_info *info)
 {
 	struct efadv_device_attr efadv_attr;
@@ -515,6 +560,23 @@ static int efa_get_device_attrs(struct efa_context *ctx, struct fi_info *info)
 	info->domain_attr->resource_mgmt	= FI_RM_DISABLED;
 	info->domain_attr->mr_cnt		= base_attr->max_mr;
 
+#if HAVE_LIBCUDA
+	if (info->ep_attr->type == FI_EP_RDM) {
+		ret = efa_get_gdr_support(ctx->ibv_ctx->device->name);
+		if (ret < 0) {
+			EFA_WARN(FI_LOG_FABRIC, "get gdr support failed!\n");
+			return ret;
+		}
+
+		if (ret == 1) {
+			info->caps			|= FI_HMEM;
+			info->tx_attr->caps		|= FI_HMEM;
+			info->rx_attr->caps		|= FI_HMEM;
+			info->domain_attr->mr_mode	|= FI_MR_HMEM;
+		}
+	}
+#endif
+
 	EFA_DBG(FI_LOG_DOMAIN, "Domain attribute :\n"
 				"\t info->domain_attr->cq_cnt		= %zu\n"
 				"\t info->domain_attr->ep_cnt		= %zu\n"
@@ -861,7 +923,6 @@ static int efa_fabric_close(fid_t fid)
 	struct efa_fabric *fab;
 	int ret;
 
-	unsetenv("RDMAV_HUGEPAGES_SAFE");
 	fab = container_of(fid, struct efa_fabric, util_fabric.fabric_fid.fid);
 	ret = ofi_fabric_close(&fab->util_fabric);
 	if (ret)
@@ -895,29 +956,6 @@ int efa_fabric(struct fi_fabric_attr *attr, struct fid_fabric **fabric_fid,
 	struct efa_fabric *fab;
 	int ret = 0;
 
-	/*
-	 * Enable rdma-core fork support and huge page support. We want call
-	 * this only when the EFA provider is selected. It is safe to call this
-	 * function again if multiple EFA fabrics are opened or if the fabric
-	 * is closed and opened again.
-	 *
-	 * TODO: allow users to disable this once the fork() to check ptrace
-	 * permissions is removed.
-	 */
-	ret = setenv("RDMAV_HUGEPAGES_SAFE", "1", 1);
-	if (ret)
-		return -errno;
-
-	ret = ibv_fork_init();
-	if (ret) {
-		EFA_WARN(FI_LOG_FABRIC, "Failed to initialize libibverbs "
-					"fork support. Please check your "
-					"application to ensure it is not "
-					"making verbs calls before "
-					"initializing EFA.\n");
-		return -ret;
-	}
-
 	fab = calloc(1, sizeof(*fab));
 	if (!fab)
 		return -FI_ENOMEM;
@@ -955,7 +993,7 @@ static void fi_efa_fini(void)
 	efa_device_free();
 #if HAVE_EFA_DL
 	smr_cleanup();
-#endif 
+#endif
 }
 
 struct fi_provider efa_prov = {
