@@ -20,7 +20,9 @@
 #include <ucs/datastruct/mpool.h>
 #include <ucs/datastruct/queue_types.h>
 #include <ucs/datastruct/bitmap.h>
+#include <ucs/datastruct/conn_match.h>
 #include <ucs/memory/memtype_cache.h>
+#include <ucs/memory/memory_type.h>
 #include <ucs/type/spinlock.h>
 #include <ucs/sys/string.h>
 #include <ucs/type/param.h>
@@ -64,8 +66,10 @@ typedef struct ucp_context_config {
     /** Segment size in the worker pre-registered memory pool */
     size_t                                 seg_size;
     /** RNDV pipeline fragment size */
-    size_t                                 rndv_frag_size;
-    /** RNDV pipline send threshold */
+    size_t                                 rndv_frag_size[UCS_MEMORY_TYPE_LAST];
+    /** Number of RNDV pipeline fragments per allocation */
+    size_t                                 rndv_num_frags[UCS_MEMORY_TYPE_LAST];
+    /** RNDV pipeline send threshold */
     size_t                                 rndv_pipeline_send_thresh;
     /** Threshold for using tag matching offload capabilities. Smaller buffers
      *  will not be posted to the transport. */
@@ -95,8 +99,6 @@ typedef struct ucp_context_config {
     size_t                                 estimated_num_eps;
     /** Estimated number of processes per node */
     size_t                                 estimated_num_ppn;
-    /** Memtype cache */
-    int                                    enable_memtype_cache;
     /** Enable flushing endpoints while flushing a worker */
     int                                    flush_worker_eps;
     /** Enable optimizations suitable for homogeneous systems */
@@ -117,18 +119,38 @@ typedef struct ucp_context_config {
     ucs_on_off_auto_value_t                proto_indirect_id;
     /** Bitmap of memory types whose allocations are registered fully */
     unsigned                               reg_whole_alloc_bitmap;
+    /** Always use flush operation in rendezvous put */
+    int                                    rndv_put_force_flush;
+    /** Maximum size of mem type direct rndv*/
+    size_t                                 rndv_memtype_direct_size;
+    /** UCP sockaddr private data format version */
+    ucp_object_version_t                   sa_client_min_hdr_version;
+    /** Remote keys with that many remote MDs or less would be allocated from a
+      * memory pool.*/
+    int                                    rkey_mpool_max_md;
+    /** Worker address format version */
+    ucp_object_version_t                   worker_addr_version;
 } ucp_context_config_t;
 
 
+typedef UCS_CONFIG_STRING_ARRAY_FIELD(names) ucp_context_config_names_t;
+
+
 struct ucp_config {
     /** Array of device lists names to use.
-     *  This array holds three lists - network devices, shared memory devices
-     *  and acceleration devices */
+     *  This array holds four lists - network devices, shared memory devices,
+     *  acceleration devices and loop-back devices */
     ucs_config_names_array_t               devices[UCT_DEVICE_TYPE_LAST];
     /** Array of transport names to use */
     ucs_config_allow_list_t                tls;
+    /** Array of protocol names to use */
+    ucs_config_allow_list_t                protos;
     /** Array of memory allocation methods */
     UCS_CONFIG_STRING_ARRAY_FIELD(methods) alloc_prio;
+    /** Array of rendezvous fragment sizes */
+    ucp_context_config_names_t             rndv_frag_sizes;
+    /** Array of rendezvous fragment elems per allocation */
+    ucp_context_config_names_t             rndv_frag_elems;
     /** Array of transports for partial worker address to pack */
     UCS_CONFIG_STRING_ARRAY_FIELD(aux_tls) sockaddr_aux_tls;
     /** Array of transports for client-server transports and port selection */
@@ -141,6 +163,10 @@ struct ucp_config {
     char                                   *selection_cmp;
     /** Configuration saved directly in the context */
     ucp_context_config_t                   ctx;
+    /** Save ucx configurations not listed in ucp_config_table **/
+    ucs_list_link_t                        cached_key_list;
+    /** Array of worker memory pool sizes */
+    UCS_CONFIG_ARRAY_FIELD(size_t, memunits) mpool_sizes;
 };
 
 
@@ -190,18 +216,22 @@ typedef struct ucp_tl_md {
  * UCP context
  */
 typedef struct ucp_context {
-
     ucp_tl_cmpt_t                 *tl_cmpts;  /* UCT components */
     ucp_rsc_index_t               num_cmpts;  /* Number of UCT components */
 
     ucp_tl_md_t                   *tl_mds;    /* Memory domain resources */
     ucp_md_index_t                num_mds;    /* Number of memory domains */
 
+    /* Map of memory domains that have valid memory keys for host memory
+       allocated with ucp_mem_mmap_common().
+       It's initialized on first access. */
+    int                           alloc_md_map_initialized;
+    ucp_md_map_t                  alloc_md_map;
+
     /* List of MDs that detect non host memory type */
     ucp_md_index_t                mem_type_detect_mds[UCS_MEMORY_TYPE_LAST];
     ucp_md_index_t                num_mem_type_detect_mds;  /* Number of mem type MDs */
     uint64_t                      mem_type_mask;            /* Supported mem type mask */
-    ucs_memtype_cache_t           *memtype_cache;           /* mem type allocation cache */
 
     ucp_tl_resource_desc_t        *tl_rscs;   /* Array of communication resources */
     ucp_tl_bitmap_t               tl_bitmap;  /* Cached map of tl resources used by workers.
@@ -212,6 +242,8 @@ typedef struct ucp_context {
     /* Mask of memory type communication resources */
     ucp_tl_bitmap_t               mem_type_access_tls[UCS_MEMORY_TYPE_LAST];
 
+    ucp_proto_id_mask_t           proto_bitmap;  /* Enabled protocols */
+
     struct {
 
         /* Bitmap of features supported by the context */
@@ -256,13 +288,19 @@ typedef struct ucp_context {
 
         /* MD to compare for transport selection scores */
         char                      *selection_cmp;
+        struct {
+           unsigned               count;
+           size_t                 *sizes;
+        } am_mpools;
     } config;
 
-    /* Configuration of multi-threadiing support */
+    /* Configuration of multi-threading support */
     ucp_mt_lock_t                 mt_lock;
 
     char                          name[UCP_ENTITY_NAME_MAX];
 
+    /* Save cached uct configurations */
+    ucs_list_link_t               cached_key_list;
 } ucp_context_t;
 
 
@@ -361,6 +399,11 @@ typedef struct ucp_tl_iface_atomic_flags {
     UCS_PARAM_VALUE(UCP_PARAM_FIELD, _params, _name, _flag, _default)
 
 
+#define UCP_ATTR_VALUE(_obj, _attrs, _name, _flag, _default) \
+    UCS_PARAM_VALUE(UCS_PP_TOKENPASTE3(UCP_, _obj, _ATTR_FIELD), _attrs, \
+                    _name, _flag, _default)
+
+
 #define ucp_assert_memtype(_context, _buffer, _length, _mem_type) \
     ucs_assert(ucp_memory_type_detect(_context, _buffer, _length) == (_mem_type))
 
@@ -444,12 +487,6 @@ ucp_tl_iface_bandwidth(ucp_context_h context, const uct_ppn_bandwidth_t *bandwid
            (bandwidth->shared / context->config.est_num_ppn);
 }
 
-static UCS_F_ALWAYS_INLINE int ucp_memory_type_cache_is_empty(ucp_context_h context)
-{
-    return (context->memtype_cache &&
-            !context->memtype_cache->pgtable.num_regions);
-}
-
 static UCS_F_ALWAYS_INLINE void
 ucp_memory_info_set_host(ucp_memory_info_t *mem_info)
 {
@@ -467,13 +504,8 @@ ucp_memory_detect_internal(ucp_context_h context, const void *address,
         goto out_host_mem;
     }
 
-    if (ucs_likely(context->memtype_cache != NULL)) {
-        if (!context->memtype_cache->pgtable.num_regions) {
-            goto out_host_mem;
-        }
-
-        status = ucs_memtype_cache_lookup(context->memtype_cache, address,
-                                          length, mem_info);
+    status = ucs_memtype_cache_lookup(address, length, mem_info);
+    if (status != UCS_ERR_NO_ELEM) {
         if (ucs_likely(status != UCS_OK)) {
             ucs_assert(status == UCS_ERR_NO_ELEM);
             goto out_host_mem;
@@ -528,4 +560,12 @@ void ucp_tl_bitmap_validate(const ucp_tl_bitmap_t *tl_bitmap,
 
 const char* ucp_context_cm_name(ucp_context_h context, ucp_rsc_index_t cm_idx);
 
+
+ucs_status_t
+ucp_config_modify_internal(ucp_config_t *config, const char *name,
+                           const char *value);
+
+
+void ucp_apply_uct_config_list(ucp_context_h context, void *config);
+
 #endif
