@@ -13,10 +13,10 @@
 #include "proto_single.h"
 
 #include <ucp/core/ucp_context.h>
-#include <ucp/core/ucp_worker.h>
 #include <ucp/dt/dt.h>
 #include <float.h>
 
+#include <ucp/core/ucp_worker.inl>
 #include <ucs/datastruct/array.inl>
 
 
@@ -25,6 +25,9 @@
  */
 #define UCP_PROTO_MSGLEN_EPSILON   0.5
 
+/* Threshold for considering two performance values as equal */
+#define UCP_PROTO_PERF_EPSILON     1e-15
+
 
 /* Parameters structure for initializing protocols for a selection parameter */
 typedef struct {
@@ -46,7 +49,8 @@ typedef struct {
 
 UCS_ARRAY_DEFINE_INLINE(ucp_proto_thresh, unsigned,
                         ucp_proto_threshold_tmp_elem_t);
-UCS_ARRAY_DEFINE_INLINE(ucp_proto_perf, unsigned, ucp_proto_perf_range_t);
+UCS_ARRAY_DEFINE_INLINE(ucp_proto_select_range, unsigned,
+                        ucp_proto_select_range_t);
 
 
 const ucp_proto_threshold_elem_t*
@@ -87,31 +91,45 @@ ucp_proto_thresholds_append(ucs_array_t(ucp_proto_thresh) *thresh_list,
     return UCS_OK;
 }
 
-static ucs_status_t
-ucp_proto_perf_append(ucs_array_t(ucp_proto_perf) *perf_list, size_t max_length,
-                      ucs_linear_func_t perf)
+static int
+ucp_proto_select_range_is_equal(const ucp_proto_select_range_t *range_elem1,
+                                const ucp_proto_select_range_t *range_elem2)
 {
-    ucp_proto_perf_range_t *perf_elem;
+    ucp_proto_perf_type_t perf_type;
+
+    for (perf_type = 0; perf_type < UCP_PROTO_PERF_TYPE_LAST; ++perf_type) {
+        if (!ucs_linear_func_is_equal(range_elem1->super.perf[perf_type],
+                                      range_elem2->super.perf[perf_type],
+                                      UCP_PROTO_PERF_EPSILON)) {
+            return 0;
+        }
+    }
+
+    return range_elem1->cfg_thresh == range_elem2->cfg_thresh;
+}
+
+static ucs_status_t
+ucp_proto_perf_append(ucs_array_t(ucp_proto_select_range) *range_list,
+                      const ucp_proto_select_range_t *new_elem)
+{
+    ucp_proto_select_range_t *range_elem;
     ucs_status_t status;
 
-    if (!ucs_array_is_empty(perf_list)) {
-        perf_elem = ucs_array_last(perf_list);
-        ucs_assert(max_length > perf_elem->max_length);
-        if (ucs_linear_func_is_equal(perf_elem->perf, perf, 1e-15)) {
-            perf_elem->max_length = max_length;
+    if (!ucs_array_is_empty(range_list)) {
+        range_elem = ucs_array_last(range_list);
+        ucs_assert(new_elem->super.max_length > range_elem->super.max_length);
+        if (ucp_proto_select_range_is_equal(range_elem, new_elem)) {
+            range_elem->super.max_length = new_elem->super.max_length;
             return UCS_OK;
         }
     }
 
-    status = ucs_array_append(ucp_proto_perf, perf_list);
+    status = ucs_array_append(ucp_proto_select_range, range_list);
     if (status != UCS_OK) {
         return status;
     }
 
-    perf_elem             = ucs_array_last(perf_list);
-    perf_elem->max_length = max_length;
-    perf_elem->perf       = perf;
-
+    *ucs_array_last(range_list) = *new_elem;
     return UCS_OK;
 }
 
@@ -119,23 +137,28 @@ static void ucp_proto_select_perf_str(const ucs_linear_func_t *perf,
                                       char *time_str, size_t time_str_max,
                                       char *bw_str, size_t bw_str_max)
 {
-    /* Estimated time */
-    snprintf(time_str, time_str_max, "%.0f + %.3f * N", perf->c * 1e9,
-             perf->m * 1e9);
+    /* Estimated time per 1024 bytes */
+    ucs_snprintf_safe(time_str, time_str_max,
+                      UCP_PROTO_PERF_FUNC_TIME_FMT,
+                      UCP_PROTO_PERF_FUNC_TIME_ARG(perf));
 
     /* Estimated bandwidth (MiB/s) */
-    snprintf(bw_str, bw_str_max, "%.2f", 1.0 / (perf->m * UCS_MBYTE));
+    ucs_snprintf_safe(bw_str, bw_str_max,
+                      UCP_PROTO_PERF_FUNC_BW_FMT,
+                      UCP_PROTO_PERF_FUNC_BW_ARG(perf));
 }
 
 
-static ucs_status_t
-ucp_proto_thresholds_select_best(ucp_proto_id_mask_t proto_mask,
-                                 const ucs_linear_func_t *proto_perf,
-                                 ucs_array_t(ucp_proto_thresh) *thresh_list,
-                                 ucs_array_t(ucp_proto_perf) *perf_list,
-                                 size_t start, size_t end)
+static ucs_status_t ucp_proto_thresholds_select_best(
+        ucp_proto_id_mask_t proto_mask, const ucp_proto_caps_t *proto_caps,
+        const ucp_proto_perf_range_t *proto_perf,
+        ucp_proto_perf_type_t perf_type,
+        ucs_array_t(ucp_proto_thresh) *thresh_list,
+        ucs_array_t(ucp_proto_select_range) *range_list, size_t start,
+        size_t end)
 {
     char time_str[64], bw_str[64], num_str[64];
+    ucp_proto_select_range_t range_elem;
     struct {
         ucp_proto_id_t proto_id;
         double         result;
@@ -144,11 +167,12 @@ ucp_proto_thresholds_select_best(ucp_proto_id_mask_t proto_mask,
     double x_intersect;
     size_t midpoint;
 
-    ucs_trace("  %-16s %-20s %-18s", "PROTOCOL", "TIME", "BANDWIDTH (MB/s)");
+    ucs_trace("  %-20s %-20s %-18s", "PROTOCOL", "TIME (ns/KB)", "BANDWIDTH (MB/s)");
     ucs_for_each_bit(curr.proto_id, proto_mask) {
-        ucp_proto_select_perf_str(&proto_perf[curr.proto_id], time_str,
-                                  sizeof(time_str), bw_str, sizeof(bw_str));
-        ucs_trace("  %-16s %-20s %-18s",
+        ucp_proto_select_perf_str(&proto_perf[curr.proto_id].perf[perf_type],
+                                  time_str, sizeof(time_str), bw_str,
+                                  sizeof(bw_str));
+        ucs_trace("  %-20s %-20s %-18s",
                   ucp_proto_id_field(curr.proto_id, name), time_str, bw_str);
     }
 
@@ -159,8 +183,9 @@ ucp_proto_thresholds_select_best(ucp_proto_id_mask_t proto_mask,
         best.result   = DBL_MAX;
         best.proto_id = UCP_PROTO_ID_INVALID;
         ucs_for_each_bit(curr.proto_id, proto_mask) {
-            curr.result = ucs_linear_func_apply(proto_perf[curr.proto_id],
-                                                start + UCP_PROTO_MSGLEN_EPSILON);
+            curr.result = ucs_linear_func_apply(
+                    proto_perf[curr.proto_id].perf[perf_type],
+                    start + UCP_PROTO_MSGLEN_EPSILON);
             ucs_assert(curr.result != DBL_MAX);
             if ((best.proto_id == UCP_PROTO_ID_INVALID) ||
                 (curr.result < best.result)) {
@@ -183,9 +208,9 @@ ucp_proto_thresholds_select_best(ucp_proto_id_mask_t proto_mask,
         midpoint    = end;
         proto_mask &= ~UCS_BIT(best.proto_id);
         ucs_for_each_bit(curr.proto_id, proto_mask) {
-            status = ucs_linear_func_intersect(proto_perf[curr.proto_id],
-                                               proto_perf[best.proto_id],
-                                               &x_intersect);
+            status = ucs_linear_func_intersect(
+                    proto_perf[curr.proto_id].perf[perf_type],
+                    proto_perf[best.proto_id].perf[perf_type], &x_intersect);
             if ((status == UCS_OK) && (x_intersect > start)) {
                 /* We care only if the intersection is after 'start', since
                  * otherwise best.proto_id is better than curr.proto_id at
@@ -210,8 +235,10 @@ ucp_proto_thresholds_select_best(ucp_proto_id_mask_t proto_mask,
             return status;
         }
 
-        status = ucp_proto_perf_append(perf_list, midpoint,
-                                       proto_perf[best.proto_id]);
+        range_elem.super            = proto_perf[best.proto_id];
+        range_elem.super.max_length = midpoint;
+        range_elem.cfg_thresh       = proto_caps[best.proto_id].cfg_thresh;
+        status = ucp_proto_perf_append(range_list, &range_elem);
         if (status != UCS_OK) {
             return status;
         }
@@ -225,15 +252,15 @@ ucp_proto_thresholds_select_best(ucp_proto_id_mask_t proto_mask,
 /*
  * Select a protocol for 'msg_length', return last message length for the proto
  */
-static ucs_status_t
-ucp_proto_thresholds_select_next(ucp_proto_id_mask_t proto_mask,
-                                 const ucp_proto_caps_t *proto_caps,
-                                 ucs_array_t(ucp_proto_thresh) *thresh_list,
-                                 ucs_array_t(ucp_proto_perf) *perf_list,
-                                 size_t msg_length, size_t *max_length_p)
+static ucs_status_t ucp_proto_thresholds_select_next(
+        ucp_proto_id_mask_t proto_mask, const ucp_proto_caps_t *proto_caps,
+        ucp_proto_perf_type_t perf_type,
+        ucs_array_t(ucp_proto_thresh) *thresh_list,
+        ucs_array_t(ucp_proto_select_range) *range_list, size_t msg_length,
+        size_t *max_length_p)
 {
     ucp_proto_id_mask_t valid_proto_mask, disabled_proto_mask;
-    ucs_linear_func_t proto_perf[UCP_PROTO_MAX_COUNT];
+    ucp_proto_perf_range_t proto_perf[UCP_PROTO_MAX_COUNT];
     ucp_proto_id_t max_prio_proto_id;
     const ucp_proto_caps_t *caps;
     unsigned max_cfg_priority;
@@ -268,7 +295,7 @@ ucp_proto_thresholds_select_next(ucp_proto_id_mask_t proto_mask,
             /* Find first (and only) range which contains 'msg_length' */
             if (msg_length <= caps->ranges[i].max_length) {
                 valid_proto_mask    |= UCS_BIT(proto_id);
-                proto_perf[proto_id] = caps->ranges[i].perf;
+                proto_perf[proto_id] = caps->ranges[i];
                 max_length           = ucs_min(max_length,
                                                caps->ranges[i].max_length);
                 break;
@@ -335,8 +362,9 @@ ucp_proto_thresholds_select_next(ucp_proto_id_mask_t proto_mask,
     }
     ucs_assert(valid_proto_mask != 0);
 
-    status = ucp_proto_thresholds_select_best(valid_proto_mask, proto_perf,
-                                              thresh_list, perf_list,
+    status = ucp_proto_thresholds_select_best(valid_proto_mask, proto_caps,
+                                              proto_perf, perf_type,
+                                              thresh_list, range_list,
                                               msg_length, max_length);
     if (status == UCS_OK) {
         *max_length_p = max_length;
@@ -346,6 +374,48 @@ ucp_proto_thresholds_select_next(ucp_proto_id_mask_t proto_mask,
     return status;
 }
 
+static void
+ucp_proto_select_init_trace_caps(ucp_proto_id_t proto_id,
+                                 const ucp_proto_init_params_t *init_params)
+{
+    ucp_proto_caps_t *proto_caps = init_params->caps;
+    const UCS_V_UNUSED ucs_linear_func_t *perf;
+    size_t UCS_V_UNUSED range_start, range_end;
+    ucs_string_buffer_t config_strb;
+    int UCS_V_UNUSED range_index;
+    char min_length_str[64];
+    char thresh_str[64];
+
+    ucs_string_buffer_init(&config_strb);
+    ucp_proto_id_call(proto_id, config_str, proto_caps->min_length, SIZE_MAX,
+                      init_params->priv, &config_strb);
+    ucs_trace("initialized protocol %s min_length %s cfg_thresh %s %s",
+              init_params->proto_name,
+              ucs_memunits_to_str(proto_caps->min_length, min_length_str,
+                                  sizeof(min_length_str)),
+              ucs_memunits_to_str(proto_caps->cfg_thresh, thresh_str,
+                                  sizeof(thresh_str)),
+              ucs_string_buffer_cstr(&config_strb));
+    ucs_string_buffer_cleanup(&config_strb);
+
+    ucs_log_indent(1);
+    range_start = 0;
+    for (range_index = 0; range_index < proto_caps->num_ranges; ++range_index) {
+        range_start = ucs_max(range_start, proto_caps->min_length);
+        range_end   = proto_caps->ranges[range_index].max_length;
+        if (range_end > range_start) {
+            perf = proto_caps->ranges[range_index].perf;
+            ucs_trace("range[%d] %s" UCP_PROTO_PERF_FUNC_TYPES_FMT, range_index,
+                      ucs_memunits_range_str(range_start, range_end, thresh_str,
+                                             sizeof(thresh_str)),
+                      UCP_PROTO_PERF_FUNC_TYPES_ARG(perf));
+        }
+        range_start = range_end + 1;
+    }
+    ucs_log_indent(-1);
+}
+
+
 static ucs_status_t
 ucp_proto_select_init_protocols(ucp_worker_h worker,
                                 ucp_worker_cfg_index_t ep_cfg_index,
@@ -358,17 +428,16 @@ ucp_proto_select_init_protocols(ucp_worker_h worker,
     ucs_string_buffer_t strb;
     size_t priv_size, offset;
     ucp_proto_id_t proto_id;
-    char min_length_str[64];
-    char thresh_str[64];
     ucs_status_t status;
     void *tmp;
 
     ucs_assert(ep_cfg_index != UCP_WORKER_CFG_INDEX_NULL);
 
-    init_params.worker        = worker;
-    init_params.select_param  = select_param;
-    init_params.ep_cfg_index  = ep_cfg_index;
-    init_params.ep_config_key = &worker->ep_config[ep_cfg_index].key;
+    init_params.worker         = worker;
+    init_params.select_param   = select_param;
+    init_params.ep_cfg_index   = ep_cfg_index;
+    init_params.rkey_cfg_index = rkey_cfg_index;
+    init_params.ep_config_key  = &worker->ep_config[ep_cfg_index].key;
 
     if (rkey_cfg_index == UCP_WORKER_CFG_INDEX_NULL) {
         init_params.rkey_config_key = NULL;
@@ -394,10 +463,10 @@ ucp_proto_select_init_protocols(ucp_worker_h worker,
     }
 
     offset = 0;
-    for (proto_id = 0; proto_id < ucp_protocols_count; ++proto_id) {
-        proto_caps            = &proto_init->caps[proto_id];
-        init_params.priv      = UCS_PTR_BYTE_OFFSET(proto_init->priv_buf,
-                                                          offset);
+    ucs_for_each_bit(proto_id, worker->context->proto_bitmap) {
+        proto_caps             = &proto_init->caps[proto_id];
+        init_params.priv       = UCS_PTR_BYTE_OFFSET(proto_init->priv_buf,
+                                                     offset);
         init_params.priv_size  = &priv_size;
         init_params.caps       = proto_caps;
         init_params.proto_name = ucp_proto_id_field(proto_id, name);
@@ -416,18 +485,7 @@ ucp_proto_select_init_protocols(ucp_worker_h worker,
             continue;
         }
 
-        ucs_string_buffer_init(&strb);
-        ucp_proto_id_call(proto_id, config_str, proto_caps->min_length,
-                          SIZE_MAX, init_params.priv, &strb);
-        ucs_trace("protocol %s has %u ranges, min_length %s, cfg_thresh %s %s",
-                  ucp_proto_id_field(proto_id, name), proto_caps->num_ranges,
-                  ucs_memunits_to_str(proto_caps->min_length, min_length_str,
-                                      sizeof(min_length_str)),
-                  ucs_memunits_to_str(proto_caps->cfg_thresh, thresh_str,
-                                      sizeof(thresh_str)),
-                  ucs_string_buffer_cstr(&strb));
-        ucs_string_buffer_cleanup(&strb);
-
+        ucp_proto_select_init_trace_caps(proto_id, &init_params);
         ucs_log_indent(-1);
 
         /* A successful protocol initialization must return non-empty
@@ -475,18 +533,18 @@ err:
 static ucs_status_t ucp_proto_select_elem_init_thresh(
         ucp_proto_select_elem_t *select_elem,
         const ucp_proto_select_init_protocols_t *proto_init,
-        ucp_worker_cfg_index_t ep_cfg_index,
+        ucp_proto_perf_type_t perf_type, ucp_worker_cfg_index_t ep_cfg_index,
         ucp_worker_cfg_index_t rkey_cfg_index)
 {
     UCS_ARRAY_DEFINE_ONSTACK(tmp_thresh_list, ucp_proto_thresh,
                              UCP_PROTO_MAX_COUNT);
-    UCS_ARRAY_DEFINE_ONSTACK(tmp_perf_list, ucp_proto_perf,
+    UCS_ARRAY_DEFINE_ONSTACK(tmp_range_list, ucp_proto_select_range,
                              UCP_PROTO_MAX_PERF_RANGES);
-    ucp_proto_perf_range_t *perf_ranges, *tmp_perf_elem;
+    ucp_proto_select_range_t *perf_ranges, *tmp_range_elem;
     ucp_proto_threshold_tmp_elem_t *tmp_thresh_elem;
     ucp_proto_threshold_elem_t *thresholds;
-    size_t msg_length, max_length;
     ucp_proto_config_t *proto_config;
+    size_t msg_length, max_length;
     ucp_proto_id_t proto_id;
     ucs_status_t status;
     size_t priv_offset;
@@ -503,9 +561,9 @@ static ucs_status_t ucp_proto_select_elem_init_thresh(
          * protocol is selected.
          */
         status = ucp_proto_thresholds_select_next(proto_init->mask,
-                                                  proto_init->caps,
+                                                  proto_init->caps, perf_type,
                                                   &tmp_thresh_list,
-                                                  &tmp_perf_list, msg_length,
+                                                  &tmp_range_list, msg_length,
                                                   &max_length);
         if (status != UCS_OK) {
             if (status == UCS_ERR_UNSUPPORTED) {
@@ -552,12 +610,13 @@ static ucs_status_t ucp_proto_select_elem_init_thresh(
         ++i;
     }
 
-    ucs_assert_always(!ucs_array_is_empty(&tmp_perf_list));
-    ucs_assert_always(ucs_array_last(&tmp_perf_list)->max_length == SIZE_MAX);
+    ucs_assert_always(!ucs_array_is_empty(&tmp_range_list));
+    ucs_assert_always(ucs_array_last(&tmp_range_list)->super.max_length ==
+                      SIZE_MAX);
 
     /* Allocate performance functions array */
-    perf_ranges = ucs_calloc(ucs_array_length(&tmp_perf_list),
-                             sizeof(*select_elem->perf_ranges), "ucp_proto_perf");
+    perf_ranges = ucs_calloc(ucs_array_length(&tmp_range_list),
+                             sizeof(*perf_ranges), "ucp_proto_perf");
     if (perf_ranges == NULL) {
         status = UCS_ERR_NO_MEMORY;
         goto err_free_thresholds;
@@ -567,8 +626,8 @@ static ucs_status_t ucp_proto_select_elem_init_thresh(
 
     /* Copy the performance elements */
     i = 0;
-    ucs_array_for_each(tmp_perf_elem, &tmp_perf_list) {
-        perf_ranges[i++] = *tmp_perf_elem;
+    ucs_array_for_each(tmp_range_elem, &tmp_range_list) {
+        perf_ranges[i++] = *tmp_range_elem;
     }
 
     return UCS_OK;
@@ -579,6 +638,17 @@ err:
     return status;
 }
 
+static ucp_proto_perf_type_t
+ucp_proto_select_param_perf_type(const ucp_proto_select_param_t *select_param)
+{
+    if (ucp_proto_select_op_attr_from_flags(select_param->op_flags) &
+        UCP_OP_ATTR_FLAG_MULTI_SEND) {
+        return UCP_PROTO_PERF_TYPE_MULTI;
+    } else {
+        return UCP_PROTO_PERF_TYPE_SINGLE;
+    }
+}
+
 static ucs_status_t
 ucp_proto_select_elem_init(ucp_worker_h worker,
                            ucp_worker_cfg_index_t ep_cfg_index,
@@ -588,6 +658,7 @@ ucp_proto_select_elem_init(ucp_worker_h worker,
 {
     UCS_STRING_BUFFER_ONSTACK(sel_param_strb, UCP_PROTO_SELECT_PARAM_STR_MAX);
     ucp_proto_select_init_protocols_t *proto_init;
+    ucp_proto_perf_type_t perf_type;
     ucs_status_t status;
 
     ucp_proto_select_param_str(select_param, &sel_param_strb);
@@ -615,8 +686,10 @@ ucp_proto_select_elem_init(ucp_worker_h worker,
         goto out_free_proto_init;
     }
 
-    status = ucp_proto_select_elem_init_thresh(select_elem, proto_init,
-                                               ep_cfg_index, rkey_cfg_index);
+    perf_type = ucp_proto_select_param_perf_type(select_param);
+    status    = ucp_proto_select_elem_init_thresh(select_elem, proto_init,
+                                                  perf_type, ep_cfg_index,
+                                                  rkey_cfg_index);
     if (status != UCS_OK) {
         goto err_cleanup_protocols;
     }
@@ -720,6 +793,7 @@ ucp_proto_select_dump_all(ucp_worker_h worker,
             "    %-18s %-18s %-20s %-18s %-12s %s\n";
     ucp_proto_select_init_protocols_t *proto_init;
     ucs_string_buffer_t config_strb;
+    ucp_proto_perf_type_t perf_type;
     size_t range_start, range_end;
     const ucp_proto_caps_t *caps;
     ucp_proto_id_t proto_id;
@@ -745,6 +819,8 @@ ucp_proto_select_dump_all(ucp_worker_h worker,
         goto out_free;
     }
 
+    perf_type = ucp_proto_select_param_perf_type(select_param);
+
     ucs_string_buffer_appendf(strb, proto_info_fmt, "PROTOCOL", "SIZE",
                               "TIME (nsec)", "BANDWIDTH (MiB/s)", "THRESHOLD",
                               "CONFIGURATION");
@@ -762,12 +838,13 @@ ucp_proto_select_dump_all(ucp_worker_h worker,
         for (i = 0; i < caps->num_ranges; ++i) {
             /* String for performance range */
             range_end = caps->ranges[i].max_length;
+
             ucs_memunits_range_str(range_start, range_end, range_str,
                                    sizeof(range_str));
+            ucp_proto_select_perf_str(&caps->ranges[i].perf[perf_type],
+                                      time_str, sizeof(time_str), bw_str,
+                                      sizeof(bw_str));
 
-            ucp_proto_select_perf_str(&caps->ranges[i].perf,
-                                      time_str, sizeof(time_str),
-                                      bw_str, sizeof(bw_str));
             /* Get protocol configuration */
             ucs_string_buffer_init(&config_strb);
             ucp_proto_id_call(proto_id, config_str, range_start, range_end,
@@ -832,25 +909,25 @@ ucp_proto_select_dump_thresholds(const ucp_proto_select_elem_t *select_elem,
 
 static void
 ucp_proto_select_dump_perf(const ucp_proto_select_elem_t *select_elem,
+                           ucp_proto_perf_type_t perf_type,
                            ucs_string_buffer_t *strb)
 {
     static const char *proto_info_fmt = "    %-16s %-20s %s\n";
-    const ucp_proto_perf_range_t *perf_elem;
+    const ucp_proto_select_range_t *range_elem;
     size_t range_start, range_end;
     char range_str[128];
     char time_str[64];
     char bw_str[64];
 
     range_start = 0;
-    perf_elem   = select_elem->perf_ranges;
+    range_elem  = select_elem->perf_ranges;
     ucs_string_buffer_appendf(strb, proto_info_fmt, "SIZE", "TIME (nsec)",
                               "BANDWIDTH (MiB/s)");
     do {
-        range_end = perf_elem->max_length;
+        range_end = range_elem->super.max_length;
 
-        ucp_proto_select_perf_str(&perf_elem->perf,
-                                  time_str, sizeof(time_str),
-                                  bw_str, sizeof(bw_str));
+        ucp_proto_select_perf_str(&range_elem->super.perf[perf_type], time_str,
+                                  sizeof(time_str), bw_str, sizeof(bw_str));
         ucs_memunits_range_str(range_start, range_end, range_str,
                                sizeof(range_str));
 
@@ -858,7 +935,7 @@ ucp_proto_select_dump_perf(const ucp_proto_select_elem_t *select_elem,
                                   bw_str);
 
         range_start = range_end + 1;
-        ++perf_elem;
+        ++range_elem;
     } while (range_end != SIZE_MAX);
 }
 
@@ -887,7 +964,9 @@ ucp_proto_select_elem_dump(ucp_worker_h worker,
     ucp_proto_select_dump_thresholds(select_elem, strb);
 
     ucs_string_buffer_appendf(strb, "\n  Performance estimation:\n");
-    ucp_proto_select_dump_perf(select_elem, strb);
+    ucp_proto_select_dump_perf(select_elem,
+                               ucp_proto_select_param_perf_type(select_param),
+                               strb);
 
     ucs_string_buffer_appendf(strb, "\n  Candidates:\n");
     status = ucp_proto_select_dump_all(worker, ep_cfg_index, rkey_cfg_index,
@@ -954,7 +1033,7 @@ void ucp_proto_select_dump_short(const ucp_proto_select_short_t *select_short,
 void ucp_proto_select_param_str(const ucp_proto_select_param_t *select_param,
                                 ucs_string_buffer_t *strb)
 {
-    char sys_dev_name[32];
+    const char *sysdev_name;
     uint32_t op_attr_mask;
 
     op_attr_mask = ucp_proto_select_op_attr_from_flags(select_param->op_flags);
@@ -974,15 +1053,19 @@ void ucp_proto_select_param_str(const ucp_proto_select_param_t *select_param,
     }
 
     if (select_param->sys_dev != UCS_SYS_DEVICE_ID_UNKNOWN) {
-        ucs_topo_sys_device_bdf_name(select_param->sys_dev, sys_dev_name,
-                                     sizeof(sys_dev_name));
-        ucs_string_buffer_appendf(strb, ", %s", sys_dev_name);
+        sysdev_name = ucs_topo_sys_device_get_name(select_param->sys_dev);
+        ucs_string_buffer_appendf(strb, ", %s", sysdev_name);
     }
 
     if (op_attr_mask & UCP_OP_ATTR_FLAG_FAST_CMPL) {
         ucs_string_buffer_appendf(strb, ", fast-completion");
     }
 
+    if (op_attr_mask & UCP_OP_ATTR_FLAG_MULTI_SEND) {
+        ucs_string_buffer_appendf(strb, ", multi");
+    }
+
+    ucs_string_buffer_rtrim(strb, ",");
     ucs_string_buffer_appendf(strb, ")");
 }
 
@@ -1008,6 +1091,7 @@ ucp_proto_select_short_init(ucp_worker_h worker, ucp_proto_select_t *proto_selec
     ucp_proto_select_param_t select_param;
     const ucp_proto_single_priv_t *spriv;
     ucp_memory_info_t mem_info;
+    ssize_t max_short_signed;
     uint32_t op_attr;
 
     ucp_memory_info_set_host(&mem_info);
@@ -1018,6 +1102,7 @@ ucp_proto_select_short_init(ucp_worker_h worker, ucp_proto_select_t *proto_selec
      * use fast-path short protocol only if the message size fits short protocol
      * in both regular mode and UCP_OP_ATTR_FLAG_FAST_CMPL mode.
      */
+    ucs_log_indent(1);
     ucs_for_each_submask(op_attr, op_attr_mask) {
         ucp_proto_select_param_init(&select_param, op_id, op_attr,
                                     UCP_DATATYPE_CONTIG, &mem_info, 1);
@@ -1034,12 +1119,19 @@ ucp_proto_select_short_init(ucp_worker_h worker, ucp_proto_select_t *proto_selec
             goto out_disable;
         }
 
+        /* If max_msg_length exceeds SSIZE_MAX, use SSIZE_MAX, since short
+           protocol thresholds are signed values */
+        max_short_signed = ucs_min(thresh->max_msg_length, SSIZE_MAX);
+
+        ucs_trace("found short protocol %s max_msg_length %zu",
+                  thresh->proto_config.proto->name, thresh->max_msg_length);
+
         /* Assume short protocol uses 'ucp_proto_single_priv_t' */
         spriv = thresh->proto_config.priv;
 
         if (proto == NULL) {
             proto                            = thresh->proto_config.proto;
-            proto_short->max_length_host_mem = thresh->max_msg_length;
+            proto_short->max_length_host_mem = max_short_signed;
             proto_short->lane                = spriv->super.lane;
             proto_short->rkey_index          = spriv->super.rkey_index;
         } else {
@@ -1051,8 +1143,8 @@ ucp_proto_select_short_init(ucp_worker_h worker, ucp_proto_select_t *proto_selec
             }
 
             /* Fast-path threshold is the minimal of all op_attr options */
-            proto_short->max_length_host_mem = ucs_min(
-                    proto_short->max_length_host_mem, thresh->max_msg_length);
+            proto_short->max_length_host_mem =
+                    ucs_min(proto_short->max_length_host_mem, max_short_signed);
         }
     }
 
@@ -1063,8 +1155,156 @@ ucp_proto_select_short_init(ucp_worker_h worker, ucp_proto_select_t *proto_selec
     ucs_assert(proto_short->max_length_host_mem >= 0);
     proto_short->max_length_unknown_mem = (context->num_mem_type_detect_mds > 0) ?
                                           -1 : proto_short->max_length_host_mem;
+    ucs_log_indent(-1);
+    ucs_trace("%s: short threshold host memory %zd unknown memory %zd",
+              ucp_operation_names[op_id], proto_short->max_length_host_mem,
+              proto_short->max_length_unknown_mem);
     return;
 
 out_disable:
+    ucs_log_indent(-1);
+    ucs_trace("%s: disabling short protocol", ucp_operation_names[op_id]);
     ucp_proto_select_short_disable(proto_short);
 }
+
+int ucp_proto_select_get_valid_range(
+        const ucp_proto_threshold_elem_t *thresholds, size_t *min_length_p,
+        size_t *max_length_p)
+{
+    const ucp_proto_threshold_elem_t *elem;
+    size_t max_msg_length;
+    int found;
+
+    found         = 0;
+    *min_length_p = 0;
+    *max_length_p = 0;
+    elem          = thresholds;
+    do {
+        max_msg_length = elem->max_msg_length;
+        if (elem->proto_config.proto->flags & UCP_PROTO_FLAG_INVALID) {
+            /* Protocol is invalid, so set range start after it */
+            if (max_msg_length < SIZE_MAX) {
+                *min_length_p = max_msg_length + 1;
+            }
+        } else {
+            /* Protocol is valid, so extend range end */
+            *max_length_p = max_msg_length;
+            found         = 1;
+        }
+
+        ++elem;
+    } while (max_msg_length < SIZE_MAX);
+
+    return found;
+}
+
+void ucp_proto_threshold_elem_str(const ucp_proto_threshold_elem_t *thresh_elem,
+                                  size_t min_length, size_t max_length,
+                                  ucs_string_buffer_t *strb)
+{
+    size_t range_start, range_end;
+    const ucp_proto_t *proto;
+    char str[64];
+
+    range_start = 0;
+    do {
+        range_end = thresh_elem->max_msg_length;
+
+        /* Print only protocols within the range provided by {min,max}_length */
+        if ((range_end >= min_length) && (range_start <= max_length)) {
+            proto = thresh_elem->proto_config.proto;
+            ucs_string_buffer_appendf(strb, "%s(", proto->name);
+            proto->config_str(ucs_max(range_start, min_length),
+                              ucs_min(range_end, max_length),
+                              thresh_elem->proto_config.priv, strb);
+            ucs_string_buffer_appendf(strb, ")");
+
+            if (range_end < max_length) {
+                ucs_memunits_to_str(thresh_elem->max_msg_length, str,
+                                    sizeof(str));
+                ucs_string_buffer_appendf(strb, "<=%s<", str);
+            }
+        }
+
+        ++thresh_elem;
+        range_start = range_end + 1;
+    } while (range_end < max_length);
+
+    ucs_string_buffer_rtrim(strb, "<");
+}
+
+ucp_proto_select_t *
+ucp_proto_select_get(ucp_worker_h worker, ucp_worker_cfg_index_t ep_cfg_index,
+                     ucp_worker_cfg_index_t rkey_cfg_index,
+                     ucp_worker_cfg_index_t *new_rkey_cfg_index)
+{
+    ucp_rkey_config_key_t rkey_config_key;
+    ucs_status_t status;
+
+    if (rkey_cfg_index == UCP_WORKER_CFG_INDEX_NULL) {
+        *new_rkey_cfg_index = UCP_WORKER_CFG_INDEX_NULL;
+        return &worker->ep_config[ep_cfg_index].proto_select;
+    } else {
+        rkey_config_key = worker->rkey_config[rkey_cfg_index].key;
+
+        rkey_config_key.ep_cfg_index = ep_cfg_index;
+        status = ucp_worker_rkey_config_get(worker, &rkey_config_key, NULL,
+                                            new_rkey_cfg_index);
+        if (status != UCS_OK) {
+            ucs_error("failed to switch to new rkey");
+            return NULL;
+        }
+
+        return &worker->rkey_config[*new_rkey_cfg_index].proto_select;
+    }
+}
+
+void ucp_proto_select_config_str(ucp_worker_h worker,
+                                 const ucp_proto_config_t *proto_config,
+                                 size_t msg_length, ucs_string_buffer_t *strb)
+{
+    const ucp_proto_select_elem_t *select_elem;
+    ucp_worker_cfg_index_t new_key_cfg_index;
+    const ucp_proto_select_range_t *range;
+    ucp_proto_select_t *proto_select;
+    ucp_proto_perf_type_t perf_type;
+    double send_time;
+
+    ucs_assert(worker->context->config.ext.proto_enable);
+
+    /* Print selection parameters */
+    ucp_proto_select_param_str(&proto_config->select_param, strb);
+    ucs_string_buffer_appendf(strb, ": %s ", proto_config->proto->name);
+    proto_config->proto->config_str(msg_length, msg_length, proto_config->priv,
+                                    strb);
+
+    /* Find protocol selection root */
+    proto_select = ucp_proto_select_get(worker, proto_config->ep_cfg_index,
+                                        proto_config->rkey_cfg_index,
+                                        &new_key_cfg_index);
+    if (proto_select == NULL) {
+        return;
+    }
+
+    /* Emulate protocol selection process */
+    ucs_assert(new_key_cfg_index == proto_config->rkey_cfg_index);
+    select_elem = ucp_proto_select_lookup_slow(worker, proto_select,
+                                               proto_config->ep_cfg_index,
+                                               proto_config->rkey_cfg_index,
+                                               &proto_config->select_param);
+    if (select_elem == NULL) {
+        return;
+    }
+
+    /* Find the relevant performance range */
+    range = select_elem->perf_ranges;
+    while (range->super.max_length < msg_length) {
+        ++range;
+    }
+
+    perf_type = ucp_proto_select_param_perf_type(&proto_config->select_param);
+    send_time = ucs_linear_func_apply(range->super.perf[perf_type], msg_length);
+    ucs_string_buffer_appendf(strb, "  %.2f MBs / %.2f us",
+                              (msg_length / send_time) / UCS_MBYTE,
+                              send_time * UCS_USEC_PER_SEC);
+}
