@@ -1,6 +1,7 @@
 
 /**
 * Copyright (C) Mellanox Technologies Ltd. 2001-2013.  ALL RIGHTS RESERVED.
+* Copyright (C) Huawei Technologies Co., Ltd. 2021.  ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -14,8 +15,9 @@
 #ifdef ENABLE_STATS
 
 static ucs_stats_class_t ucs_frag_list_stats_class = {
-    .name = "frag_list",
-    .num_counters = UCS_FRAG_LIST_STAT_LAST,
+    .name          = "frag_list",
+    .num_counters  = UCS_FRAG_LIST_STAT_LAST,
+    .class_id      = UCS_STATS_CLASS_ID_INVALID,
     .counter_names = {
             [UCS_FRAG_LIST_STAT_GAPS]              = "gaps",
             [UCS_FRAG_LIST_STAT_GAP_LEN]           = "gap_len",
@@ -44,8 +46,9 @@ ucs_status_t ucs_frag_list_init(ucs_frag_list_sn_t initial_sn, ucs_frag_list_t *
 #ifdef ENABLE_STATS
     frag_list->prev_sn = initial_sn;
 #endif
-    status = UCS_STATS_NODE_ALLOC(&frag_list->stats, &ucs_frag_list_stats_class,
-                                 stats_parent);
+    status = UCS_STATS_NODE_ALLOC(&frag_list->stats,
+                                  &ucs_frag_list_stats_class,
+                                  stats_parent, "-%p", frag_list);
     return status;
 }
 
@@ -59,14 +62,14 @@ void ucs_frag_list_cleanup(ucs_frag_list_t *frag_list)
 }
 
 /*
- prevh--- h --- .. --- 
+ prevh--- h --- .. ---
           |
           e
           |
           e
  replace h with new_h:
 
-prevh --- new_h --- .. --- 
+prevh --- new_h --- .. ---
           |
           h
           |
@@ -86,7 +89,7 @@ frag_list_replace_head(ucs_frag_list_t *frag_list, ucs_frag_list_elem_t *prevh,
 
     new_h->head.first_sn = h->head.first_sn-1;
     new_h->head.last_sn = h->head.last_sn;
-    /* add new_h before h in holes list */ 
+    /* add new_h before h in holes list */
     /* take h from holes list */
     if (prevh == NULL) {
         e = ucs_queue_pull_elem_non_empty(&frag_list->list, ucs_frag_list_elem_t, list);
@@ -107,13 +110,13 @@ frag_list_replace_head(ucs_frag_list_t *frag_list, ucs_frag_list_elem_t *prevh,
 }
 
 /*
-    ..--- h --- .. --- 
+    ..--- h --- .. ---
           |
           e
 
   add new element to h:
 
-    ..--- h --- .. --- 
+    ..--- h --- .. ---
           |
           |
           e
@@ -130,14 +133,14 @@ static inline void frag_list_add_tail(ucs_frag_list_elem_t *h, ucs_frag_list_ele
     ucs_queue_push(&h->head.list, &elem->list);
 }
 
-/* 
+/*
   merge h2 into h1. Before:
 
-    ..--- h1 --- h2 --- 
+    ..--- h1 --- h2 ---
           |     |
           e     e2
  after:
-    ..--- h1 --- .. --- 
+    ..--- h1 --- .. ---
           |      |
           e      e
           |
@@ -160,9 +163,9 @@ static inline void frag_list_merge_heads(ucs_frag_list_t *head, ucs_frag_list_el
     ucs_queue_splice(&h1->head.list, &h2->head.list);
 }
 
-/* 
+/*
   insert new_h into h1. Before:
- prevh--- h --- .. --- 
+ prevh--- h --- .. ---
           |     |
           e     e
           |
@@ -173,7 +176,7 @@ static inline void frag_list_merge_heads(ucs_frag_list_t *head, ucs_frag_list_el
                     |      |
                     e      e
  */
-static inline void frag_list_insert_head(ucs_frag_list_t *head, 
+static inline void frag_list_insert_head(ucs_frag_list_t *head,
         ucs_frag_list_elem_t *prevh, ucs_frag_list_elem_t *h, ucs_frag_list_elem_t *new_h,  ucs_frag_list_sn_t sn)
 {
 
@@ -191,9 +194,9 @@ static inline void frag_list_insert_head(ucs_frag_list_t *head,
 }
 
 
-/* 
+/*
   insert new_h into h1. Before:
-   ..--- prevh --- h --- 
+   ..--- prevh --- h ---
           |        |
           e        e
           |
@@ -285,7 +288,7 @@ ucs_frag_list_insert_slow(ucs_frag_list_t *head, ucs_frag_list_elem_t *elem,
 
         if (UCS_FRAG_LIST_SN_CMP(sn+1, ==, h->head.first_sn)) {
             frag_list_replace_head(head, prevh, h, elem);
-            /* no need to check merge here. merge iff prev->last_sn+1==sn & sn+1 == h->first_sn 
+            /* no need to check merge here. merge iff prev->last_sn+1==sn & sn+1 == h->first_sn
              * the condition is handled in next if */
             head->elem_count++;
             return UCS_FRAG_LIST_INSERT_SLOW;
@@ -313,7 +316,7 @@ ucs_frag_list_insert_slow(ucs_frag_list_t *head, ucs_frag_list_elem_t *elem,
             if (prevh) {
                 ucs_assert(UCS_FRAG_LIST_SN_CMP(prevh->head.last_sn+1, <, sn));
             }
-            UCS_STATS_UPDATE_COUNTER(head->stats, UCS_FRAG_LIST_STAT_GAP_LEN, 
+            UCS_STATS_UPDATE_COUNTER(head->stats, UCS_FRAG_LIST_STAT_GAP_LEN,
                                      prevh ? sn-prevh->head.last_sn : sn-head->head_sn);
             UCS_STATS_UPDATE_COUNTER(head->stats, UCS_FRAG_LIST_STAT_GAPS, 1);
             frag_list_insert_head(head, prevh, h, elem, sn);
@@ -331,7 +334,7 @@ ucs_frag_list_insert_slow(ucs_frag_list_t *head, ucs_frag_list_elem_t *elem,
 
     head->elem_count++;
     head->list_count++;
-    UCS_STATS_UPDATE_COUNTER(head->stats, UCS_FRAG_LIST_STAT_GAP_LEN, 
+    UCS_STATS_UPDATE_COUNTER(head->stats, UCS_FRAG_LIST_STAT_GAP_LEN,
                              sn-head->head_sn);
     UCS_STATS_UPDATE_COUNTER(head->stats, UCS_FRAG_LIST_STAT_GAPS, 1);
     return UCS_FRAG_LIST_INSERT_SLOW;
@@ -344,7 +347,7 @@ ucs_frag_list_insert_slow(ucs_frag_list_t *head, ucs_frag_list_elem_t *elem,
         e
 
  * mode of action
- *  - check if we have elements on ready list, if we do take one from there 
+ *  - check if we have elements on ready list, if we do take one from there
  *  - see if h is ready for extraction (sn check), extract firt, move rest to the ready list
  */
 
@@ -380,14 +383,14 @@ void ucs_frag_list_dump(ucs_frag_list_t *head, int how)
     elem_count = 0;
 
     ucs_queue_for_each(e, &head->ready_list, list) {
-       elem_count++; 
+       elem_count++;
     }
 
     ucs_queue_for_each(h, &head->list, list) {
         list_count++;
         cnt = 0;
         ucs_queue_for_each(e, &h->head.list, list) {
-           cnt++; 
+           cnt++;
            elem_count++;
         }
         elem_count++;
@@ -407,4 +410,3 @@ void ucs_frag_list_dump(ucs_frag_list_t *head, int how)
     ucs_assert(head->elem_count == elem_count);
     ucs_assert(head->list_count == list_count);
 }
-
