@@ -197,19 +197,19 @@ static inline unsigned long ucs_list_length(ucs_list_link_t *head)
  * Iterate over members of the list.
  */
 #define ucs_list_for_each(_elem, _head, _member) \
-    for (_elem = ucs_container_of((_head)->next, typeof(*_elem), _member); \
+    for (_elem = ucs_container_of((_head)->next, ucs_typeof(*_elem), _member); \
         &(_elem)->_member != (_head); \
-        _elem = ucs_container_of((_elem)->_member.next, typeof(*_elem), _member))
+        _elem = ucs_container_of((_elem)->_member.next, ucs_typeof(*_elem), _member))
 
 /**
  * Iterate over members of the list, the user may invalidate the current entry.
  */
 #define ucs_list_for_each_safe(_elem, _telem, _head, _member) \
-    for (_elem = ucs_container_of((_head)->next, typeof(*_elem), _member), \
-        _telem = ucs_container_of(_elem->_member.next, typeof(*_elem), _member); \
+    for (_elem = ucs_container_of((_head)->next, ucs_typeof(*_elem), _member), \
+        _telem = ucs_container_of(_elem->_member.next, ucs_typeof(*_elem), _member); \
         &_elem->_member != (_head); \
         _elem = _telem, \
-        _telem = ucs_container_of(_telem->_member.next, typeof(*_telem), _member))
+        _telem = ucs_container_of(_telem->_member.next, ucs_typeof(*_telem), _member))
 
 /**
  * Extract list head
