@@ -40,14 +40,16 @@ typedef ssize_t (*ucs_socket_iov_func_t)(int fd, const struct msghdr *msg,
                                          int flags);
 
 
-static void ucs_socket_print_error_info(int sys_errno)
+static void ucs_socket_print_error_info(const char *err_str, int sys_errno)
 {
     if (sys_errno == EMFILE) {
-        ucs_error("the maximal number of files that could be opened "
+        ucs_error("%s: the maximal number of files that could be opened "
                   "simultaneously was reached, try to increase the limit "
                   "by setting the max open files limit (ulimit -n) to "
                   "a greater value (current: %d)",
-                  ucs_sys_max_open_files());
+                  err_str, ucs_sys_max_open_files());
+    } else {
+        ucs_error("%s: %m", err_str);
     }
 }
 
@@ -98,22 +100,73 @@ out:
     return status;
 }
 
-int ucs_netif_is_active(const char *if_name)
+ucs_status_t ucs_netif_get_addr(const char *if_name, sa_family_t af,
+                                struct sockaddr *saddr,
+                                struct sockaddr *netmask)
 {
-    ucs_status_t status;
-    struct ifreq ifr;
+    ucs_status_t status = UCS_ERR_NO_DEVICE;
+    struct ifaddrs *ifa;
+    struct ifaddrs *ifaddrs;
+    const struct sockaddr_in6 *saddr6;
+    size_t addrlen;
 
-    status = ucs_netif_ioctl(if_name, SIOCGIFADDR, &ifr);
-    if (status != UCS_OK) {
-        return 0;
+    if (getifaddrs(&ifaddrs)) {
+        ucs_warn("getifaddrs error: %m");
+        status = UCS_ERR_IO_ERROR;
+        goto out;
     }
 
-    status = ucs_netif_ioctl(if_name, SIOCGIFFLAGS, &ifr);
-    if (status != UCS_OK) {
-        return 0;
+    for (ifa = ifaddrs; ifa != NULL; ifa = ifa->ifa_next) {
+        if ((if_name != NULL) && (0 != strcmp(if_name, ifa->ifa_name))) {
+            continue;
+        }
+
+        if ((ifa->ifa_addr == NULL) ||
+            ((ifa->ifa_addr->sa_family != AF_INET) &&
+             (ifa->ifa_addr->sa_family != AF_INET6))) {
+            continue;
+        }
+
+        if (!ucs_netif_flags_is_active(ifa->ifa_flags)) {
+            continue;
+        }
+
+        if (ifa->ifa_addr->sa_family == AF_INET6) {
+            saddr6 = (const struct sockaddr_in6*)ifa->ifa_addr;
+            if (IN6_IS_ADDR_LOOPBACK(&saddr6->sin6_addr) ||
+                IN6_IS_ADDR_LINKLOCAL(&saddr6->sin6_addr)) {
+                continue;
+            }
+        }
+
+        if ((af == AF_UNSPEC) || (ifa->ifa_addr->sa_family == af)) {
+            status = ucs_sockaddr_sizeof(ifa->ifa_addr, &addrlen);
+            if (status != UCS_OK) {
+                goto out_free_ifaddr;
+            }
+
+            if (saddr != NULL) {
+                memcpy(saddr, ifa->ifa_addr, addrlen);
+            }
+
+            if (netmask != NULL) {
+                memcpy(netmask, ifa->ifa_netmask, addrlen);
+            }
+
+            status = UCS_OK;
+            break;
+        }
     }
 
-    return ucs_netif_flags_is_active(ifr.ifr_flags);
+out_free_ifaddr:
+    freeifaddrs(ifaddrs);
+out:
+    return status;
+}
+
+int ucs_netif_is_active(const char *if_name, sa_family_t af)
+{
+    return ucs_netif_get_addr(if_name, af, NULL, NULL) == UCS_OK;
 }
 
 unsigned ucs_netif_bond_ad_num_ports(const char *bond_name)
@@ -125,7 +178,7 @@ unsigned ucs_netif_bond_ad_num_ports(const char *bond_name)
                                   UCS_NETIF_BOND_AD_NUM_PORTS_FMT, bond_name);
     if ((status != UCS_OK) || (ad_num_ports <= 0) ||
         (ad_num_ports > UINT_MAX)) {
-        ucs_diag("failed to read from " UCS_NETIF_BOND_AD_NUM_PORTS_FMT ": %m, "
+        ucs_trace("failed to read from " UCS_NETIF_BOND_AD_NUM_PORTS_FMT ": %m, "
                  "assuming 802.3ad bonding is disabled", bond_name);
         return 1;
     }
@@ -137,8 +190,7 @@ ucs_status_t ucs_socket_create(int domain, int type, int *fd_p)
 {
     int fd = socket(domain, type, 0);
     if (fd < 0) {
-        ucs_error("socket create failed: %m");
-        ucs_socket_print_error_info(errno);
+        ucs_socket_print_error_info("socket create failed", errno);
         return UCS_ERR_IO_ERROR;
     }
 
@@ -181,17 +233,30 @@ ucs_status_t ucs_socket_getopt(int fd, int level, int optname,
     return UCS_OK;
 }
 
+ucs_status_t ucs_socket_getname(int fd, struct sockaddr_storage *sock_addr,
+                                socklen_t *addr_len)
+{
+    int ret;
+
+    *addr_len = sizeof(struct sockaddr_storage);
+    ret       = getsockname(fd, (struct sockaddr*)sock_addr,
+                            addr_len);
+    if (ret < 0) {
+        ucs_error("getsockname(fd=%d) failed: %m", fd);
+        return UCS_ERR_IO_ERROR;
+    }
+
+    return UCS_OK;
+}
+
 const char *ucs_socket_getname_str(int fd, char *str, size_t max_size)
 {
     struct sockaddr_storage sock_addr = {0}; /* Suppress Clang false-positive */
     socklen_t addr_size;
-    int ret;
+    ucs_status_t status;
 
-    addr_size = sizeof(sock_addr);
-    ret       = getsockname(fd, (struct sockaddr*)&sock_addr,
-                            &addr_size);
-    if (ret < 0) {
-        ucs_debug("getsockname(fd=%d) failed: %m", fd);
+    status = ucs_socket_getname(fd, &sock_addr, &addr_size);
+    if (status != UCS_OK) {
         ucs_strncpy_safe(str, "-", max_size);
         return str;
     }
@@ -280,6 +345,7 @@ ucs_status_t ucs_socket_accept(int fd, struct sockaddr *addr, socklen_t *length_
 {
     ucs_status_t status;
     char ip_port_str[UCS_SOCKADDR_STRING_LEN];
+    UCS_STRING_BUFFER_ONSTACK(strb, 128);
 
     *accept_fd = accept(fd, addr, length_ptr);
     if (*accept_fd < 0) {
@@ -288,10 +354,10 @@ ucs_status_t ucs_socket_accept(int fd, struct sockaddr *addr, socklen_t *length_
             return status;
         }
 
-        ucs_error("accept() failed (client addr %s): %m",
-                  ucs_sockaddr_str(addr, ip_port_str, UCS_SOCKADDR_STRING_LEN));
-
-        ucs_socket_print_error_info(errno);
+        ucs_string_buffer_appendf(&strb, "accept() failed (client addr %s)",
+                                  ucs_sockaddr_str(addr, ip_port_str,
+                                                   UCS_SOCKADDR_STRING_LEN));
+        ucs_socket_print_error_info(ucs_string_buffer_cstr(&strb), errno);
 
         return status;
     }
@@ -807,6 +873,27 @@ int ucs_sockaddr_ip_cmp(const struct sockaddr *sa1, const struct sockaddr *sa2)
                   UCS_IPV4_ADDR_LEN : UCS_IPV6_ADDR_LEN);
 }
 
+ucs_status_t ucs_sockaddr_set_inaddr_any(struct sockaddr *saddr, sa_family_t af)
+{
+    struct sockaddr_in *sa_in;
+    struct sockaddr_in6 *sa_in6;
+
+    switch (af) {
+    case AF_INET:
+        sa_in                  = (struct sockaddr_in*)saddr;
+        sa_in->sin_addr.s_addr = INADDR_ANY;
+        break;
+    case AF_INET6:
+        sa_in6            = (struct sockaddr_in6*)saddr;
+        sa_in6->sin6_addr = in6addr_any;
+        break;
+    default:
+        ucs_debug("invalid address family: %d", af);
+        return UCS_ERR_INVALID_PARAM;
+    }
+    return UCS_OK;
+}
+
 int ucs_sockaddr_is_inaddr_any(const struct sockaddr *addr)
 {
     switch (addr->sa_family) {
@@ -875,7 +962,7 @@ ucs_status_t ucs_sockaddr_get_ifname(int fd, char *ifname_str, size_t max_strlen
         return UCS_ERR_INVALID_PARAM;
     }
 
-    ucs_debug("check ifname for socket on %s", 
+    ucs_debug("check ifname for socket on %s",
               ucs_sockaddr_str(my_addr, str_local_addr, UCS_SOCKADDR_STRING_LEN));
 
     if (getifaddrs(&ifaddrs)) {
@@ -891,7 +978,7 @@ ucs_status_t ucs_sockaddr_get_ifname(int fd, char *ifname_str, size_t max_strlen
             continue;
         }
 
-        if (((sa->sa_family == AF_INET) ||(sa->sa_family == AF_INET6)) && 
+        if (((sa->sa_family == AF_INET) ||(sa->sa_family == AF_INET6)) &&
             (!ucs_sockaddr_cmp(sa, my_addr, NULL))) {
             ucs_debug("matching ip found iface on %s", ifa->ifa_name);
             ucs_strncpy_safe(ifname_str, ifa->ifa_name, max_strlen);
@@ -905,40 +992,6 @@ ucs_status_t ucs_sockaddr_get_ifname(int fd, char *ifname_str, size_t max_strlen
     return status;
 }
 
-static ucs_status_t ucs_sockaddr_ifreq(const char *if_name,
-                                       unsigned long request,
-                                       struct sockaddr_in *dest)
-{
-    ucs_status_t status;
-    struct ifreq ifr;
-
-    status = ucs_netif_ioctl(if_name, request, &ifr);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    if (ifr.ifr_addr.sa_family != AF_INET) {
-        ucs_error("%s address is not INET", if_name);
-        return UCS_ERR_INVALID_ADDR;
-    }
-
-    memcpy(dest, &ifr.ifr_addr, sizeof(*dest));
-
-    return UCS_OK;
-}
-
-ucs_status_t
-ucs_sockaddr_get_ifaddr(const char *if_name, struct sockaddr_in *addr)
-{
-    return ucs_sockaddr_ifreq(if_name, SIOCGIFADDR, addr);
-}
-
-ucs_status_t
-ucs_sockaddr_get_ifmask(const char *if_name, struct sockaddr_in *mask)
-{
-    return ucs_sockaddr_ifreq(if_name, SIOCGIFNETMASK, mask);
-}
-
 const char *ucs_sockaddr_address_family_str(sa_family_t af)
 {
     switch (af) {
