@@ -264,9 +264,8 @@ int ucs_get_first_cpu()
     int first_cpu, total_cpus, ret;
     ucs_sys_cpuset_t mask;
 
-    ret = ucs_sysconf(_SC_NPROCESSORS_CONF);
+    ret = ucs_sys_get_num_cpus();
     if (ret < 0) {
-        ucs_error("failed to get local cpu count: %m");
         return ret;
     }
     total_cpus = ret;
@@ -399,7 +398,7 @@ static ssize_t ucs_read_file_vararg(char *buffer, size_t max, int silent,
     ssize_t read_bytes;
     int fd;
 
-    vsnprintf(filename, MAXPATHLEN, filename_fmt, ap);
+    ucs_vsnprintf_safe(filename, MAXPATHLEN, filename_fmt, ap);
 
     fd = open(filename, O_RDONLY);
     if (fd < 0) {
@@ -560,6 +559,7 @@ void ucs_sys_iterate_vm(void *address, size_t size, ucs_sys_vma_cb_t cb,
             }
         }
 
+        /* coverity[tainted_data] */
         cb(&info, ctx);
     }
 
@@ -873,7 +873,7 @@ ucs_status_t ucs_sysv_alloc(size_t *size, size_t max_size, void **address_p,
 #ifdef SHM_HUGETLB
             if (!(flags & SHM_HUGETLB))
 #endif
-	    {
+            {
                 ucs_error("%s", error_string);
             }
             return UCS_ERR_NO_MEMORY;
@@ -924,7 +924,7 @@ ucs_status_t ucs_sysv_alloc(size_t *size, size_t max_size, void **address_p,
         }
     }
 
-    ucs_memtrack_allocated(ptr, alloc_size UCS_MEMTRACK_VAL);
+    ucs_memtrack_allocated(ptr, alloc_size, alloc_name);
     *address_p = ptr;
     *size      = alloc_size;
     return UCS_OK;
@@ -945,7 +945,7 @@ ucs_status_t ucs_sysv_free(void *address)
 }
 
 ucs_status_t ucs_mmap_alloc(size_t *size, void **address_p,
-                            int flags UCS_MEMTRACK_ARG)
+                            int flags, const char *alloc_name)
 {
     size_t alloc_length;
     void *addr;
@@ -953,7 +953,7 @@ ucs_status_t ucs_mmap_alloc(size_t *size, void **address_p,
     alloc_length = ucs_align_up_pow2(*size, ucs_get_page_size());
 
     addr = ucs_mmap(*address_p, alloc_length, PROT_READ | PROT_WRITE,
-                    MAP_PRIVATE | MAP_ANON | flags, -1, 0 UCS_MEMTRACK_VAL);
+                    MAP_PRIVATE | MAP_ANON | flags, -1, 0, alloc_name);
     if (addr == MAP_FAILED) {
         return UCS_ERR_NO_MEMORY;
     }
@@ -1487,3 +1487,57 @@ ucs_status_t ucs_sys_get_file_time(const char *name, ucs_sys_file_time_t type,
         return UCS_ERR_INVALID_PARAM;
     }
 }
+
+ucs_status_t ucs_sys_check_fd_limit_per_process()
+{
+    int fd;
+
+    fd = open("/dev/null", O_RDONLY);
+    if ((fd == -1) && (errno == EMFILE)) {
+        return UCS_ERR_EXCEEDS_LIMIT;
+    }
+
+    if (fd != -1) {
+        close(fd);
+    }
+
+    return UCS_OK;
+}
+
+ucs_status_t ucs_pthread_create(pthread_t *thread_id_p,
+                                void *(*start_routine)(void*), void *arg,
+                                const char *fmt, ...)
+{
+    char thread_name[NAME_MAX];
+    pthread_t thread_id;
+    va_list ap;
+    int ret;
+
+    ret = pthread_create(&thread_id, NULL, start_routine, arg);
+    if (ret != 0) {
+        ucs_error("pthread_create() failed: %m");
+        return UCS_ERR_IO_ERROR;
+    }
+
+    va_start(ap, fmt);
+    ucs_vsnprintf_safe(thread_name, sizeof(thread_name), fmt, ap);
+    va_end(ap);
+
+    pthread_setname_np(thread_id, thread_name);
+    *thread_id_p = thread_id;
+    return UCS_OK;
+}
+
+long ucs_sys_get_num_cpus()
+{
+    static long num_cpus = 0;
+
+    if (num_cpus == 0) {
+        num_cpus = ucs_sysconf(_SC_NPROCESSORS_CONF);
+        if (num_cpus == -1) {
+            ucs_error("failed to get local cpu count: %m");
+        }
+    }
+
+    return num_cpus;
+}
