@@ -16,6 +16,11 @@
 #include <ucs/sys/string.h>
 
 
+#define UCT_GDR_COPY_IFACE_DEFAULT_BANDWIDTH (6911.0 * UCS_MBYTE)
+#define UCT_GDR_COPY_IFACE_OVERHEAD          0
+#define UCT_GDR_COPY_IFACE_LATENCY           ucs_linear_func_make(1e-6, 0)
+
+
 static ucs_config_field_t uct_gdr_copy_iface_config_table[] = {
 
     {"", "", NULL,
@@ -86,7 +91,7 @@ uct_gdr_copy_iface_query(uct_iface_h tl_iface, uct_iface_attr_t *iface_attr)
     iface_attr->cap.am.max_hdr          = 0;
     iface_attr->cap.am.max_iov          = 1;
 
-    iface_attr->latency                 = ucs_linear_func_make(1e-6, 0);
+    iface_attr->latency                 = UCT_GDR_COPY_IFACE_LATENCY;
     iface_attr->bandwidth.dedicated     = 0;
     iface_attr->bandwidth.shared        = UCT_GDR_COPY_IFACE_DEFAULT_BANDWIDTH;
     iface_attr->overhead                = UCT_GDR_COPY_IFACE_OVERHEAD;
@@ -102,11 +107,11 @@ uct_gdr_copy_estimate_perf(uct_iface_h iface, uct_perf_attr_t *perf_attr)
         perf_attr->bandwidth.dedicated = 0;
         if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_OPERATION) {
             switch (perf_attr->operation) {
-            case UCT_OP_GET_SHORT:
-            case UCT_OP_GET_ZCOPY:
+            case UCT_EP_OP_GET_SHORT:
+            case UCT_EP_OP_GET_ZCOPY:
                 perf_attr->bandwidth.shared = 440.0 * UCS_MBYTE;
                 break;
-            case UCT_OP_PUT_SHORT:
+            case UCT_EP_OP_PUT_SHORT:
                 perf_attr->bandwidth.shared = 10200.0 * UCS_MBYTE;
                 break;
             default:
@@ -118,8 +123,20 @@ uct_gdr_copy_estimate_perf(uct_iface_h iface, uct_perf_attr_t *perf_attr)
         }
     }
 
-    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_OVERHEAD) {
-        perf_attr->overhead = UCT_GDR_COPY_IFACE_OVERHEAD;
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_SEND_PRE_OVERHEAD) {
+        perf_attr->send_pre_overhead = UCT_GDR_COPY_IFACE_OVERHEAD;
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_SEND_POST_OVERHEAD) {
+        perf_attr->send_post_overhead = 0;
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_RECV_OVERHEAD) {
+        perf_attr->recv_overhead = 0;
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_LATENCY) {
+        perf_attr->latency = UCT_GDR_COPY_IFACE_LATENCY;
     }
 
     return UCS_OK;
@@ -149,6 +166,7 @@ static uct_iface_ops_t uct_gdr_copy_iface_ops = {
 static uct_iface_internal_ops_t uct_gdr_copy_iface_internal_ops = {
     .iface_estimate_perf = uct_gdr_copy_estimate_perf,
     .iface_vfs_refresh   = (uct_iface_vfs_refresh_func_t)ucs_empty_function,
+    .ep_query            = (uct_ep_query_func_t)ucs_empty_function_return_unsupported
 };
 
 static UCS_CLASS_INIT_FUNC(uct_gdr_copy_iface_t, uct_md_h md, uct_worker_h worker,
