@@ -21,6 +21,7 @@
 #include <ucs/time/time.h>
 #include <ucm/api/ucm.h>
 #include <ucs/datastruct/string_buffer.h>
+#include <ucs/vfs/base/vfs_obj.h>
 #include <pthread.h>
 #ifdef HAVE_PTHREAD_NP_H
 #include <pthread_np.h>
@@ -150,7 +151,7 @@ static ucs_config_field_t uct_ib_md_config_table[] = {
 
     {"GPU_DIRECT_RDMA", "try",
      "Use GPU Direct RDMA for HCA to access GPU pages directly\n",
-     ucs_offsetof(uct_ib_md_config_t, ext.enable_gpudirect_rdma), UCS_CONFIG_TYPE_TERNARY},
+     ucs_offsetof(uct_ib_md_config_t, enable_gpudirect_rdma), UCS_CONFIG_TYPE_TERNARY},
 
 #ifdef HAVE_EXP_UMR
     {"MAX_INLINE_KLM_LIST", "inf",
@@ -196,8 +197,9 @@ static ucs_config_field_t uct_ib_md_config_table[] = {
 
 #ifdef ENABLE_STATS
 static ucs_stats_class_t uct_ib_md_stats_class = {
-    .name           = "",
-    .num_counters   = UCT_IB_MD_STAT_LAST,
+    .name          = "",
+    .num_counters  = UCT_IB_MD_STAT_LAST,
+    .class_id      = UCS_STATS_CLASS_ID_INVALID,
     .counter_names = {
         [UCT_IB_MD_STAT_MEM_ALLOC]   = "mem_alloc",
         [UCT_IB_MD_STAT_MEM_REG]     = "mem_reg"
@@ -285,19 +287,6 @@ typedef struct {
     int           silent;
 } uct_ib_md_mem_reg_thread_t;
 
-static void uct_ib_check_gpudirect_driver(uct_ib_md_t *md, uct_md_attr_t *md_attr,
-                                          const char *file,
-                                          ucs_memory_type_t mem_type)
-{
-    if (!access(file, F_OK)) {
-        md_attr->cap.reg_mem_types |= UCS_BIT(mem_type);
-    }
-
-    ucs_debug("%s: %s GPUDirect RDMA is %s",
-              uct_ib_device_name(&md->dev), ucs_memory_type_names[mem_type],
-              md_attr->cap.reg_mem_types & UCS_BIT(mem_type) ?
-              "enabled" : "disabled");
-}
 
 static ucs_status_t uct_ib_md_query(uct_md_h uct_md, uct_md_attr_t *md_attr)
 {
@@ -314,28 +303,9 @@ static ucs_status_t uct_ib_md_query(uct_md_h uct_md, uct_md_attr_t *md_attr)
     md_attr->cap.alloc_mem_types  = 0;
     md_attr->cap.access_mem_types = UCS_BIT(UCS_MEMORY_TYPE_HOST);
     md_attr->cap.detect_mem_types = 0;
-
-    if (md->config.enable_gpudirect_rdma != UCS_NO) {
-        /* check if GDR driver is loaded */
-        uct_ib_check_gpudirect_driver(md, md_attr,
-                                      "/sys/kernel/mm/memory_peers/nv_mem/version",
-                                      UCS_MEMORY_TYPE_CUDA);
-
-        /* check if ROCM KFD driver is loaded */
-        uct_ib_check_gpudirect_driver(md, md_attr, "/dev/kfd",
-                                      UCS_MEMORY_TYPE_ROCM);
-
-        if (!(md_attr->cap.reg_mem_types & ~UCS_MEMORY_TYPES_CPU_ACCESSIBLE) &&
-            (md->config.enable_gpudirect_rdma == UCS_YES)) {
-                ucs_error("%s: Couldn't enable GPUDirect RDMA. Please make sure"
-                          " nv_peer_mem or amdgpu plugin installed correctly.",
-                          uct_ib_device_name(&md->dev));
-                return UCS_ERR_UNSUPPORTED;
-        }
-    }
-
-    md_attr->rkey_packed_size = UCT_IB_MD_PACKED_RKEY_SIZE;
-    md_attr->reg_cost         = md->reg_cost;
+    md_attr->cap.reg_mem_types    = md->reg_mem_types;
+    md_attr->rkey_packed_size     = UCT_IB_MD_PACKED_RKEY_SIZE;
+    md_attr->reg_cost             = md->reg_cost;
     ucs_sys_cpuset_copy(&md_attr->local_cpus, &md->dev.local_cpus);
 
     return UCS_OK;
@@ -809,7 +779,7 @@ static ucs_status_t uct_ib_mem_reg_internal(uct_md_h uct_md, void *address,
         memh->flags |= UCT_IB_MEM_FLAG_RELAXED_ORDERING;
     }
 
-    ucs_debug("registered memory %p..%p on %s lkey 0x%x rkey 0x%x "
+    ucs_trace("registered memory %p..%p on %s lkey 0x%x rkey 0x%x "
               "access 0x%lx flags 0x%x", address,
               UCS_PTR_BYTE_OFFSET(address, length),
               uct_ib_device_name(&md->dev), memh->lkey, memh->rkey,
@@ -867,8 +837,8 @@ static ucs_status_t uct_ib_mem_dereg(uct_md_h uct_md,
     uct_ib_memh_free(ib_memh);
     if (UCT_MD_MEM_DEREG_FIELD_VALUE(params, flags, FIELD_FLAGS, 0) &
         UCT_MD_MEM_DEREG_FLAG_INVALIDATE) {
-        ucs_assert(params->cb != NULL); /* suppress coverity false-positive */
-        params->cb(UCT_MD_MEM_DEREG_FIELD_VALUE(params, arg, FIELD_ARG, NULL));
+        ucs_assert(params->comp != NULL); /* suppress coverity false-positive */
+        uct_invoke_completion(params->comp, UCS_OK);
     }
 
     return status;
@@ -1037,12 +1007,18 @@ static ucs_status_t uct_ib_mem_rcache_reg(uct_md_h uct_md, void *address,
     return UCS_OK;
 }
 
+static void uct_ib_mem_region_invalidate_cb(void *arg)
+{
+    uct_completion_t *comp = arg;
+
+    uct_invoke_completion(comp, UCS_OK);
+}
+
 static ucs_status_t
 uct_ib_mem_rcache_dereg(uct_md_h uct_md,
                         const uct_md_mem_dereg_params_t *params)
 {
     uct_ib_md_t *md = ucs_derived_of(uct_md, uct_ib_md_t);
-    void *arg;
     uct_ib_rcache_region_t *region;
 
     UCT_MD_MEM_DEREG_CHECK_PARAMS(params, 1);
@@ -1050,9 +1026,9 @@ uct_ib_mem_rcache_dereg(uct_md_h uct_md,
     region = uct_ib_rcache_region_from_memh(params->memh);
     if (UCT_MD_MEM_DEREG_FIELD_VALUE(params, flags, FIELD_FLAGS, 0) &
         UCT_MD_MEM_DEREG_FLAG_INVALIDATE) {
-        arg = UCT_MD_MEM_DEREG_FIELD_VALUE(params, arg, FIELD_ARG, NULL);
-        ucs_rcache_region_invalidate(md->rcache, &region->super, params->cb,
-                                     arg);
+        ucs_rcache_region_invalidate(md->rcache, &region->super,
+                                     uct_ib_mem_region_invalidate_cb,
+                                     params->comp);
     }
 
     ucs_rcache_region_put(md->rcache, &region->super);
@@ -1165,6 +1141,10 @@ uct_ib_mem_global_odp_dereg(uct_md_h uct_md,
 
     ib_memh = params->memh;
     status  = uct_ib_memh_dereg(md, ib_memh);
+    if (status != UCS_OK) {
+        return status;
+    }
+
     uct_ib_memh_free(ib_memh);
     return status;
 }
@@ -1179,11 +1159,33 @@ static uct_md_ops_t UCS_V_UNUSED uct_ib_md_global_odp_ops = {
     .detect_memory_type = ucs_empty_function_return_unsupported,
 };
 
+int uct_ib_device_is_accessible(struct ibv_device *device)
+{
+    /* Enough place to hold the full path */
+    char device_path[IBV_SYSFS_PATH_MAX];
+    struct stat st;
+
+    ucs_snprintf_safe(device_path, sizeof(device_path), "%s%s",
+                      "/dev/infiniband/", device->dev_name);
+
+    /* Could not stat the path or
+       the path is not a char device file or
+       the device cannot be accessed for read & write
+    */
+    if ((stat(device_path, &st) != 0) || !S_ISCHR(st.st_mode) ||
+        (access(device_path, R_OK | W_OK) != 0)) {
+        return 0;
+    }
+
+    return 1;
+}
+
 static ucs_status_t uct_ib_query_md_resources(uct_component_t *component,
                                               uct_md_resource_desc_t **resources_p,
                                               unsigned *num_resources_p)
 {
     UCS_MODULE_FRAMEWORK_DECLARE(uct_ib);
+    int num_resources = 0;
     uct_md_resource_desc_t *resources;
     struct ibv_device **device_list;
     ucs_status_t status;
@@ -1207,12 +1209,19 @@ static ucs_status_t uct_ib_query_md_resources(uct_component_t *component,
     }
 
     for (i = 0; i < num_devices; ++i) {
-        ucs_snprintf_zero(resources[i].md_name, sizeof(resources[i].md_name),
+        /* Skip non-existent and non-accessible devices */
+        if (!uct_ib_device_is_accessible(device_list[i])) {
+            continue;
+        }
+
+        ucs_snprintf_zero(resources[num_resources].md_name,
+                          sizeof(resources[num_resources].md_name),
                           "%s", ibv_get_device_name(device_list[i]));
+        num_resources++;
     }
 
     *resources_p     = resources;
-    *num_resources_p = num_devices;
+    *num_resources_p = num_resources;
     status = UCS_OK;
 
 out_free_device_list:
@@ -1284,7 +1293,7 @@ err:
 }
 
 static ucs_status_t
-uct_ib_md_parse_reg_methods(uct_ib_md_t *md, uct_md_attr_t *md_attr,
+uct_ib_md_parse_reg_methods(uct_ib_md_t *md,
                             const uct_ib_md_config_t *md_config)
 {
     ucs_rcache_params_t rcache_params;
@@ -1298,8 +1307,8 @@ uct_ib_md_parse_reg_methods(uct_ib_md_t *md, uct_md_attr_t *md_attr,
                                                md->memh_struct_size;
             rcache_params.max_alignment      = ucs_get_page_size();
             rcache_params.ucm_events         = UCM_EVENT_VM_UNMAPPED;
-            if (md_attr->cap.reg_mem_types & ~UCS_BIT(UCS_MEMORY_TYPE_HOST)) {
-                rcache_params.ucm_events     |= UCM_EVENT_MEM_TYPE_FREE;
+            if (md->reg_mem_types & ~UCS_BIT(UCS_MEMORY_TYPE_HOST)) {
+                rcache_params.ucm_events    |= UCM_EVENT_MEM_TYPE_FREE;
             }
             rcache_params.context            = md;
             rcache_params.ops                = &uct_ib_rcache_ops;
@@ -1315,7 +1324,8 @@ uct_ib_md_parse_reg_methods(uct_ib_md_t *md, uct_md_attr_t *md_attr,
             }
 
             md->super.ops = &uct_ib_md_rcache_ops;
-            md->reg_cost  = ucs_linear_func_make(md_config->rcache.overhead, 0);
+            md->reg_cost  = ucs_linear_func_make(
+                    uct_md_rcache_overhead(&md_config->rcache), 0);
             ucs_debug("%s: using registration cache",
                       uct_ib_device_name(&md->dev));
             return UCS_OK;
@@ -1352,9 +1362,10 @@ uct_ib_md_parse_reg_methods(uct_ib_md_t *md, uct_md_attr_t *md_attr,
 static ucs_status_t
 uct_ib_md_parse_device_config(uct_ib_md_t *md, const uct_ib_md_config_t *md_config)
 {
+    char *flags_str = NULL;
     uct_ib_device_spec_t *spec;
     ucs_status_t status;
-    char *flags_str, *p;
+    char *p;
     unsigned i, count;
     int nfields;
 
@@ -1386,6 +1397,9 @@ uct_ib_md_parse_device_config(uct_ib_md_t *md, const uct_ib_md_config_t *md_conf
         }
 
         if (nfields >= 4) {
+            /* Check that 'flags_str' is not NULL to suppress the Coverity warning */
+            ucs_assert(flags_str != NULL);
+
             for (p = flags_str; *p != 0; ++p) {
                 if (*p == '4') {
                     spec->flags |= UCT_IB_DEVICE_FLAG_MLX4_PRM;
@@ -1400,11 +1414,13 @@ uct_ib_md_parse_device_config(uct_ib_md_t *md, const uct_ib_md_config_t *md_conf
                 } else {
                     ucs_error("invalid device flag: '%c'", *p);
                     free(flags_str);
+                    flags_str = NULL;
                     status = UCS_ERR_INVALID_PARAM;
                     goto err_free;
                 }
             }
             free(flags_str);
+            flags_str = NULL;
         }
 
         ucs_trace("added device '%s' vendor_id 0x%x device_id %d flags %c%c prio %d",
@@ -1645,11 +1661,22 @@ void uct_ib_md_parse_relaxed_order(uct_ib_md_t *md,
     }
 }
 
+static void uct_ib_check_gpudirect_driver(uct_ib_md_t *md, const char *file,
+                                          ucs_memory_type_t mem_type)
+{
+    if (!access(file, F_OK)) {
+        md->reg_mem_types |= UCS_BIT(mem_type);
+    }
+
+    ucs_debug("%s: %s GPUDirect RDMA is %s", uct_ib_device_name(&md->dev),
+              ucs_memory_type_names[mem_type],
+              md->reg_mem_types & UCS_BIT(mem_type) ? "enabled" : "disabled");
+}
+
 ucs_status_t uct_ib_md_open_common(uct_ib_md_t *md,
                                    struct ibv_device *ib_device,
                                    const uct_ib_md_config_t *md_config)
 {
-    uct_md_attr_t md_attr;
     ucs_status_t status;
 
     md->super.ops       = &uct_ib_md_ops;
@@ -1661,8 +1688,8 @@ ucs_status_t uct_ib_md_open_common(uct_ib_md_t *md,
 
     /* Create statistics */
     status = UCS_STATS_NODE_ALLOC(&md->stats, &uct_ib_md_stats_class,
-                                  ucs_stats_get_root(),
-                                  "%s-%p", ibv_get_device_name(ib_device), md);
+                                  ucs_stats_get_root(), "%s-%p",
+                                  ibv_get_device_name(ib_device), md);
     if (status != UCS_OK) {
         goto err;
     }
@@ -1697,22 +1724,39 @@ ucs_status_t uct_ib_md_open_common(uct_ib_md_t *md,
         goto err_cleanup_device;
     }
 
-    status = uct_md_query(&md->super, &md_attr);
-    if (status != UCS_OK) {
-        goto err_dealloc_pd;
+    /* Check for GPU-direct support */
+    md->reg_mem_types = UCS_BIT(UCS_MEMORY_TYPE_HOST);
+    if (md_config->enable_gpudirect_rdma != UCS_NO) {
+        /* check if GDR driver is loaded */
+        uct_ib_check_gpudirect_driver(
+                md, "/sys/kernel/mm/memory_peers/nv_mem/version",
+                UCS_MEMORY_TYPE_CUDA);
+
+        /* check if ROCM KFD driver is loaded */
+        uct_ib_check_gpudirect_driver(md, "/dev/kfd", UCS_MEMORY_TYPE_ROCM);
+
+        if (!(md->reg_mem_types & ~UCS_MEMORY_TYPES_CPU_ACCESSIBLE) &&
+            (md_config->enable_gpudirect_rdma == UCS_YES)) {
+            ucs_error("%s: Couldn't enable GPUDirect RDMA. Please make sure"
+                      " nv_peer_mem or amdgpu plugin installed correctly.",
+                      uct_ib_device_name(&md->dev));
+            status = UCS_ERR_UNSUPPORTED;
+            goto err_dealloc_pd;
+        }
     }
 
-    status = uct_ib_md_parse_reg_methods(md, &md_attr, md_config);
+    status = uct_ib_md_parse_reg_methods(md, md_config);
     if (status != UCS_OK) {
         goto err_dealloc_pd;
     }
 
     md->dev.max_zcopy_log_sge = INT_MAX;
-    if (md_attr.cap.reg_mem_types & ~UCS_BIT(UCS_MEMORY_TYPE_HOST)) {
+    if (md->reg_mem_types & ~UCS_BIT(UCS_MEMORY_TYPE_HOST)) {
         md->dev.max_zcopy_log_sge = 1;
     }
 
     md->pci_bw = uct_ib_md_pci_bw(md_config, ib_device);
+
     return UCS_OK;
 
 err_dealloc_pd:
@@ -1815,6 +1859,15 @@ err:
     return status;
 }
 
+static void uct_ib_md_vfs_init(uct_md_h md)
+{
+    uct_ib_md_t *ib_md = ucs_derived_of(md, uct_ib_md_t);
+
+    if (ib_md->rcache != NULL) {
+        ucs_vfs_obj_add_sym_link(md, ib_md->rcache, "rcache");
+    }
+}
+
 static uct_ib_md_ops_t uct_ib_verbs_md_ops = {
     .open                = uct_ib_verbs_md_open,
     .cleanup             = (uct_ib_md_cleanup_func_t)ucs_empty_function,
@@ -1846,6 +1899,7 @@ uct_component_t uct_ib_component = {
     },
     .cm_config          = UCS_CONFIG_EMPTY_GLOBAL_LIST_ENTRY,
     .tl_list            = UCT_COMPONENT_TL_LIST_INITIALIZER(&uct_ib_component),
-    .flags              = 0
+    .flags              = 0,
+    .md_vfs_init        = uct_ib_md_vfs_init
 };
 UCT_COMPONENT_REGISTER(&uct_ib_component);
