@@ -76,11 +76,13 @@
 #if HAVE_DECL_MLX5DV_UAR_ALLOC_TYPE_BF
 #  define UCT_IB_MLX5_UAR_ALLOC_TYPE_WC MLX5DV_UAR_ALLOC_TYPE_BF
 #else
-#  define UCT_IB_MLX5_UAR_ALLOC_TYPE_WC 0
+#  define UCT_IB_MLX5_UAR_ALLOC_TYPE_WC 0x0
 #endif
 
 #if HAVE_DECL_MLX5DV_UAR_ALLOC_TYPE_NC
 #  define UCT_IB_MLX5_UAR_ALLOC_TYPE_NC MLX5DV_UAR_ALLOC_TYPE_NC
+#else
+#  define UCT_IB_MLX5_UAR_ALLOC_TYPE_NC 0x1
 #endif
 
 #define UCT_IB_MLX5_OPMOD_EXT_ATOMIC(_log_arg_size) \
@@ -197,6 +199,16 @@ enum {
     UCT_IB_MLX5_SRQ_TOPO_CYCLIC_MP_RQ = 0x3
 };
 
+
+enum {
+#if UCS_ENABLE_ASSERT
+    UCT_IB_MLX5_TXWQ_FLAG_FAILED = UCS_BIT(0)
+#else
+    UCT_IB_MLX5_TXWQ_FLAG_FAILED = 0
+#endif
+};
+
+
 #if HAVE_DEVX
 typedef struct uct_ib_mlx5_devx_umem {
     struct mlx5dv_devx_umem  *mem;
@@ -383,7 +395,8 @@ typedef struct uct_ib_mlx5_txwq {
     uint16_t                    bb_max;
     uint16_t                    sig_pi;     /* PI for last signaled WQE */
 #if UCS_ENABLE_ASSERT
-    uint16_t                    hw_ci;
+    uint16_t                    hw_ci; /* First BB index of last completed WQE */
+    uint8_t                     flags; /* Debug flags */
 #endif
     uct_ib_fence_info_t         fi;
 } uct_ib_mlx5_txwq_t;
@@ -544,6 +557,13 @@ ucs_status_t uct_ib_mlx5_txwq_init(uct_priv_worker_t *worker,
                                    uct_ib_mlx5_mmio_mode_t cfg_mmio_mode,
                                    uct_ib_mlx5_txwq_t *txwq, struct ibv_qp *verbs_qp);
 
+/* Get pointer to a WQE by producer index */
+void *uct_ib_mlx5_txwq_get_wqe(const uct_ib_mlx5_txwq_t *txwq, uint16_t pi);
+
+/* Count how many WQEs are currently posted */
+uint16_t uct_ib_mlx5_txwq_num_posted_wqes(const uct_ib_mlx5_txwq_t *txwq,
+                                          uint16_t outstanding);
+
 void uct_ib_mlx5_qp_mmio_cleanup(uct_ib_mlx5_qp_t *qp,
                                  uct_ib_mlx5_mmio_reg_t *reg);
 
@@ -587,6 +607,16 @@ ucs_status_t uct_ib_mlx5_devx_uar_init(uct_ib_mlx5_devx_uar_t *uar,
 
 void uct_ib_mlx5_devx_uar_cleanup(uct_ib_mlx5_devx_uar_t *uar);
 
+/**
+ * Check whether the interface uses AR.
+ */
+int uct_ib_mlx5_iface_has_ar(uct_ib_iface_t *iface);
+
+
+void uct_ib_mlx5_txwq_validate_always(uct_ib_mlx5_txwq_t *wq, uint16_t num_bb,
+                                      int hw_ci_updated);
+
+
 /**
  * DEVX QP API
  */
