@@ -1,5 +1,6 @@
 /**
 * Copyright (C) Mellanox Technologies Ltd. 2001-2018.  ALL RIGHTS RESERVED.
+* Copyright (C) Huawei Technologies Co., Ltd. 2021.  ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -41,7 +42,7 @@ ucs_config_field_t uct_rc_mlx5_common_config_table[] = {
    ucs_offsetof(uct_rc_mlx5_iface_common_config_t, tm.seg_size),
    UCS_CONFIG_TYPE_MEMUNITS},
 
-  {"TM_MP_SRQ_ENABLE", "try",
+  {"TM_MP_SRQ_ENABLE", "no",
    "Enable multi-packet SRQ support. Relevant for hardware tag-matching only.",
    ucs_offsetof(uct_rc_mlx5_iface_common_config_t, tm.mp_enable),
    UCS_CONFIG_TYPE_TERNARY},
@@ -62,27 +63,79 @@ ucs_config_field_t uct_rc_mlx5_common_config_table[] = {
    ucs_offsetof(uct_rc_mlx5_iface_common_config_t, exp_backoff),
    UCS_CONFIG_TYPE_UINT},
 
-  {"CYCLIC_SRQ_ENABLE", "try",
-   "Enable using the \"cyclic\" SRQ type (SRQ is organized as a continuous \n"
-   "array of WQEs), otherwise - using the \"list\" SRQ type (SRQ is organized \n"
-   "as a buffer containing linked list of WQEs.",
-   ucs_offsetof(uct_rc_mlx5_iface_common_config_t, cyclic_srq_enable),
-   UCS_CONFIG_TYPE_TERNARY},
+  {"SRQ_TOPO", "cyclic,cyclic_emulated",
+   "List of SRQ topology types in order of preference. Supported types are:\n"
+   "\n"
+   "list              SRQ is organized as a buffer containing linked list of WQEs.\n"
+   "\n"
+   "cyclic            SRQ is organized as a continuous array of WQEs. Requires DEVX.\n"
+   "\n"
+   "cyclic_emulated   SRQ is organized as a continuous array of WQEs, but HW\n"
+   "                  treats it as a linked list. Doesn`t require DEVX.",
+   ucs_offsetof(uct_rc_mlx5_iface_common_config_t, srq_topo),
+   UCS_CONFIG_TYPE_STRING_ARRAY},
+
+  {"LOG_ACK_REQ_FREQ", "8",
+   "Log of the ack frequency for requests, when using DevX. Valid values are: 0-"
+    UCS_PP_MAKE_STRING(UCT_RC_MLX5_MAX_LOG_ACK_REQ_FREQ) ".",
+   ucs_offsetof(uct_rc_mlx5_iface_common_config_t, log_ack_req_freq),
+   UCS_CONFIG_TYPE_UINT},
 
   {NULL}
 };
 
 
+static UCS_F_ALWAYS_INLINE ucs_status_t
+uct_rc_mlx5_iface_srq_set_seg(uct_rc_mlx5_iface_common_t *iface,
+                              uct_ib_mlx5_srq_seg_t *seg)
+{
+    uct_ib_iface_recv_desc_t *desc;
+    uint64_t desc_map;
+    void *hdr;
+    int i;
+
+    desc_map = ~seg->srq.ptr_mask & UCS_MASK(iface->tm.mp.num_strides);
+    ucs_for_each_bit(i, desc_map) {
+        UCT_TL_IFACE_GET_RX_DESC(&iface->super.super.super, &iface->super.rx.mp,
+                                 desc, return UCS_ERR_NO_MEMORY);
+
+        /* Set receive data segment pointer. Length is pre-initialized. */
+        hdr                = uct_ib_iface_recv_desc_hdr(&iface->super.super,
+                                                        desc);
+        seg->srq.ptr_mask |= UCS_BIT(i);
+        seg->srq.desc      = desc; /* Optimization for non-MP case (1 stride) */
+        seg->dptr[i].lkey  = htonl(desc->lkey);
+        seg->dptr[i].addr  = htobe64((uintptr_t)hdr);
+        VALGRIND_MAKE_MEM_NOACCESS(hdr, iface->super.super.config.seg_size);
+    }
+
+    return UCS_OK;
+}
+
+/* Update resources and write doorbell record */
+static UCS_F_ALWAYS_INLINE void
+uct_rc_mlx5_iface_update_srq_res(uct_rc_iface_t *iface, uct_ib_mlx5_srq_t *srq,
+                                 uint16_t wqe_index, uint16_t count)
+{
+    ucs_assert(iface->rx.srq.available >= count);
+
+    if (count == 0) {
+        return;
+    }
+
+    srq->ready_idx              = wqe_index;
+    srq->sw_pi                 += count;
+    iface->rx.srq.available    -= count;
+    ucs_memory_cpu_store_fence();
+    *srq->db                    = htonl(srq->sw_pi);
+}
+
 unsigned uct_rc_mlx5_iface_srq_post_recv(uct_rc_mlx5_iface_common_t *iface)
 {
     uct_ib_mlx5_srq_t *srq   = &iface->rx.srq;
     uct_rc_iface_t *rc_iface = &iface->super;
     uct_ib_mlx5_srq_seg_t *seg;
-    uct_ib_iface_recv_desc_t *desc;
     uint16_t count, wqe_index, next_index;
-    uint64_t desc_map;
-    void *hdr;
-    int i;
 
     /* Make sure the union is right */
     UCS_STATIC_ASSERT(ucs_offsetof(uct_ib_mlx5_srq_seg_t, mlx5_srq.next_wqe_index) ==
@@ -91,6 +144,7 @@ unsigned uct_rc_mlx5_iface_srq_post_recv(uct_rc_mlx5_iface_common_t *iface)
                       sizeof(struct mlx5_wqe_srq_next_seg));
 
     ucs_assert(UCS_CIRCULAR_COMPARE16(srq->ready_idx, <=, srq->free_idx));
+    ucs_assert(rc_iface->rx.srq.available > 0);
 
     wqe_index = srq->ready_idx;
     for (;;) {
@@ -106,40 +160,59 @@ unsigned uct_rc_mlx5_iface_srq_post_recv(uct_rc_mlx5_iface_common_t *iface)
             srq->free_idx  = next_index;
         }
 
-        desc_map = ~seg->srq.ptr_mask & UCS_MASK(iface->tm.mp.num_strides);
-        ucs_for_each_bit(i, desc_map) {
-            UCT_TL_IFACE_GET_RX_DESC(&rc_iface->super.super, &rc_iface->rx.mp,
-                                     desc, goto out);
-
-            /* Set receive data segment pointer. Length is pre-initialized. */
-            hdr                = uct_ib_iface_recv_desc_hdr(&rc_iface->super, desc);
-            seg->srq.ptr_mask |= UCS_BIT(i);
-            seg->srq.desc      = desc; /* Optimization for non-MP case (1 stride) */
-            seg->dptr[i].lkey  = htonl(desc->lkey);
-            seg->dptr[i].addr  = htobe64((uintptr_t)hdr);
-            VALGRIND_MAKE_MEM_NOACCESS(hdr, rc_iface->super.config.seg_size);
+        if (uct_rc_mlx5_iface_srq_set_seg(iface, seg) != UCS_OK) {
+            break;
         }
 
         wqe_index = next_index;
     }
 
-out:
     count = wqe_index - srq->sw_pi;
-    ucs_assert(rc_iface->rx.srq.available >= count);
+    uct_rc_mlx5_iface_update_srq_res(rc_iface, srq, wqe_index, count);
+    ucs_assert(uct_ib_mlx5_srq_get_wqe(srq, srq->mask)->srq.next_wqe_index == 0);
+    return count;
+}
 
-    if (count > 0) {
-        srq->ready_idx              = wqe_index;
-        srq->sw_pi                  = wqe_index;
-        rc_iface->rx.srq.available -= count;
-        ucs_memory_cpu_store_fence();
-        *srq->db                    = htonl(srq->sw_pi);
-        ucs_assert(uct_ib_mlx5_srq_get_wqe(srq, srq->mask)->srq.next_wqe_index == 0);
+unsigned uct_rc_mlx5_iface_srq_post_recv_ll(uct_rc_mlx5_iface_common_t *iface)
+{
+    uct_ib_mlx5_srq_t *srq     = &iface->rx.srq;
+    uct_rc_iface_t *rc_iface   = &iface->super;
+    uct_ib_mlx5_srq_seg_t *seg = NULL;
+    uint16_t count             = 0;
+    uint16_t wqe_index, next_index;
+
+    ucs_assert(rc_iface->rx.srq.available > 0);
+
+    wqe_index = srq->ready_idx;
+    seg       = uct_ib_mlx5_srq_get_wqe(srq, wqe_index);
+
+    for (;;) {
+        next_index = ntohs(seg->srq.next_wqe_index);
+        if (next_index == (srq->free_idx & srq->mask)) {
+            break;
+        }
+        seg = uct_ib_mlx5_srq_get_wqe(srq, next_index);
+
+        if (uct_rc_mlx5_iface_srq_set_seg(iface, seg) != UCS_OK) {
+            break;
+        }
+
+        wqe_index = next_index;
+        count++;
     }
+
+    uct_rc_mlx5_iface_update_srq_res(rc_iface, srq, wqe_index, count);
     return count;
 }
 
 void uct_rc_mlx5_iface_common_prepost_recvs(uct_rc_mlx5_iface_common_t *iface)
 {
+    /* prepost recvs only if quota available (recvs were not preposted
+     * before) */
+    if (iface->super.rx.srq.quota == 0) {
+        return;
+    }
+
     iface->super.rx.srq.available = iface->super.rx.srq.quota;
     iface->super.rx.srq.quota     = 0;
     uct_rc_mlx5_iface_srq_post_recv(iface);
@@ -173,8 +246,9 @@ UCT_RC_MLX5_DEFINE_ATOMIC_LE_HANDLER(64)
 #if IBV_HW_TM
 #  ifdef ENABLE_STATS
 static ucs_stats_class_t uct_rc_mlx5_tag_stats_class = {
-    .name = "tag",
-    .num_counters = UCT_RC_MLX5_STAT_TAG_LAST,
+    .name          = "tag",
+    .num_counters  = UCT_RC_MLX5_STAT_TAG_LAST,
+    .class_id      = UCS_STATS_CLASS_ID_INVALID,
     .counter_names = {
         [UCT_RC_MLX5_STAT_TAG_RX_EXP]            = "rx_exp",
         [UCT_RC_MLX5_STAT_TAG_RX_EAGER_UNEXP]    = "rx_unexp_eager",
@@ -202,6 +276,7 @@ uct_rc_mlx5_devx_create_cmd_qp(uct_rc_mlx5_iface_common_t *iface)
 
     ucs_assert(iface->tm.cmd_wq.super.super.type == UCT_IB_MLX5_OBJ_TYPE_LAST);
 
+    attr.super.qp_type          = IBV_QPT_RC;
     attr.super.cap.max_send_wr  = iface->tm.cmd_qp_len;
     attr.super.cap.max_send_sge = 1;
     attr.super.ibv.pd           = md->super.pd;
@@ -225,7 +300,7 @@ uct_rc_mlx5_devx_create_cmd_qp(uct_rc_mlx5_iface_common_t *iface)
     status = uct_rc_mlx5_iface_common_devx_connect_qp(
             iface, &iface->tm.cmd_wq.super.super,
             iface->tm.cmd_wq.super.super.qp_num, &ah_attr,
-            iface->super.super.config.path_mtu);
+            iface->super.super.config.path_mtu, 0);
     if (status != UCS_OK) {
         goto err_destroy_qp;
     }
@@ -385,8 +460,9 @@ ucs_status_t uct_rc_mlx5_iface_common_tag_init(uct_rc_mlx5_iface_common_t *iface
     iface->tm.head = &iface->tm.list[0];
     iface->tm.tail = &iface->tm.list[i];
 
-    status = UCS_STATS_NODE_ALLOC(&iface->tm.stats, &uct_rc_mlx5_tag_stats_class,
-                                  iface->stats);
+    status = UCS_STATS_NODE_ALLOC(&iface->tm.stats,
+                                  &uct_rc_mlx5_tag_stats_class,
+                                  iface->stats, "");
     if (status != UCS_OK) {
         ucs_error("Failed to allocate tag stats: %s", ucs_status_string(status));
         goto err_cmd_wq_free;
@@ -416,7 +492,8 @@ void uct_rc_mlx5_iface_common_tag_cleanup(uct_rc_mlx5_iface_common_t *iface)
     }
 
     uct_ib_mlx5_destroy_qp(md, &iface->tm.cmd_wq.super.super);
-    uct_ib_mlx5_txwq_cleanup(&iface->tm.cmd_wq.super);
+    uct_ib_mlx5_qp_mmio_cleanup(&iface->tm.cmd_wq.super.super,
+                                iface->tm.cmd_wq.super.reg);
     ucs_free(iface->tm.list);
     ucs_free(iface->tm.cmd_wq.ops);
     uct_rc_mlx5_tag_cleanup(iface);
@@ -458,6 +535,7 @@ void uct_rc_mlx5_iface_fill_attr(uct_rc_mlx5_iface_common_t *iface,
         break;
     case UCT_IB_MLX5_OBJ_TYPE_DEVX:
         uct_rc_iface_fill_attr(&iface->super, &qp_attr->super, max_send_wr, NULL);
+        qp_attr->mmio_mode = iface->tx.mmio_mode;
         break;
     case UCT_IB_MLX5_OBJ_TYPE_LAST:
         break;
@@ -466,18 +544,6 @@ void uct_rc_mlx5_iface_fill_attr(uct_rc_mlx5_iface_common_t *iface,
     qp_attr->super.srq_num = srq->srq_num;
 }
 
-static ucs_status_t
-uct_rc_mlx5_iface_check_no_devx_rx(uct_rc_mlx5_iface_common_t *iface)
-{
-    if (iface->config.cyclic_srq_enable == UCS_YES) {
-        ucs_error(UCT_IB_IFACE_FMT ": cyclic SRQ type is not supported",
-                  UCT_IB_IFACE_ARG(&iface->super.super));
-        return UCS_ERR_UNSUPPORTED;
-    }
-
-    return UCS_OK;
-}
-
 ucs_status_t
 uct_rc_mlx5_common_iface_init_rx(uct_rc_mlx5_iface_common_t *iface,
                                  const uct_rc_iface_common_config_t *rc_config)
@@ -485,10 +551,7 @@ uct_rc_mlx5_common_iface_init_rx(uct_rc_mlx5_iface_common_t *iface,
     uct_ib_mlx5_md_t *md = ucs_derived_of(iface->super.super.super.md, uct_ib_mlx5_md_t);
     ucs_status_t status;
 
-    status = uct_rc_mlx5_iface_check_no_devx_rx(iface);
-    if (status != UCS_OK) {
-        return status;
-    }
+    ucs_assert(iface->config.srq_topo != UCT_RC_MLX5_SRQ_TOPO_CYCLIC);
 
     status = uct_rc_iface_init_rx(&iface->super, rc_config,
                                   &iface->rx.srq.verbs.srq);
@@ -548,18 +611,16 @@ void uct_rc_mlx5_release_desc(uct_recv_desc_t *self, void *desc)
 void uct_rc_mlx5_handle_unexp_rndv(uct_rc_mlx5_iface_common_t *iface,
                                    struct ibv_tmh *tmh, uct_tag_t tag,
                                    struct mlx5_cqe64 *cqe, unsigned flags,
-                                   unsigned byte_len)
+                                   unsigned byte_len, int poll_flags)
 {
     uct_rc_mlx5_tmh_priv_data_t *priv = (uct_rc_mlx5_tmh_priv_data_t*)tmh->reserved;
-    uct_ib_md_t *ib_md                = uct_ib_iface_md(&iface->super.super);
     struct ibv_rvh *rvh;
     unsigned tm_hdrs_len;
     unsigned rndv_usr_hdr_len;
     size_t rndv_data_len;
     void *rndv_usr_hdr;
-    void *rb;
     ucs_status_t status;
-    char packed_rkey[UCT_COMPONENT_NAME_MAX + UCT_IB_MD_PACKED_RKEY_SIZE];
+    char packed_rkey[UCT_IB_MD_PACKED_RKEY_SIZE];
 
     rvh = (struct ibv_rvh*)(tmh + 1);
 
@@ -583,8 +644,7 @@ void uct_rc_mlx5_handle_unexp_rndv(uct_rc_mlx5_iface_common_t *iface,
     memcpy((char*)rndv_usr_hdr - priv->length, &priv->data, priv->length);
 
     /* Create "packed" rkey to pass it in the callback */
-    rb = uct_md_fill_md_name(&ib_md->super, packed_rkey);
-    uct_ib_md_pack_rkey(ntohl(rvh->rkey), UCT_IB_INVALID_RKEY, rb);
+    uct_ib_md_pack_rkey(ntohl(rvh->rkey), UCT_IB_INVALID_RKEY, packed_rkey);
 
     /* Do not pass flags to cb, because rkey is allocated on stack */
     status = iface->tm.rndv_unexp.cb(iface->tm.rndv_unexp.arg, 0, tag,
@@ -595,7 +655,8 @@ void uct_rc_mlx5_handle_unexp_rndv(uct_rc_mlx5_iface_common_t *iface,
 
     uct_rc_mlx5_iface_unexp_consumed(iface, iface->tm.rndv_desc.offset,
                                      &iface->tm.rndv_desc.super, cqe,
-                                     status, ntohs(cqe->wqe_counter));
+                                     status, ntohs(cqe->wqe_counter),
+                                     poll_flags);
 
     UCT_RC_MLX5_TM_STAT(iface, RX_RNDV_UNEXP);
 }
@@ -633,7 +694,8 @@ static ucs_mpool_ops_t uct_dm_iface_mpool_ops = {
     .chunk_alloc   = uct_rc_mlx5_iface_common_dm_mpool_chunk_malloc,
     .chunk_release = ucs_mpool_chunk_free,
     .obj_init      = uct_rc_mlx5_iface_common_dm_mp_obj_init,
-    .obj_cleanup   = NULL
+    .obj_cleanup   = NULL,
+    .obj_str       = NULL
 };
 
 
@@ -796,10 +858,7 @@ ucs_status_t uct_rc_mlx5_init_rx_tm(uct_rc_mlx5_iface_common_t *iface,
     uct_ib_md_t *md = uct_ib_iface_md(&iface->super.super);
     ucs_status_t status;
 
-    status = uct_rc_mlx5_iface_check_no_devx_rx(iface);
-    if (status != UCS_OK) {
-        return status;
-    }
+    ucs_assert(iface->config.srq_topo != UCT_RC_MLX5_SRQ_TOPO_CYCLIC);
 
     uct_rc_mlx5_init_rx_tm_common(iface, config, rndv_hdr_len);
 
@@ -878,7 +937,7 @@ void uct_rc_mlx5_tag_cleanup(uct_rc_mlx5_iface_common_t *iface)
 {
 #if IBV_HW_TM
     if (UCT_RC_MLX5_TM_ENABLED(iface)) {
-        ucs_ptr_array_cleanup(&iface->tm.rndv_comps);
+        ucs_ptr_array_cleanup(&iface->tm.rndv_comps, 1);
         UCS_STATS_NODE_FREE(iface->tm.stats);
     }
 #endif
@@ -962,6 +1021,44 @@ void uct_rc_mlx5_iface_common_dm_cleanup(uct_rc_mlx5_iface_common_t *iface)
 #endif
 }
 
+#if HAVE_DECL_MLX5DV_CREATE_QP
+void uct_rc_mlx5_common_fill_dv_qp_attr(uct_rc_mlx5_iface_common_t *iface,
+                                        struct ibv_qp_init_attr_ex *qp_attr,
+                                        struct mlx5dv_qp_init_attr *dv_attr,
+                                        unsigned scat2cqe_dir_mask)
+{
+#if HAVE_DECL_MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE
+    dv_attr->comp_mask   |= MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS;
+    dv_attr->create_flags = 0;
+
+    if ((scat2cqe_dir_mask & UCS_BIT(UCT_IB_DIR_RX)) &&
+        (iface->super.super.config.max_inl_cqe[UCT_IB_DIR_RX] == 0)) {
+        /* make sure responder scatter2cqe is disabled */
+        dv_attr->create_flags |= MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE;
+    }
+#endif
+
+    if (scat2cqe_dir_mask & UCS_BIT(UCT_IB_DIR_TX)) {
+        if (iface->super.super.config.max_inl_cqe[UCT_IB_DIR_TX] == 0) {
+            /* Tell the driver to not signal all send WRs, so it will disable
+             * requester scatter2cqe. Set this also for older driver which
+             * doesn't support specific scatter2cqe flags.
+             */
+            qp_attr->sq_sig_all = 0;
+        }
+#if HAVE_DECL_MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE
+        else if (!(dv_attr->create_flags &
+                   MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE)) {
+            /* force-enable requester scatter2cqe, regardless of SIGNAL_ALL_WR,
+             * unless it was already disabled on responder side (otherwise
+             * mlx5 driver check fails) */
+            dv_attr->create_flags |= MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE;
+        }
+#endif
+    }
+}
+#endif
+
 void uct_rc_mlx5_iface_common_query(uct_ib_iface_t *ib_iface,
                                     uct_iface_attr_t *iface_attr,
                                     size_t max_inline, size_t max_tag_eager_iov)
@@ -1051,6 +1148,29 @@ void uct_rc_mlx5_iface_common_sync_cqs_ci(uct_rc_mlx5_iface_common_t *iface,
 #endif
 }
 
+ucs_status_t
+uct_rc_mlx5_iface_common_arm_cq(uct_ib_iface_t *ib_iface, uct_ib_dir_t dir,
+                                int solicited_only)
+{
+    uct_rc_mlx5_iface_common_t *iface = ucs_derived_of(ib_iface,
+                                                       uct_rc_mlx5_iface_common_t);
+#if HAVE_DECL_MLX5DV_INIT_OBJ
+    return uct_ib_mlx5dv_arm_cq(&iface->cq[dir], solicited_only);
+#else
+    uct_ib_mlx5_update_cq_ci(iface->super.super.cq[dir],
+                             iface->cq[dir].cq_ci);
+    return uct_ib_iface_arm_cq(ib_iface, dir, solicited_only);
+#endif
+}
+
+void uct_rc_mlx5_iface_common_event_cq(uct_ib_iface_t *ib_iface,
+                                       uct_ib_dir_t dir)
+{
+    uct_rc_mlx5_iface_common_t *iface = ucs_derived_of(ib_iface,
+                                                       uct_rc_mlx5_iface_common_t);
+    iface->cq[dir].cq_sn++;
+}
+
 int uct_rc_mlx5_iface_commom_clean(uct_ib_mlx5_cq_t *mlx5_cq,
                                    uct_ib_mlx5_srq_t *srq, uint32_t qpn)
 {
@@ -1102,8 +1222,8 @@ int uct_rc_mlx5_iface_commom_clean(uct_ib_mlx5_cq_t *mlx5_cq,
         }
     }
 
-    mlx5_cq->cq_ci             += nfreed;
+    mlx5_cq->cq_ci += nfreed;
+    uct_ib_mlx5_update_db_cq_ci(mlx5_cq);
 
     return nfreed;
 }
-
