@@ -58,6 +58,13 @@ typedef struct uct_rc_verbs_ep {
 } uct_rc_verbs_ep_t;
 
 
+/* Context for cleaning QP */
+typedef struct {
+    uct_rc_iface_qp_cleanup_ctx_t super;
+    struct ibv_qp                 *qp;
+} uct_rc_verbs_iface_qp_cleanup_ctx_t;
+
+
 /**
  * RC verbs interface configuration.
  */
@@ -153,6 +160,8 @@ ucs_status_t uct_rc_verbs_ep_fence(uct_ep_h tl_ep, unsigned flags);
 
 void uct_rc_verbs_ep_post_check(uct_ep_h tl_ep);
 
+void uct_rc_verbs_ep_vfs_populate(uct_rc_ep_t *rc_ep);
+
 ucs_status_t uct_rc_verbs_ep_fc_ctrl(uct_ep_t *tl_ep, unsigned op,
                                      uct_rc_pending_req_t *req);
 
@@ -162,6 +171,4 @@ ucs_status_t uct_rc_verbs_ep_connect_to_ep(uct_ep_h tl_ep,
                                            const uct_device_addr_t *dev_addr,
                                            const uct_ep_addr_t *ep_addr);
 
-unsigned uct_rc_verbs_ep_cleanup_qp(void *arg);
-
 #endif
