@@ -9,12 +9,17 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <sys/time.h>
+#include <unistd.h>
 #include <string.h>
 #include <assert.h>
+#include <errno.h>
+#include <malloc.h>
 
 #include <algorithm>
 #include <limits>
 
+#include <ucs/debug/memtrack.h>
+
 
 #define AM_MSG_ID 0
 
@@ -27,16 +32,7 @@ struct ucx_request {
     uint32_t                     conn_id;
     size_t                       recv_length;
     ucs_list_link_t              pos;
-};
-
-// Holds details of arrived AM message
-struct UcxAmDesc {
-    UcxAmDesc(void *data, const ucp_am_recv_param_t *param) :
-        _data(data), _param(param) {
-    }
-
-    void                         *_data;
-    const ucp_am_recv_param_t    *_param;
+    const char                   *what;
 };
 
 UcxCallback::~UcxCallback()
@@ -55,7 +51,8 @@ EmptyCallback* EmptyCallback::get() {
 
 bool UcxLog::use_human_time = false;
 
-UcxLog::UcxLog(const char* prefix, bool enable)
+UcxLog::UcxLog(const char* prefix, bool enable, std::ostream *os, bool abort) :
+        _os(os), _abort(abort)
 {
     if (!enable) {
         _ss = NULL;
@@ -80,9 +77,12 @@ UcxLog::UcxLog(const char* prefix, bool enable)
 UcxLog::~UcxLog()
 {
     if (_ss != NULL) {
-        (*_ss) << std::endl;
-        std::cout << (*_ss).str();
+        (*_os) << (*_ss).str() << std::endl;
         delete _ss;
+
+        if (_abort) {
+            abort();
+        }
     }
 }
 
@@ -98,37 +98,39 @@ void UcxContext::UcxAcceptCallback::operator()(ucs_status_t status)
 {
     if (status == UCS_OK) {
         _context.dispatch_connection_accepted(&_connection);
+    } else if ((status != UCS_ERR_CANCELED) ||
+               !_connection.is_disconnecting()) {
+        _connection.disconnect(new UcxDisconnectCallback());
     }
 
     delete this;
 }
 
-UcxContext::UcxDisconnectCallback::UcxDisconnectCallback(UcxConnection &conn)
-    : _conn(&conn)
-{
-}
-
-UcxContext::UcxDisconnectCallback::~UcxDisconnectCallback()
-{
-    delete _conn;
-}
-
 void UcxContext::UcxDisconnectCallback::operator()(ucs_status_t status)
 {
     delete this;
 }
 
-UcxContext::UcxContext(size_t iomsg_size, double connect_timeout, bool use_am) :
+UcxContext::UcxContext(size_t iomsg_size, double connect_timeout, bool use_am,
+                       bool use_epoll) :
     _context(NULL), _worker(NULL), _listener(NULL), _iomsg_recv_request(NULL),
     _iomsg_buffer(iomsg_size, '\0'), _connect_timeout(connect_timeout),
-    _use_am(use_am)
+    _use_am(use_am), _worker_fd(-1), _epoll_fd(-1)
 {
+    if (use_epoll) {
+        _epoll_fd = epoll_create(1);
+        assert(_epoll_fd >= 0);
+    }
 }
 
 UcxContext::~UcxContext()
 {
-    destroy_connections();
-    destroy_listener();
+    assert(_conns.empty());
+    assert(_conns_in_progress.empty());
+    assert(_conn_requests.empty());
+    assert(_disconnecting_conns.empty());
+    assert(_failed_conns.empty());
+
     destroy_worker();
     if (_context) {
         ucp_cleanup(_context);
@@ -149,6 +151,9 @@ bool UcxContext::init()
                               UCP_PARAM_FIELD_REQUEST_SIZE;
     ucp_params.features     = _use_am ? UCP_FEATURE_AM :
                                         UCP_FEATURE_TAG | UCP_FEATURE_STREAM;
+    if (_epoll_fd != -1) {
+        ucp_params.features |= UCP_FEATURE_WAKEUP;
+    }
     ucp_params.request_init = request_init;
     ucp_params.request_size = sizeof(ucx_request);
     ucs_status_t status = ucp_init(&ucp_params, NULL, &_context);
@@ -172,6 +177,14 @@ bool UcxContext::init()
         return false;
     }
 
+    status = epoll_init();
+    if (status != UCS_OK) {
+        destroy_worker();
+        ucp_cleanup(_context);
+        _context = NULL;
+        return false;
+    }
+
     UCX_LOG << "created worker " << _worker;
 
     if (_use_am) {
@@ -206,10 +219,14 @@ bool UcxContext::listen(const struct sockaddr* saddr, size_t addrlen)
     return true;
 }
 
-void UcxContext::progress()
+void UcxContext::progress(unsigned count)
 {
-    ucp_worker_progress(_worker);
-    progress_io_message();
+    int i = 0;
+
+    while ((i++ < count) && progress_worker_event()) {
+        progress_io_message();
+    }
+
     progress_timed_out_conns();
     progress_conn_requests();
     progress_failed_connections();
@@ -329,12 +346,67 @@ int UcxContext::is_timeout_elapsed(struct timeval const *tv_prior, double timeou
     return ((elapsed.tv_sec + (elapsed.tv_usec * 1e-6)) > timeout);
 }
 
+ucs_status_t UcxContext::epoll_init()
+{
+    ucs_status_t status;
+    epoll_event  ev;
+
+    if (_epoll_fd == -1) {
+        return UCS_OK;
+    }
+
+    status = ucp_worker_get_efd(_worker, &_worker_fd);
+    if (status != UCS_OK) {
+        UCX_LOG << "failed to get ucp_worker fd to be epoll monitored";
+        return status;
+    }
+
+    status = ucp_worker_arm(_worker);
+    if (status == UCS_ERR_BUSY) {
+        UCX_LOG << "some events are arrived already";
+    } else if (status != UCS_OK) {
+        UCX_LOG << "ucp_epoll error: " << ucs_status_string(status);
+        return status;
+    }
+
+    memset(&ev, 0, sizeof(ev));
+    ev.events  = EPOLLIN;
+    ev.data.fd = _worker_fd;
+
+    if (epoll_ctl(_epoll_fd, EPOLL_CTL_ADD, _worker_fd, &ev) == -1) {
+        UCX_LOG << "epoll add worker fd: " << _worker_fd << " failed.";
+        return UCS_ERR_UNSUPPORTED;
+    }
+
+    return UCS_OK;
+}
+
+bool UcxContext::progress_worker_event()
+{
+    int         ret;
+    epoll_event ev;
+
+    if (ucp_worker_progress(_worker)) {
+        return true;
+    }
+
+    if ((_epoll_fd == -1) || (ucp_worker_arm(_worker) == UCS_ERR_BUSY)) {
+        return false;
+    }
+
+    do {
+         ret = epoll_wait(_epoll_fd, &ev, 1, -1);
+    } while ((ret == -1) && (errno == EINTR || errno == EAGAIN));
+
+    assert(ev.data.fd == _worker_fd);
+    return false;
+}
+
 void UcxContext::progress_timed_out_conns()
 {
     while (!_conns_in_progress.empty() &&
            (get_time() > _conns_in_progress.begin()->first)) {
         UcxConnection *conn = _conns_in_progress.begin()->second;
-        _conns_in_progress.erase(_conns_in_progress.begin());
         conn->handle_connection_error(UCS_ERR_TIMED_OUT);
     }
 }
@@ -379,8 +451,10 @@ void UcxContext::progress_io_message()
             return;
         }
 
-        dispatch_io_message(conn, &_iomsg_buffer[0],
-                            _iomsg_recv_request->recv_length);
+        if (conn->ucx_status() == UCS_OK) {
+            dispatch_io_message(conn, &_iomsg_buffer[0],
+                                _iomsg_recv_request->recv_length);
+        }
     }
     request_release(_iomsg_recv_request);
     recv_io_message();
@@ -402,6 +476,7 @@ void UcxContext::progress_disconnected_connections()
         UcxConnection *conn = *it;
         if (conn->disconnect_progress()) {
             it = _disconnecting_conns.erase(it);
+            delete conn;
         } else {
             ++it;
         }
@@ -473,24 +548,36 @@ void UcxContext::remove_connection(UcxConnection *conn)
     }
 }
 
-void UcxContext::remove_connection_inprogress(UcxConnection *conn)
+UcxContext::timeout_conn_t::iterator
+UcxContext::find_connection_inprogress(UcxConnection *conn)
 {
     // we expect to remove connections from the list close to the same order
     // as created, so this linear search should be pretty fast
-    timeout_conn_t::iterator i;
+    UcxContext::timeout_conn_t::iterator i;
+
     for (i = _conns_in_progress.begin(); i != _conns_in_progress.end(); ++i) {
         if (i->second == conn) {
-            _conns_in_progress.erase(i);
-            return;
+            return i;
         }
     }
+
+    return _conns_in_progress.end();
+}
+
+void UcxContext::remove_connection_inprogress(UcxConnection *conn)
+{
+    timeout_conn_t::iterator i = find_connection_inprogress(conn);
+
+    if (i != _conns_in_progress.end()) {
+        _conns_in_progress.erase(i);
+    }
 }
 
 void UcxContext::move_connection_to_disconnecting(UcxConnection *conn)
 {
-    remove_connection(conn);
-    assert(std::find(_disconnecting_conns.begin(), _disconnecting_conns.end(),
-                     conn) == _disconnecting_conns.end());
+    assert(_conns.find(conn->id()) == _conns.end());
+    assert(find_connection_inprogress(conn) == _conns_in_progress.end());
+    assert(!is_in_disconnecting_list(conn));
     _disconnecting_conns.push_back(conn);
 }
 
@@ -505,34 +592,43 @@ void UcxContext::handle_connection_error(UcxConnection *conn)
     _failed_conns.push_back(conn);
 }
 
-void UcxContext::destroy_connections()
+void UcxContext::wait_disconnected_connections()
 {
-    while (!_conn_requests.empty()) {
-        ucp_conn_request_h conn_req = _conn_requests.front().conn_request;
-        UCX_LOG << "reject connection request " << conn_req;
-        ucp_listener_reject(_listener, conn_req);
-        _conn_requests.pop_front();
-    }
-
-    while (!_conns_in_progress.empty()) {
-        UcxConnection &conn = *_conns_in_progress.begin()->second;
-        _conns_in_progress.erase(_conns_in_progress.begin());
-        conn.disconnect(new UcxDisconnectCallback(conn));
-    }
-
-    UCX_LOG << "destroy_connections";
-    while (!_conns.empty()) {
-        UcxConnection &conn = *_conns.begin()->second;
-        _conns.erase(_conns.begin());
-        conn.disconnect(new UcxDisconnectCallback(conn));
-    }
-
     while (!_disconnecting_conns.empty()) {
         ucp_worker_progress(_worker);
         progress_disconnected_connections();
     }
 }
 
+void UcxContext::destroy_connections()
+{
+    // wait for all failed connections being disconnected, call progress to
+    // discover new failed connections if any
+    while (!_failed_conns.empty()) {
+        progress_failed_connections();
+        ucp_worker_progress(_worker);
+    }
+
+    // remove all connections which are in-progress, this is optimization to
+    // avoid checking existence of a connection in _conns_in_progress vector
+    // for each connection passed to disconnect() method
+    while (!_conns_in_progress.empty()) {
+        UcxConnection &conn = *_conns_in_progress.begin()->second;
+        _conns_in_progress.erase(_conns_in_progress.begin());
+        conn.disconnect(new UcxDisconnectCallback());
+    }
+
+    // disconnect all connections which are not in-progress
+    UCX_LOG << "destroy connections";
+    while (!_conns.empty()) {
+        UcxConnection &conn = *_conns.begin()->second;
+        conn.disconnect(new UcxDisconnectCallback());
+    }
+
+    // wait for all connections being completely disconnected
+    wait_disconnected_connections();
+}
+
 double UcxContext::get_time() {
     struct timeval tv;
     gettimeofday(&tv, NULL);
@@ -541,9 +637,19 @@ double UcxContext::get_time() {
 
 void UcxContext::destroy_listener()
 {
-    if (_listener) {
-        ucp_listener_destroy(_listener);
+    if (_listener == NULL) {
+        return;
     }
+
+    // reject all connection requests saved _conn_requests deque
+    while (!_conn_requests.empty()) {
+        ucp_conn_request_h conn_req = _conn_requests.front().conn_request;
+        UCX_LOG << "reject connection request " << conn_req;
+        ucp_listener_reject(_listener, conn_req);
+        _conn_requests.pop_front();
+    }
+
+    ucp_listener_destroy(_listener);
 }
 
 void UcxContext::destroy_worker()
@@ -558,6 +664,11 @@ void UcxContext::destroy_worker()
     }
 
     ucp_worker_destroy(_worker);
+    if (_epoll_fd >= 0) {
+        close(_epoll_fd);
+        _epoll_fd = -1;
+    }
+    _worker = NULL;
 }
 
 ucs_status_t UcxContext::am_recv_callback(void *arg, const void *header,
@@ -581,10 +692,16 @@ ucs_status_t UcxContext::am_recv_callback(void *arg, const void *header,
 
     UcxConnection *conn = iter->second;
 
-    UcxAmDesc data_desc(data, param);
-
-    self->dispatch_am_message(conn, header, header_length, data_desc);
+    if (conn->ucx_status() == UCS_OK) {
+        UcxAmDesc data_desc(data, param);
+        self->dispatch_am_message(conn, header, header_length, data_desc);
+    }
 
+    /* In the current example, dispatch_am_message is always processing
+     * the received data internally and never needs it later.
+     * If data is going to be used outside this callback, UCS_INPROGRESS
+     * should be returned. Call ucp_am_data_release() when data is not needed.
+     */
     return UCS_OK;
 }
 
@@ -601,6 +718,32 @@ void UcxContext::set_am_handler(ucp_am_recv_callback_t cb, void *arg)
     ucp_worker_set_am_recv_handler(_worker, &param);
 }
 
+void *UcxContext::malloc(size_t size, const char *name)
+{
+    void *ptr;
+
+    ptr = ::malloc(size);
+    ucs_memtrack_allocated(ptr, size, name);
+
+    return ptr;
+}
+
+void *UcxContext::memalign(size_t alignment, size_t size, const char *name)
+{
+    void *ptr;
+
+    ptr = ::memalign(alignment, size);
+    ucs_memtrack_allocated(ptr, size, name);
+
+    return ptr;
+}
+
+void UcxContext::free(void *ptr)
+{
+    ucs_memtrack_releasing(ptr);
+    ::free(ptr);
+}
+
 #define UCX_CONN_LOG UcxLog(_log_prefix, true)
 
 unsigned UcxConnection::_num_instances = 0;
@@ -678,28 +821,39 @@ void UcxConnection::accept(ucp_conn_request_h conn_req, UcxCallback *callback)
 
 void UcxConnection::disconnect(UcxCallback *callback)
 {
-    /* establish cb must be destroyed earlier since it accesses
-     * the connection */
-    assert(_establish_cb == NULL);
+    UCX_CONN_LOG << "disconnect, ep is " << _ep;
+
     assert(_disconnect_cb == NULL);
-    assert(_ep != NULL);
-
-    UCX_CONN_LOG << "destroying, ep is " << _ep;
-    ep_close(UCP_EP_CLOSE_MODE_FORCE);
-
     _disconnect_cb = callback;
-    if (ucs_list_is_empty(&_all_requests)) {
-        _context.move_connection_to_disconnecting(this);
-    } else {
-        cancel_all();
+
+    _context.remove_connection(this);
+
+    if (!is_established()) {
+        assert(_ucx_status == UCS_INPROGRESS);
+        established(UCS_ERR_CANCELED);
+    } else if (_ucx_status == UCS_OK) {
+        _ucx_status = UCS_ERR_NOT_CONNECTED;
     }
+
+    assert(UCS_STATUS_IS_ERR(_ucx_status));
+
+    _context.move_connection_to_disconnecting(this);
+
+    cancel_all();
+
+    // close the EP after cancelling all outstanding operations to purge all
+    // requests scheduled on the EP which could wait for the acknowledgments
+    ep_close(UCP_EP_CLOSE_MODE_FORCE);
 }
 
 bool UcxConnection::disconnect_progress()
 {
-    assert(_ep == NULL);
     assert(_disconnect_cb != NULL);
 
+    if (!ucs_list_is_empty(&_all_requests)) {
+        return false;
+    }
+
     if (UCS_PTR_IS_PTR(_close_request)) {
         if (ucp_request_check_status(_close_request) == UCS_INPROGRESS) {
             return false;
@@ -709,11 +863,9 @@ bool UcxConnection::disconnect_progress()
         }
     }
 
-    assert(ucs_list_is_empty(&_all_requests));
-    UcxCallback *cb = _disconnect_cb;
-    _disconnect_cb  = NULL;
-    // invoke last since it can delete this object
-    (*cb)(UCS_OK);
+    UCX_CONN_LOG << "disconnection completed";
+
+    invoke_callback(_disconnect_cb, UCS_OK);
     return true;
 }
 
@@ -735,6 +887,7 @@ bool UcxConnection::recv_data(void *buffer, size_t length, uint32_t sn,
                               UcxCallback* callback)
 {
     if (_ep == NULL) {
+        (*callback)(UCS_ERR_CANCELED);
         return false;
     }
 
@@ -752,6 +905,7 @@ bool UcxConnection::send_am(const void *meta, size_t meta_length,
                             UcxCallback* callback)
 {
     if (_ep == NULL) {
+        (*callback)(UCS_ERR_CANCELED);
         return false;
     }
 
@@ -774,7 +928,6 @@ bool UcxConnection::recv_am_data(void *buffer, size_t length,
     assert(_ep != NULL);
 
     if (!(data_desc._param->recv_attr & UCP_AM_RECV_ATTR_FLAG_RNDV)) {
-        memcpy(buffer, data_desc._data, length);
         (*callback)(UCS_OK);
         return true;
     }
@@ -799,7 +952,8 @@ void UcxConnection::cancel_all()
     unsigned     count = 0;
     ucs_list_for_each_safe(request, tmp, &_all_requests, pos) {
         ++count;
-        UCX_CONN_LOG << "canceling " << request << " request #" << count;
+        UCX_CONN_LOG << "canceling " << request->what << " request " << request
+                     << " #" << count;
         ucp_request_cancel(_context.worker(), request);
     }
 }
@@ -814,19 +968,23 @@ ucp_tag_t UcxConnection::make_iomsg_tag(uint32_t conn_id, uint32_t sn)
     return UcxContext::IOMSG_TAG | make_data_tag(conn_id, sn);
 }
 
-void UcxConnection::stream_send_callback(void *request, ucs_status_t status)
-{
-}
-
 void UcxConnection::stream_recv_callback(void *request, ucs_status_t status,
                                          size_t recv_len)
 {
     ucx_request *r      = reinterpret_cast<ucx_request*>(request);
     UcxConnection *conn = r->conn;
 
+    assert(!r->completed);
+    r->status = status;
+
     if (!conn->is_established()) {
         assert(conn->_establish_cb == r->callback);
-        conn->established(status);
+
+        if (status == UCS_OK) {
+            conn->established(status);
+        } else {
+            conn->handle_connection_error(status);
+        }
     } else {
         assert(UCS_STATUS_IS_ERR(conn->ucx_status()));
     }
@@ -881,8 +1039,9 @@ void UcxConnection::set_log_prefix(const struct sockaddr* saddr,
                                    socklen_t addrlen)
 {
     std::stringstream ss;
-    ss << "[UCX-connection #" << _conn_id << " " <<
-          UcxContext::sockaddr_str(saddr, addrlen) << "]";
+    _remote_address = UcxContext::sockaddr_str(saddr, addrlen);
+    ss << "[UCX-connection " << this << ": #" << _conn_id << " "
+       << _remote_address << "]";
     memset(_log_prefix, 0, MAX_LOG_PREFIX_SIZE);
     int length = ss.str().length();
     if (length >= MAX_LOG_PREFIX_SIZE) {
@@ -905,7 +1064,10 @@ void UcxConnection::connect_tag(UcxCallback *callback)
         _context._conns_in_progress.push_back(std::make_pair(
                 UcxContext::get_time() + _context._connect_timeout, this));
     } else {
-        established(UCS_PTR_STATUS(rreq));
+        ucs_status_t status = UCS_PTR_STATUS(rreq);
+
+        assert(status != UCS_ERR_CANCELED);
+        established(status);
         if (rreq != NULL) {
             // failed to receive
             return;
@@ -914,12 +1076,10 @@ void UcxConnection::connect_tag(UcxCallback *callback)
 
     // send local connection id
     void *sreq = ucp_stream_send_nb(_ep, &_conn_id, 1, dt_int,
-                                    stream_send_callback, 0);
+                                    common_request_callback, 0);
     // we do not have to check the status here, in case if the endpoint is
     // failed we should handle it in ep_params.err_handler.cb set above
-    if (UCS_PTR_IS_PTR(sreq)) {
-        ucp_request_free(sreq);
-    }
+    process_request("conn_id send", sreq, EmptyCallback::get());
 }
 
 void UcxConnection::connect_am(UcxCallback *callback)
@@ -964,25 +1124,50 @@ void UcxConnection::connect_common(ucp_ep_params_t &ep_params,
 
 void UcxConnection::established(ucs_status_t status)
 {
-    if (!_use_am && (status == UCS_OK)) {
-        assert(_remote_conn_id != 0);
-        UCX_CONN_LOG << "Remote id is " << _remote_conn_id;
-    }
+    ucp_ep_attr_t ep_attr;
+    std::string local_address;
+    std::string remote_address;
 
     _ucx_status = status;
-    _context.remove_connection_inprogress(this);
 
-    (*_establish_cb)(status);
-    _establish_cb = NULL;
+    if (!_use_am && (status == UCS_OK)) {
+        assert(_remote_conn_id != 0);
+
+        ep_attr.field_mask = UCP_EP_ATTR_FIELD_LOCAL_SOCKADDR |
+                             UCP_EP_ATTR_FIELD_REMOTE_SOCKADDR;
+        status = ucp_ep_query(_ep, &ep_attr);
+        if (status != UCS_OK) {
+            UCX_CONN_LOG << "Remote id is " << _remote_conn_id;
+            UCX_CONN_LOG << "ucp_ep_query() failed: "
+                         << ucs_status_string(status);
+        } else {
+            local_address  = UcxContext::sockaddr_str(
+                                 (const struct sockaddr*)&ep_attr.local_sockaddr,
+                                 sizeof(ep_attr.local_sockaddr));
+            remote_address = UcxContext::sockaddr_str(
+                                 (const struct sockaddr*)&ep_attr.remote_sockaddr,
+                                 sizeof(ep_attr.remote_sockaddr));
+            UCX_CONN_LOG << "Remote id is "    << _remote_conn_id
+                         << ", endpoint "      << _ep
+                         << ", local address " << local_address
+                         << " remote address " << remote_address;
+        }
+    }
+
+    _context.remove_connection_inprogress(this);
+    invoke_callback(_establish_cb, status);
 }
 
 bool UcxConnection::send_common(const void *buffer, size_t length, ucp_tag_t tag,
                                 UcxCallback* callback)
 {
     if (_ep == NULL) {
+        (*callback)(UCS_ERR_CANCELED);
         return false;
     }
 
+    assert(_ucx_status == UCS_OK);
+
     ucs_status_ptr_t ptr_status = ucp_tag_send_nb(_ep, buffer, length,
                                                   ucp_dt_make_contig(1), tag,
                                                   common_request_callback);
@@ -1000,19 +1185,16 @@ void UcxConnection::request_completed(ucx_request *r)
     ucs_list_del(&r->pos);
 
     if (_disconnect_cb != NULL) {
-        UCX_CONN_LOG << "completing request " << r << " with status \""
-                     << ucs_status_string(r->status) << "\" (" << r->status
-                     << ")" << " during disconnect";
-
-        if (ucs_list_is_empty(&_all_requests)) {
-            _context.move_connection_to_disconnecting(this);
-        }
+        UCX_CONN_LOG << "completing " << r->what << " request " << r
+                     << " with status \"" << ucs_status_string(r->status)
+                     << "\" (" << r->status << ")" << " during disconnect";
+        assert(_context.is_in_disconnecting_list(this));
     }
 }
 
 void UcxConnection::handle_connection_error(ucs_status_t status)
 {
-    if (UCS_STATUS_IS_ERR(_ucx_status)) {
+    if (UCS_STATUS_IS_ERR(_ucx_status) || is_disconnecting()) {
         return;
     }
 
@@ -1023,8 +1205,8 @@ void UcxConnection::handle_connection_error(ucs_status_t status)
     if (is_established()) {
         _context.handle_connection_error(this);
     } else {
-        (*_establish_cb)(status);
-        _establish_cb = NULL;
+        _context.remove_connection_inprogress(this);
+        invoke_callback(_establish_cb, status);
     }
 }
 
@@ -1036,7 +1218,7 @@ void UcxConnection::ep_close(enum ucp_ep_close_mode mode)
         return;
     }
 
-    assert(!_close_request);
+    assert(_close_request == NULL);
 
     UCX_CONN_LOG << "closing ep " << _ep << " mode " << mode_str[mode];
     _close_request = ucp_ep_close_nb(_ep, mode);
@@ -1072,8 +1254,16 @@ bool UcxConnection::process_request(const char *what,
             // will be completed by callback
             r->callback = callback;
             r->conn     = this;
+            r->what     = what;
             request_started(r);
             return true;
         }
     }
 }
+
+void UcxConnection::invoke_callback(UcxCallback *&callback, ucs_status_t status)
+{
+    UcxCallback *cb = callback;
+    callback        = NULL;
+    (*cb)(status);
+}
