@@ -36,7 +36,7 @@ public:
 
     virtual void init() {
         ucs::skip_on_address_sanitizer();
-        if (get_variant_value() == VARIANT_PROTO_ENABLE) {
+        if (enable_proto()) {
             modify_config("PROTO_ENABLE", "y");
         }
         ucp_test::init();
@@ -75,6 +75,7 @@ protected:
     void test_length0(unsigned flags);
     void test_rkey_management(ucp_mem_h memh, bool is_dummy,
                               bool expect_rma_offload);
+    bool enable_proto() const;
 
 private:
     void expect_same_distance(const ucs_sys_dev_distance_t &dist1,
@@ -189,36 +190,47 @@ void test_ucp_mmap::test_rkey_management(ucp_mem_h memh, bool is_dummy,
     /* rkey->md_map is a subset of all possible keys */
     EXPECT_TRUE(ucs_test_all_flags(memh->md_map, rkey->md_map));
 
-    bool have_rma              = resolve_rma(&receiver(), rkey);
-    bool have_amo              = resolve_amo(&receiver(), rkey);
-    bool have_rma_bw_get_zcopy = resolve_rma_bw_get_zcopy(&receiver(), rkey);
-    bool have_rma_bw_put_zcopy = resolve_rma_bw_put_zcopy(&receiver(), rkey);
+    /* Test remote key protocols selection */
+    if (enable_proto()) {
+        test_rkey_proto(memh);
+    } else {
+        bool have_rma              = resolve_rma(&receiver(), rkey);
+        bool have_amo              = resolve_amo(&receiver(), rkey);
+        bool have_rma_bw_get_zcopy = resolve_rma_bw_get_zcopy(&receiver(),
+                                                              rkey);
+        bool have_rma_bw_put_zcopy = resolve_rma_bw_put_zcopy(&receiver(),
+                                                              rkey);
 
-    /* Test that lane resolution on the remote key returns consistent results */
-    for (int i = 0; i < 10; ++i) {
-        switch (ucs::rand() % 4) {
-        case 0:
-            EXPECT_EQ(have_rma, resolve_rma(&receiver(), rkey));
-            break;
-        case 1:
-            EXPECT_EQ(have_amo, resolve_amo(&receiver(), rkey));
-            break;
-        case 2:
-            EXPECT_EQ(have_rma_bw_get_zcopy,
-                      resolve_rma_bw_get_zcopy(&receiver(), rkey));
-            break;
-        case 3:
-            EXPECT_EQ(have_rma_bw_put_zcopy,
-                      resolve_rma_bw_put_zcopy(&receiver(), rkey));
-            break;
+        /* Test that lane resolution on the remote key returns consistent results */
+        for (int i = 0; i < 10; ++i) {
+            switch (ucs::rand() % 4) {
+            case 0:
+                EXPECT_EQ(have_rma, resolve_rma(&receiver(), rkey));
+                break;
+            case 1:
+                EXPECT_EQ(have_amo, resolve_amo(&receiver(), rkey));
+                break;
+            case 2:
+                EXPECT_EQ(have_rma_bw_get_zcopy,
+                          resolve_rma_bw_get_zcopy(&receiver(), rkey));
+                break;
+            case 3:
+                EXPECT_EQ(have_rma_bw_put_zcopy,
+                          resolve_rma_bw_put_zcopy(&receiver(), rkey));
+                break;
+            }
         }
-    }
 
-    if (expect_rma_offload) {
-        if (is_dummy) {
-            EXPECT_EQ(&ucp_rma_sw_proto, rkey->cache.rma_proto);
-        } else {
-            EXPECT_EQ(&ucp_rma_basic_proto, rkey->cache.rma_proto);
+        if (expect_rma_offload) {
+            if (is_dummy) {
+                EXPECT_EQ(&ucp_rma_sw_proto,
+                          UCP_RKEY_RMA_PROTO(rkey->cache.rma_proto_index));
+            } else {
+                ucs_assert(&ucp_rma_basic_proto ==
+                           UCP_RKEY_RMA_PROTO(rkey->cache.rma_proto_index));
+                EXPECT_EQ(&ucp_rma_basic_proto,
+                          UCP_RKEY_RMA_PROTO(rkey->cache.rma_proto_index));
+            }
         }
     }
 
@@ -233,9 +245,11 @@ void test_ucp_mmap::test_rkey_management(ucp_mem_h memh, bool is_dummy,
 
     ucp_rkey_destroy(rkey);
     ucp_rkey_buffer_release(rkey_buffer);
+}
 
-    /* Test remote key protocols selection */
-    test_rkey_proto(memh);
+bool test_ucp_mmap::enable_proto() const
+{
+    return get_variant_value() == VARIANT_PROTO_ENABLE;
 }
 
 void test_ucp_mmap::expect_same_distance(const ucs_sys_dev_distance_t &dist1,
@@ -258,9 +272,14 @@ void test_ucp_mmap::test_rkey_proto(ucp_mem_h memh)
     EXPECT_EQ(memh->mem_type, mem_info.type);
 
     /* Collect distances from all devices in the system */
-    uint64_t sys_dev_map = UCS_MASK(ucs_topo_num_devices());
+    ucp_sys_dev_map_t sys_dev_map = UCS_MASK(ucs_topo_num_devices());
     std::vector<ucs_sys_dev_distance_t> sys_distance(ucs_topo_num_devices());
     for (unsigned i = 0; i < sys_distance.size(); ++i) {
+        if (std::string(ucs_topo_sys_device_get_name(i)).find("test") == 0) {
+            /* Dummy device created by test */
+            continue;
+        }
+
         status = ucs_topo_get_distance(mem_info.sys_dev, i, &sys_distance[i]);
         ASSERT_UCS_OK(status);
     }
@@ -283,7 +302,7 @@ void test_ucp_mmap::test_rkey_proto(ucp_mem_h memh)
     ASSERT_UCS_OK(status);
 
     /* Check rkey configuration */
-    if (receiver().ucph()->config.ext.proto_enable) {
+    if (enable_proto()) {
         ucp_rkey_config_t *rkey_config = ucp_rkey_config(receiver().worker(),
                                                          rkey);
         ucp_ep_config_t *ep_config     = ucp_ep_config(receiver().ep());
@@ -306,42 +325,14 @@ void test_ucp_mmap::test_rkey_proto(ucp_mem_h memh)
     ucp_rkey_destroy(rkey);
 }
 
-UCS_TEST_P(test_ucp_mmap, alloc) {
-    ucs_status_t status;
-    bool is_dummy;
-
-    for (int i = 0; i < 1000 / ucs::test_time_multiplier(); ++i) {
-        size_t size = ucs::rand() % (UCS_MBYTE);
-
-        ucp_mem_h memh;
-        ucp_mem_map_params_t params;
-
-        params.field_mask = UCP_MEM_MAP_PARAM_FIELD_ADDRESS |
-                            UCP_MEM_MAP_PARAM_FIELD_LENGTH |
-                            UCP_MEM_MAP_PARAM_FIELD_FLAGS;
-        params.address    = NULL;
-        params.length     = size;
-        params.flags      = mem_map_flags() | UCP_MEM_MAP_ALLOCATE;
-
-        status = ucp_mem_map(sender().ucph(), &params, &memh);
-        ASSERT_UCS_OK(status);
-
-        is_dummy = (size == 0);
-        test_rkey_management(memh, is_dummy, is_tl_rdma() || is_tl_shm());
-
-        status = ucp_mem_unmap(sender().ucph(), memh);
-        ASSERT_UCS_OK(status);
-    }
-}
-
 UCS_TEST_P(test_ucp_mmap, alloc_mem_type) {
-    std::vector<ucs_memory_type_t> mem_types = mem_buffer::supported_mem_types();
+    const std::vector<ucs_memory_type_t> &mem_types =
+            mem_buffer::supported_mem_types();
     ucs_status_t status;
     bool is_dummy;
     bool expect_rma_offload;
 
-    for (std::vector<ucs_memory_type_t>::iterator mem_type = mem_types.begin();
-         mem_type != mem_types.end(); ++mem_type) {
+    for (auto mem_type : mem_types) {
         for (int i = 0; i < (100 / ucs::test_time_multiplier()); ++i) {
             size_t size = ucs::rand() % (UCS_MBYTE);
 
@@ -352,7 +343,7 @@ UCS_TEST_P(test_ucp_mmap, alloc_mem_type) {
                                  UCP_MEM_MAP_PARAM_FIELD_FLAGS   |
                                  UCP_MEM_MAP_PARAM_FIELD_MEMORY_TYPE;
             params.address     = NULL;
-            params.memory_type = *mem_type;
+            params.memory_type = mem_type;
             params.length      = size;
             params.flags       = UCP_MEM_MAP_ALLOCATE;
 
@@ -361,8 +352,9 @@ UCS_TEST_P(test_ucp_mmap, alloc_mem_type) {
             ASSERT_UCS_OK(status);
 
             is_dummy           = (size == 0);
-            expect_rma_offload = ((*mem_type != UCS_MEMORY_TYPE_CUDA_MANAGED) &&
-                                  (is_tl_rdma() || is_tl_shm()));
+            expect_rma_offload = !UCP_MEM_IS_CUDA_MANAGED(mem_type) &&
+                                 (is_tl_rdma() || is_tl_shm()) &&
+                                 check_reg_mem_types(sender(), mem_type);
             test_rkey_management(memh, is_dummy, expect_rma_offload);
 
             status = ucp_mem_unmap(sender().ucph(), memh);
@@ -371,48 +363,16 @@ UCS_TEST_P(test_ucp_mmap, alloc_mem_type) {
     }
 }
 
-UCS_TEST_P(test_ucp_mmap, reg) {
-    ucs_status_t status;
-    bool is_dummy;
-
-    for (int i = 0; i < 1000 / ucs::test_time_multiplier(); ++i) {
-        size_t size = ucs::rand() % (UCS_MBYTE);
-
-        void *ptr = malloc(size);
-        ucs::fill_random(ptr, size);
-
-        ucp_mem_h memh;
-        ucp_mem_map_params_t params;
-
-        params.field_mask = UCP_MEM_MAP_PARAM_FIELD_ADDRESS |
-                            UCP_MEM_MAP_PARAM_FIELD_LENGTH |
-                            UCP_MEM_MAP_PARAM_FIELD_FLAGS;
-        params.address    = ptr;
-        params.length     = size;
-        params.flags      = mem_map_flags();
-
-        status = ucp_mem_map(sender().ucph(), &params, &memh);
-        ASSERT_UCS_OK(status);
-
-        is_dummy = (size == 0);
-        test_rkey_management(memh, is_dummy, is_tl_rdma());
-
-        status = ucp_mem_unmap(sender().ucph(), memh);
-        ASSERT_UCS_OK(status);
-
-        free(ptr);
-    }
-}
-
 UCS_TEST_P(test_ucp_mmap, reg_mem_type) {
-    std::vector<ucs_memory_type_t> mem_types = mem_buffer::supported_mem_types();
+    const std::vector<ucs_memory_type_t> &mem_types =
+            mem_buffer::supported_mem_types();
     ucs_status_t status;
     bool is_dummy;
+    bool expect_rma_offload;
     ucs_memory_type_t alloc_mem_type;
 
     for (int i = 0; i < 1000 / ucs::test_time_multiplier(); ++i) {
-        size_t size = ucs::rand() % (UCS_MBYTE);
-
+        size_t size    = ucs::rand() % UCS_MBYTE;
         alloc_mem_type = mem_types.at(ucs::rand() % mem_types.size());
         mem_buffer buf(size, alloc_mem_type);
         mem_buffer::pattern_fill(buf.ptr(), size, 0, alloc_mem_type);
@@ -436,10 +396,12 @@ UCS_TEST_P(test_ucp_mmap, reg_mem_type) {
         if (!is_dummy) {
             EXPECT_EQ(alloc_mem_type, memh->mem_type);
         }
-        test_rkey_management(memh, is_dummy,
+
+        expect_rma_offload = !UCP_MEM_IS_CUDA_MANAGED(alloc_mem_type) &&
+                             !UCP_MEM_IS_ROCM_MANAGED(alloc_mem_type) &&
                              is_tl_rdma() &&
-                                     !UCP_MEM_IS_CUDA_MANAGED(alloc_mem_type) &&
-                                     !UCP_MEM_IS_ROCM_MANAGED(alloc_mem_type));
+                             check_reg_mem_types(sender(), alloc_mem_type);
+        test_rkey_management(memh, is_dummy, expect_rma_offload);
 
         status = ucp_mem_unmap(sender().ucph(), memh);
         ASSERT_UCS_OK(status);
@@ -496,7 +458,8 @@ UCS_TEST_P(test_ucp_mmap, alloc_advise) {
     ucs_status_t status;
     bool is_dummy;
 
-    size_t size = 128 * UCS_MBYTE;
+    const size_t size = ucs_max(UCS_KBYTE,
+                                128 * UCS_MBYTE / ucs::test_time_multiplier());
 
     ucp_mem_h memh;
     ucp_mem_map_params_t params;
@@ -540,9 +503,9 @@ UCS_TEST_P(test_ucp_mmap, reg_advise) {
     ucs_status_t status;
     bool is_dummy;
 
-    size_t size = 128 * UCS_MBYTE;
-
-    void *ptr = malloc(size);
+    const size_t size = ucs_max(UCS_KBYTE,
+                                128 * UCS_MBYTE / ucs::test_time_multiplier());
+    void *ptr         = malloc(size);
     ucs::fill_random(ptr, size);
 
     ucp_mem_h               memh;
