@@ -4,7 +4,9 @@
 * See file LICENSE for terms.
 */
 
-#include "test_ucp_tag.h"
+#include <common/test.h>
+
+#include "ucp_test.h"
 #include "ucp_datatype.h"
 
 extern "C" {
@@ -17,17 +19,18 @@ class test_ucp_peer_failure : public ucp_test {
 public:
     test_ucp_peer_failure();
 
-    static std::vector<ucp_test_param>
-    enum_test_params(const ucp_params_t& ctx_params, const std::string& name,
-                     const std::string& test_case_name, const std::string& tls);
+    static void get_test_variants(std::vector<ucp_test_variant>& variants);
 
     ucp_ep_params_t get_ep_params();
 
 protected:
+    static const int AM_ID = 0;
+
     enum {
-        TEST_TAG = UCS_BIT(0),
+        TEST_AM  = UCS_BIT(0),
         TEST_RMA = UCS_BIT(1),
-        FAIL_IMM = UCS_BIT(2)
+        FAIL_IMM = UCS_BIT(2),
+        WAKEUP   = UCS_BIT(3)
     };
 
     enum {
@@ -37,14 +40,16 @@ protected:
 
     typedef ucs::handle<ucp_mem_h, ucp_context_h> mem_handle_t;
 
-    void set_timeouts();
+    void set_am_handler(entity &e);
+    static ucs_status_t
+    am_callback(void *arg, const void *header, size_t header_length, void *data,
+                size_t length, const ucp_am_recv_param_t *param);
     static void err_cb(void *arg, ucp_ep_h ep, ucs_status_t status);
     ucp_ep_h stable_sender();
     ucp_ep_h failing_sender();
     entity& stable_receiver();
     entity& failing_receiver();
     void *send_nb(ucp_ep_h ep, ucp_rkey_h rkey);
-    void *recv_nb(entity& e);
     static ucs_log_func_rc_t
     warn_unreleased_rdesc_handler(const char *file, unsigned line,
                                   const char *function,
@@ -57,56 +62,40 @@ protected:
     void get_rkey(ucp_ep_h ep, entity& dst, mem_handle_t& memh,
                   ucs::handle<ucp_rkey_h>& rkey);
     void set_rkeys();
-    static void send_cb(void *request, ucs_status_t status);
-    static void recv_cb(void *request, ucs_status_t status,
-                        ucp_tag_recv_info_t *info);
+    static void send_cb(void *request, ucs_status_t status, void *user_data);
 
     virtual void cleanup();
 
     void do_test(size_t msg_size, int pre_msg_count, bool force_close,
                  bool request_must_fail);
 
+    size_t                              m_am_rx_count;
     size_t                              m_err_count;
     ucs_status_t                        m_err_status;
     std::string                         m_sbuf, m_rbuf;
     mem_handle_t                        m_stable_memh, m_failing_memh;
     ucs::handle<ucp_rkey_h>             m_stable_rkey, m_failing_rkey;
-    ucs::ptr_vector<ucs::scoped_setenv> m_env;
 };
 
 UCP_INSTANTIATE_TEST_CASE(test_ucp_peer_failure)
 
 
-test_ucp_peer_failure::test_ucp_peer_failure() : m_err_count(0), m_err_status(UCS_OK) {
+test_ucp_peer_failure::test_ucp_peer_failure() :
+    m_am_rx_count(0), m_err_count(0), m_err_status(UCS_OK)
+{
     ucs::fill_random(m_sbuf);
-    set_timeouts();
+    set_tl_small_timeouts();
 }
 
-std::vector<ucp_test_param>
-test_ucp_peer_failure::enum_test_params(const ucp_params_t& ctx_params,
-                                        const std::string& name,
-                                        const std::string& test_case_name,
-                                        const std::string& tls)
+void test_ucp_peer_failure::get_test_variants(
+        std::vector<ucp_test_variant> &variants)
 {
-    std::vector<ucp_test_param> result;
-
-    ucp_params_t params = ucp_test::get_ctx_params();
-
-    params.field_mask  |= UCP_PARAM_FIELD_FEATURES;
-
-    params.features = UCP_FEATURE_TAG;
-    generate_test_params_variant(params, name, test_case_name + "/tag", tls,
-                                 TEST_TAG, result);
-    generate_test_params_variant(params, name, test_case_name + "/tag_fail_imm",
-                                 tls, TEST_TAG | FAIL_IMM, result);
-
-    params.features = UCP_FEATURE_RMA;
-    generate_test_params_variant(params, name, test_case_name + "/rma", tls,
-                                 TEST_RMA, result);
-    generate_test_params_variant(params, name, test_case_name + "/rma_fail_imm",
-                                 tls, TEST_RMA | FAIL_IMM, result);
-
-    return result;
+    add_variant_with_value(variants, UCP_FEATURE_AM, TEST_AM, "am");
+    add_variant_with_value(variants, UCP_FEATURE_RMA, TEST_RMA, "rma");
+    add_variant_with_value(variants, UCP_FEATURE_AM, TEST_AM | FAIL_IMM,
+                           "am_fail_imm");
+    add_variant_with_value(variants, UCP_FEATURE_RMA, TEST_RMA | FAIL_IMM,
+                           "rma_fail_imm");
 }
 
 ucp_ep_params_t test_ucp_peer_failure::get_ep_params() {
@@ -120,20 +109,45 @@ ucp_ep_params_t test_ucp_peer_failure::get_ep_params() {
     return params;
 }
 
-void test_ucp_peer_failure::set_timeouts() {
-    /* Set small TL timeouts to reduce testing time */
-    m_env.push_back(new ucs::scoped_setenv("UCX_RC_TIMEOUT",     "10ms"));
-    m_env.push_back(new ucs::scoped_setenv("UCX_RC_RNR_TIMEOUT", "10ms"));
-    m_env.push_back(new ucs::scoped_setenv("UCX_RC_RETRY_COUNT", "2"));
+void test_ucp_peer_failure::set_am_handler(entity &e)
+{
+    if (!(get_variant_value() & TEST_AM)) {
+        return;
+    }
+
+    ucp_am_handler_param_t param;
+    param.field_mask = UCP_AM_HANDLER_PARAM_FIELD_ID |
+                       UCP_AM_HANDLER_PARAM_FIELD_CB |
+                       UCP_AM_HANDLER_PARAM_FIELD_ARG;
+    param.cb         = am_callback;
+    param.arg        = this;
+    param.id         = AM_ID;
+
+    ucs_status_t status = ucp_worker_set_am_recv_handler(e.worker(), &param);
+    ASSERT_UCS_OK(status);
+}
+
+ucs_status_t
+test_ucp_peer_failure::am_callback(void *arg, const void *header,
+                                   size_t header_length, void *data,
+                                   size_t length,
+                                   const ucp_am_recv_param_t *param)
+{
+    test_ucp_peer_failure *self = reinterpret_cast<test_ucp_peer_failure*>(arg);
+    ++self->m_am_rx_count;
+    return UCS_OK;
 }
 
 void test_ucp_peer_failure::err_cb(void *arg, ucp_ep_h ep, ucs_status_t status) {
     test_ucp_peer_failure *self = reinterpret_cast<test_ucp_peer_failure*>(arg);
-    EXPECT_EQ(UCS_ERR_ENDPOINT_TIMEOUT, status);
+    EXPECT_TRUE((UCS_ERR_CONNECTION_RESET == status) ||
+                (UCS_ERR_ENDPOINT_TIMEOUT == status));
     self->m_err_status = status;
     ++self->m_err_count;
 }
 
+/* stable pair: sender = ep(0), receiver: entity(size - 1)
+ * failing pair: sender = ep(1), receiver: entity(size - 2)*/
 ucp_ep_h test_ucp_peer_failure::stable_sender() {
     return sender().ep(0, STABLE_EP_INDEX);
 }
@@ -143,32 +157,26 @@ ucp_ep_h test_ucp_peer_failure::failing_sender() {
 }
 
 ucp_test::entity& test_ucp_peer_failure::stable_receiver() {
-    return m_entities.at(m_entities.size() - 2);
+    return m_entities.at(m_entities.size() - 1 - STABLE_EP_INDEX);
 }
 
 ucp_test::entity& test_ucp_peer_failure::failing_receiver() {
-    return m_entities.at(m_entities.size() - 1);
+    return m_entities.at(m_entities.size() - 1 - FAILING_EP_INDEX);
 }
 
-void *test_ucp_peer_failure::send_nb(ucp_ep_h ep, ucp_rkey_h rkey) {
-    if (GetParam().variant & TEST_TAG) {
-        return ucp_tag_send_nb(ep, &m_sbuf[0], m_sbuf.size(), DATATYPE, 0,
-                               send_cb);
-    } else if (GetParam().variant & TEST_RMA) {
-        return ucp_put_nb(ep, &m_sbuf[0], m_sbuf.size(), (uintptr_t)&m_rbuf[0],
-                          rkey, send_cb);
-    } else {
-        ucs_fatal("invalid test case");
-    }
-}
-
-void *test_ucp_peer_failure::recv_nb(entity& e) {
-    ucs_assert(m_rbuf.size() >= m_sbuf.size());
-    if (GetParam().variant & TEST_TAG) {
-        return ucp_tag_recv_nb(e.worker(), &m_rbuf[0], m_rbuf.size(), DATATYPE, 0,
-                               0, recv_cb);
-    } else if (GetParam().variant & TEST_RMA) {
-        return NULL;
+void *test_ucp_peer_failure::send_nb(ucp_ep_h ep, ucp_rkey_h rkey)
+{
+    ucp_request_param_t param;
+    param.op_attr_mask = UCP_OP_ATTR_FIELD_DATATYPE |
+                         UCP_OP_ATTR_FIELD_CALLBACK;
+    param.datatype     = DATATYPE;
+    param.cb.send      = send_cb;
+    if (get_variant_value() & TEST_AM) {
+        return ucp_am_send_nbx(ep, AM_ID, NULL, 0, &m_sbuf[0], m_sbuf.size(),
+                               &param);
+    } else if (get_variant_value() & TEST_RMA) {
+        return ucp_put_nbx(ep, &m_sbuf[0], m_sbuf.size(), (uintptr_t)&m_rbuf[0],
+                           rkey, &param);
     } else {
         ucs_fatal("invalid test case");
     }
@@ -211,13 +219,27 @@ void test_ucp_peer_failure::fail_receiver() {
     }
 }
 
-void test_ucp_peer_failure::smoke_test(bool stable_pair) {
-    void *rreq = recv_nb(stable_pair ? stable_receiver() : failing_receiver());
-    void *sreq = send_nb(stable_pair ? stable_sender()   : failing_sender(),
-                         stable_pair ? m_stable_rkey     : m_failing_rkey);
-    wait(sreq);
-    wait(rreq);
-    EXPECT_EQ(m_sbuf, m_rbuf);
+void test_ucp_peer_failure::smoke_test(bool stable_pair)
+{
+    ucp_ep_h send_ep = stable_pair ? stable_sender() : failing_sender();
+    size_t am_count  = m_am_rx_count;
+
+    // Send and wait for completion
+    void *sreq = send_nb(send_ep, stable_pair ? m_stable_rkey : m_failing_rkey);
+    request_wait(sreq);
+
+    if (get_variant_value() & TEST_AM) {
+        // Wait for active message to be received
+        while (m_am_rx_count < am_count) {
+            progress();
+        }
+    } else if (get_variant_value() & TEST_RMA) {
+        // Flush the sender and expect data to arrive on receiver
+        void *freq = ucp_ep_flush_nb(send_ep, 0,
+                                     (ucp_send_callback_t)ucs_empty_function);
+        request_wait(freq);
+        EXPECT_EQ(m_sbuf, m_rbuf);
+    }
 }
 
 void test_ucp_peer_failure::unmap_memh(ucp_mem_h memh, ucp_context_h context)
@@ -258,7 +280,7 @@ void test_ucp_peer_failure::get_rkey(ucp_ep_h ep, entity& dst, mem_handle_t& mem
 
 void test_ucp_peer_failure::set_rkeys() {
 
-    if (GetParam().variant & TEST_RMA) {
+    if (get_variant_value() & TEST_RMA) {
         get_rkey(failing_sender(), failing_receiver(), m_failing_memh,
                  m_failing_rkey);
         get_rkey(stable_sender(), stable_receiver(), m_stable_memh,
@@ -266,12 +288,8 @@ void test_ucp_peer_failure::set_rkeys() {
     }
 }
 
-void test_ucp_peer_failure::send_cb(void *request, ucs_status_t status)
-{
-}
-
-void test_ucp_peer_failure::recv_cb(void *request, ucs_status_t status,
-                                    ucp_tag_recv_info_t *info)
+void test_ucp_peer_failure::send_cb(void *request, ucs_status_t status,
+                                    void *user_data)
 {
 }
 
@@ -295,6 +313,8 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
     create_entity();
     sender().connect(&stable_receiver(),  get_ep_params(), STABLE_EP_INDEX);
     sender().connect(&failing_receiver(), get_ep_params(), FAILING_EP_INDEX);
+    set_am_handler(stable_receiver());
+    set_am_handler(failing_receiver());
 
     set_rkeys();
 
@@ -303,7 +323,7 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
      * run traffic on a stable pair to connect it */
     smoke_test(true);
 
-    if (!(GetParam().variant & FAIL_IMM)) {
+    if (!(get_variant_value() & FAIL_IMM)) {
         /* if not fail immediately, run traffic on failing pair to connect it */
         smoke_test(false);
     }
@@ -319,8 +339,9 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
         }
     }
 
+    flush_ep(sender(), 0, FAILING_EP_INDEX);
     EXPECT_EQ(UCS_OK, m_err_status);
-    
+
     /* Since UCT/UD EP has a SW implementation of reliablity on which peer
      * failure mechanism is based, we should set small UCT/UD EP timeout
      * for UCT/UD EPs for sender's UCP EP to reduce testing time */
@@ -332,7 +353,7 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
         fail_receiver();
 
         void *sreq = send_nb(failing_sender(), m_failing_rkey);
-
+        flush_ep(sender(), 0, FAILING_EP_INDEX);
         while (!m_err_count) {
             progress();
         }
@@ -347,7 +368,8 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
             ucs_status_t status = ucp_request_check_status(sreq);
             EXPECT_NE(UCS_INPROGRESS, status);
             if (request_must_fail) {
-                EXPECT_EQ(m_err_status, status);
+                EXPECT_TRUE((m_err_status == status) ||
+                            (UCS_ERR_CANCELED == status));
             } else {
                 EXPECT_TRUE((m_err_status == status) || (UCS_OK == status));
             }
@@ -355,9 +377,10 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
         }
 
         /* Additional sends must fail */
-        void *sreq2 = send_nb(failing_sender(), m_failing_rkey);
-        EXPECT_FALSE(UCS_PTR_IS_PTR(sreq2));
-        EXPECT_EQ(m_err_status, UCS_PTR_STATUS(sreq2));
+        void *sreq2         = send_nb(failing_sender(), m_failing_rkey);
+        ucs_status_t status = request_wait(sreq2);
+        EXPECT_TRUE(UCS_STATUS_IS_ERR(status));
+        EXPECT_EQ(m_err_status, status);
 
         if (force_close) {
             unsigned allocd_eps_before =
@@ -368,12 +391,13 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
             m_failing_rkey.reset();
 
             void *creq = ucp_ep_close_nb(ep, UCP_EP_CLOSE_MODE_FORCE);
-            wait(creq);
+            request_wait(creq);
+            short_progress_loop(); /* allow discard lanes & complete destroy EP */
 
             unsigned allocd_eps_after =
                     ucs_strided_alloc_inuse_count(&sender().worker()->ep_alloc);
 
-            if (!(GetParam().variant & FAIL_IMM)) {
+            if (!(get_variant_value() & FAIL_IMM)) {
                 EXPECT_LT(allocd_eps_after, allocd_eps_before);
             }
         }
@@ -397,15 +421,8 @@ void test_ucp_peer_failure::do_test(size_t msg_size, int pre_msg_count,
     /* Check that TX polling is working well */
     while (sender().progress());
 
-    /* Destroy rkeys before destroying the worker (which also destroys the
-     * endpoints) */
+    /* Destroy rkey for failing pair */
     m_failing_rkey.reset();
-    m_stable_rkey.reset();
-
-    /* When all requests on sender are done we need to prevent LOCAL_FLUSH
-     * in test teardown. Receiver is killed and doesn't respond on FC requests
-     */
-    sender().destroy_worker();
 }
 
 UCS_TEST_P(test_ucp_peer_failure, basic) {
@@ -415,17 +432,11 @@ UCS_TEST_P(test_ucp_peer_failure, basic) {
             false /* must_fail */);
 }
 
-UCS_TEST_P(test_ucp_peer_failure, rndv_disable) {
-    const size_t size_max = std::numeric_limits<size_t>::max();
-
-    sender().connect(&receiver(), get_ep_params(), STABLE_EP_INDEX);
-    EXPECT_EQ(size_max, ucp_ep_config(sender().ep())->tag.rndv.am_thresh);
-    EXPECT_EQ(size_max, ucp_ep_config(sender().ep())->tag.rndv.rma_thresh);
-    EXPECT_EQ(size_max, ucp_ep_config(sender().ep())->tag.rndv_send_nbr.am_thresh);
-    EXPECT_EQ(size_max, ucp_ep_config(sender().ep())->tag.rndv_send_nbr.rma_thresh);
-}
-
-UCS_TEST_P(test_ucp_peer_failure, zcopy, "ZCOPY_THRESH=1023") {
+UCS_TEST_P(test_ucp_peer_failure, zcopy, "ZCOPY_THRESH=1023",
+           /* to catch failure with TCP during progressing multi AM Zcopy,
+            * since `must_fail=true` */
+           "TCP_SNDBUF?=1k", "TCP_RCVBUF?=128",
+           "TCP_RX_SEG_SIZE?=512", "TCP_TX_SEG_SIZE?=256") {
     do_test(UCS_KBYTE, /* msg_size */
             0, /* pre_msg_cnt */
             false, /* force_close */
@@ -439,32 +450,97 @@ UCS_TEST_P(test_ucp_peer_failure, bcopy_multi, "SEG_SIZE?=512", "RC_TM_ENABLE?=n
             false /* must_fail */);
 }
 
-UCS_TEST_P(test_ucp_peer_failure, force_close, "RC_FC_ENABLE?=n") {
+UCS_TEST_P(test_ucp_peer_failure, force_close, "RC_FC_ENABLE?=n",
+           /* To catch unexpected descriptors leak, for multi-fragment protocol
+              with TCP */
+           "TCP_RX_SEG_SIZE?=1024", "TCP_TX_SEG_SIZE?=1024")
+{
     do_test(16000, /* msg_size */
             1000, /* pre_msg_cnt */
             true, /* force_close */
             false /* must_fail */);
 }
 
-UCS_TEST_SKIP_COND_P(test_ucp_peer_failure, disable_sync_send,
-                     !(GetParam().variant & TEST_TAG)) {
-    const size_t        max_size = UCS_MBYTE;
-    std::vector<char>   buf(max_size, 0);
-    void                *req;
+class test_ucp_peer_failure_keepalive : public test_ucp_peer_failure
+{
+public:
+    test_ucp_peer_failure_keepalive() {
+        m_sbuf.resize(1 * UCS_MBYTE);
+        m_rbuf.resize(1 * UCS_MBYTE);
 
-    sender().connect(&receiver(), get_ep_params());
-
-    /* Make sure API is disabled for any size and data type */
-    for (size_t size = 1; size <= max_size; size *= 2) {
-        req = ucp_tag_send_sync_nb(sender().ep(), buf.data(), size, DATATYPE,
-                                   0x111337, NULL);
-        EXPECT_FALSE(UCS_PTR_IS_PTR(req));
-        EXPECT_EQ(UCS_ERR_UNSUPPORTED, UCS_PTR_STATUS(req));
-
-        ucp::data_type_desc_t dt_desc(DATATYPE_IOV, buf.data(), size);
-        req = ucp_tag_send_sync_nb(sender().ep(), dt_desc.buf(), dt_desc.count(),
-                                   dt_desc.dt(), 0x111337, NULL);
-        EXPECT_FALSE(UCS_PTR_IS_PTR(req));
-        EXPECT_EQ(UCS_ERR_UNSUPPORTED, UCS_PTR_STATUS(req));
+        m_env.push_back(new ucs::scoped_setenv("UCX_TCP_KEEPIDLE", "inf"));
+        m_env.push_back(new ucs::scoped_setenv("UCX_UD_TIMEOUT", "3s"));
     }
+
+    void init() {
+        test_ucp_peer_failure::init();
+        create_entity();
+        sender().connect(&stable_receiver(), get_ep_params(), STABLE_EP_INDEX);
+        sender().connect(&failing_receiver(), get_ep_params(), FAILING_EP_INDEX);
+        stable_receiver().connect(&sender(), get_ep_params());
+        failing_receiver().connect(&sender(), get_ep_params());
+        set_am_handler(failing_receiver());
+        set_am_handler(stable_receiver());
+    }
+
+    static void get_test_variants(std::vector<ucp_test_variant>& variants) {
+        add_variant_with_value(variants, UCP_FEATURE_AM, TEST_AM, "am");
+        add_variant_with_value(variants, UCP_FEATURE_AM | UCP_FEATURE_WAKEUP,
+                               TEST_AM | WAKEUP, "am_wakeup");
+    }
+};
+
+UCS_TEST_P(test_ucp_peer_failure_keepalive, kill_receiver,
+           "KEEPALIVE_INTERVAL=0.3", "KEEPALIVE_NUM_EPS=inf") {
+    /* TODO: wireup is not tested yet */
+
+    scoped_log_handler err_handler(wrap_errors_logger);
+    scoped_log_handler warn_handler(hide_warns_logger);
+
+    /* initiate p2p pairing */
+    ucp_ep_resolve_remote_id(failing_sender(), 0);
+    smoke_test(true); /* allow wireup to complete */
+    smoke_test(false);
+
+    if (ucp_ep_config(stable_sender())->key.ep_check_map == 0) {
+        UCS_TEST_SKIP_R("Unsupported");
+    }
+
+    /* ensure both pair have ep_check map */
+    ASSERT_NE(0, ucp_ep_config(failing_sender())->key.ep_check_map);
+
+    /* aux (ud) transport doesn't support keepalive feature and
+     * we are assuming that wireup/connect procedure is done */
+
+    EXPECT_EQ(0, m_err_count); /* ensure no errors are detected */
+
+    /* flush all outstanding ops to allow keepalive to run */
+    flush_worker(sender());
+    if (get_variant_value() & WAKEUP) {
+        wait_for_wakeup({ sender().worker(), failing_receiver().worker() },
+                        100, true);
+    }
+
+    /* kill EPs & ifaces */
+    failing_receiver().close_all_eps(*this, 0, UCP_EP_CLOSE_MODE_FORCE);
+    if (get_variant_value() & WAKEUP) {
+        wait_for_wakeup({ sender().worker() });
+    }
+    wait_for_flag(&m_err_count);
+
+    /* dump warnings */
+    int warn_count = m_warnings.size();
+    for (int i = 0; i < warn_count; ++i) {
+        UCS_TEST_MESSAGE << "< " << m_warnings[i] << " >";
+    }
+
+    EXPECT_NE(0, m_err_count);
+
+    /* check if stable receiver is still works */
+    m_err_count = 0;
+    smoke_test(true);
+
+    EXPECT_EQ(0, m_err_count); /* ensure no errors are detected */
 }
+
+UCP_INSTANTIATE_TEST_CASE(test_ucp_peer_failure_keepalive)
