@@ -19,7 +19,7 @@ using namespace ucs; /* For vector<char> serialization */
 class test_ucp_tag_match : public test_ucp_tag {
 public:
     enum {
-        ENABLE_PROTO = UCS_BIT(2)
+        ENABLE_PROTO = UCS_BIT(8)
     };
 
     test_ucp_tag_match() {
@@ -63,6 +63,8 @@ public:
     }
 
 protected:
+    void test_iov(const size_t *iov_sizes, size_t iov_count);
+
     static void recv_callback_release_req(void *request, ucs_status_t status,
                                           ucp_tag_recv_info_t *info)
     {
@@ -141,6 +143,71 @@ UCS_TEST_P(test_ucp_tag_match, send_recv_exp_medium) {
     request_free(my_recv_req);
 }
 
+void test_ucp_tag_match::test_iov(const size_t *iov_sizes, size_t iov_count)
+{
+    std::stringstream ss;
+    for (size_t i = 0; i < iov_count; ++i) {
+        ss << iov_sizes[i] << " ";
+    }
+    UCS_TEST_MESSAGE << "{ " << ss.str() << "}";
+
+    std::vector<std::string> sendbufs(iov_count);
+    std::vector<std::string> recvbufs(iov_count);
+
+    std::string send_flat_data;
+    ucp_dt_iov_t send_iov[iov_count], recv_iov[iov_count];
+    for (size_t i = 0; i < iov_count; ++i) {
+        sendbufs[i].resize(iov_sizes[i]);
+        ucs::fill_random(sendbufs[i]);
+        send_iov[i].buffer = &sendbufs[i][0];
+        send_iov[i].length = sendbufs[i].size();
+        send_flat_data    += sendbufs[i];
+
+        recvbufs[i].resize(iov_sizes[iov_count - i - 1], 'x');
+        recv_iov[i].buffer = &recvbufs[i][0];
+        recv_iov[i].length = recvbufs[i].size();
+    }
+
+    request *my_recv_req = recv_nb(recv_iov, iov_count, UCP_DATATYPE_IOV,
+                                   0x1337, 0xffff);
+    ASSERT_TRUE(!UCS_PTR_IS_ERR(my_recv_req));
+
+    send_b(send_iov, iov_count, UCP_DATATYPE_IOV, 0x111337);
+    wait(my_recv_req);
+
+    EXPECT_EQ(send_flat_data.size(), my_recv_req->info.length);
+    EXPECT_EQ((ucp_tag_t)0x111337, my_recv_req->info.sender_tag);
+    request_free(my_recv_req);
+
+    /* Compare data */
+    std::string recv_flat_data;
+    for (size_t i = 0; i < iov_count; ++i) {
+        recv_flat_data += recvbufs[i];
+    }
+    EXPECT_EQ(send_flat_data, recv_flat_data);
+}
+
+UCS_TEST_P(test_ucp_tag_match, send_recv_exp_iov, "RNDV_THRESH=inf")
+{
+    test_iov(NULL, 0);
+
+    static const size_t iov_sizes0[] = {0};
+    test_iov(iov_sizes0, ucs_static_array_size(iov_sizes0));
+
+    static const size_t iov_sizes1[] = {1000, 3000, 10000};
+    test_iov(iov_sizes1, ucs_static_array_size(iov_sizes1));
+
+    static const size_t iov_sizes2[] = {1000, 0, 10000, 0};
+    test_iov(iov_sizes2, ucs_static_array_size(iov_sizes2));
+
+    static const size_t iov_sizes3[] = {0, 3000, 0, 10000, 0, 0, 0};
+    test_iov(iov_sizes3, ucs_static_array_size(iov_sizes3));
+
+    static const size_t iov_sizes4[] = {32, 16, 18, 15, 0, 0, 1, 78, 54, 198,
+                                        234354, 1, 10, 100000, 0, 6};
+    test_iov(iov_sizes4, ucs_static_array_size(iov_sizes4));
+}
+
 UCS_TEST_P(test_ucp_tag_match, send2_nb_recv_exp_medium) {
     static const size_t size = 50000;
     request user_data;
@@ -478,14 +545,18 @@ public:
         RNDV_SCHEME_AUTO = 0,
         RNDV_SCHEME_PUT_ZCOPY,
         RNDV_SCHEME_GET_ZCOPY,
-        RNDV_SCHEME_LAST
+        RNDV_SCHEME_LAST,
+        PUT_ZCOPY_FLUSH = ENABLE_PROTO << 1
     };
 
     static const std::string rndv_schemes[];
 
     void init() {
         ASSERT_LE(rndv_scheme(), (int)RNDV_SCHEME_GET_ZCOPY);
+        UCS_STATIC_ASSERT(!(ENABLE_PROTO & UCS_MASK(RNDV_SCHEME_LAST)));
+        modify_config("RNDV_THRESH", "0");
         modify_config("RNDV_SCHEME", rndv_schemes[rndv_scheme()]);
+        modify_config("RNDV_PUT_FORCE_FLUSH", force_flush() ? "y" : "n");
         test_ucp_tag_match::init();
     }
 
@@ -493,15 +564,16 @@ public:
         for (int rndv_scheme = 0; rndv_scheme < RNDV_SCHEME_LAST; ++rndv_scheme) {
             add_variant_with_value(variants, get_ctx_params(), rndv_scheme,
                                    "rndv_" + rndv_schemes[rndv_scheme]);
+            add_variant_with_value(variants, get_ctx_params(),
+                                   rndv_scheme | ENABLE_PROTO,
+                                   rndv_schemes[rndv_scheme] + ",proto");
         }
 
-        // Generate variants with and new protocols
+        // Add variant with force flush
         add_variant_with_value(variants, get_ctx_params(),
-                               RNDV_SCHEME_AUTO | ENABLE_PROTO,
-                               "rndv_auto,proto");
-        add_variant_with_value(variants, get_ctx_params(),
-                               RNDV_SCHEME_GET_ZCOPY | ENABLE_PROTO,
-                               "rndv_get_zcopy,proto");
+                               RNDV_SCHEME_PUT_ZCOPY | ENABLE_PROTO |
+                                       PUT_ZCOPY_FLUSH,
+                               "rndv_put_flush,proto");
     }
 
 protected:
@@ -511,13 +583,18 @@ protected:
         ucs_assert(!(mask & ENABLE_PROTO));
         return get_variant_value() & mask;
     }
+
+    bool force_flush() const
+    {
+        return get_variant_value() & PUT_ZCOPY_FLUSH;
+    }
 };
 
 const std::string test_ucp_tag_match_rndv::rndv_schemes[] = { "auto",
                                                               "put_zcopy",
                                                               "get_zcopy" };
 
-UCS_TEST_P(test_ucp_tag_match_rndv, length0, "RNDV_THRESH=0")
+UCS_TEST_P(test_ucp_tag_match_rndv, length0)
 {
     request *my_send_req = send_nb((void*)0xdeadbeef, 0, DATATYPE, 1);
     ASSERT_TRUE(!UCS_PTR_IS_ERR(my_send_req));
@@ -529,7 +606,8 @@ UCS_TEST_P(test_ucp_tag_match_rndv, length0, "RNDV_THRESH=0")
     wait_and_validate(my_send_req);
 }
 
-UCS_TEST_P(test_ucp_tag_match_rndv, sync_send_unexp, "RNDV_THRESH=1048576") {
+UCS_TEST_P(test_ucp_tag_match_rndv, sync_send_unexp)
+{
     static const size_t size = 1148576;
     request             *my_send_req;
     ucp_tag_recv_info_t info;
@@ -564,7 +642,8 @@ UCS_TEST_P(test_ucp_tag_match_rndv, sync_send_unexp, "RNDV_THRESH=1048576") {
     request_free(my_send_req);
 }
 
-UCS_TEST_P(test_ucp_tag_match_rndv, req_exp, "RNDV_THRESH=1048576") {
+UCS_TEST_P(test_ucp_tag_match_rndv, req_exp)
+{
     static const size_t size = 1148576;
     request *my_send_req, *my_recv_req;
 
@@ -598,7 +677,8 @@ UCS_TEST_P(test_ucp_tag_match_rndv, req_exp, "RNDV_THRESH=1048576") {
     request_free(my_recv_req);
 }
 
-UCS_TEST_P(test_ucp_tag_match_rndv, rts_unexp, "RNDV_THRESH=1048576") {
+UCS_TEST_P(test_ucp_tag_match_rndv, rts_unexp)
+{
     static const size_t size = 1148576;
     request             *my_send_req;
     ucp_tag_recv_info_t info;
@@ -630,7 +710,8 @@ UCS_TEST_P(test_ucp_tag_match_rndv, rts_unexp, "RNDV_THRESH=1048576") {
     EXPECT_EQ(sendbuf, recvbuf);
 }
 
-UCS_TEST_P(test_ucp_tag_match_rndv, truncated, "RNDV_THRESH=1048576") {
+UCS_TEST_P(test_ucp_tag_match_rndv, truncated)
+{
     static const size_t size = 1148576;
     request *my_send_req;
     ucp_tag_recv_info_t info;
@@ -658,7 +739,8 @@ UCS_TEST_P(test_ucp_tag_match_rndv, truncated, "RNDV_THRESH=1048576") {
     wait_and_validate(my_send_req);
 }
 
-UCS_TEST_P(test_ucp_tag_match_rndv, post_larger_recv, "RNDV_THRESH=0") {
+UCS_TEST_P(test_ucp_tag_match_rndv, post_larger_recv)
+{
     /* small send size should probably be lower than minimum GET Zcopy
      * size supported by IB TLs */
     static const size_t small_send_size = 16;
@@ -736,17 +818,20 @@ UCS_TEST_P(test_ucp_tag_match_rndv, req_exp_auto_thresh, "RNDV_THRESH=auto") {
 }
 
 UCS_TEST_P(test_ucp_tag_match_rndv, exp_huge_mix) {
-    const size_t sizes[] = { 1000, 2000, 8000, 2500ul * UCS_MBYTE };
+    const std::vector<size_t> sizes = {1000, 2000, 8000, 2500ul * UCS_MBYTE,
+                                       UCS_GBYTE + 32};
 
     /* small sizes should warm-up tag cache */
-    for (unsigned i = 0; i < ucs_static_array_size(sizes); ++i) {
-        const size_t size = sizes[i] / ucs::test_time_multiplier();
+    for (auto c_size : sizes) {
+        const size_t size = c_size / ucs::test_time_multiplier() /
+                            ucs::test_time_multiplier();
         request *my_send_req, *my_recv_req;
 
         std::vector<char> sendbuf(size, 0);
         std::vector<char> recvbuf(size, 0);
 
         ucs::fill_random(sendbuf);
+        VALGRIND_MAKE_MEM_UNDEFINED(&recvbuf[0], recvbuf.size());
 
         my_recv_req = recv_nb(&recvbuf[0], recvbuf.size(), DATATYPE, 0x1337, 0xffff);
         ASSERT_TRUE(!UCS_PTR_IS_ERR(my_recv_req));
@@ -767,7 +852,8 @@ UCS_TEST_P(test_ucp_tag_match_rndv, exp_huge_mix) {
     }
 }
 
-UCS_TEST_P(test_ucp_tag_match_rndv, bidir_multi_exp_post, "RNDV_THRESH=0") {
+UCS_TEST_P(test_ucp_tag_match_rndv, bidir_multi_exp_post)
+{
     const size_t sizes[] = { 8 * UCS_KBYTE, 128 * UCS_KBYTE, 512 * UCS_KBYTE,
                              8 * UCS_MBYTE, 128 * UCS_MBYTE, 512 * UCS_MBYTE };
 
