@@ -8,6 +8,7 @@
 extern "C" {
 #include <ucs/arch/atomic.h>
 }
+#include <list>
 
 class uct_flush_test : public uct_test {
 public:
@@ -121,6 +122,12 @@ public:
         return am_req->test->am_send_pending(am_req);
     }
 
+    static void purge_cb(uct_pending_req_t *self, void *arg)
+    {
+        test_req_t *req = ucs_container_of(self, test_req_t, uct);
+        --req->comp.count;
+    }
+
     static ucs_status_t flush_progress(uct_pending_req_t *req)
     {
         test_req_t *flush_req = ucs_container_of(req, test_req_t, uct);
@@ -176,8 +183,9 @@ public:
                                  UCT_CB_FLAG_ASYNC);
 
         uct_completion_t zcomp;
-        zcomp.count = 2;
-        zcomp.func  = NULL;
+        zcomp.count  = 2;
+        zcomp.status = UCS_OK;
+        zcomp.func   = NULL;
 
         ucs_status_t status;
         UCS_TEST_GET_BUFFER_IOV(iov, iovcnt, sendbuf.ptr(), sendbuf.length(),
@@ -263,8 +271,9 @@ public:
     void flush_ep_nb() {
         uct_completion_t comp;
         ucs_status_t status;
-        comp.count = 2;
-        comp.func  = NULL;
+        comp.count  = 2;
+        comp.status = UCS_OK;
+        comp.func   = NULL;
         do {
             progress();
             status = uct_ep_flush(sender().ep(0), m_flush_flags, &comp);
@@ -345,11 +354,12 @@ void uct_flush_test::test_flush_am_pending(flush_func_t flush, bool destroy_ep)
      std::vector<test_req_t> reqs;
      reqs.resize(10);
      for (std::vector<test_req_t>::iterator it = reqs.begin(); it != reqs.end();) {
-         it->sendbuf    = &sendbuf;
-         it->test       = this;
-         it->uct.func   = am_progress;
-         it->comp.count = 2;
-         it->comp.func  = NULL;
+         it->sendbuf     = &sendbuf;
+         it->test        = this;
+         it->uct.func    = am_progress;
+         it->comp.count  = 2;
+         it->comp.func   = NULL;
+         it->comp.status = UCS_OK;
          status = uct_ep_pending_add(sender().ep(0), &it->uct, 0);
          if (UCS_ERR_BUSY == status) {
              /* User advised to retry the send. It means no requests added
@@ -364,10 +374,15 @@ void uct_flush_test::test_flush_am_pending(flush_func_t flush, bool destroy_ep)
          ASSERT_UCS_OK(status);
      }
 
+     if (is_flush_cancel()) {
+         uct_ep_pending_purge(sender().ep(0), purge_cb, NULL);
+     }
+
      /* Try to start a flush */
      test_req_t flush_req;
-     flush_req.comp.count = 2;
-     flush_req.comp.func  = NULL;
+     flush_req.comp.count  = 2;
+     flush_req.comp.status = UCS_OK;
+     flush_req.comp.func   = NULL;
 
      for (;;) {
          status = uct_ep_flush(sender().ep(0), m_flush_flags, &flush_req.comp);
@@ -394,11 +409,7 @@ void uct_flush_test::test_flush_am_pending(flush_func_t flush, bool destroy_ep)
      EXPECT_EQ(1, flush_req.comp.count);
 
      while (!reqs.empty()) {
-         if (is_flush_cancel()) {
-            EXPECT_EQ(2, reqs.back().comp.count);
-         } else {
-            EXPECT_EQ(1, reqs.back().comp.count);
-         }
+         EXPECT_EQ(1, reqs.back().comp.count);
          reqs.pop_back();
      }
 
@@ -538,3 +549,265 @@ UCS_TEST_SKIP_COND_P(uct_flush_test, am_pending_flush_nb,
 }
 
 UCT_INSTANTIATE_TEST_CASE(uct_flush_test)
+
+class uct_cancel_test : public uct_test {
+public:
+    static const size_t BUF_SIZE    = 8 * 1024;
+    static const size_t BUF_SIZE_DC = 1 * 1024;
+
+    class peer {
+    public:
+        peer(uct_cancel_test &test) :
+            m_e(NULL), m_buf(NULL), m_buf32(NULL), m_peer(NULL), m_test(test)
+        {
+            m_e = m_test.uct_test::create_entity(0, error_handler_cb);
+            m_test.m_entities.push_back(m_e);
+
+            m_buf.reset(new mapped_buffer(BUF_SIZE, 0, *m_e));
+            m_buf32.reset(new mapped_buffer(32, 0, *m_e));
+            uct_iface_set_am_handler(m_e->iface(), 0, am_cb, &m_test,
+                                     UCT_CB_FLAG_ASYNC);
+        }
+
+        void connect() {
+            m_e->destroy_eps();
+            m_peer->m_e->destroy_eps();
+            m_e->connect(0, *m_peer->m_e, 0);
+            m_peer->m_e->connect(0, *m_e, 0);
+        }
+
+        entity                       *m_e;
+        ucs::auto_ptr<mapped_buffer> m_buf;
+        ucs::auto_ptr<mapped_buffer> m_buf32;
+        peer                         *m_peer;
+
+    private:
+        uct_cancel_test &m_test;
+    };
+
+    uct_cancel_test() :
+        uct_test(), m_s0(NULL), m_s1(NULL), m_err_count(0)
+    {
+    }
+
+    ucs_status_t am_bcopy(peer *s) {
+        mapped_buffer &sendbuf = *s->m_buf32;
+        ssize_t packed_len;
+
+        packed_len = uct_ep_am_bcopy(s->m_e->ep(0), 0, mapped_buffer::pack,
+                                     (void*)&sendbuf, 0);
+        if (packed_len >= 0) {
+            EXPECT_EQ(sendbuf.length(), (size_t)packed_len);
+            return UCS_OK;
+        } else {
+            return (ucs_status_t)packed_len;
+        }
+    }
+
+    ucs_status_t am_zcopy(peer *s) {
+        size_t size = ucs_min(BUF_SIZE, s->m_e->iface_attr().cap.am.max_zcopy);
+        mapped_buffer &sendbuf = *s->m_buf;
+        size_t header_length = 0;
+        uct_iov_t iov;
+
+        if (has_transport("dc_mlx5")) {
+            size = ucs_min(BUF_SIZE_DC, size);
+        }
+
+        iov.buffer = (char*)sendbuf.ptr() + header_length;
+        iov.count  = 1;
+        iov.length = size - header_length;
+        iov.memh   = sendbuf.memh();
+        return uct_ep_am_zcopy(s->m_e->ep(0), 0, sendbuf.ptr(), header_length,
+                               &iov, 1, 0, NULL);
+    }
+
+    ucs_status_t get_zcopy(peer *s) {
+        size_t size = ucs_min(BUF_SIZE, s->m_e->iface_attr().cap.get.max_zcopy);
+        mapped_buffer &sendbuf = *s->m_buf;
+        mapped_buffer &recvbuf = *s->m_peer->m_buf;
+
+        if (has_transport("dc_mlx5")) {
+            size = ucs_min(BUF_SIZE_DC, size);
+        }
+
+        UCS_TEST_GET_BUFFER_IOV(iov, iovcnt, sendbuf.ptr(), size,
+                                sendbuf.memh(), s->m_e->iface_attr().cap.get.max_iov);
+
+        return uct_ep_get_zcopy(s->m_e->ep(0), iov, iovcnt, recvbuf.addr(),
+                                recvbuf.rkey(), NULL);
+    }
+
+    ucs_status_t get_bcopy(peer *s) {
+        mapped_buffer &sendbuf = *s->m_buf32;
+        mapped_buffer &recvbuf = *s->m_peer->m_buf32;
+
+        return uct_ep_get_bcopy(s->m_e->ep(0), (uct_unpack_callback_t)memcpy,
+                                sendbuf.ptr(), sendbuf.length(),
+                                recvbuf.addr(), recvbuf.rkey(), NULL);
+    }
+
+    void flush_and_reconnect() {
+        std::list<entity *> flushing;
+        ucs_status_t status = UCS_OK;
+        uct_completion_t done;
+
+        m_err_count = 0;
+        flushing.push_back(m_s0->m_e);
+        flushing.push_back(m_s1->m_e);
+        done.count  = flushing.size() + 1;
+        done.status = UCS_OK;
+        done.func   = NULL;
+        ucs_time_t loop_end_limit = ucs_get_time() + ucs_time_from_sec(200.0);
+        while (!flushing.empty() && (ucs_get_time() < loop_end_limit)) {
+            std::list<entity *>::iterator iter = flushing.begin();
+            while (iter != flushing.end()) {
+                status = uct_ep_flush((*iter)->ep(0), UCT_FLUSH_FLAG_CANCEL, &done);
+                if (status == UCS_ERR_NO_RESOURCE) {
+                    iter++;
+                } else {
+                    ASSERT_UCS_OK_OR_INPROGRESS(status);
+                    iter = flushing.erase(iter);
+                    if (status == UCS_OK) {
+                        done.count--;
+                    }
+                }
+            }
+
+            short_progress_loop();
+        }
+        ASSERT_UCS_OK_OR_INPROGRESS(status);
+        double holdup = 200.0 - ucs_time_to_sec(loop_end_limit - ucs_get_time());
+        if (holdup > 10.0) {
+            UCS_TEST_MESSAGE << "flush took " << holdup << " sec";
+        }
+
+        /* coverity[loop_condition] */
+        while (done.count != 1) {
+            progress();
+        }
+
+        m_s0->connect();
+
+        /* there is a chance that one side getting disconect error before
+         * calling flush(CANCEL) */
+        EXPECT_LE(m_err_count, 1);
+    }
+
+    typedef ucs_status_t (uct_cancel_test::* send_func_t)(peer *s);
+
+    void fill(send_func_t send) {
+        ucs_status_t status;
+        std::list<peer *> filling;
+
+        filling.push_back(m_s0);
+        filling.push_back(m_s1);
+        while (!filling.empty()) {
+            std::list<peer *>::iterator iter = filling.begin();
+            while (iter != filling.end()) {
+                status = (this->*send)(*iter);
+                if (status == UCS_ERR_NO_RESOURCE) {
+                    iter = filling.erase(iter);
+                } else {
+                    ASSERT_UCS_OK_OR_INPROGRESS(status);
+                    iter++;
+                }
+            }
+        }
+    }
+
+    int count() {
+        return 100;
+    }
+
+    void do_test(send_func_t send) {
+        ucs_time_t loop_end_limit = ucs_get_time() + ucs_time_from_sec(300.0);
+        for (int i = 0; (i < count()) && (ucs_get_time() < loop_end_limit); ++i) {
+            fill(send);
+            flush_and_reconnect();
+        }
+    }
+
+protected:
+
+    ucs::auto_ptr<peer> m_s0;
+    ucs::auto_ptr<peer> m_s1;
+    int m_err_count;
+
+    virtual void init() {
+        uct_test::init();
+
+        m_s0.reset(new peer(*this));
+        check_skip_test_tl();
+        m_s1.reset(new peer(*this));
+
+        m_s0->m_peer = m_s1;
+        m_s1->m_peer = m_s0;
+
+        m_s0->connect();
+        flush();
+    }
+
+    virtual void cleanup() {
+        flush();
+
+        m_s0.reset();
+        m_s1.reset();
+
+        uct_test::cleanup();
+    }
+
+    static ucs_status_t
+    error_handler_cb(void *arg, uct_ep_h ep, ucs_status_t status) {
+        uct_cancel_test *test = reinterpret_cast<uct_cancel_test*>(arg);
+        return test->error_handler(ep, status);
+    }
+
+    static ucs_status_t am_cb(void *arg, void *data, size_t length, unsigned flags) {
+        uct_cancel_test *test = reinterpret_cast<uct_cancel_test*>(arg);
+        return test->am(data, length, flags);
+    }
+
+    ucs_status_t error_handler(uct_ep_h ep, ucs_status_t status) {
+        EXPECT_EQ(UCS_ERR_ENDPOINT_TIMEOUT, status);
+        m_err_count++;
+        return UCS_OK;
+    }
+
+    ucs_status_t am(void *data, size_t length, unsigned flags) {
+        return UCS_OK;
+    }
+
+    void check_skip_test_tl() {
+        if ((GetParam()->tl_name != "dc_mlx5") &&
+            (GetParam()->tl_name != "rc_verbs") &&
+            (GetParam()->tl_name != "rc_mlx5")) {
+            UCS_TEST_SKIP_R("not supported yet");
+        }
+
+        check_skip_test();
+    }
+
+};
+
+UCS_TEST_SKIP_COND_P(uct_cancel_test, am_zcopy,
+                     !check_caps(UCT_IFACE_FLAG_AM_ZCOPY)) {
+    do_test(&uct_cancel_test::am_zcopy);
+}
+
+UCS_TEST_SKIP_COND_P(uct_cancel_test, am_bcopy,
+                     !check_caps(UCT_IFACE_FLAG_AM_BCOPY)) {
+    do_test(&uct_cancel_test::am_bcopy);
+}
+
+UCS_TEST_SKIP_COND_P(uct_cancel_test, get_bcopy,
+                     !check_caps(UCT_IFACE_FLAG_GET_BCOPY)) {
+    do_test(&uct_cancel_test::get_bcopy);
+}
+
+UCS_TEST_SKIP_COND_P(uct_cancel_test, get_zcopy,
+                     !check_caps(UCT_IFACE_FLAG_GET_ZCOPY)) {
+    do_test(&uct_cancel_test::get_zcopy);
+}
+
+UCT_INSTANTIATE_TEST_CASE(uct_cancel_test)
