@@ -1,7 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2004 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "create_f90_util.h"
@@ -41,10 +40,6 @@ typedef struct realModel {
     MPI_Datatype dtype;
 } realModel;
 
-#undef FUNCNAME
-#define FUNCNAME MPI_Type_create_f90_complex
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /*@
    MPI_Type_create_f90_complex - Return a predefined type that matches
    the specified range
@@ -77,12 +72,11 @@ int MPI_Type_create_f90_complex(int precision, int range, MPI_Datatype * newtype
         {MPIR_F90_REAL_MODEL, MPI_COMPLEX},
         {MPIR_F90_DOUBLE_MODEL, MPI_DOUBLE_COMPLEX}
     };
-    MPIR_FUNC_TERSE_STATE_DECL(MPID_STATE_MPI_TYPE_CREATE_F90_COMPLEX);
 
     MPIR_ERRTEST_INITIALIZED_ORDIE();
 
     MPID_THREAD_CS_ENTER(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
-    MPIR_FUNC_TERSE_ENTER(MPID_STATE_MPI_TYPE_CREATE_F90_COMPLEX);
+    MPIR_FUNC_TERSE_ENTER;
 
     /* ... body of routine ...  */
     /* MPI 2.1, Section 16.2, page 473 lines 12-27 make it clear that
@@ -96,8 +90,7 @@ int MPI_Type_create_f90_complex(int precision, int range, MPI_Datatype * newtype
                                                        f90_real_model[i].digits,
                                                        f90_real_model[i].exponent,
                                                        &f90_real_model[i].dtype);
-            if (mpi_errno)
-                MPIR_ERR_POP(mpi_errno);
+            MPIR_ERR_CHECK(mpi_errno);
         }
     }
 
@@ -120,13 +113,12 @@ int MPI_Type_create_f90_complex(int precision, int range, MPI_Datatype * newtype
                                                    MPI_COMBINER_F90_COMPLEX, range, precision,
                                                    newtype);
     }
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     /* ... end of body of routine ... */
 
   fn_exit:
-    MPIR_FUNC_TERSE_EXIT(MPID_STATE_MPI_TYPE_CREATE_F90_COMPLEX);
+    MPIR_FUNC_TERSE_EXIT;
     MPID_THREAD_CS_EXIT(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
     return mpi_errno;
   fn_fail:
@@ -134,12 +126,12 @@ int MPI_Type_create_f90_complex(int precision, int range, MPI_Datatype * newtype
 #ifdef HAVE_ERROR_CHECKING
     {
         mpi_errno =
-            MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER,
+            MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, __func__, __LINE__, MPI_ERR_OTHER,
                                  "**mpi_type_create_f90_complex",
                                  "**mpi_type_create_f90_complex %d %d", precision, range);
     }
 #endif
-    mpi_errno = MPIR_Err_return_comm(0, FCNAME, mpi_errno);
+    mpi_errno = MPIR_Err_return_comm(0, __func__, mpi_errno);
     goto fn_exit;
     /* --END ERROR HANDLING-- */
 }
