@@ -1,7 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2011 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 /* This file contains glue code that ROMIO needs to call/use in order to access
@@ -19,6 +18,36 @@ int MPIR_Ext_dbg_romio_terse_enabled = 0;
 int MPIR_Ext_dbg_romio_typical_enabled = 0;
 int MPIR_Ext_dbg_romio_verbose_enabled = 0;
 
+/* NOTE: we'll lazily initialize this mutex */
+static MPL_thread_mutex_t romio_mutex;
+static MPL_atomic_int_t romio_mutex_initialized = MPL_ATOMIC_INT_T_INITIALIZER(0);
+
+void MPIR_Ext_mutex_init(void)
+{
+#if defined(MPICH_IS_THREADED)
+    if (!MPL_atomic_load_int(&romio_mutex_initialized)) {
+        int err;
+        MPL_thread_mutex_create(&romio_mutex, &err);
+        MPIR_Assert(err == 0);
+
+        MPL_atomic_store_int(&romio_mutex_initialized, 1);
+    }
+#endif
+}
+
+void MPIR_Ext_mutex_finalize(void)
+{
+#if defined(MPICH_IS_THREADED)
+    if (MPL_atomic_load_int(&romio_mutex_initialized)) {
+        int err;
+        MPL_thread_mutex_destroy(&romio_mutex, &err);
+        MPIR_Assert(err == 0);
+
+        MPL_atomic_store_int(&romio_mutex_initialized, 0);
+    }
+#endif
+}
+
 /* to be called early by ROMIO's initialization process in order to setup init-time
  * glue code that cannot be initialized statically */
 int MPIR_Ext_init(void)
@@ -53,12 +82,27 @@ int MPIR_Ext_assert_fail(const char *cond, const char *file_name, int line_num)
  * threading strategies. */
 void MPIR_Ext_cs_enter(void)
 {
-    MPID_THREAD_CS_ENTER(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
+#if defined(MPICH_IS_THREADED)
+    if (MPIR_ThreadInfo.isThreaded) {
+        /* lazily initialize the mutex */
+        MPIR_Ext_mutex_init();
+
+        int err;
+        MPL_thread_mutex_lock(&romio_mutex, &err, MPL_THREAD_PRIO_HIGH);
+        MPIR_Assert(err == 0);
+    }
+#endif
 }
 
 void MPIR_Ext_cs_exit(void)
 {
-    MPID_THREAD_CS_EXIT(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
+#if defined(MPICH_IS_THREADED)
+    if (MPIR_ThreadInfo.isThreaded) {
+        int err;
+        MPL_thread_mutex_unlock(&romio_mutex, &err);
+        MPIR_Assert(err == 0);
+    }
+#endif
 }
 
 /* This routine is for a thread to yield control when the thread is waiting for
@@ -67,33 +111,28 @@ void MPIR_Ext_cs_exit(void)
 void MPIR_Ext_cs_yield(void)
 {
     /* TODO: check whether the progress engine is blocked */
-    MPID_THREAD_CS_YIELD(GLOBAL, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
+    MPIR_Ext_cs_exit();
+    MPL_thread_yield();
+    MPIR_Ext_cs_enter();
 }
 
 /* will consider MPI_DATATYPE_NULL to be an error */
-#undef FUNCNAME
-#define FUNCNAME MPIR_Ext_datatype_iscommitted
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIR_Ext_datatype_iscommitted(MPI_Datatype datatype)
 {
     int mpi_errno = MPI_SUCCESS;
 
     MPIR_ERRTEST_DATATYPE(datatype, "datatype", mpi_errno);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
-    if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN) {
+    if (!HANDLE_IS_BUILTIN(datatype)) {
         MPIR_Datatype *datatype_ptr = NULL;
         MPIR_Datatype_get_ptr(datatype, datatype_ptr);
 
         MPIR_Datatype_valid_ptr(datatype_ptr, mpi_errno);
-        if (mpi_errno)
-            MPIR_ERR_POP(mpi_errno);
+        MPIR_ERR_CHECK(mpi_errno);
 
         MPIR_Datatype_committed_ptr(datatype_ptr, mpi_errno);
-        if (mpi_errno)
-            MPIR_ERR_POP(mpi_errno);
+        MPIR_ERR_CHECK(mpi_errno);
     }
 
   fn_fail:
