@@ -30,9 +30,10 @@
   S*/
 typedef struct MPIR_Datatype_contents {
     int combiner;
-    int nr_ints;
-    int nr_aints;
-    int nr_types;
+    MPI_Aint nr_ints;
+    MPI_Aint nr_aints;
+    MPI_Aint nr_counts;
+    MPI_Aint nr_types;
     /* space allocated beyond structure used to store the types[],
      * ints[], and aints[], in that order.
      */
@@ -62,7 +63,7 @@ typedef struct MPIR_Datatype_contents {
   Most were not chosen because any benefit in performance or memory
   efficiency was outweighed by the added complexity of the implementation.
 
-  A number of fields contain only boolean inforation ('is_contig',
+  A number of fields contain only boolean information ('is_contig',
   'is_committed').  These
   could be combined and stored in a single bit vector.
 
@@ -133,7 +134,7 @@ struct MPIR_Datatype {
     /* handle to the backend datatype engine + some content that we
      * query from it and cache over here for performance reasons */
     struct {
-        void *handle;
+        MPIR_TYPEREP_HANDLE_TYPE handle;
         MPI_Aint num_contig_blocks;     /* contig blocks in one datatype element */
     } typerep;
 
@@ -299,7 +300,7 @@ void MPIR_Datatype_get_flattened(MPI_Datatype type, void **flattened, int *flatt
      if (datatype_ptr->free_fn) {                                           \
          mpi_errno = (datatype_ptr->free_fn)(datatype_ptr);               \
           if (mpi_errno) {                                                  \
-           MPIR_FUNC_TERSE_EXIT(MPID_STATE_MPI_TYPE_FREE);                  \
+           MPIR_FUNC_EXIT;                  \
            return MPIR_Err_return_comm(0, __func__, mpi_errno);             \
           }                                                                 \
      } */                                                                   \
@@ -351,30 +352,11 @@ void MPIR_Datatype_get_flattened(MPI_Datatype type, void **flattened, int *flatt
     }                                                               \
     } while (0)
 
-static inline void MPIR_Datatype_free_contents(MPIR_Datatype * dtp)
-{
-    int i, struct_sz = sizeof(MPIR_Datatype_contents);
-    int epsilon;
-    MPIR_Datatype *old_dtp;
-    MPI_Datatype *array_of_types;
-
-    if ((epsilon = struct_sz % MAX_ALIGNMENT)) {
-        struct_sz += MAX_ALIGNMENT - epsilon;
-    }
-
-    /* note: relies on types being first after structure */
-    array_of_types = (MPI_Datatype *) ((char *) dtp->contents + struct_sz);
-
-    for (i = 0; i < dtp->contents->nr_types; i++) {
-        if (!HANDLE_IS_BUILTIN(array_of_types[i])) {
-            MPIR_Datatype_get_ptr(array_of_types[i], old_dtp);
-            MPIR_Datatype_ptr_release(old_dtp);
-        }
-    }
-
-    MPL_free(dtp->contents);
-    dtp->contents = NULL;
-}
+#define MPIR_DATATYPE_IS_PREDEFINED(type) \
+    ((HANDLE_IS_BUILTIN((type))) || \
+     (type == MPI_FLOAT_INT) || (type == MPI_DOUBLE_INT) || \
+     (type == MPI_LONG_INT) || (type == MPI_SHORT_INT) || \
+     (type == MPI_LONG_DOUBLE_INT))
 
 /*@
   MPIR_Datatype_set_contents - store contents information for use in
@@ -384,15 +366,17 @@ static inline void MPIR_Datatype_free_contents(MPIR_Datatype * dtp)
   @*/
 static inline int MPIR_Datatype_set_contents(MPIR_Datatype * new_dtp,
                                              int combiner,
-                                             int nr_ints,
-                                             int nr_aints,
-                                             int nr_types,
+                                             MPI_Aint nr_ints,
+                                             MPI_Aint nr_aints,
+                                             MPI_Aint nr_counts,
+                                             MPI_Aint nr_types,
                                              int array_of_ints[],
                                              const MPI_Aint array_of_aints[],
+                                             const MPI_Aint array_of_counts[],
                                              const MPI_Datatype array_of_types[])
 {
-    int i, contents_size, epsilon, mpi_errno;
-    int struct_sz, ints_sz, aints_sz, types_sz;
+    int mpi_errno;
+    MPI_Aint struct_sz, ints_sz, aints_sz, counts_sz, types_sz, contents_size;
     MPIR_Datatype_contents *cp;
     MPIR_Datatype *old_dtp;
     char *ptr;
@@ -401,12 +385,15 @@ static inline int MPIR_Datatype_set_contents(MPIR_Datatype * new_dtp,
     types_sz = nr_types * sizeof(MPI_Datatype);
     ints_sz = nr_ints * sizeof(int);
     aints_sz = nr_aints * sizeof(MPI_Aint);
+    counts_sz = nr_counts * sizeof(MPI_Aint);
 
     /* pad the struct, types, and ints before we allocate.
      *
-     * note: it's not necessary that we pad the aints,
+     * note: it's not necessary that we pad the counts,
      *       because they are last in the region.
+     *       Padding it anyway for readability.
      */
+    MPI_Aint epsilon;
     if ((epsilon = struct_sz % MAX_ALIGNMENT)) {
         struct_sz += MAX_ALIGNMENT - epsilon;
     }
@@ -416,8 +403,14 @@ static inline int MPIR_Datatype_set_contents(MPIR_Datatype * new_dtp,
     if ((epsilon = ints_sz % MAX_ALIGNMENT)) {
         ints_sz += MAX_ALIGNMENT - epsilon;
     }
+    if ((epsilon = aints_sz % MAX_ALIGNMENT)) {
+        aints_sz += MAX_ALIGNMENT - epsilon;
+    }
+    if ((epsilon = counts_sz % MAX_ALIGNMENT)) {
+        counts_sz += MAX_ALIGNMENT - epsilon;
+    }
 
-    contents_size = struct_sz + types_sz + ints_sz + aints_sz;
+    contents_size = struct_sz + types_sz + ints_sz + aints_sz + counts_sz;
 
     cp = (MPIR_Datatype_contents *) MPL_malloc(contents_size, MPL_MEM_DATATYPE);
     /* --BEGIN ERROR HANDLING-- */
@@ -434,6 +427,7 @@ static inline int MPIR_Datatype_set_contents(MPIR_Datatype * new_dtp,
     cp->nr_ints = nr_ints;
     cp->nr_aints = nr_aints;
     cp->nr_types = nr_types;
+    cp->nr_counts = nr_counts;
 
     /* arrays are stored in the following order: types, ints, aints,
      * following the structure itself.
@@ -453,11 +447,17 @@ static inline int MPIR_Datatype_set_contents(MPIR_Datatype * new_dtp,
     if (nr_aints > 0) {
         MPIR_Memcpy(ptr, array_of_aints, nr_aints * sizeof(MPI_Aint));
     }
+
+    ptr = ((char *) cp) + struct_sz + types_sz + ints_sz + aints_sz;
+    if (nr_counts > 0) {
+        MPIR_Memcpy(ptr, array_of_counts, nr_counts * sizeof(MPI_Aint));
+    }
+
     new_dtp->contents = cp;
     new_dtp->flattened = NULL;
 
     /* increment reference counts on all the derived types used here */
-    for (i = 0; i < nr_types; i++) {
+    for (MPI_Aint i = 0; i < nr_types; i++) {
         if (!HANDLE_IS_BUILTIN(array_of_types[i])) {
             MPIR_Datatype_get_ptr(array_of_types[i], old_dtp);
             MPIR_Datatype_ptr_add_ref(old_dtp);
@@ -467,6 +467,76 @@ static inline int MPIR_Datatype_set_contents(MPIR_Datatype * new_dtp,
     return MPI_SUCCESS;
 }
 
+MPL_STATIC_INLINE_PREFIX void MPIR_Datatype_access_contents(MPIR_Datatype_contents * cp,
+                                                            int **p_ints,
+                                                            MPI_Aint ** p_aints,
+                                                            MPI_Aint ** p_counts,
+                                                            MPI_Datatype ** p_types)
+{
+    MPI_Aint struct_sz, ints_sz, aints_sz, counts_sz, types_sz;
+
+    struct_sz = sizeof(MPIR_Datatype_contents);
+    types_sz = cp->nr_types * sizeof(MPI_Datatype);
+    ints_sz = cp->nr_ints * sizeof(int);
+    aints_sz = cp->nr_aints * sizeof(MPI_Aint);
+    counts_sz = cp->nr_counts * sizeof(MPI_Aint);
+
+    MPI_Aint epsilon;
+    if ((epsilon = struct_sz % MAX_ALIGNMENT)) {
+        struct_sz += MAX_ALIGNMENT - epsilon;
+    }
+    if ((epsilon = types_sz % MAX_ALIGNMENT)) {
+        types_sz += MAX_ALIGNMENT - epsilon;
+    }
+    if ((epsilon = ints_sz % MAX_ALIGNMENT)) {
+        ints_sz += MAX_ALIGNMENT - epsilon;
+    }
+    if ((epsilon = ints_sz % MAX_ALIGNMENT)) {
+        aints_sz += MAX_ALIGNMENT - epsilon;
+    }
+    if ((epsilon = counts_sz % MAX_ALIGNMENT)) {
+        counts_sz += MAX_ALIGNMENT - epsilon;
+    }
+
+    *p_types = (void *) ((char *) cp + struct_sz);
+    *p_ints = (void *) ((char *) cp + struct_sz + types_sz);
+    *p_aints = (void *) ((char *) cp + struct_sz + types_sz + ints_sz);
+    *p_counts = (void *) ((char *) cp + struct_sz + types_sz + ints_sz + aints_sz);
+}
+
+MPL_STATIC_INLINE_PREFIX void MPIR_Datatype_free_contents(MPIR_Datatype * dtp)
+{
+    MPIR_Datatype_contents *cp = dtp->contents;
+
+    int *ints;
+    MPI_Aint *aints, *counts;
+    MPI_Datatype *types;
+
+    MPIR_Datatype_access_contents(cp, &ints, &aints, &counts, &types);
+
+    for (int i = 0; i < cp->nr_types; i++) {
+        if (!HANDLE_IS_BUILTIN(types[i])) {
+            MPIR_Datatype *old_dtp;
+            MPIR_Datatype_get_ptr(types[i], old_dtp);
+            MPIR_Datatype_ptr_release(old_dtp);
+        }
+    }
+
+    MPL_free(cp);
+    dtp->contents = NULL;
+}
+
+MPL_STATIC_INLINE_PREFIX int MPIR_Type_get_combiner(MPI_Datatype datatype)
+{
+    if (MPIR_DATATYPE_IS_PREDEFINED(datatype)) {
+        return MPI_COMBINER_NAMED;
+    } else {
+        MPIR_Datatype *dtp;
+        MPIR_Datatype_get_ptr(datatype, dtp);
+        return dtp->contents->combiner;
+    }
+}
+
 MPL_STATIC_INLINE_PREFIX MPI_Datatype MPIR_Datatype_predefined_get_type(uint32_t index)
 {
     MPIR_Assert(index < MPIR_DATATYPE_N_PREDEFINED);
@@ -496,85 +566,40 @@ MPL_STATIC_INLINE_PREFIX int MPIR_Datatype_predefined_get_index(MPI_Datatype dat
     return dtype_index;
 }
 
-/* contents accessor functions */
-void MPIR_Type_access_contents(MPI_Datatype type, int **ints_p, MPI_Aint ** aints_p,
-                               MPI_Datatype ** types_p);
-void MPIR_Type_release_contents(MPI_Datatype type, int **ints_p, MPI_Aint ** aints_p,
-                                MPI_Datatype ** types_p);
-
 /* This routine is used to install an attribute free routine for datatypes
    at finalize-time */
 void MPII_Datatype_attr_finalize(void);
 int MPII_Type_zerolen(MPI_Datatype * newtype);
 
-#define MPIR_DATATYPE_IS_PREDEFINED(type) \
-    ((HANDLE_IS_BUILTIN((type))) || \
-     (type == MPI_FLOAT_INT) || (type == MPI_DOUBLE_INT) || \
-     (type == MPI_LONG_INT) || (type == MPI_SHORT_INT) || \
-     (type == MPI_LONG_DOUBLE_INT))
-
 int MPIR_Get_elements_x_impl(MPI_Count * bytes, MPI_Datatype datatype, MPI_Count * elements);
-int MPIR_Status_set_elements_x_impl(MPI_Status * status, MPI_Datatype datatype, MPI_Count count);
 void MPIR_Type_get_extent_x_impl(MPI_Datatype datatype, MPI_Count * lb, MPI_Count * extent);
 void MPIR_Type_get_true_extent_x_impl(MPI_Datatype datatype, MPI_Count * true_lb,
                                       MPI_Count * true_extent);
-int MPIR_Type_size_x_impl(MPI_Datatype datatype, MPI_Count * size);
-
-void MPIR_Get_count_impl(const MPI_Status * status, MPI_Datatype datatype, MPI_Aint * count);
-int MPIR_Type_commit_impl(MPI_Datatype * datatype);
-int MPIR_Type_create_struct_impl(int count,
-                                 const int array_of_blocklengths[],
-                                 const MPI_Aint array_of_displacements[],
-                                 const MPI_Datatype array_of_types[], MPI_Datatype * newtype);
-int MPIR_Type_create_indexed_block_impl(int count,
-                                        int blocklength,
-                                        const int array_of_displacements[],
-                                        MPI_Datatype oldtype, MPI_Datatype * newtype);
-int MPIR_Type_create_hindexed_block_impl(int count, int blocklength,
-                                         const MPI_Aint array_of_displacements[],
-                                         MPI_Datatype oldtype, MPI_Datatype * newtype);
-int MPIR_Type_contiguous_impl(int count, MPI_Datatype old_type, MPI_Datatype * new_type_p);
 int MPIR_Type_contiguous_x_impl(MPI_Count count, MPI_Datatype old_type, MPI_Datatype * new_type_p);
-void MPIR_Type_get_extent_impl(MPI_Datatype datatype, MPI_Aint * lb, MPI_Aint * extent);
-void MPIR_Type_get_true_extent_impl(MPI_Datatype datatype, MPI_Aint * true_lb,
-                                    MPI_Aint * true_extent);
-void MPIR_Type_get_envelope(MPI_Datatype datatype, int *num_integers, int *num_addresses,
-                            int *num_datatypes, int *combiner);
-int MPIR_Type_hvector_impl(int count, int blocklen, MPI_Aint stride, MPI_Datatype old_type,
-                           MPI_Datatype * newtype_p);
-int MPIR_Type_indexed_impl(int count, const int blocklens[], const int indices[],
-                           MPI_Datatype old_type, MPI_Datatype * newtype);
 void MPIR_Type_free_impl(MPI_Datatype * datatype);
-int MPIR_Type_vector_impl(int count, int blocklength, int stride, MPI_Datatype old_type,
-                          MPI_Datatype * newtype_p);
-int MPIR_Type_struct_impl(int count, const int blocklens[], const MPI_Aint indices[],
-                          const MPI_Datatype old_types[], MPI_Datatype * newtype);
-void MPIR_Pack_size_impl(int incount, MPI_Datatype datatype, MPI_Aint * size);
+void MPIR_Pack_size(MPI_Aint incount, MPI_Datatype datatype, MPI_Aint * size);
 void MPIR_Type_lb_impl(MPI_Datatype datatype, MPI_Aint * displacement);
 
 /* Datatype functions */
-int MPIR_Type_dup(MPI_Datatype oldtype, MPI_Datatype * newtype);
-int MPIR_Type_struct(int count, const int *blocklength_array, const MPI_Aint * displacement_array,
-                     const MPI_Datatype * oldtype_array, MPI_Datatype * newtype);
-int MPIR_Type_indexed(int count, const int *blocklength_array, const void *displacement_array,
-                      int dispinbytes, MPI_Datatype oldtype, MPI_Datatype * newtype);
-int MPIR_Type_vector(int count, int blocklength, MPI_Aint stride, int strideinbytes,
-                     MPI_Datatype oldtype, MPI_Datatype * newtype);
-int MPIR_Type_contiguous(int count, MPI_Datatype oldtype, MPI_Datatype * newtype);
-int MPIR_Type_blockindexed(int count,
-                           int blocklength,
-                           const void *displacement_array,
-                           int dispinbytes, MPI_Datatype oldtype, MPI_Datatype * newtype);
-
-int MPIR_Type_commit(MPI_Datatype * type);
 int MPII_Type_zerolen(MPI_Datatype * newtype);
-int MPIR_Type_create_resized(MPI_Datatype oldtype, MPI_Aint lb, MPI_Aint extent,
-                             MPI_Datatype * newtype);
-int MPIR_Type_get_contents(MPI_Datatype datatype, int max_integers, int max_addresses,
-                           int max_datatypes, int array_of_integers[],
-                           MPI_Aint array_of_addresses[], MPI_Datatype array_of_datatypes[]);
 int MPIR_Type_create_pairtype(MPI_Datatype datatype, MPIR_Datatype * new_dtp);
 
+int MPIR_Type_contiguous(MPI_Aint count, MPI_Datatype oldtype, MPI_Datatype * newtype);
+int MPIR_Type_vector(MPI_Aint count, MPI_Aint blocklength, MPI_Aint stride, bool strideinbytes,
+                     MPI_Datatype oldtype, MPI_Datatype * newtype);
+int MPIR_Type_blockindexed(MPI_Aint count, MPI_Aint blocklength,
+                           const MPI_Aint displacement_array[],
+                           bool dispinbytes, MPI_Datatype oldtype, MPI_Datatype * newtype);
+int MPIR_Type_indexed(MPI_Aint count, const MPI_Aint * blocklength_array,
+                      const MPI_Aint * displacement_array,
+                      bool dispinbytes, MPI_Datatype oldtype, MPI_Datatype * newtype);
+int MPIR_Type_struct(MPI_Aint count, const MPI_Aint * blocklength_array,
+                     const MPI_Aint * displacement_array,
+                     const MPI_Datatype * oldtype_array, MPI_Datatype * newtype);
+int MPIR_Type_create_resized(MPI_Datatype oldtype, MPI_Aint lb, MPI_Aint extent,
+                             MPI_Datatype * new_type);
+int MPIR_Type_dup(MPI_Datatype oldtype, MPI_Datatype * newtype);
+
 /* debugging helper functions */
 char *MPIR_Datatype_builtin_to_string(MPI_Datatype type);
 char *MPIR_Datatype_combiner_to_string(int combiner);
@@ -582,12 +607,12 @@ void MPIR_Datatype_debug(MPI_Datatype type, int array_ct);
 
 MPI_Aint MPII_Datatype_indexed_count_contig(MPI_Aint count,
                                             const MPI_Aint * blocklength_array,
-                                            const void *displacement_array,
+                                            const MPI_Aint * displacement_array,
                                             int dispinbytes, MPI_Aint old_extent);
 
 MPI_Aint MPII_Datatype_blockindexed_count_contig(MPI_Aint count,
                                                  MPI_Aint blklen,
-                                                 const void *disp_array,
+                                                 const MPI_Aint disp_array[],
                                                  int dispinbytes, MPI_Aint old_extent);
 
 #endif /* MPIR_DATATYPE_H_INCLUDED */
