@@ -1,11 +1,6 @@
 /*
- *  (C) 2006 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
- *
- *  Portions of this code were written by Intel Corporation.
- *  Copyright (C) 2011-2018 Intel Corporation.  Intel provides this material
- *  to Argonne National Laboratory subject to Software Grant and Corporate
- *  Contributor License Agreement dated February 8, 2012.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "mpiimpl.h"
@@ -45,10 +40,6 @@ int MPII_Recexchalgo_comm_cleanup(MPIR_Comm * comm)
  * participating in Step 2. In Step 3, the ranks that participated in Step 2 send
  * the final data to non-partcipating ranks.
 */
-#undef FUNCNAME
-#define FUNCNAME MPII_Recexchalgo_get_neighbors
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPII_Recexchalgo_get_neighbors(int rank, int nranks, int *k_,
                                    int *step1_sendto, int **step1_recvfrom_, int *step1_nrecvs,
                                    int ***step2_nbrs_, int *step2_nphases, int *p_of_k_, int *T_)
@@ -59,8 +50,7 @@ int MPII_Recexchalgo_get_neighbors(int rank, int nranks, int *k_,
     int **step2_nbrs;
     int *step1_recvfrom;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPII_RECEXCHALGO_GET_NEIGHBORS);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPII_RECEXCHALGO_GET_NEIGHBORS);
+    MPIR_FUNC_ENTER;
 
     k = *k_;
     if (nranks < k)     /* If size of the communicator is less than k, reduce the value of k */
@@ -139,21 +129,23 @@ int MPII_Recexchalgo_get_neighbors(int rank, int nranks, int *k_,
     MPL_DBG_MSG_FMT(MPIR_DBG_COLL, VERBOSE, (MPL_DBG_FDEST, "step 1 nbr computation completed"));
 
     /* Step 2 */
-    if (*step1_sendto == -1) {  /* calulate step2_nbrs only for participating ranks */
+    if (*step1_sendto == -1) {  /* calculate step2_nbrs only for participating ranks */
         int *digit = (int *) MPL_malloc(sizeof(int) * log_p_of_k, MPL_MEM_COLL);
         MPIR_Assert(digit != NULL);
-        int temprank = newrank, index = 0, remainder;
+        int temprank = newrank;
         int mask = 0x1;
         int phase = 0, cbit, cnt, nbr, power;
 
         /* calculate the digits in base k representation of newrank */
         for (i = 0; i < log_p_of_k; i++)
             digit[i] = 0;
+
+        int remainder, i_digit = 0;
         while (temprank != 0) {
             remainder = temprank % k;
             temprank = temprank / k;
-            digit[index] = remainder;
-            index++;
+            digit[i_digit] = remainder;
+            i_digit++;
         }
 
         while (mask < p_of_k) {
@@ -190,45 +182,35 @@ int MPII_Recexchalgo_get_neighbors(int rank, int nranks, int *k_,
         MPL_free(digit);
     }
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPII_RECEXCHALGO_GET_NEIGHBORS);
+    MPIR_FUNC_EXIT;
 
     return mpi_errno;
 }
 
 
-#undef FUNCNAME
-#define FUNCNAME MPII_Recexchalgo_origrank_to_step2rank
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPII_Recexchalgo_origrank_to_step2rank(int rank, int rem, int T, int k)
 {
     int step2rank;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPII_RECEXCHALGO_ORGRANK_TO_STEP2RANK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPII_RECEXCHALGO_ORGRANK_TO_STEP2RANK);
+    MPIR_FUNC_ENTER;
 
     step2rank = (rank < T) ? rank / k : rank - rem;
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPII_RECEXCHALGO_ORGRANK_TO_STEP2RANK);
+    MPIR_FUNC_EXIT;
 
     return step2rank;
 }
 
 
-#undef FUNCNAME
-#define FUNCNAME MPII_Recexchalgo_step2rank_to_origrank
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPII_Recexchalgo_step2rank_to_origrank(int rank, int rem, int T, int k)
 {
     int orig_rank;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPII_RECEXCHALGO_STEP2RANK_TO_ORGRANK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPII_RECEXCHALGO_STEP2RANK_TO_ORGRANK);
+    MPIR_FUNC_ENTER;
 
     orig_rank = (rank < rem / (k - 1)) ? (rank * k) + (k - 1) : rank + rem;
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPII_RECEXCHALGO_STEP2RANK_TO_ORGRANK);
+    MPIR_FUNC_EXIT;
 
     return orig_rank;
 }
@@ -238,10 +220,6 @@ int MPII_Recexchalgo_step2rank_to_origrank(int rank, int rem, int T, int k)
  * phase in recursive exchange algorithms in collective operations like Allgather,
  * Allgatherv, Reducescatter.
 */
-#undef FUNCNAME
-#define FUNCNAME MPII_Recexchalgo_get_count_and_offset
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPII_Recexchalgo_get_count_and_offset(int rank, int phase, int k, int nranks, int *count,
                                           int *offset)
 {
@@ -250,8 +228,7 @@ int MPII_Recexchalgo_get_count_and_offset(int rank, int phase, int k, int nranks
     int k_power_phase = 1;
     int p_of_k = 1, rem, T;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPII_RECEXCHALGO_GET_COUNT_AND_OFFSET);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPII_RECEXCHALGO_GET_COUNT_AND_OFFSET);
+    MPIR_FUNC_ENTER;
 
     /* p_of_k is the largest power of k that is less than nranks */
     while (p_of_k <= nranks) {
@@ -279,7 +256,7 @@ int MPII_Recexchalgo_get_count_and_offset(int rank, int phase, int k, int nranks
     *count = orig_max - orig_min;
     *offset = orig_min + 1;
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPII_RECEXCHALGO_GET_COUNT_AND_OFFSET);
+    MPIR_FUNC_EXIT;
 
     return mpi_errno;
 }
@@ -290,26 +267,21 @@ int MPII_Recexchalgo_get_count_and_offset(int rank, int phase, int k, int nranks
  * 2. Calculates the digit reversed (in base 'k' representation) rank of the Step 2 rank
  * 3. Convert the digit reversed rank in the previous Step to the original rank.
 */
-#undef FUNCNAME
-#define FUNCNAME MPII_Recexchalgo_reverse_digits_step2
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPII_Recexchalgo_reverse_digits_step2(int rank, int comm_size, int k)
 {
     int i, T, rem, power, step2rank, step2_reverse_rank = 0;
     int pofk = 1, log_pofk = 0;
     int *digit, *digit_reverse;
-    int remainder, index = 0;
-    int mpi_errno = MPI_SUCCESS;
+    int mpi_errno ATTRIBUTE((unused)) = MPI_SUCCESS;
     MPIR_CHKLMEM_DECL(2);
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPII_RECEXCHALGO_REVERSE_DIGITS);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPII_RECEXCHALGO_REVERSE_DIGITS);
+    MPIR_FUNC_ENTER;
 
     while (pofk <= comm_size) {
         pofk *= k;
         log_pofk++;
     }
+    MPIR_Assert(log_pofk > 0);
     pofk /= k;
     log_pofk--;
 
@@ -326,11 +298,13 @@ int MPII_Recexchalgo_reverse_digits_step2(int rank, int comm_size, int k)
                         mpi_errno, "digit_reverse buffer", MPL_MEM_COLL);
     for (i = 0; i < log_pofk; i++)
         digit[i] = 0;
+
+    int remainder, i_digit = 0;
     while (step2rank != 0) {
         remainder = step2rank % k;
         step2rank = step2rank / k;
-        digit[index] = remainder;
-        index++;
+        digit[i_digit] = remainder;
+        i_digit++;
     }
 
     /* reverse the number in base k representation to get the step2_reverse_rank
@@ -351,7 +325,7 @@ int MPII_Recexchalgo_reverse_digits_step2(int rank, int comm_size, int k)
     MPL_DBG_MSG_FMT(MPIR_DBG_COLL, VERBOSE,
                     (MPL_DBG_FDEST, "reverse_rank is %d", step2_reverse_rank));
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPII_RECEXCHALGO_REVERSE_DIGITS);
+    MPIR_FUNC_EXIT;
 
   fn_exit:
     MPIR_CHKLMEM_FREEALL();
