@@ -15,10 +15,10 @@
  * where n is total size of data gathered on each process.
  */
 int MPIR_Allgather_intra_brucks(const void *sendbuf,
-                                int sendcount,
+                                MPI_Aint sendcount,
                                 MPI_Datatype sendtype,
                                 void *recvbuf,
-                                int recvcount,
+                                MPI_Aint recvcount,
                                 MPI_Datatype recvtype,
                                 MPIR_Comm * comm_ptr, MPIR_Errflag_t * errflag)
 {
