@@ -23,10 +23,10 @@
  * other processes.
  */
 int MPIR_Alltoall_intra_pairwise(const void *sendbuf,
-                                 int sendcount,
+                                 MPI_Aint sendcount,
                                  MPI_Datatype sendtype,
                                  void *recvbuf,
-                                 int recvcount,
+                                 MPI_Aint recvcount,
                                  MPI_Datatype recvtype,
                                  MPIR_Comm * comm_ptr, MPIR_Errflag_t * errflag)
 {
@@ -36,9 +36,6 @@ int MPIR_Alltoall_intra_pairwise(const void *sendbuf,
     int mpi_errno_ret = MPI_SUCCESS;
     MPI_Status status;
 
-    if (recvcount == 0)
-        return MPI_SUCCESS;
-
     comm_size = comm_ptr->local_size;
     rank = comm_ptr->rank;
 
