@@ -6,10 +6,6 @@
 #ifndef COLL_TYPES_H_INCLUDED
 #define COLL_TYPES_H_INCLUDED
 
-/* include types from transports */
-#include "tsp_stubtran_types.h"
-#include "tsp_gentran_types.h"
-
 /* Type definitions from all algorithms */
 #include "stubalgo_types.h"
 #include "treealgo_types.h"
