@@ -5,7 +5,7 @@
 
 #include "mpiimpl.h"
 
-int MPIR_Reduce_intra_smp(const void *sendbuf, void *recvbuf, int count,
+int MPIR_Reduce_intra_smp(const void *sendbuf, void *recvbuf, MPI_Aint count,
                           MPI_Datatype datatype, MPI_Op op, int root, MPIR_Comm * comm_ptr,
                           MPIR_Errflag_t * errflag)
 {
