@@ -5,23 +5,21 @@
 
 #include "mpiimpl.h"
 
-int MPIR_Reduce_scatter_block_allcomm_nb(const void *sendbuf, void *recvbuf, int recvcount,
+int MPIR_Reduce_scatter_block_allcomm_nb(const void *sendbuf, void *recvbuf, MPI_Aint recvcount,
                                          MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm_ptr,
                                          MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPI_Request req = MPI_REQUEST_NULL;
     MPIR_Request *req_ptr = NULL;
 
     /* just call the nonblocking version and wait on it */
     mpi_errno =
         MPIR_Ireduce_scatter_block(sendbuf, recvbuf, recvcount, datatype, op, comm_ptr, &req_ptr);
     MPIR_ERR_CHECK(mpi_errno);
-    if (req_ptr)
-        req = req_ptr->handle;
 
-    mpi_errno = MPIR_Wait(&req, MPI_STATUS_IGNORE);
+    mpi_errno = MPIC_Wait(req_ptr, errflag);
     MPIR_ERR_CHECK(mpi_errno);
+    MPIR_Request_free(req_ptr);
 
   fn_exit:
     return mpi_errno;
