@@ -6,7 +6,7 @@
 #include "mpiimpl.h"
 
 
-int MPIR_Scan_intra_smp(const void *sendbuf, void *recvbuf, int count,
+int MPIR_Scan_intra_smp(const void *sendbuf, void *recvbuf, MPI_Aint count,
                         MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm_ptr,
                         MPIR_Errflag_t * errflag)
 {
@@ -140,7 +140,7 @@ int MPIR_Scan_intra_smp(const void *sendbuf, void *recvbuf, int count,
      * scan result. for example, to node 3, it will have reduce result
      * of rank 1,2,3,4,5,6 in tempbuf.
      * then we should broadcast this result in the local node, and
-     * reduce it with recvbuf to get final result if nessesary. */
+     * reduce it with recvbuf to get final result if necessary. */
 
     if (comm_ptr->node_comm != NULL) {
         mpi_errno = MPIR_Bcast(&noneed, 1, MPI_INT, 0, comm_ptr->node_comm, errflag);
