@@ -5,13 +5,12 @@
 
 #include "mpiimpl.h"
 
-int MPIR_Scatterv_allcomm_nb(const void *sendbuf, const int *sendcounts, const int *displs,
-                             MPI_Datatype sendtype, void *recvbuf, int recvcount,
-                             MPI_Datatype recvtype, int root, MPIR_Comm * comm_ptr,
-                             MPIR_Errflag_t * errflag)
+int MPIR_Scatterv_allcomm_nb(const void *sendbuf, const MPI_Aint * sendcounts,
+                             const MPI_Aint * displs, MPI_Datatype sendtype, void *recvbuf,
+                             MPI_Aint recvcount, MPI_Datatype recvtype, int root,
+                             MPIR_Comm * comm_ptr, MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPI_Request req = MPI_REQUEST_NULL;
     MPIR_Request *req_ptr = NULL;
 
     /* just call the nonblocking version and wait on it */
@@ -19,11 +18,10 @@ int MPIR_Scatterv_allcomm_nb(const void *sendbuf, const int *sendcounts, const i
         MPIR_Iscatterv(sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root,
                        comm_ptr, &req_ptr);
     MPIR_ERR_CHECK(mpi_errno);
-    if (req_ptr)
-        req = req_ptr->handle;
 
-    mpi_errno = MPIR_Wait(&req, MPI_STATUS_IGNORE);
+    mpi_errno = MPIC_Wait(req_ptr, errflag);
     MPIR_ERR_CHECK(mpi_errno);
+    MPIR_Request_free(req_ptr);
 
   fn_exit:
     return mpi_errno;
