@@ -1,13 +1,10 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *
- *  (C) 2001 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "mpiimpl.h"
-#include "errcodes.h"
-
+#include "uthash.h"
 #include <string.h>
 
 /*
@@ -50,8 +47,23 @@ static const char *(user_class_msgs[ERROR_MAX_NCLASS]) = {
 static const char *(user_code_msgs[ERROR_MAX_NCODE]) = {
 0};
 
-static int first_free_class = 1;        /* class 0 is reserved */
-static int first_free_code = 1; /* code 0 is reserved */
+/* a container for integer objects that can be used as a linked list
+ * or as a hashmap */
+struct intcnt {
+    int val;
+
+    struct intcnt *next;
+    struct intcnt *prev;
+
+    UT_hash_handle hh;
+};
+
+static struct {
+    int next;
+    struct intcnt *free;
+    struct intcnt *used;
+} err_class, err_code;
+
 static const char empty_error_string[1] = { 0 };
 
 /* Forward reference */
@@ -82,6 +94,13 @@ static void MPIR_Init_err_dyncodes(void)
     /* FIXME: Does this need a thread-safe init? */
     not_initialized = 0;
 
+    err_class.next = 1; /* class 0 is reserved */
+    err_class.free = NULL;
+    err_class.used = NULL;
+    err_code.next = 1;  /* code 0 is reserved */
+    err_code.free = NULL;
+    err_code.used = NULL;
+
     for (i = 0; i < ERROR_MAX_NCLASS; i++) {
         user_class_msgs[i] = 0;
     }
@@ -96,21 +115,13 @@ static void MPIR_Init_err_dyncodes(void)
     MPIR_Add_finalize(MPIR_Dynerrcodes_finalize, (void *) 0, 9);
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIR_Err_set_msg
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /*
-  MPIR_Err_set_msg - Change the message for an error code or class
-
-Input Parameters:
-+ code - Error code or class
-- msg  - New message to use
+  MPIR_Add_error_string_impl - Change the message for an error code or class
 
   Notes:
   This routine is needed to implement 'MPI_Add_error_string'.
 */
-int MPIR_Err_set_msg(int code, const char *msg_string)
+int MPIR_Add_error_string_impl(int code, const char *msg_string)
 {
     int errcode, errclass;
     size_t msg_len;
@@ -123,7 +134,7 @@ int MPIR_Err_set_msg(int code, const char *msg_string)
          * an error (see MPI_Add_error_string in the standard) */
         MPIR_Init_err_dyncodes();
         return MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
-                                    "MPIR_Err_set_msg", __LINE__,
+                                    __func__, __LINE__,
                                     MPI_ERR_ARG, "**argerrcode", "**argerrcode %d", code);
     }
     /* --END ERROR HANDLING-- */
@@ -137,7 +148,7 @@ int MPIR_Err_set_msg(int code, const char *msg_string)
     if (code & ~(ERROR_CLASS_MASK | ERROR_DYN_MASK | ERROR_GENERIC_MASK)) {
         /* Check for invalid error code */
         return MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
-                                    FCNAME, __LINE__,
+                                    __func__, __LINE__,
                                     MPI_ERR_ARG, "**argerrcode", "**argerrcode %d", code);
     }
     /* --END ERROR HANDLING-- */
@@ -148,28 +159,29 @@ int MPIR_Err_set_msg(int code, const char *msg_string)
     /* --BEGIN ERROR HANDLING-- */
     if (!str) {
         return MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
-                                    FCNAME, __LINE__, MPI_ERR_OTHER,
+                                    __func__, __LINE__, MPI_ERR_OTHER,
                                     "**nomem", "**nomem %s %d", "error message string", msg_len);
     }
     /* --END ERROR HANDLING-- */
 
     /* --------------------------------------------------------------------- */
     MPL_strncpy(str, msg_string, msg_len + 1);
+
     if (errcode) {
-        if (errcode < first_free_code) {
-            if (user_code_msgs[errcode]) {
-                MPL_free((void *) (user_code_msgs[errcode]));
-            }
+        struct intcnt *s;
+        HASH_FIND_INT(err_code.used, &errcode, s);
+        if (s) {
+            MPL_free((void *) (user_code_msgs[errcode]));
             user_code_msgs[errcode] = (const char *) str;
         } else {
             /* FIXME : Unallocated error code? */
             MPL_free(str);
         }
     } else {
-        if (errclass < first_free_class) {
-            if (user_class_msgs[errclass]) {
-                MPL_free((void *) (user_class_msgs[errclass]));
-            }
+        struct intcnt *s;
+        HASH_FIND_INT(err_class.used, &errclass, s);
+        if (s) {
+            MPL_free((void *) (user_class_msgs[errclass]));
             user_class_msgs[errclass] = (const char *) str;
         } else {
             /* FIXME : Unallocated error code? */
@@ -181,10 +193,51 @@ int MPIR_Err_set_msg(int code, const char *msg_string)
 }
 
 /*
-  MPIR_Err_add_class - Creata a new error class
+  MPIR_Delete_error_string_impl - Delete the string associated with an error code or class
 
-  Return value:
-  An error class.  Returns -1 if no more classes are available.
+  Notes:
+  This is used to implement 'MPI_Delete_error_string'; it may also be used by a
+  device to delete device-specific error strings.
+*/
+int MPIR_Delete_error_string_impl(int code)
+{
+    int mpi_errno = MPI_SUCCESS;
+    int errcode = (int) (((unsigned int) code & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT);
+    int errclass = code & ERROR_CLASS_MASK;
+
+    if (not_initialized)
+        MPIR_Init_err_dyncodes();
+
+    if (errcode) {
+        struct intcnt *s;
+        HASH_FIND_INT(err_code.used, &errcode, s);
+        if (s) {
+            MPL_free((void *) (user_code_msgs[errcode]));
+            user_code_msgs[errcode] = NULL;
+        } else {
+            mpi_errno = MPI_ERR_OTHER;
+            goto fn_fail;
+        }
+    } else {
+        struct intcnt *s;
+        HASH_FIND_INT(err_class.used, &errclass, s);
+        if (s) {
+            MPL_free((void *) (user_class_msgs[errclass]));
+            user_class_msgs[errclass] = NULL;
+        } else {
+            mpi_errno = MPI_ERR_OTHER;
+            goto fn_fail;
+        }
+    }
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
+
+/*
+  MPIR_Add_error_class_impl - Creata a new error class
 
   Notes:
   This is used to implement 'MPI_Add_error_class'; it may also be used by a
@@ -196,48 +249,96 @@ int MPIR_Err_set_msg(int code, const char *msg_string)
 
   This routine should be run within a SINGLE_CS in the multithreaded case.
 */
-int MPIR_Err_add_class(void)
+int MPIR_Add_error_class_impl(int *errorclass)
 {
+    int mpi_errno = MPI_SUCCESS;
     int new_class;
 
     if (not_initialized)
         MPIR_Init_err_dyncodes();
 
     /* Get new class */
-    new_class = first_free_class;
-    ++first_free_class;
-
-    /* --BEGIN ERROR HANDLING-- */
-    if (new_class >= ERROR_MAX_NCLASS) {
-        /* Fail if out of classes */
-        return -1;
+    struct intcnt *s;
+    if (err_class.free) {
+        s = err_class.free;
+        DL_DELETE(err_class.free, s);
+        HASH_ADD_INT(err_class.used, val, s, MPL_MEM_BUFFER);
+    } else {
+        s = (struct intcnt *) MPL_malloc(sizeof(struct intcnt), MPL_MEM_BUFFER);
+        s->val = err_class.next++;
+        HASH_ADD_INT(err_class.used, val, s, MPL_MEM_BUFFER);
     }
-    /* --END ERROR HANDLING-- */
+    new_class = s->val;
+
+    /* Fail if out of classes */
+    MPIR_ERR_CHKANDJUMP(new_class >= ERROR_MAX_NCLASS, mpi_errno, MPI_ERR_OTHER, "**noerrclasses");
 
     /* Note that the MPI interface always adds an error class without
      * a string.  */
     user_class_msgs[new_class] = 0;
 
-    return (new_class | ERROR_DYN_MASK);
+    new_class |= ERROR_DYN_MASK;
+
+    if (new_class > MPIR_Process.attrs.lastusedcode) {
+        MPIR_Process.attrs.lastusedcode = new_class;
+    }
+
+    *errorclass = new_class;
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
 }
 
 /*
-  MPIR_Err_add_code - Create a new error code that is associated with an
+  MPIR_Delete_error_class_impl - Delete an error class
+
+  Notes:
+  This is used to implement 'MPI_Delete_error_class'; it may also be used by a
+  device to delete device-specific error classes.
+
+  Predefined classes are handled directly; this routine is not used to
+  delete the predefined MPI error classes.  This is done to reduce the
+  number of steps that must be executed when starting an MPI program.
+
+  This routine should be run within a SINGLE_CS in the multithreaded case.
+*/
+int MPIR_Delete_error_class_impl(int user_errclass)
+{
+    int mpi_errno = MPI_SUCCESS;
+
+    if (not_initialized)
+        MPIR_Init_err_dyncodes();
+
+    int errclass = user_errclass & ~ERROR_DYN_MASK;
+    struct intcnt *s;
+    HASH_FIND_INT(err_class.used, &errclass, s);
+
+    MPIR_ERR_CHKANDJUMP(s == NULL, mpi_errno, MPI_ERR_OTHER, "**predeferrclass");
+
+    HASH_DEL(err_class.used, s);
+    DL_APPEND(err_class.free, s);
+    MPL_free((char *) user_class_msgs[s->val]);
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
+
+/*
+  MPIR_Add_error_code_impl - Create a new error code that is associated with an
   existing error class
 
-Input Parameters:
-. class - Error class to which the code belongs.
-
-  Return value:
-  An error code.
-
   Notes:
   This is used to implement 'MPI_Add_error_code'; it may also be used by a
   device to add device-specific error codes.
 
   */
-int MPIR_Err_add_code(int class)
+int MPIR_Add_error_code_impl(int class, int *code)
 {
+    int mpi_errno = MPI_SUCCESS;
     int new_code;
 
     /* Note that we can add codes to existing classes, so we may
@@ -246,22 +347,62 @@ int MPIR_Err_add_code(int class)
         MPIR_Init_err_dyncodes();
 
     /* Get the new code */
-    new_code = first_free_code;
-    ++first_free_code;
-
-    /* --BEGIN ERROR HANDLING-- */
-    if (new_code >= ERROR_MAX_NCODE) {
-        /* Fail if out of codes */
-        return -1;
+    struct intcnt *s;
+    if (err_code.free) {
+        s = err_code.free;
+        DL_DELETE(err_code.free, s);
+        HASH_ADD_INT(err_code.used, val, s, MPL_MEM_BUFFER);
+    } else {
+        s = (struct intcnt *) MPL_malloc(sizeof(struct intcnt), MPL_MEM_BUFFER);
+        s->val = err_code.next++;
+        HASH_ADD_INT(err_code.used, val, s, MPL_MEM_BUFFER);
     }
-    /* --END ERROR HANDLING-- */
+    new_code = s->val;
+
+    /* Fail if out of codes */
+    MPIR_ERR_CHKANDJUMP(new_code >= ERROR_MAX_NCODE, mpi_errno, MPI_ERR_OTHER, "**noerrcodes");
 
     /* Create the full error code */
     new_code = class | (new_code << ERROR_GENERIC_SHIFT);
 
     /* FIXME: For robustness, we should make sure that the associated string
      * is initialized to null */
-    return new_code;
+    *code = new_code;
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
+
+/*
+  MPIR_Delete_error_code_impl - Delete an error code
+
+  Notes:
+  This is used to implement 'MPI_Delete_error_code'; it may also be used by a
+  device to delete device-specific error codes.
+*/
+int MPIR_Delete_error_code_impl(int code)
+{
+    int mpi_errno = MPI_SUCCESS;
+    int errcode = (int) (((unsigned int) code & ERROR_GENERIC_MASK) >> ERROR_GENERIC_SHIFT);
+
+    if (not_initialized)
+        MPIR_Init_err_dyncodes();
+
+    struct intcnt *s;
+    HASH_FIND_INT(err_code.used, &errcode, s);
+
+    MPIR_ERR_CHKANDJUMP(s == NULL, mpi_errno, MPI_ERR_OTHER, "**predeferrcode");
+
+    HASH_DEL(err_code.used, s);
+    DL_APPEND(err_code.free, s);
+    MPL_free((char *) user_code_msgs[s->val]);
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
 }
 
 /*
@@ -270,7 +411,7 @@ int MPIR_Err_add_code(int class)
 
 Input Parameters:
 + code - An error class or code.  If a code, it must have been created by
-  'MPIR_Err_create_code'.
+  'MPIR_Add_error_code_impl'.
 
   Return value:
   A pointer to a null-terminated text string with the corresponding error
@@ -297,13 +438,17 @@ static const char *get_dynerr_string(int code)
     }
 
     if (errcode) {
-        if (errcode < first_free_code) {
+        struct intcnt *s;
+        HASH_FIND_INT(err_code.used, &errcode, s);
+        if (s) {
             errstr = user_code_msgs[errcode];
             if (!errstr)
                 errstr = empty_error_string;
         }
     } else {
-        if (errclass < first_free_class) {
+        struct intcnt *s;
+        HASH_FIND_INT(err_class.used, &errclass, s);
+        if (s) {
             errstr = user_class_msgs[errclass];
             if (!errstr)
                 errstr = empty_error_string;
@@ -316,20 +461,33 @@ static const char *get_dynerr_string(int code)
 
 static int MPIR_Dynerrcodes_finalize(void *p ATTRIBUTE((unused)))
 {
-    int i;
-
     MPL_UNREFERENCED_ARG(p);
 
     if (not_initialized == 0) {
+        struct intcnt *s, *tmp;
 
-        for (i = 0; i < first_free_class; i++) {
-            if (user_class_msgs[i])
-                MPL_free((char *) user_class_msgs[i]);
+        HASH_ITER(hh, err_class.used, s, tmp) {
+            MPL_free((char *) user_class_msgs[s->val]);
+            HASH_DEL(err_class.used, s);
+            MPL_free(s);
         }
 
-        for (i = 0; i < first_free_code; i++) {
-            if (user_code_msgs[i])
-                MPL_free((char *) user_code_msgs[i]);
+        DL_FOREACH_SAFE(err_class.free, s, tmp) {
+            MPL_free((char *) user_class_msgs[s->val]);
+            DL_DELETE(err_class.free, s);
+            MPL_free(s);
+        }
+
+        HASH_ITER(hh, err_code.used, s, tmp) {
+            MPL_free((char *) user_code_msgs[s->val]);
+            HASH_DEL(err_code.used, s);
+            MPL_free(s);
+        }
+
+        DL_FOREACH_SAFE(err_code.free, s, tmp) {
+            MPL_free((char *) user_code_msgs[s->val]);
+            DL_DELETE(err_code.free, s);
+            MPL_free(s);
         }
     }
     return 0;
