@@ -1,12 +1,9 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2001 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "mpiimpl.h"
-#include "mpii_bsend.h"
-#include "bsendutil.h"
 
 /*
  * Miscellaneous comments
@@ -55,12 +52,12 @@ static void MPIR_Bsend_dump(void);
  * taking advantage of the "alignpad"), but this would require more changes.
  */
 static struct BsendBuffer {
-    void *buffer;               /* Pointer to the begining of the user-
+    void *buffer;               /* Pointer to the beginning of the user-
                                  * provided buffer */
-    size_t buffer_size;         /* Size of the user-provided buffer */
+    MPI_Aint buffer_size;       /* Size of the user-provided buffer */
     void *origbuffer;           /* Pointer to the buffer provided by
                                  * the user */
-    size_t origbuffer_size;     /* Size of the buffer as provided
+    MPI_Aint origbuffer_size;   /* Size of the buffer as provided
                                  * by the user */
     MPII_Bsend_data_t *avail;   /* Pointer to the first available block
                                  * of space */
@@ -88,11 +85,7 @@ static void MPIR_Bsend_free_segment(MPII_Bsend_data_t *);
  * Attach a buffer.  This checks for the error conditions and then
  * initialized the avail buffer.
  */
-#undef FUNCNAME
-#define FUNCNAME MPIR_Bsend_attach
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
-int MPIR_Bsend_attach(void *buffer, int buffer_size)
+int MPIR_Bsend_attach(void *buffer, MPI_Aint buffer_size)
 {
     MPII_Bsend_data_t *p;
     size_t offset, align_sz;
@@ -122,6 +115,7 @@ int MPIR_Bsend_attach(void *buffer, int buffer_size)
     }
 #endif /* HAVE_ERROR_CHECKING */
 
+    MPID_THREAD_CS_ENTER(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
     if (!initialized) {
         initialized = 1;
         MPIR_Add_finalize(MPIR_Bsend_finalize, (void *) 0, 10);
@@ -155,6 +149,7 @@ int MPIR_Bsend_attach(void *buffer, int buffer_size)
     p->next = p->prev = NULL;
     p->msg.msgbuf = (char *) p + BSENDDATA_HEADER_TRUE_SIZE;
 
+    MPID_THREAD_CS_EXIT(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
     return MPI_SUCCESS;
 }
 
@@ -164,19 +159,17 @@ int MPIR_Bsend_attach(void *buffer, int buffer_size)
  * argument as an "int" (the definition predates that of ssize_t as a
  * standard type).
  */
-#undef FUNCNAME
-#define FUNCNAME MPIR_Bsend_detach
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
-int MPIR_Bsend_detach(void *bufferp, int *size)
+int MPIR_Bsend_detach(void *bufferp, MPI_Aint * size)
 {
     int mpi_errno = MPI_SUCCESS;
 
+    MPID_THREAD_CS_ENTER(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
     if (BsendBuffer.pending) {
         /* FIXME: Process pending bsend requests in detach */
-        return MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
-                                    "MPIR_Bsend_detach", __LINE__, MPI_ERR_OTHER, "**bsendpending",
-                                    0);
+        mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
+                                         "MPIR_Bsend_detach", __LINE__, MPI_ERR_OTHER,
+                                         "**bsendpending", 0);
+        goto fn_fail;
     }
     if (BsendBuffer.active) {
         /* Loop through each active element and wait on it */
@@ -185,17 +178,14 @@ int MPIR_Bsend_detach(void *bufferp, int *size)
         while (p) {
             MPI_Request r = p->request->handle;
             mpi_errno = MPIR_Wait(&r, MPI_STATUS_IGNORE);
-            if (mpi_errno)
-                MPIR_ERR_POP(mpi_errno);
+            MPIR_ERR_CHECK(mpi_errno);
             p = p->next;
         }
     }
 
 /* Note that this works even when the buffer does not exist */
     *(void **) bufferp = BsendBuffer.origbuffer;
-    /* This cast to int will work because the user must use an int to describe
-     * the buffer size */
-    *size = (int) BsendBuffer.origbuffer_size;
+    *size = (MPI_Aint) BsendBuffer.origbuffer_size;
     BsendBuffer.origbuffer = NULL;
     BsendBuffer.origbuffer_size = 0;
     BsendBuffer.buffer = 0;
@@ -205,6 +195,7 @@ int MPIR_Bsend_detach(void *bufferp, int *size)
     BsendBuffer.pending = 0;
 
   fn_exit:
+    MPID_THREAD_CS_EXIT(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -213,13 +204,8 @@ int MPIR_Bsend_detach(void *bufferp, int *size)
 /*
  * Initiate an ibsend.  We'll used this for Bsend as well.
  */
-#undef FUNCNAME
-#define FUNCNAME MPIR_Bsend_isend
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
-                     int dest, int tag, MPIR_Comm * comm_ptr,
-                     MPII_Bsend_kind_t kind, MPIR_Request ** request)
+                     int dest, int tag, MPIR_Comm * comm_ptr, MPIR_Request ** request)
 {
     int mpi_errno = MPI_SUCCESS;
     MPII_Bsend_data_t *p;
@@ -227,10 +213,8 @@ int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
     MPI_Aint packsize;
     int pass;
 
-    /* Find a free segment and copy the data into it.  If we could
-     * have, we would already have used tBsend to send the message with
-     * no copying.
-     *
+    MPID_THREAD_CS_ENTER(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
+    /*
      * We may want to decide here whether we need to pack at all
      * or if we can just use (a MPIR_Memcpy) of the buffer.
      */
@@ -239,11 +223,10 @@ int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
     /* We check the active buffer first.  This helps avoid storage
      * fragmentation */
     mpi_errno = MPIR_Bsend_check_active();
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     if (dtype != MPI_PACKED)
-        MPIR_Pack_size_impl(count, dtype, &packsize);
+        MPIR_Pack_size(count, dtype, &packsize);
     else
         packsize = count;
 
@@ -268,14 +251,16 @@ int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
 
             /* Pack the data into the buffer */
             /* We may want to optimize for the special case of
-             * either primative or contiguous types, and just
+             * either primitive or contiguous types, and just
              * use MPIR_Memcpy and the provided datatype */
             msg->count = 0;
             if (dtype != MPI_PACKED) {
+                MPI_Aint actual_pack_bytes;
+                void *pbuf = (void *) ((char *) p->msg.msgbuf + p->msg.count);
                 mpi_errno =
-                    MPIR_Pack_impl(buf, count, dtype, p->msg.msgbuf, packsize, &p->msg.count);
-                if (mpi_errno)
-                    MPIR_ERR_POP(mpi_errno);
+                    MPIR_Typerep_pack(buf, count, dtype, 0, pbuf, packsize, &actual_pack_bytes);
+                MPIR_ERR_CHECK(mpi_errno);
+                p->msg.count += actual_pack_bytes;
             } else {
                 MPIR_Memcpy(p->msg.msgbuf, buf, count);
                 p->msg.count = count;
@@ -297,12 +282,15 @@ int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
                  * to do this was commented out and probably did not match
                  * the current request internals */
                 MPIR_Bsend_take_buffer(p, p->msg.count);
-                p->kind = kind;
-                *request = p->request;
+                if (request) {
+                    /* Add 1 ref_count for MPI_Wait/Test */
+                    MPIR_Request_add_ref(p->request);
+                    *request = p->request;
+                }
             }
             break;
         }
-        /* If we found a buffer or we're in the seccond pass, then break.
+        /* If we found a buffer or we're in the second pass, then break.
          * Note that the test on phere is redundant, as the code breaks
          * out of the loop in the test above if a block p is found. */
         if (p || pass == 1)
@@ -325,6 +313,7 @@ int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
     }
 
   fn_exit:
+    MPID_THREAD_CS_EXIT(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -333,15 +322,13 @@ int MPIR_Bsend_isend(const void *buf, int count, MPI_Datatype dtype,
 /*
  * The following routine looks up the segment used by request req
  * and frees it. The request is assumed to be completed. This routine
- * is called by only MPIR_Ibsend_free.
+ * is called by only MPIR_Ibsend_cancel.
  */
-#undef FUNCNAME
-#define FUNCNAME MPIR_Bsend_free_seg
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIR_Bsend_free_req_seg(MPIR_Request * req)
 {
     int mpi_errno = MPI_ERR_INTERN;
+    MPID_THREAD_CS_ENTER(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
+
     MPII_Bsend_data_t *active = BsendBuffer.active;
 
     MPL_DBG_MSG_P(MPIR_DBG_BSEND, TYPICAL, "Checking active starting at %p", active);
@@ -357,6 +344,7 @@ int MPIR_Bsend_free_req_seg(MPIR_Request * req)
         MPL_DBG_MSG_P(MPIR_DBG_BSEND, TYPICAL, "Next active is %p", active);
     }
 
+    MPID_THREAD_CS_EXIT(VCI, MPIR_THREAD_VCI_BSEND_MUTEX);
     return mpi_errno;
 }
 
@@ -370,10 +358,6 @@ int MPIR_Bsend_free_req_seg(MPIR_Request * req)
 /* Add block p to the free list. Merge into adjacent blocks.  Used only
    within the check_active */
 
-#undef FUNCNAME
-#define FUNCNAME MPIR_Bsend_free_segment
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 static void MPIR_Bsend_free_segment(MPII_Bsend_data_t * p)
 {
     MPII_Bsend_data_t *prev = p->prev, *avail = BsendBuffer.avail, *avail_prev;
@@ -384,7 +368,7 @@ static void MPIR_Bsend_free_segment(MPII_Bsend_data_t * p)
                                               ((char *) p) + p->total_size));
 
     MPL_DBG_MSG_D(MPIR_DBG_BSEND, TYPICAL,
-                  "At the begining of free_segment with size %llu:",
+                  "At the beginning of free_segment with size %llu:",
                   (unsigned long long) p->total_size);
     MPL_DBG_STMT(MPIR_DBG_BSEND, TYPICAL, MPIR_Bsend_dump());
 
@@ -462,57 +446,37 @@ static void MPIR_Bsend_free_segment(MPII_Bsend_data_t * p)
  * track of the type of MPI routine (ibsend, bsend, or bsend_init/start)
  * that created the bsend entry.
  */
-#undef FUNCNAME
-#define FUNCNAME MPIR_Bsend_check_active
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
+
+/* TODO: make it as a progress_hook. The critical section need be made more granular.
+ * Or, does it matter? */
+static int MPIR_Bsend_progress(void)
+{
+    int mpi_errno = MPI_SUCCESS;
+
+    MPII_Bsend_data_t *active = BsendBuffer.active;
+    while (active) {
+        MPII_Bsend_data_t *next_active = active->next;
+        MPIR_Request *req = active->request;
+        if (MPIR_Request_is_complete(req)) {
+            MPIR_Bsend_free_segment(active);
+            if (!MPIR_Request_is_persistent(req)) {
+                MPIR_Request_free(req);
+            }
+        }
+        active = next_active;
+    }
+
+    return mpi_errno;
+}
+
 static int MPIR_Bsend_check_active(void)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPII_Bsend_data_t *active = BsendBuffer.active, *next_active;
 
-    MPL_DBG_MSG_P(MPIR_DBG_BSEND, TYPICAL, "Checking active starting at %p", active);
-    while (active) {
-        MPI_Request r = active->request->handle;
-        int flag;
-
-        next_active = active->next;
-
-        if (active->kind == IBSEND) {
-            /* We handle ibsend specially to allow for the user
-             * to attempt and cancel the request. Also, to allow
-             * for a cancel attempt (which must be attempted before
-             * a successful test or wait), we only start
-             * testing when the user has successfully released
-             * the request (it is a grequest, the free call will do it) */
-            flag = 0;
-            /* XXX DJG FIXME-MT should we be checking this? */
-            if (MPIR_Object_get_ref(active->request) == 1) {
-                mpi_errno = MPIR_Test(&r, &flag, MPI_STATUS_IGNORE);
-                if (mpi_errno)
-                    MPIR_ERR_POP(mpi_errno);
-            } else {
-                /* We need to invoke the progress engine in case we
-                 * need to advance other, incomplete communication.  */
-                MPID_Progress_state progress_state;
-                MPID_Progress_start(&progress_state);
-                mpi_errno = MPID_Progress_test();
-                MPID_Progress_end(&progress_state);
-                if (mpi_errno)
-                    MPIR_ERR_POP(mpi_errno);
-            }
-        } else {
-            mpi_errno = MPIR_Test(&r, &flag, MPI_STATUS_IGNORE);
-            if (mpi_errno)
-                MPIR_ERR_POP(mpi_errno);
-        }
-        if (flag) {
-            /* We're done.  Remove this segment */
-            MPL_DBG_MSG_P(MPIR_DBG_BSEND, TYPICAL, "Removing segment %p", active);
-            MPIR_Bsend_free_segment(active);
-        }
-        active = next_active;
-        MPL_DBG_MSG_P(MPIR_DBG_BSEND, TYPICAL, "Next active is %p", active);
+    if (BsendBuffer.active) {
+        mpi_errno = MPID_Progress_test(NULL);
+        MPIR_ERR_CHECK(mpi_errno);
+        MPIR_Bsend_progress();
     }
 
   fn_exit:
@@ -571,8 +535,9 @@ static void MPIR_Bsend_take_buffer(MPII_Bsend_data_t * p, size_t size)
     /* Compute the remaining size.  This must include any padding
      * that must be added to make the new block properly aligned */
     alloc_size = size;
-    if (alloc_size & 0x7)
-        alloc_size += (8 - (alloc_size & 0x7));
+    if (alloc_size & (MAX_ALIGNMENT - 1)) {
+        alloc_size += (MAX_ALIGNMENT - (alloc_size & (MAX_ALIGNMENT - 1)));
+    }
     /* alloc_size is the amount of space (out of size) that we will
      * allocate for this buffer. */
 
@@ -635,8 +600,9 @@ static void MPIR_Bsend_take_buffer(MPII_Bsend_data_t * p, size_t size)
 
 static int MPIR_Bsend_finalize(void *p ATTRIBUTE((unused)))
 {
+    /* No lock since this is inside MPI_Finalize */
     void *b;
-    int s;
+    MPI_Aint s;
 
     MPL_UNREFERENCED_ARG(p);
 
@@ -684,3 +650,8 @@ static void MPIR_Bsend_dump(void)
     MPL_DBG_MSG(MPIR_DBG_BSEND, TYPICAL, "end of list");
 }
 #endif
+
+int MPIR_Buffer_detach_impl(void *buffer_addr, MPI_Aint * size)
+{
+    return MPIR_Bsend_detach(buffer_addr, size);
+}
