@@ -28,6 +28,33 @@ static ADIOI_Flatlist_node *flatlist_node_new(MPI_Datatype datatype, MPI_Count c
     return flat;
 }
 
+/*
+ * I don't really expect this to ever trigger, but without the below safety
+ * valve, the design relies on the Count function coming out >= whatever
+ * the Flatten function comes up with.  There are enough differences between
+ * the two that it's hard to be positive this will always be true.  So every
+ * time something's added to flat's arrays, let's make sure they're big enough
+ * and re-alloc if not.
+ */
+static void flatlist_node_grow(ADIOI_Flatlist_node * flat, int idx)
+{
+    if (idx >= flat->count) {
+        ADIO_Offset *new_blocklens;
+        ADIO_Offset *new_indices;
+        int new_count = (flat->count * 1.25 + 4);
+        new_blocklens = (ADIO_Offset *) ADIOI_Calloc(new_count * 2, sizeof(ADIO_Offset));
+        new_indices = new_blocklens + new_count;
+        if (flat->count) {
+            memcpy(new_blocklens, flat->blocklens, flat->count * sizeof(ADIO_Offset));
+            memcpy(new_indices, flat->indices, flat->count * sizeof(ADIO_Offset));
+            ADIOI_Free(flat->blocklens);
+        }
+        flat->blocklens = new_blocklens;
+        flat->indices = new_indices;
+        flat->count = new_count;
+    }
+}
+
 void ADIOI_Optimize_flattened(ADIOI_Flatlist_node * flat_type);
 /* flatten datatype and add it to Flatlist */
 ADIOI_Flatlist_node *ADIOI_Flatten_datatype(MPI_Datatype datatype)
@@ -83,16 +110,26 @@ ADIOI_Flatlist_node *ADIOI_Flatten_datatype(MPI_Datatype datatype)
         DBG_FPRINTF(stderr, "ADIOI_Flatten_datatype:: ADIOI_Flatten\n");
 #endif
 
+/*
+ * Setting flat->count to curr_index, since curr_index is the most fundamentally
+ * correct updated value that represents what's in the indices/blocklens arrays.
+ * It would be nice if the counter function and the flatten function were in sync,
+ * but the numerous cases that decrement flat->count in the flatten function show
+ * that syncing them is a hack, and as long as the counter doesn't under-count
+ * it's good enough.
+ */
+        flat->count = curr_index;
+
         ADIOI_Optimize_flattened(flat);
 /* debug */
 #ifdef FLATTEN_DEBUG
         {
             int i;
-            for (i = 0; i < flat->count; i++)
+            for (i = 0; i < flat->count; i++) {
                 DBG_FPRINTF(stderr,
                             "ADIOI_Flatten_datatype:: i %#X, blocklens %#llX, indices %#llX\n", i,
-                            flat->blocklens[i], flat->indices[i]
-);
+                            (long long) flat->blocklens[i], (long long) flat->indices[i]);
+            }
         }
 #endif
     }
@@ -116,7 +153,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
      * avoid >2G integer arithmetic problems */
     ADIO_Offset top_count;
     MPI_Count i, j, old_size, prev_index, basic_num, num, nonzeroth;
-    MPI_Aint old_extent;        /* Assume extents are non-negative */
+    MPI_Aint lb, old_extent;    /* Assume extents are non-negative */
     int *ints;
     MPI_Aint *adds;             /* Make no assumptions about +/- sign on these */
     MPI_Datatype *types;
@@ -127,8 +164,8 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
     MPI_Type_get_contents(datatype, nints, nadds, ntypes, ints, adds, types);
 
 #ifdef FLATTEN_DEBUG
-    DBG_FPRINTF(stderr, "ADIOI_Flatten:: st_offset %#llX, curr_index %#llX\n", st_offset,
-                *curr_index);
+    DBG_FPRINTF(stderr, "ADIOI_Flatten:: st_offset %#llX, curr_index %#llX\n",
+                (long long) st_offset, (long long) *curr_index);
     DBG_FPRINTF(stderr, "ADIOI_Flatten:: nints %#X, nadds %#X, ntypes %#X\n", nints, nadds, ntypes);
     for (i = 0; i < nints; ++i) {
         DBG_FPRINTF(stderr, "ADIOI_Flatten:: ints[%lld]=%#X\n", (long long) i, ints[i]);
@@ -201,13 +238,15 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
 #ifdef FLATTEN_DEBUG
                 DBG_FPRINTF(stderr,
                             "ADIOI_Flatten:: MPI_COMBINER_DARRAY <ADIOI_Flatten(dtype, flat->indices[%#X] %#llX, flat->blocklens[%#X] %#llX, st_offset %#llX, curr_index %#llX);\n",
-                            0, flat->indices[0], 0, flat->blocklens[0], st_offset, *curr_index);
+                            0, (long long) flat->indices[0], 0, (long long) flat->blocklens[0],
+                            (long long) st_offset, (long long) *curr_index);
 #endif
                 ADIOI_Flatten(dtype, flat, st_offset, curr_index);
 #ifdef FLATTEN_DEBUG
                 DBG_FPRINTF(stderr,
                             "ADIOI_Flatten:: MPI_COMBINER_DARRAY >ADIOI_Flatten(dtype, flat->indices[%#X] %#llX, flat->blocklens[%#X] %#llX, st_offset %#llX, curr_index %#llX);\n",
-                            0, flat->indices[0], 0, flat->blocklens[0], st_offset, *curr_index);
+                            0, (long long) flat->indices[0], 0, (long long) flat->blocklens[0],
+                            (long long) st_offset, (long long) *curr_index);
 #endif
                 MPI_Type_free(&dtype);
             }
@@ -228,13 +267,15 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             if (prev_index == *curr_index) {
 /* simplest case, made up of basic or contiguous types */
                 j = *curr_index;
+                flatlist_node_grow(flat, j);
                 flat->indices[j] = st_offset;
                 MPI_Type_size_x(types[0], &old_size);
                 flat->blocklens[j] = top_count * old_size;
 #ifdef FLATTEN_DEBUG
                 DBG_FPRINTF(stderr,
                             "ADIOI_Flatten:: simple flat->indices[%#llX] %#llX, flat->blocklens[%#llX] %#llX\n",
-                            j, flat->indices[j], j, flat->blocklens[j]);
+                            (long long) j, (long long) flat->indices[j],
+                            (long long) j, (long long) flat->blocklens[j]);
 #endif
                 (*curr_index)++;
             } else {
@@ -243,16 +284,18 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 num = *curr_index - prev_index;
 
 /* The noncontiguous types have to be replicated count times */
-                MPI_Type_extent(types[0], &old_extent);
+                MPI_Type_get_extent(types[0], &lb, &old_extent);
                 for (m = 1; m < top_count; m++) {
                     for (i = 0; i < num; i++) {
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] =
                             flat->indices[j - num] + ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         flat->blocklens[j] = flat->blocklens[j - num];
 #ifdef FLATTEN_DEBUG
                         DBG_FPRINTF(stderr,
                                     "ADIOI_Flatten:: derived flat->indices[%#llX] %#llX, flat->blocklens[%#llX] %#llX\n",
-                                    j, flat->indices[j], j, flat->blocklens[j]);
+                                    (long long) j, (long long) flat->indices[j],
+                                    (long long) j, (long long) flat->blocklens[j]);
 #endif
                         j++;
                     }
@@ -279,10 +322,12 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                  * avoid >2G integer arithmetic problems */
                 ADIO_Offset blocklength = ints[1], stride = ints[2];
                 j = *curr_index;
+                flatlist_node_grow(flat, j);
                 flat->indices[j] = st_offset;
                 MPI_Type_size_x(types[0], &old_size);
                 flat->blocklens[j] = blocklength * old_size;
                 for (i = j + 1; i < j + top_count; i++) {
+                    flatlist_node_grow(flat, i);
                     flat->indices[i] = flat->indices[i - 1] + stride * old_size;
                     flat->blocklens[i] = flat->blocklens[j];
                 }
@@ -298,9 +343,10 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
 
 /* The noncontiguous types have to be replicated blocklen times
    and then strided. Replicate the first one. */
-                MPI_Type_extent(types[0], &old_extent);
+                MPI_Type_get_extent(types[0], &lb, &old_extent);
                 for (m = 1; m < blocklength; m++) {
                     for (i = 0; i < num; i++) {
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] =
                             flat->indices[j - num] + ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         flat->blocklens[j] = flat->blocklens[j - num];
@@ -313,6 +359,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 num = *curr_index - prev_index;
                 for (i = 1; i < top_count; i++) {
                     for (m = 0; m < num; m++) {
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] =
                             flat->indices[j - num] + stride * ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         flat->blocklens[j] = flat->blocklens[j - num];
@@ -342,10 +389,12 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                  * avoid >2G integer arithmetic problems */
                 ADIO_Offset blocklength = ints[1];
                 j = *curr_index;
+                flatlist_node_grow(flat, j);
                 flat->indices[j] = st_offset;
                 MPI_Type_size_x(types[0], &old_size);
                 flat->blocklens[j] = blocklength * old_size;
                 for (i = j + 1; i < j + top_count; i++) {
+                    flatlist_node_grow(flat, i);
                     flat->indices[i] = flat->indices[i - 1] + adds[0];
                     flat->blocklens[i] = flat->blocklens[j];
                 }
@@ -361,9 +410,10 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
 
 /* The noncontiguous types have to be replicated blocklen times
    and then strided. Replicate the first one. */
-                MPI_Type_extent(types[0], &old_extent);
+                MPI_Type_get_extent(types[0], &lb, &old_extent);
                 for (m = 1; m < blocklength; m++) {
                     for (i = 0; i < num; i++) {
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] =
                             flat->indices[j - num] + ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         flat->blocklens[j] = flat->blocklens[j - num];
@@ -376,6 +426,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 num = *curr_index - prev_index;
                 for (i = 1; i < top_count; i++) {
                     for (m = 0; m < num; m++) {
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] = flat->indices[j - num] + adds[0];
                         flat->blocklens[j] = flat->blocklens[j - num];
                         j++;
@@ -392,7 +443,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             top_count = ints[0];
             MPI_Type_get_envelope(types[0], &old_nints, &old_nadds, &old_ntypes, &old_combiner);
             ADIOI_Datatype_iscontig(types[0], &old_is_contig);
-            MPI_Type_extent(types[0], &old_extent);
+            MPI_Type_get_extent(types[0], &lb, &old_extent);
 
             prev_index = *curr_index;
             if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig)) {
@@ -412,16 +463,15 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                      * avoid >2G integer arithmetic problems */
                     ADIO_Offset blocklength = ints[1 + i - j], stride = ints[top_count + 1 + i - j];
                     if (blocklength > 0) {
+                        flatlist_node_grow(flat, nonzeroth);
                         flat->indices[nonzeroth] =
                             st_offset + stride * ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         flat->blocklens[nonzeroth] =
                             blocklength * ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         nonzeroth++;
-                    } else {
-                        flat->count--;  /* don't count/consider any zero-length blocklens */
                     }
                 }
-                *curr_index = i;
+                *curr_index = nonzeroth;
             } else {
 /* indexed type made up of noncontiguous derived types */
 
@@ -434,14 +484,13 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 for (m = 1; m < ints[1]; m++) {
                     for (i = 0, nonzeroth = j; i < num; i++) {
                         if (flat->blocklens[j - num] > 0) {
+                            flatlist_node_grow(flat, nonzeroth);
                             flat->indices[nonzeroth] =
                                 flat->indices[nonzeroth - num] +
                                 ADIOI_AINT_CAST_TO_OFFSET old_extent;
                             flat->blocklens[nonzeroth] = flat->blocklens[nonzeroth - num];
                             j++;
                             nonzeroth++;
-                        } else {
-                            flat->count--;
                         }
                     }
                 }
@@ -456,28 +505,26 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                          * avoid >2G integer arithmetic problems */
                         ADIO_Offset stride = ints[top_count + 1 + i] - ints[top_count + i];
                         if (flat->blocklens[j - num] > 0) {
+                            flatlist_node_grow(flat, nonzeroth);
                             flat->indices[nonzeroth] =
                                 flat->indices[j - num] +
                                 stride * ADIOI_AINT_CAST_TO_OFFSET old_extent;
                             flat->blocklens[nonzeroth] = flat->blocklens[j - num];
                             j++;
                             nonzeroth++;
-                        } else {
-                            flat->count--;
                         }
                     }
                     *curr_index = j;
                     for (m = 1; m < ints[1 + i]; m++) {
                         for (k = 0, nonzeroth = j; k < basic_num; k++) {
                             if (flat->blocklens[j - basic_num] > 0) {
+                                flatlist_node_grow(flat, nonzeroth);
                                 flat->indices[nonzeroth] =
                                     flat->indices[j - basic_num] +
                                     ADIOI_AINT_CAST_TO_OFFSET old_extent;
                                 flat->blocklens[nonzeroth] = flat->blocklens[j - basic_num];
                                 j++;
                                 nonzeroth++;
-                            } else {
-                                flat->count--;
                             }
                         }
                     }
@@ -489,9 +536,8 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
 #if defined HAVE_DECL_MPI_COMBINER_HINDEXED_BLOCK && HAVE_DECL_MPI_COMBINER_HINDEXED_BLOCK
         case MPI_COMBINER_HINDEXED_BLOCK:
             is_hindexed_block = 1;
-            /* deliberate fall-through */
-            MPL_FALLTHROUGH;
 #endif
+            /* fall through */
         case MPI_COMBINER_INDEXED_BLOCK:
 #ifdef FLATTEN_DEBUG
             DBG_FPRINTF(stderr, "ADIOI_Flatten:: MPI_COMBINER_INDEXED_BLOCK\n");
@@ -499,7 +545,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             top_count = ints[0];
             MPI_Type_get_envelope(types[0], &old_nints, &old_nadds, &old_ntypes, &old_combiner);
             ADIOI_Datatype_iscontig(types[0], &old_is_contig);
-            MPI_Type_extent(types[0], &old_extent);
+            MPI_Type_get_extent(types[0], &lb, &old_extent);
 
             prev_index = *curr_index;
             if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig)) {
@@ -523,9 +569,11 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                      * avoid >2G integer arithmetic problems */
                     ADIO_Offset blocklength = ints[1];
                     if (is_hindexed_block) {
+                        flatlist_node_grow(flat, i);
                         flat->indices[i] = st_offset + adds[i - j];
                     } else {
                         ADIO_Offset stride = ints[1 + 1 + i - j];
+                        flatlist_node_grow(flat, i);
                         flat->indices[i] = st_offset +
                             stride * ADIOI_AINT_CAST_TO_OFFSET old_extent;
                     }
@@ -545,8 +593,9 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                         if (is_hindexed_block) {
                             /* this is the one place the hindexed case uses the
                              * extent of a type */
-                            MPI_Type_extent(types[0], &old_extent);
+                            MPI_Type_get_extent(types[0], &lb, &old_extent);
                         }
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] = flat->indices[j - num] +
                             ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         flat->blocklens[j] = flat->blocklens[j - num];
@@ -560,11 +609,13 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 for (i = 1; i < top_count; i++) {
                     for (m = 0; m < num; m++) {
                         if (is_hindexed_block) {
+                            flatlist_node_grow(flat, j);
                             flat->indices[j] = flat->indices[j - num] + adds[i] - adds[i - 1];
                         } else {
                             /* By using ADIO_Offset we preserve +/- sign and
                              * avoid >2G integer arithmetic problems */
                             ADIO_Offset stride = ints[2 + i] - ints[1 + i];
+                            flatlist_node_grow(flat, j);
                             flat->indices[j] = flat->indices[j - num] +
                                 stride * ADIOI_AINT_CAST_TO_OFFSET old_extent;
                         }
@@ -599,14 +650,13 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                         /* By using ADIO_Offset we preserve +/- sign and
                          * avoid >2G integer arithmetic problems */
                         ADIO_Offset blocklength = ints[1 + i - j];
+                        flatlist_node_grow(flat, nonzeroth);
                         flat->indices[nonzeroth] = st_offset + adds[i - j];
                         flat->blocklens[nonzeroth] = blocklength * old_size;
                         nonzeroth++;
-                    } else {
-                        flat->count--;
                     }
                 }
-                *curr_index = i;
+                *curr_index = nonzeroth;
             } else {
 /* indexed type made up of noncontiguous derived types */
 
@@ -616,17 +666,16 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
 
 /* The noncontiguous types have to be replicated blocklens[i] times
    and then strided. Replicate the first one. */
-                MPI_Type_extent(types[0], &old_extent);
+                MPI_Type_get_extent(types[0], &lb, &old_extent);
                 for (m = 1; m < ints[1]; m++) {
                     for (i = 0, nonzeroth = j; i < num; i++) {
                         if (flat->blocklens[j - num] > 0) {
+                            flatlist_node_grow(flat, nonzeroth);
                             flat->indices[nonzeroth] =
                                 flat->indices[j - num] + ADIOI_AINT_CAST_TO_OFFSET old_extent;
                             flat->blocklens[nonzeroth] = flat->blocklens[j - num];
                             j++;
                             nonzeroth++;
-                        } else {
-                            flat->count--;
                         }
                     }
                 }
@@ -638,19 +687,19 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                     prev_index = *curr_index;
                     for (m = 0, nonzeroth = j; m < basic_num; m++) {
                         if (flat->blocklens[j - num] > 0) {
+                            flatlist_node_grow(flat, nonzeroth);
                             flat->indices[nonzeroth] =
                                 flat->indices[j - num] + adds[i] - adds[i - 1];
                             flat->blocklens[nonzeroth] = flat->blocklens[j - num];
                             j++;
                             nonzeroth++;
-                        } else {
-                            flat->count--;
                         }
                     }
                     *curr_index = j;
                     for (m = 1; m < ints[1 + i]; m++) {
                         for (k = 0, nonzeroth = j; k < basic_num; k++) {
                             if (flat->blocklens[j - basic_num] > 0) {
+                                flatlist_node_grow(flat, nonzeroth);
                                 flat->indices[nonzeroth] =
                                     flat->indices[j - basic_num] +
                                     ADIOI_AINT_CAST_TO_OFFSET old_extent;
@@ -686,6 +735,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                     if (ints[1 + n] > 0 || types[n] == MPI_LB || types[n] == MPI_UB) {
                         ADIO_Offset blocklength = ints[1 + n];
                         j = *curr_index;
+                        flatlist_node_grow(flat, j);
                         flat->indices[j] = st_offset + adds[n];
                         MPI_Type_size_x(types[n], &old_size);
                         flat->blocklens[j] = blocklength * old_size;
@@ -700,9 +750,6 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                                     n, adds[n], j, flat->indices[j], j, flat->blocklens[j]);
 #endif
                         (*curr_index)++;
-                    } else {
-                        flat->count--;  /* don't count/consider any zero-length
-                                         * blocklens */
                     }
                 } else {
 /* current type made up of noncontiguous derived types */
@@ -711,9 +758,10 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                     num = *curr_index - prev_index;
 
 /* The current type has to be replicated blocklens[n] times */
-                    MPI_Type_extent(types[n], &old_extent);
+                    MPI_Type_get_extent(types[n], &lb, &old_extent);
                     for (m = 1; m < ints[1 + n]; m++) {
                         for (i = 0; i < num; i++) {
+                            flatlist_node_grow(flat, j);
                             flat->indices[j] =
                                 flat->indices[j - num] + ADIOI_AINT_CAST_TO_OFFSET old_extent;
                             flat->blocklens[j] = flat->blocklens[j - num];
@@ -721,7 +769,8 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                             DBG_FPRINTF(stderr,
                                         "ADIOI_Flatten:: simple old_extent " MPI_AINT_FMT_HEX_SPEC
                                         ", flat->indices[%#llX] %#llX, flat->blocklens[%#llX] %#llX\n",
-                                        old_extent, j, flat->indices[j], j, flat->blocklens[j]);
+                                        old_extent, (long long) j, (long long) flat->indices[j],
+                                        (long long) j, (long long) flat->blocklens[j]);
 #endif
                             j++;
                         }
@@ -746,6 +795,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
              * bound based on the inner type, but the lower bound based on the
              * upper type.  check both lb and ub to prevent mixing updates */
             if (flat->lb_idx == -1 && flat->ub_idx == -1) {
+                flatlist_node_grow(flat, j);
                 flat->indices[j] = st_offset + adds[0];
                 /* this zero-length blocklens[] element, unlike elsewhere in the
                  * flattening code, is correct and is used to indicate a lower bound
@@ -758,14 +808,14 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 DBG_FPRINTF(stderr,
                             "ADIOI_Flatten:: simple adds[%#X] " MPI_AINT_FMT_HEX_SPEC
                             ", flat->indices[%#llX] %#llX, flat->blocklens[%#llX] %#llX\n", 0,
-                            adds[0], j, flat->indices[j], j, flat->blocklens[j]);
+                            adds[0], (long long) j, (long long) flat->indices[j],
+                            (long long) j, (long long) flat->blocklens[j]);
 #endif
 
                 (*curr_index)++;
             } else {
                 /* skipped over this chunk because something else higher-up in the
                  * type construction set this for us already */
-                flat->count--;
                 st_offset -= adds[0];
             }
 
@@ -779,6 +829,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             } else {
                 /* current type is basic or contiguous */
                 j = *curr_index;
+                flatlist_node_grow(flat, j);
                 flat->indices[j] = st_offset;
                 MPI_Type_size_x(types[0], &old_size);
                 flat->blocklens[j] = old_size;
@@ -787,7 +838,8 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
                 DBG_FPRINTF(stderr,
                             "ADIOI_Flatten:: simple adds[%#X] " MPI_AINT_FMT_HEX_SPEC
                             ", flat->indices[%#llX] %#llX, flat->blocklens[%#llX] %#llX\n", 0,
-                            adds[0], j, flat->indices[j], j, flat->blocklens[j]);
+                            adds[0], (long long) j, (long long) flat->indices[j],
+                            (long long) j, (long long) flat->blocklens[j]);
 #endif
 
                 (*curr_index)++;
@@ -797,6 +849,7 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             /* see note above about mixing updates for why we check lb and ub */
             if ((flat->lb_idx == -1 && flat->ub_idx == -1) || lb_updated) {
                 j = *curr_index;
+                flatlist_node_grow(flat, j);
                 flat->indices[j] = st_offset + adds[0] + adds[1];
                 /* again, zero-element ok: an upper-bound marker explicitly set by the
                  * constructor of this resized type */
@@ -805,7 +858,6 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             } else {
                 /* skipped over this chunk because something else higher-up in the
                  * type construction set this for us already */
-                flat->count--;
                 (*curr_index)--;
             }
 
@@ -813,7 +865,8 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
             DBG_FPRINTF(stderr,
                         "ADIOI_Flatten:: simple adds[%#X] " MPI_AINT_FMT_HEX_SPEC
                         ", flat->indices[%#llX] %#llX, flat->blocklens[%#llX] %#llX\n", 1, adds[1],
-                        j, flat->indices[j], j, flat->blocklens[j]);
+                        (long long) j, (long long) flat->indices[j],
+                        (long long) j, (long long) flat->blocklens[j]);
 #endif
 
             (*curr_index)++;
@@ -837,8 +890,8 @@ void ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node * flat,
     ADIOI_Free(types);
 
 #ifdef FLATTEN_DEBUG
-    DBG_FPRINTF(stderr, "ADIOI_Flatten:: return st_offset %#llX, curr_index %#llX\n", st_offset,
-                *curr_index);
+    DBG_FPRINTF(stderr, "ADIOI_Flatten:: return st_offset %#llX, curr_index %#llX\n",
+                (long long) st_offset, (long long) *curr_index);
 #endif
 
 }
