@@ -4,6 +4,8 @@
  */
 
 #include "mpioimpl.h"
+#include <limits.h>
+#include <assert.h>
 
 #ifdef HAVE_WEAK_SYMBOLS
 
@@ -20,6 +22,19 @@ int MPI_File_iwrite_all(MPI_File fh, const void *buf, int count, MPI_Datatype da
     __attribute__ ((weak, alias("PMPI_File_iwrite_all")));
 #endif
 
+#if defined(HAVE_PRAGMA_WEAK)
+#pragma weak MPI_File_iwrite_all_c = PMPI_File_iwrite_all_c
+#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
+#pragma _HP_SECONDARY_DEF PMPI_File_iwrite_all_c MPI_File_iwrite_all_c
+#elif defined(HAVE_PRAGMA_CRI_DUP)
+#pragma _CRI duplicate MPI_File_iwrite_all_c as PMPI_File_iwrite_all_c
+/* end of weak pragmas */
+#elif defined(HAVE_WEAK_ATTRIBUTE)
+int MPI_File_iwrite_all_c(MPI_File fh, const void *buf, MPI_Count count, MPI_Datatype datatype,
+                          MPI_Request * request)
+    __attribute__ ((weak, alias("PMPI_File_iwrite_all_c")));
+#endif
+
 /* Include mapping from MPI->PMPI */
 #define MPIO_BUILD_PROFILING
 #include "mpioprof.h"
@@ -64,6 +79,49 @@ int MPI_File_iwrite_all(MPI_File fh, ROMIO_CONST void *buf, int count,
     return error_code;
 }
 
+/* large count function */
+
+
+#ifdef HAVE_MPI_GREQUEST
+#include "mpiu_greq.h"
+#endif
+
+/*@
+    MPI_File_iwrite_all_c - Nonblocking collective write using individual file pointer
+
+Input Parameters:
+. fh - file handle (handle)
+. buf - initial address of buffer (choice)
+. count - number of elements in buffer (nonnegative integer)
+. datatype - datatype of each buffer element (handle)
+
+Output Parameters:
+. request - request object (handle)
+
+.N fortran
+@*/
+int MPI_File_iwrite_all_c(MPI_File fh, ROMIO_CONST void *buf, MPI_Count count,
+                          MPI_Datatype datatype, MPI_Request * request)
+{
+    assert(count <= INT_MAX);
+    int error_code;
+    static char myname[] = "MPI_FILE_IWRITE_ALL";
+#ifdef MPI_hpux
+    int fl_xmpi;
+
+    HPMP_IO_START(fl_xmpi, BLKMPIFILEIWRITEALL, TRDTBLOCK, fh, datatype, count);
+#endif /* MPI_hpux */
+
+    error_code = MPIOI_File_iwrite_all(fh, (MPI_Offset) 0,
+                                       ADIO_INDIVIDUAL, buf, count, datatype, myname, request);
+
+#ifdef MPI_hpux
+    HPMP_IO_END(fl_xmpi, fh, datatype, count);
+#endif /* MPI_hpux */
+
+    return error_code;
+}
+
 /* Note: MPIOI_File_iwrite_all also used by MPI_File_iwrite_at_all */
 /* prevent multiple definitions of this routine */
 #ifdef MPIO_BUILD_PROFILING
@@ -73,6 +131,7 @@ int MPIOI_File_iwrite_all(MPI_File fh,
                           const void *buf,
                           int count, MPI_Datatype datatype, char *myname, MPI_Request * request)
 {
+    assert(count <= INT_MAX);
     int error_code;
     MPI_Count datatype_size;
     ADIO_File adio_fh;
