@@ -1,10 +1,11 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *   Copyright (C) 1997 University of Chicago.
- *   See COPYRIGHT notice in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "mpioimpl.h"
+#include <limits.h>
+#include <assert.h>
 
 #ifdef HAVE_WEAK_SYMBOLS
 
@@ -20,6 +21,18 @@ int MPI_File_write(MPI_File fh, const void *buf, int count, MPI_Datatype datatyp
                    MPI_Status * status) __attribute__ ((weak, alias("PMPI_File_write")));
 #endif
 
+#if defined(HAVE_PRAGMA_WEAK)
+#pragma weak MPI_File_write_c = PMPI_File_write_c
+#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
+#pragma _HP_SECONDARY_DEF PMPI_File_write_c MPI_File_write_c
+#elif defined(HAVE_PRAGMA_CRI_DUP)
+#pragma _CRI duplicate MPI_File_write_c as PMPI_File_write_c
+/* end of weak pragmas */
+#elif defined(HAVE_WEAK_ATTRIBUTE)
+int MPI_File_write_c(MPI_File fh, const void *buf, MPI_Count count, MPI_Datatype datatype,
+                     MPI_Status * status) __attribute__ ((weak, alias("PMPI_File_write_c")));
+#endif
+
 /* Include mapping from MPI->PMPI */
 #define MPIO_BUILD_PROFILING
 #include "mpioprof.h"
@@ -62,6 +75,46 @@ int MPI_File_write(MPI_File fh, ROMIO_CONST void *buf, int count,
     return error_code;
 }
 
+/* large count function */
+
+
+
+/*@
+    MPI_File_write_c - Write using individual file pointer
+
+Input Parameters:
+. fh - file handle (handle)
+. buf - initial address of buffer (choice)
+. count - number of elements in buffer (nonnegative integer)
+. datatype - datatype of each buffer element (handle)
+
+Output Parameters:
+. status - status object (Status)
+
+.N fortran
+@*/
+int MPI_File_write_c(MPI_File fh, ROMIO_CONST void *buf, MPI_Count count,
+                     MPI_Datatype datatype, MPI_Status * status)
+{
+    assert(count <= INT_MAX);
+    int error_code;
+    static char myname[] = "MPI_FILE_WRITE";
+#ifdef MPI_hpux
+    int fl_xmpi;
+
+    HPMP_IO_START(fl_xmpi, BLKMPIFILEWRITE, TRDTBLOCK, fh, datatype, count);
+#endif /* MPI_hpux */
+
+    error_code = MPIOI_File_write(fh, (MPI_Offset) 0, ADIO_INDIVIDUAL, buf,
+                                  count, datatype, myname, status);
+
+#ifdef MPI_hpux
+    HPMP_IO_END(fl_xmpi, fh, datatype, count);
+#endif /* MPI_hpux */
+
+    return error_code;
+}
+
 /* prevent multiple definitions of this routine */
 #ifdef MPIO_BUILD_PROFILING
 int MPIOI_File_write(MPI_File fh,
@@ -70,6 +123,7 @@ int MPIOI_File_write(MPI_File fh,
                      const void *buf,
                      int count, MPI_Datatype datatype, char *myname, MPI_Status * status)
 {
+    assert(count <= INT_MAX);
     int error_code, buftype_is_contig, filetype_is_contig;
     MPI_Count datatype_size;
     ADIO_Offset off, bufsize;
