@@ -27,22 +27,21 @@ MPID_nem_ofi_send_callback(cq_tagged_entry_t * wc ATTRIBUTE((unused)), MPIR_Requ
 #include "ofi_tagged_template.c"
 
 /* ------------------------------------------------------------------------ */
-/* Receive callback called after sending a syncronous send acknowledgement. */
+/* Receive callback called after sending a synchronous send acknowledgement. */
 /* ------------------------------------------------------------------------ */
 static inline int MPID_nem_ofi_sync_recv_callback(cq_tagged_entry_t * wc ATTRIBUTE((unused)),
                                                   MPIR_Request * rreq)
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPID_NEM_OFI_SYNC_RECV_CALLBACK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPID_NEM_OFI_SYNC_RECV_CALLBACK);
+    MPIR_FUNC_ENTER;
 
     MPIDI_CH3U_Recvq_DP(REQ_OFI(rreq)->parent);
     MPIDI_CH3I_NM_OFI_RC(MPID_Request_complete(REQ_OFI(rreq)->parent));
     MPIDI_CH3I_NM_OFI_RC(MPID_Request_complete(rreq));
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPID_NEM_OFI_SYNC_RECV_CALLBACK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -56,23 +55,21 @@ static inline int MPID_nem_ofi_send_callback(cq_tagged_entry_t * wc ATTRIBUTE((u
                                              MPIR_Request * sreq)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPID_NEM_OFI_SEND_CALLBACK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPID_NEM_OFI_SEND_CALLBACK);
+    MPIR_FUNC_ENTER;
     MPL_free(REQ_OFI(sreq)->pack_buffer);
     MPIDI_CH3I_NM_OFI_RC(MPID_Request_complete(sreq));
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPID_NEM_OFI_SEND_CALLBACK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
 }
 
 /* Use macro as the two functions share common body */
-#define DO_CANCEL(req, _FCID)                           \
+#define DO_CANCEL(req) \
   int mpi_errno = MPI_SUCCESS;                          \
   int ret;                                              \
-  MPIR_FUNC_VERBOSE_STATE_DECL(_FCID);                  \
-  MPIR_FUNC_VERBOSE_ENTER(_FCID);                       \
+  MPIR_FUNC_ENTER;                                      \
   MPID_nem_ofi_poll(MPID_NONBLOCKING_POLL);             \
   ret = fi_cancel((fid_t)gl_data.endpoint,              \
                   &(REQ_OFI(req)->ofi_context));        \
@@ -88,17 +85,17 @@ static inline int MPID_nem_ofi_send_callback(cq_tagged_entry_t * wc ATTRIBUTE((u
   } else {                                              \
     MPIR_STATUS_SET_CANCEL_BIT(req->status, FALSE);     \
   }                                                     \
-  MPIR_FUNC_VERBOSE_EXIT(_FCID);                        \
+  MPIR_FUNC_EXIT;                                       \
   return mpi_errno;
 
 int MPID_nem_ofi_cancel_send(struct MPIDI_VC *vc ATTRIBUTE((unused)), struct MPIR_Request *sreq)
 {
-    DO_CANCEL(sreq, MPID_NEM_OFI_CANCEL_SEND);
+    DO_CANCEL(sreq);
 }
 
 int MPID_nem_ofi_cancel_recv(struct MPIDI_VC *vc ATTRIBUTE((unused)), struct MPIR_Request *rreq)
 {
-    DO_CANCEL(rreq, MPID_NEM_OFI_CANCEL_RECV);
+    DO_CANCEL(rreq);
 }
 
 
@@ -106,8 +103,7 @@ int MPID_nem_ofi_anysource_matched(MPIR_Request * rreq)
 {
     int matched = FALSE;
     int ret;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPID_NEM_OFI_ANYSOURCE_MATCHED);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPID_NEM_OFI_ANYSOURCE_MATCHED);
+    MPIR_FUNC_ENTER;
     /* ----------------------------------------------------- */
     /* Netmod has notified us that it has matched an any     */
     /* source request on another device.  We have the chance */
@@ -115,7 +111,7 @@ int MPID_nem_ofi_anysource_matched(MPIR_Request * rreq)
     /* ----------------------------------------------------- */
     ret = fi_cancel((fid_t) gl_data.endpoint, &(REQ_OFI(rreq)->ofi_context));
     if (ret == 0) {
-        /* Cancel succeded. This means that the actual message has been
+        /* Cancel succeeded. This means that the actual message has been
          *  received via nemesis shared memory. We need to return
          *  matched=False. The request will be completed at the nemesis level.
          *
@@ -130,6 +126,6 @@ int MPID_nem_ofi_anysource_matched(MPIR_Request * rreq)
          */
         matched = TRUE;
     }
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPID_NEM_OFI_ANYSOURCE_MATCHED);
+    MPIR_FUNC_EXIT;
     return matched;
 }
