@@ -20,7 +20,7 @@ static inline int MPID_Barrier(MPIR_Comm * comm, MPIR_Errflag_t * errflag)
     return MPIR_Barrier_impl(comm, errflag);
 }
 
-static inline int MPID_Bcast(void *buffer, int count, MPI_Datatype datatype, int root,
+static inline int MPID_Bcast(void *buffer, MPI_Aint count, MPI_Datatype datatype, int root,
                              MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
 #ifdef HAVE_LIBHCOLL
@@ -30,7 +30,7 @@ static inline int MPID_Bcast(void *buffer, int count, MPI_Datatype datatype, int
     return MPIR_Bcast_impl(buffer, count, datatype, root, comm, errflag);
 }
 
-static inline int MPID_Allreduce(const void *sendbuf, void *recvbuf, int count,
+static inline int MPID_Allreduce(const void *sendbuf, void *recvbuf, MPI_Aint count,
                                  MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                                  MPIR_Errflag_t * errflag)
 {
@@ -41,8 +41,8 @@ static inline int MPID_Allreduce(const void *sendbuf, void *recvbuf, int count,
     return MPIR_Allreduce_impl(sendbuf, recvbuf, count, datatype, op, comm, errflag);
 }
 
-static inline int MPID_Allgather(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                 void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Allgather(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                 void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                  MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
 #ifdef HAVE_LIBHCOLL
@@ -54,8 +54,8 @@ static inline int MPID_Allgather(const void *sendbuf, int sendcount, MPI_Datatyp
                                     recvcount, recvtype, comm, errflag);
 }
 
-static inline int MPID_Allgatherv(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                  void *recvbuf, const int *recvcounts, const int *displs,
+static inline int MPID_Allgatherv(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                  void *recvbuf, const MPI_Aint * recvcounts, const MPI_Aint * displs,
                                   MPI_Datatype recvtype, MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -67,8 +67,8 @@ static inline int MPID_Allgatherv(const void *sendbuf, int sendcount, MPI_Dataty
     return mpi_errno;
 }
 
-static inline int MPID_Scatter(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                               void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Scatter(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                               void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                int root, MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -79,8 +79,8 @@ static inline int MPID_Scatter(const void *sendbuf, int sendcount, MPI_Datatype
     return mpi_errno;
 }
 
-static inline int MPID_Scatterv(const void *sendbuf, const int *sendcounts, const int *displs,
-                                MPI_Datatype sendtype, void *recvbuf, int recvcount,
+static inline int MPID_Scatterv(const void *sendbuf, const MPI_Aint * sendcounts, const MPI_Aint * displs,
+                                MPI_Datatype sendtype, void *recvbuf, MPI_Aint recvcount,
                                 MPI_Datatype recvtype, int root, MPIR_Comm * comm,
                                 MPIR_Errflag_t * errflag)
 {
@@ -93,8 +93,8 @@ static inline int MPID_Scatterv(const void *sendbuf, const int *sendcounts, cons
     return mpi_errno;
 }
 
-static inline int MPID_Gather(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                              void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Gather(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                              void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                               int root, MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -105,8 +105,8 @@ static inline int MPID_Gather(const void *sendbuf, int sendcount, MPI_Datatype s
     return mpi_errno;
 }
 
-static inline int MPID_Gatherv(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                               void *recvbuf, const int *recvcounts, const int *displs,
+static inline int MPID_Gatherv(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                               void *recvbuf, const MPI_Aint * recvcounts, const MPI_Aint * displs,
                                MPI_Datatype recvtype, int root, MPIR_Comm * comm,
                                MPIR_Errflag_t * errflag)
 {
@@ -119,8 +119,8 @@ static inline int MPID_Gatherv(const void *sendbuf, int sendcount, MPI_Datatype
     return mpi_errno;
 }
 
-static inline int MPID_Alltoall(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Alltoall(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                 MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -131,9 +131,9 @@ static inline int MPID_Alltoall(const void *sendbuf, int sendcount, MPI_Datatype
     return mpi_errno;
 }
 
-static inline int MPID_Alltoallv(const void *sendbuf, const int *sendcounts, const int *sdispls,
-                                 MPI_Datatype sendtype, void *recvbuf, const int *recvcounts,
-                                 const int *rdispls, MPI_Datatype recvtype, MPIR_Comm * comm,
+static inline int MPID_Alltoallv(const void *sendbuf, const MPI_Aint * sendcounts, const MPI_Aint * sdispls,
+                                 MPI_Datatype sendtype, void *recvbuf, const MPI_Aint * recvcounts,
+                                 const MPI_Aint * rdispls, MPI_Datatype recvtype, MPIR_Comm * comm,
                                  MPIR_Errflag_t * errflag)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -145,9 +145,9 @@ static inline int MPID_Alltoallv(const void *sendbuf, const int *sendcounts, con
     return mpi_errno;
 }
 
-static inline int MPID_Alltoallw(const void *sendbuf, const int sendcounts[], const int sdispls[],
+static inline int MPID_Alltoallw(const void *sendbuf, const MPI_Aint sendcounts[], const MPI_Aint sdispls[],
                                  const MPI_Datatype sendtypes[], void *recvbuf,
-                                 const int recvcounts[], const int rdispls[],
+                                 const MPI_Aint recvcounts[], const MPI_Aint rdispls[],
                                  const MPI_Datatype recvtypes[], MPIR_Comm * comm_ptr,
                                  MPIR_Errflag_t * errflag)
 {
@@ -160,7 +160,7 @@ static inline int MPID_Alltoallw(const void *sendbuf, const int sendcounts[], co
     return mpi_errno;
 }
 
-static inline int MPID_Reduce(const void *sendbuf, void *recvbuf, int count,
+static inline int MPID_Reduce(const void *sendbuf, void *recvbuf, MPI_Aint count,
                               MPI_Datatype datatype, MPI_Op op, int root,
                               MPIR_Comm * comm, MPIR_Errflag_t * errflag)
 {
@@ -172,7 +172,7 @@ static inline int MPID_Reduce(const void *sendbuf, void *recvbuf, int count,
     return mpi_errno;
 }
 
-static inline int MPID_Reduce_scatter(const void *sendbuf, void *recvbuf, const int recvcounts[],
+static inline int MPID_Reduce_scatter(const void *sendbuf, void *recvbuf, const MPI_Aint recvcounts[],
                                       MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm_ptr,
                                       MPIR_Errflag_t * errflag)
 {
@@ -185,7 +185,7 @@ static inline int MPID_Reduce_scatter(const void *sendbuf, void *recvbuf, const
 }
 
 static inline int MPID_Reduce_scatter_block(const void *sendbuf, void *recvbuf,
-                                            int recvcount, MPI_Datatype datatype,
+                                            MPI_Aint recvcount, MPI_Datatype datatype,
                                             MPI_Op op, MPIR_Comm * comm_ptr,
                                             MPIR_Errflag_t * errflag)
 {
@@ -198,7 +198,7 @@ static inline int MPID_Reduce_scatter_block(const void *sendbuf, void *recvbuf,
     return mpi_errno;
 }
 
-static inline int MPID_Scan(const void *sendbuf, void *recvbuf, int count,
+static inline int MPID_Scan(const void *sendbuf, void *recvbuf, MPI_Aint count,
                             MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                             MPIR_Errflag_t * errflag)
 {
@@ -210,7 +210,7 @@ static inline int MPID_Scan(const void *sendbuf, void *recvbuf, int count,
     return mpi_errno;
 }
 
-static inline int MPID_Exscan(const void *sendbuf, void *recvbuf, int count,
+static inline int MPID_Exscan(const void *sendbuf, void *recvbuf, MPI_Aint count,
                               MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                               MPIR_Errflag_t * errflag)
 {
@@ -222,8 +222,8 @@ static inline int MPID_Exscan(const void *sendbuf, void *recvbuf, int count,
     return mpi_errno;
 }
 
-static inline int MPID_Neighbor_allgather(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                          void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Neighbor_allgather(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                          void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                           MPIR_Comm * comm)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -235,9 +235,9 @@ static inline int MPID_Neighbor_allgather(const void *sendbuf, int sendcount, MP
     return mpi_errno;
 }
 
-static inline int MPID_Neighbor_allgatherv(const void *sendbuf, int sendcount,
+static inline int MPID_Neighbor_allgatherv(const void *sendbuf, MPI_Aint sendcount,
                                            MPI_Datatype sendtype, void *recvbuf,
-                                           const int recvcounts[], const int displs[],
+                                           const MPI_Aint recvcounts[], const MPI_Aint displs[],
                                            MPI_Datatype recvtype, MPIR_Comm * comm)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -249,10 +249,10 @@ static inline int MPID_Neighbor_allgatherv(const void *sendbuf, int sendcount,
     return mpi_errno;
 }
 
-static inline int MPID_Neighbor_alltoallv(const void *sendbuf, const int sendcounts[],
-                                          const int sdispls[], MPI_Datatype sendtype,
-                                          void *recvbuf, const int recvcounts[],
-                                          const int rdispls[], MPI_Datatype recvtype,
+static inline int MPID_Neighbor_alltoallv(const void *sendbuf, const MPI_Aint sendcounts[],
+                                          const MPI_Aint sdispls[], MPI_Datatype sendtype,
+                                          void *recvbuf, const MPI_Aint recvcounts[],
+                                          const MPI_Aint rdispls[], MPI_Datatype recvtype,
                                           MPIR_Comm * comm)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -264,9 +264,9 @@ static inline int MPID_Neighbor_alltoallv(const void *sendbuf, const int sendcou
     return mpi_errno;
 }
 
-static inline int MPID_Neighbor_alltoallw(const void *sendbuf, const int sendcounts[],
+static inline int MPID_Neighbor_alltoallw(const void *sendbuf, const MPI_Aint sendcounts[],
                                           const MPI_Aint sdispls[], const MPI_Datatype sendtypes[],
-                                          void *recvbuf, const int recvcounts[],
+                                          void *recvbuf, const MPI_Aint recvcounts[],
                                           const MPI_Aint rdispls[], const MPI_Datatype recvtypes[],
                                           MPIR_Comm * comm)
 {
@@ -279,8 +279,8 @@ static inline int MPID_Neighbor_alltoallw(const void *sendbuf, const int sendcou
     return mpi_errno;
 }
 
-static inline int MPID_Neighbor_alltoall(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                         void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Neighbor_alltoall(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                         void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                          MPIR_Comm * comm)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -292,8 +292,8 @@ static inline int MPID_Neighbor_alltoall(const void *sendbuf, int sendcount, MPI
     return mpi_errno;
 }
 
-static inline int MPID_Ineighbor_allgather(const void *sendbuf, int sendcount,
-                                           MPI_Datatype sendtype, void *recvbuf, int recvcount,
+static inline int MPID_Ineighbor_allgather(const void *sendbuf, MPI_Aint sendcount,
+                                           MPI_Datatype sendtype, void *recvbuf, MPI_Aint recvcount,
                                            MPI_Datatype recvtype, MPIR_Comm * comm,
                                            MPIR_Request **request)
 {
@@ -306,9 +306,9 @@ static inline int MPID_Ineighbor_allgather(const void *sendbuf, int sendcount,
     return mpi_errno;
 }
 
-static inline int MPID_Ineighbor_allgatherv(const void *sendbuf, int sendcount,
+static inline int MPID_Ineighbor_allgatherv(const void *sendbuf, MPI_Aint sendcount,
                                             MPI_Datatype sendtype, void *recvbuf,
-                                            const int recvcounts[], const int displs[],
+                                            const MPI_Aint recvcounts[], const MPI_Aint displs[],
                                             MPI_Datatype recvtype, MPIR_Comm * comm,
                                             MPIR_Request **request)
 {
@@ -321,8 +321,8 @@ static inline int MPID_Ineighbor_allgatherv(const void *sendbuf, int sendcount,
     return mpi_errno;
 }
 
-static inline int MPID_Ineighbor_alltoall(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                          void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Ineighbor_alltoall(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                          void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                           MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -334,10 +334,10 @@ static inline int MPID_Ineighbor_alltoall(const void *sendbuf, int sendcount, MP
     return mpi_errno;
 }
 
-static inline int MPID_Ineighbor_alltoallv(const void *sendbuf, const int sendcounts[],
-                                           const int sdispls[], MPI_Datatype sendtype,
-                                           void *recvbuf, const int recvcounts[],
-                                           const int rdispls[], MPI_Datatype recvtype,
+static inline int MPID_Ineighbor_alltoallv(const void *sendbuf, const MPI_Aint sendcounts[],
+                                           const MPI_Aint sdispls[], MPI_Datatype sendtype,
+                                           void *recvbuf, const MPI_Aint recvcounts[],
+                                           const MPI_Aint rdispls[], MPI_Datatype recvtype,
                                            MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -350,9 +350,9 @@ static inline int MPID_Ineighbor_alltoallv(const void *sendbuf, const int sendco
     return mpi_errno;
 }
 
-static inline int MPID_Ineighbor_alltoallw(const void *sendbuf, const int sendcounts[],
+static inline int MPID_Ineighbor_alltoallw(const void *sendbuf, const MPI_Aint sendcounts[],
                                            const MPI_Aint sdispls[], const MPI_Datatype sendtypes[],
-                                           void *recvbuf, const int recvcounts[],
+                                           void *recvbuf, const MPI_Aint recvcounts[],
                                            const MPI_Aint rdispls[], const MPI_Datatype recvtypes[],
                                            MPIR_Comm * comm, MPIR_Request **request)
 {
@@ -375,7 +375,7 @@ static inline int MPID_Ibarrier(MPIR_Comm * comm, MPIR_Request **request)
     return mpi_errno;
 }
 
-static inline int MPID_Ibcast(void *buffer, int count, MPI_Datatype datatype, int root,
+static inline int MPID_Ibcast(void *buffer, MPI_Aint count, MPI_Datatype datatype, int root,
                               MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -385,8 +385,8 @@ static inline int MPID_Ibcast(void *buffer, int count, MPI_Datatype datatype, in
     return mpi_errno;
 }
 
-static inline int MPID_Iallgather(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                  void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Iallgather(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                  void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                   MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -397,8 +397,8 @@ static inline int MPID_Iallgather(const void *sendbuf, int sendcount, MPI_Dataty
     return mpi_errno;
 }
 
-static inline int MPID_Iallgatherv(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                   void *recvbuf, const int *recvcounts, const int *displs,
+static inline int MPID_Iallgatherv(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                   void *recvbuf, const MPI_Aint * recvcounts, const MPI_Aint * displs,
                                    MPI_Datatype recvtype, MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -410,7 +410,7 @@ static inline int MPID_Iallgatherv(const void *sendbuf, int sendcount, MPI_Datat
     return mpi_errno;
 }
 
-static inline int MPID_Iallreduce(const void *sendbuf, void *recvbuf, int count,
+static inline int MPID_Iallreduce(const void *sendbuf, void *recvbuf, MPI_Aint count,
                                   MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                                   MPIR_Request **request)
 {
@@ -422,8 +422,8 @@ static inline int MPID_Iallreduce(const void *sendbuf, void *recvbuf, int count,
     return mpi_errno;
 }
 
-static inline int MPID_Ialltoall(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                 void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Ialltoall(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                 void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                  MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -434,10 +434,10 @@ static inline int MPID_Ialltoall(const void *sendbuf, int sendcount, MPI_Datatyp
     return mpi_errno;
 }
 
-static inline int MPID_Ialltoallv(const void *sendbuf, const int sendcounts[],
-                                  const int sdispls[], MPI_Datatype sendtype,
-                                  void *recvbuf, const int recvcounts[],
-                                  const int rdispls[], MPI_Datatype recvtype,
+static inline int MPID_Ialltoallv(const void *sendbuf, const MPI_Aint sendcounts[],
+                                  const MPI_Aint sdispls[], MPI_Datatype sendtype,
+                                  void *recvbuf, const MPI_Aint recvcounts[],
+                                  const MPI_Aint rdispls[], MPI_Datatype recvtype,
                                   MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -449,10 +449,10 @@ static inline int MPID_Ialltoallv(const void *sendbuf, const int sendcounts[],
     return mpi_errno;
 }
 
-static inline int MPID_Ialltoallw(const void *sendbuf, const int sendcounts[],
-                                  const int sdispls[], const MPI_Datatype sendtypes[],
-                                  void *recvbuf, const int recvcounts[],
-                                  const int rdispls[], const MPI_Datatype recvtypes[],
+static inline int MPID_Ialltoallw(const void *sendbuf, const MPI_Aint sendcounts[],
+                                  const MPI_Aint sdispls[], const MPI_Datatype sendtypes[],
+                                  void *recvbuf, const MPI_Aint recvcounts[],
+                                  const MPI_Aint rdispls[], const MPI_Datatype recvtypes[],
                                   MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -464,7 +464,7 @@ static inline int MPID_Ialltoallw(const void *sendbuf, const int sendcounts[],
     return mpi_errno;
 }
 
-static inline int MPID_Iexscan(const void *sendbuf, void *recvbuf, int count,
+static inline int MPID_Iexscan(const void *sendbuf, void *recvbuf, MPI_Aint count,
                                MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                                MPIR_Request **request)
 {
@@ -476,8 +476,8 @@ static inline int MPID_Iexscan(const void *sendbuf, void *recvbuf, int count,
     return mpi_errno;
 }
 
-static inline int MPID_Igather(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                               void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Igather(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                               void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                int root, MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -488,8 +488,8 @@ static inline int MPID_Igather(const void *sendbuf, int sendcount, MPI_Datatype
     return mpi_errno;
 }
 
-static inline int MPID_Igatherv(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                void *recvbuf, const int *recvcounts, const int *displs,
+static inline int MPID_Igatherv(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                void *recvbuf, const MPI_Aint * recvcounts, const MPI_Aint * displs,
                                 MPI_Datatype recvtype, int root, MPIR_Comm * comm,
                                 MPIR_Request **request)
 {
@@ -502,7 +502,7 @@ static inline int MPID_Igatherv(const void *sendbuf, int sendcount, MPI_Datatype
     return mpi_errno;
 }
 
-static inline int MPID_Ireduce_scatter_block(const void *sendbuf, void *recvbuf, int recvcount,
+static inline int MPID_Ireduce_scatter_block(const void *sendbuf, void *recvbuf, MPI_Aint recvcount,
                                              MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                                              MPIR_Request **request)
 {
@@ -514,7 +514,7 @@ static inline int MPID_Ireduce_scatter_block(const void *sendbuf, void *recvbuf,
     return mpi_errno;
 }
 
-static inline int MPID_Ireduce_scatter(const void *sendbuf, void *recvbuf, const int recvcounts[],
+static inline int MPID_Ireduce_scatter(const void *sendbuf, void *recvbuf, const MPI_Aint recvcounts[],
                                        MPI_Datatype datatype, MPI_Op op, MPIR_Comm * comm,
                                        MPIR_Request **request)
 {
@@ -526,7 +526,7 @@ static inline int MPID_Ireduce_scatter(const void *sendbuf, void *recvbuf, const
     return mpi_errno;
 }
 
-static inline int MPID_Ireduce(const void *sendbuf, void *recvbuf, int count, MPI_Datatype datatype,
+static inline int MPID_Ireduce(const void *sendbuf, void *recvbuf, MPI_Aint count, MPI_Datatype datatype,
                                MPI_Op op, int root, MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -537,7 +537,7 @@ static inline int MPID_Ireduce(const void *sendbuf, void *recvbuf, int count, MP
     return mpi_errno;
 }
 
-static inline int MPID_Iscan(const void *sendbuf, void *recvbuf, int count, MPI_Datatype datatype,
+static inline int MPID_Iscan(const void *sendbuf, void *recvbuf, MPI_Aint count, MPI_Datatype datatype,
                              MPI_Op op, MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -548,8 +548,8 @@ static inline int MPID_Iscan(const void *sendbuf, void *recvbuf, int count, MPI_
     return mpi_errno;
 }
 
-static inline int MPID_Iscatter(const void *sendbuf, int sendcount, MPI_Datatype sendtype,
-                                void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Iscatter(const void *sendbuf, MPI_Aint sendcount, MPI_Datatype sendtype,
+                                void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                 int root, MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -560,9 +560,9 @@ static inline int MPID_Iscatter(const void *sendbuf, int sendcount, MPI_Datatype
     return mpi_errno;
 }
 
-static inline int MPID_Iscatterv(const void *sendbuf, const int *sendcounts,
-                                 const int *displs, MPI_Datatype sendtype,
-                                 void *recvbuf, int recvcount, MPI_Datatype recvtype,
+static inline int MPID_Iscatterv(const void *sendbuf, const MPI_Aint * sendcounts,
+                                 const MPI_Aint * displs, MPI_Datatype sendtype,
+                                 void *recvbuf, MPI_Aint recvcount, MPI_Datatype recvtype,
                                  int root, MPIR_Comm * comm, MPIR_Request **request)
 {
     int mpi_errno = MPI_SUCCESS;
