@@ -25,9 +25,8 @@
    use macros in mpidimpl.h *instead* of this routine */
 void MPID_Request_create_hook(MPIR_Request *req)
 {
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPID_REQUEST_INIT);
 
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPID_REQUEST_INIT);
+    MPIR_FUNC_ENTER;
     
     req->dev.datatype_ptr	   = NULL;
     req->dev.msg_offset         = 0;
@@ -83,9 +82,8 @@ int MPIDI_CH3U_Request_load_send_iov(MPIR_Request * const sreq,
 {
     MPI_Aint last;
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_SEND_IOV);
 
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_SEND_IOV);
+    MPIR_FUNC_ENTER;
 
     last = sreq->dev.msgsize;
     MPL_DBG_MSG_FMT(MPIDI_CH3_DBG_CHANNEL,VERBOSE,(MPL_DBG_FDEST,
@@ -181,11 +179,8 @@ int MPIDI_CH3U_Request_load_send_iov(MPIR_Request * const sreq,
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_SEND_IOV);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
-
-  fn_fail:
-    goto fn_exit;
 }
 
 
@@ -201,9 +196,8 @@ int MPIDI_CH3U_Request_load_recv_iov(MPIR_Request * const rreq)
 {
     MPI_Aint last;
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_RECV_IOV);
 
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_RECV_IOV);
+    MPIR_FUNC_ENTER;
 
     if (rreq->dev.orig_msg_offset == MPIDI_LOAD_RECV_IOV_ORIG_MSG_OFFSET_UNSET) {
         rreq->dev.orig_msg_offset = rreq->dev.msg_offset;
@@ -286,7 +280,7 @@ int MPIDI_CH3U_Request_load_recv_iov(MPIR_Request * const rreq)
 	/* --BEGIN ERROR HANDLING-- */
 	if (rreq->dev.iov_count == 0)
 	{
-	    /* If the data can't be unpacked, the we have a mis-match between
+	    /* If the data can't be unpacked, the we have a mismatch between
 	       the datatype and the amount of data received.  Adjust
 	       the segment info so that the remaining data is received and 
 	       thrown away. */
@@ -398,7 +392,7 @@ int MPIDI_CH3U_Request_load_recv_iov(MPIR_Request * const rreq)
     }
     
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_RECV_IOV);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
@@ -412,9 +406,8 @@ int MPIDI_CH3U_Request_unpack_srbuf(MPIR_Request * rreq)
     MPI_Aint last;
     int tmpbuf_last;
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_SRBUF);
     
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_SRBUF);
+    MPIR_FUNC_ENTER;
 
     tmpbuf_last = (int)(rreq->dev.msg_offset + rreq->dev.tmpbuf_sz);
     if (rreq->dev.msgsize < tmpbuf_last)
@@ -475,7 +468,7 @@ int MPIDI_CH3U_Request_unpack_srbuf(MPIR_Request * rreq)
 	rreq->dev.msg_offset = last;
     }
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_SRBUF);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
@@ -492,10 +485,8 @@ int MPIDI_CH3U_Request_unpack_uebuf(MPIR_Request * rreq)
     MPIR_Datatype * dt_ptr;
     intptr_t unpack_sz;
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_UEBUF);
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MEMCPY);
 
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_UEBUF);
+    MPIR_FUNC_ENTER;
 
     MPIDI_Datatype_get_info(rreq->dev.user_count, rreq->dev.datatype, 
 			    dt_contig, userbuf_sz, dt_ptr, dt_true_lb);
@@ -527,10 +518,10 @@ int MPIDI_CH3U_Request_unpack_uebuf(MPIR_Request * rreq)
 	    /* TODO - check that amount of data is consistent with
 	       datatype.  If not we should return an error (unless
 	       configured with --enable-fast) */
-	    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MEMCPY);
-	    MPIR_Memcpy((char *)rreq->dev.user_buf + dt_true_lb, rreq->dev.tmpbuf,
+	    MPIR_FUNC_ENTER;
+	    MPIR_Memcpy(MPIR_get_contig_ptr(rreq->dev.user_buf, dt_true_lb), rreq->dev.tmpbuf,
 		   unpack_sz);
-	    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MEMCPY);
+	    MPIR_FUNC_EXIT;
 	}
 	else
 	{
@@ -554,28 +545,25 @@ int MPIDI_CH3U_Request_unpack_uebuf(MPIR_Request * rreq)
 	}
     }
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_UEBUF);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 int MPID_Request_complete(MPIR_Request *req)
 {
-    int incomplete, notify_counter;
+    int incomplete;
     int mpi_errno = MPI_SUCCESS;
 
     MPIDI_CH3U_Request_decrement_cc(req, &incomplete);
     if (!incomplete) {
         /* decrement completion_notification counter */
         if (req->completion_notification)
-            MPIR_cc_decr(req->completion_notification, &notify_counter);
+            MPIR_cc_dec(req->completion_notification);
 
 	MPIR_Request_free(req);
     }
 
- fn_exit:
     return mpi_errno;
- fn_fail:
-    goto fn_exit;
 }
 
 void MPID_Request_free_hook(MPIR_Request *req)
@@ -587,7 +575,8 @@ void MPID_Request_free_hook(MPIR_Request *req)
 
     /* trigger request_completed callback function */
     if (req->dev.request_completed_cb != NULL && MPIR_Request_is_complete(req)) {
-        int mpi_errno = req->dev.request_completed_cb(req);
+        MPIR_AssertDeclValue(int mpi_errno, MPI_SUCCESS);
+        mpi_errno = req->dev.request_completed_cb(req);
         MPIR_Assert(mpi_errno == MPI_SUCCESS);
 
         req->dev.request_completed_cb = NULL;
