@@ -22,9 +22,8 @@ int MPID_Issend(const void * buf, int count, MPI_Datatype datatype, int rank, in
 #endif    
     int eager_threshold = -1;
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPID_ISSEND);
 
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPID_ISSEND);
+    MPIR_FUNC_ENTER;
 
     MPL_DBG_MSG_FMT(MPIDI_CH3_DBG_OTHER,VERBOSE,(MPL_DBG_FDEST,
                  "rank=%d, tag=%d, context=%d", 
@@ -116,6 +115,6 @@ int MPID_Issend(const void * buf, int count, MPI_Datatype datatype, int rank, in
 		  )
     
   fn_fail:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPID_ISSEND);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
