@@ -430,6 +430,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_get_set_number(const char *set_name)
 #define MPIDI_OFI_ENABLE_PT2PT_NOPACK_GNI       MPIDI_OFI_ON
 #define MPIDI_OFI_ENABLE_HMEM_GNI               MPIDI_OFI_OFF
 #define MPIDI_OFI_NUM_AM_BUFFERS_GNI            MPIDI_OFI_MAX_NUM_AM_BUFFERS
+#define MPIDI_OFI_NUM_OPTIMIZED_MEMORY_REGIONS_GNI (0)
 #define MPIDI_OFI_CONTEXT_BITS_GNI              (16)
 #define MPIDI_OFI_SOURCE_BITS_GNI               (24)
 #define MPIDI_OFI_TAG_BITS_GNI                  (20)
@@ -457,6 +458,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_get_set_number(const char *set_name)
 #define MPIDI_OFI_ENABLE_PT2PT_NOPACK       MPIDI_OFI_ENABLE_PT2PT_NOPACK_GNI
 #define MPIDI_OFI_ENABLE_HMEM               MPIDI_OFI_ENABLE_HMEM_GNI
 #define MPIDI_OFI_NUM_AM_BUFFERS            MPIDI_OFI_NUM_AM_BUFFERS_GNI
+#define MPIDI_OFI_NUM_OPTIMIZED_MEMORY_REGIONS  MPIDI_OFI_NUM_OPTIMIZED_MEMORY_REGIONS_GNI
 #define MPIDI_OFI_PROTOCOL_MASK             (0xE000000000000000ULL)
 #define MPIDI_OFI_CONTEXT_MASK              (0x0FFFF00000000000ULL)
 #define MPIDI_OFI_SOURCE_MASK               (0x00000FFFFFF00000ULL)     /* GNI does not support immediate data
