@@ -15,20 +15,20 @@ static int win_init_sep(MPIR_Win * win);
 static int win_init_stx(MPIR_Win * win);
 static int win_init_global(MPIR_Win * win);
 static int win_init(MPIR_Win * win);
+static void win_init_am(MPIR_Win * win);
 
 static void load_acc_hint(MPIR_Win * win)
 {
     int op_index = 0, i;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_LOAD_ACC_HINT);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_LOAD_ACC_HINT);
+    MPIR_FUNC_ENTER;
 
     if (!MPIDI_OFI_WIN(win).acc_hint)
         MPIDI_OFI_WIN(win).acc_hint = MPL_malloc(sizeof(MPIDI_OFI_win_acc_hint_t), MPL_MEM_RMA);
 
     /* TODO: we assume all pes pass the same hint. Allreduce is needed to check
      * the maximal value or the spec must define it as same value on all pes.
-     * Now we assume the spec requries same value on all pes. */
+     * Now we assume the spec requires same value on all pes. */
 
     /* We translate the atomic op hints to max count allowed for all possible atomics with each
      * datatype. We do not need more specific info (e.g., <datatype, op>, because any process may use
@@ -73,7 +73,7 @@ static void load_acc_hint(MPIR_Win * win)
         }
     }
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_LOAD_ACC_HINT);
+    MPIR_FUNC_EXIT;
 }
 
 /* Set OFI attributes and capabilities for RMA. */
@@ -124,9 +124,9 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
     MPIR_Errflag_t errflag = MPIR_ERR_NONE;
     MPIR_Comm *comm_ptr = win->comm_ptr;
     MPIDI_OFI_win_targetinfo_t *winfo;
+    int nic = 0;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_WIN_ALLGATHER);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_WIN_ALLGATHER);
+    MPIR_FUNC_ENTER;
 
     if (!MPIDI_OFI_ENABLE_MR_PROV_KEY) {
         MPIDI_OFI_WIN(win).mr_key = MPIDI_OFI_WIN(win).win_id;
@@ -134,6 +134,10 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
         MPIDI_OFI_WIN(win).mr_key = 0;
     }
 
+    /* we need register mr on the correct domain for the vni */
+    int vni = MPIDI_OFI_get_win_vni(win);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+
     /* Register the allocated win buffer or MPI_BOTTOM (NULL) for dynamic win.
      * It is clear that we cannot register NULL when FI_MR_ALLOCATED is set, thus
      * we skip trial and return immediately. When FI_MR_ALLOCATED is not set, however,
@@ -144,13 +148,20 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
     int rc = 0, allrc = 0;
     MPIDI_OFI_WIN(win).mr = NULL;
     if (base || (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC && !MPIDI_OFI_ENABLE_MR_ALLOCATED)) {
+        size_t len;
+        if (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC) {
+            len = UINTPTR_MAX - (uintptr_t) base;
+        } else {
+            len = (size_t) win->size;
+        }
+
         /* device buffers are not currently supported */
         if (MPIR_GPU_query_pointer_is_dev(base))
             rc = -1;
         else
-            MPIDI_OFI_CALL_RETURN(fi_mr_reg(MPIDI_OFI_global.ctx[0].domain,     /* In:  Domain Object */
+            MPIDI_OFI_CALL_RETURN(fi_mr_reg(MPIDI_OFI_global.ctx[ctx_idx].domain,       /* In:  Domain Object */
                                             base,       /* In:  Lower memory address */
-                                            win->size,  /* In:  Length              */
+                                            len,        /* In:  Length              */
                                             FI_REMOTE_READ | FI_REMOTE_WRITE,   /* In:  Expose MR for read  */
                                             0ULL,       /* In:  offset(not used)    */
                                             MPIDI_OFI_WIN(win).mr_key,  /* In:  requested key       */
@@ -158,7 +169,11 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
                                             &MPIDI_OFI_WIN(win).mr,     /* Out: memregion object    */
                                             NULL), rc); /* In:  context             */
     } else if (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC) {
+        /* We may still do native atomics with collective attach, let's load acc_hint */
+        load_acc_hint(win);
         goto fn_exit;
+    } else {
+        /* FIXME: what's in the branch? If it can't happen, add an assertion here */
     }
 
     /* Check if any process fails to register. If so, release local MR and force AM path. */
@@ -207,7 +222,7 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
     load_acc_hint(win);
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_WIN_ALLGATHER);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -225,18 +240,17 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
 static int win_set_per_win_sync(MPIR_Win * win)
 {
     int ret, mpi_errno = MPI_SUCCESS;
+    int nic = 0;
+    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, MPIDI_OFI_WIN(win).vni, nic);
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_WIN_SET_PER_WIN_SYNC);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_WIN_SET_PER_WIN_SYNC);
+    MPIR_FUNC_ENTER;
 
     struct fi_cntr_attr cntr_attr;
     memset(&cntr_attr, 0, sizeof(cntr_attr));
     cntr_attr.events = FI_CNTR_EVENTS_COMP;
     cntr_attr.wait_obj = FI_WAIT_UNSPEC;
-    MPIDI_OFI_CALL_RETURN(fi_cntr_open(MPIDI_OFI_global.ctx[0].domain,  /* In:  Domain Object */
-                                       &cntr_attr,      /* In:  Configuration object */
-                                       &MPIDI_OFI_WIN(win).cmpl_cntr,   /* Out: Counter Object */
-                                       NULL), ret);     /* Context: counter events   */
+    MPIDI_OFI_CALL_RETURN(fi_cntr_open(MPIDI_OFI_global.ctx[ctx_idx].domain, &cntr_attr,
+                                       &MPIDI_OFI_WIN(win).cmpl_cntr, NULL), ret);
     if (ret < 0) {
         MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE, "Failed to open completion counter.\n");
         mpi_errno = MPIDI_OFI_ENAVAIL;
@@ -259,12 +273,17 @@ static int win_set_per_win_sync(MPIR_Win * win)
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_WIN_SET_PER_WIN_SYNC);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
 }
 
+static void win_init_am(MPIR_Win * win)
+{
+    MPIDI_WIN(win, am_vci) %= MPIDI_OFI_global.num_vnis;
+}
+
 /*
  * Initialize endpoint using scalable endpoint.
  *
@@ -278,59 +297,62 @@ static int win_init_sep(MPIR_Win * win)
 {
     int i, ret, mpi_errno = MPI_SUCCESS;
     struct fi_info *finfo;
+    int nic = 0;
+    int vni = MPIDI_OFI_WIN(win).vni;
+    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_WIN_INIT_SEP);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_WIN_INIT_SEP);
+    MPIR_FUNC_ENTER;
 
-    finfo = fi_dupinfo(MPIDI_OFI_global.prov_use);
+    finfo = fi_dupinfo(MPIDI_OFI_global.prov_use[0]);
     MPIR_Assert(finfo);
 
+
     /* Initialize scalable EP when first window is created. */
-    if (MPIDI_OFI_global.rma_sep == NULL) {
-        /* Specify number of transmit context according user input and provider limit. */
+    if (MPIDI_OFI_global.ctx[ctx_idx].rma_sep == NULL) {
+        /* NOTE: if MPIDI_OFI_VNI_USE_SEPCTX, we could share rma_stx_ctx across vnis */
+
         MPIDI_OFI_global.max_rma_sep_tx_cnt =
-            MPL_MIN(MPIDI_OFI_global.prov_use->domain_attr->max_ep_tx_ctx,
+            MPL_MIN(MPIDI_OFI_global.prov_use[0]->domain_attr->max_ep_tx_ctx,
                     MPIR_CVAR_CH4_OFI_MAX_RMA_SEP_CTX);
         finfo->ep_attr->tx_ctx_cnt = MPIDI_OFI_global.max_rma_sep_tx_cnt;
 
-        MPIDI_OFI_CALL_RETURN(fi_scalable_ep
-                              (MPIDI_OFI_global.ctx[0].domain, finfo, &MPIDI_OFI_global.rma_sep,
-                               NULL), ret);
+        MPIDI_OFI_CALL_RETURN(fi_scalable_ep(MPIDI_OFI_global.ctx[ctx_idx].domain, finfo,
+                                             &MPIDI_OFI_global.ctx[ctx_idx].rma_sep, NULL), ret);
         if (ret < 0) {
             MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE, "Failed to create scalable endpoint.\n");
-            MPIDI_OFI_global.rma_sep = NULL;
+            MPIDI_OFI_global.ctx[ctx_idx].rma_sep = NULL;
             mpi_errno = MPIDI_OFI_ENAVAIL;
             goto fn_fail;
         }
 
-        MPIDI_OFI_CALL_RETURN(fi_scalable_ep_bind
-                              (MPIDI_OFI_global.rma_sep, &(MPIDI_OFI_global.ctx[0].av->fid), 0),
+        MPIDI_OFI_CALL_RETURN(fi_scalable_ep_bind(MPIDI_OFI_global.ctx[ctx_idx].rma_sep,
+                                                  &(MPIDI_OFI_global.ctx[ctx_idx].av->fid), 0),
                               ret);
         if (ret < 0) {
             MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
                         "Failed to bind scalable endpoint to address vector.\n");
-            MPIDI_OFI_global.rma_sep = NULL;
+            MPIDI_OFI_global.ctx[ctx_idx].rma_sep = NULL;
             mpi_errno = MPIDI_OFI_ENAVAIL;
             goto fn_fail;
         }
 
-        /* Allocate and initilize tx index array. */
-        utarray_new(MPIDI_OFI_global.rma_sep_idx_array, &ut_int_icd, MPL_MEM_RMA);
+        /* Allocate and initialize tx index array. */
+        utarray_new(MPIDI_OFI_global.ctx[ctx_idx].rma_sep_idx_array, &ut_int_icd, MPL_MEM_RMA);
         for (i = 0; i < MPIDI_OFI_global.max_rma_sep_tx_cnt; i++) {
-            utarray_push_back(MPIDI_OFI_global.rma_sep_idx_array, &i, MPL_MEM_RMA);
+            utarray_push_back(MPIDI_OFI_global.ctx[ctx_idx].rma_sep_idx_array, &i, MPL_MEM_RMA);
         }
     }
     /* Set per window transmit attributes. */
     set_rma_fi_info(win, finfo);
     /* Get available transmit context index. */
-    int *idx = (int *) utarray_back(MPIDI_OFI_global.rma_sep_idx_array);
+    int *idx = (int *) utarray_back(MPIDI_OFI_global.ctx[ctx_idx].rma_sep_idx_array);
     if (idx == NULL) {
         mpi_errno = MPIDI_OFI_ENAVAIL;
         goto fn_fail;
     }
     /* Retrieve transmit context on scalable EP. */
     MPIDI_OFI_CALL_RETURN(fi_tx_context
-                          (MPIDI_OFI_global.rma_sep, *idx, finfo->tx_attr,
+                          (MPIDI_OFI_global.ctx[ctx_idx].rma_sep, *idx, finfo->tx_attr,
                            &(MPIDI_OFI_WIN(win).ep), NULL), ret);
     if (ret < 0) {
         MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
@@ -341,10 +363,10 @@ static int win_init_sep(MPIR_Win * win)
 
     MPIDI_OFI_WIN(win).sep_tx_idx = *idx;
     /* Pop this index out of reserving array. */
-    utarray_pop_back(MPIDI_OFI_global.rma_sep_idx_array);
+    utarray_pop_back(MPIDI_OFI_global.ctx[ctx_idx].rma_sep_idx_array);
 
     MPIDI_OFI_CALL_RETURN(fi_ep_bind(MPIDI_OFI_WIN(win).ep,
-                                     &MPIDI_OFI_global.ctx[0].cq->fid,
+                                     &MPIDI_OFI_global.ctx[ctx_idx].cq->fid,
                                      FI_TRANSMIT | FI_SELECTIVE_COMPLETION), ret);
     if (ret < 0) {
         MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
@@ -371,13 +393,13 @@ static int win_init_sep(MPIR_Win * win)
 
   fn_exit:
     fi_freeinfo(finfo);
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_WIN_INIT_SEP);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     if (MPIDI_OFI_WIN(win).sep_tx_idx != -1) {
         /* Push tx idx back into available pool. */
-        utarray_push_back(MPIDI_OFI_global.rma_sep_idx_array, &MPIDI_OFI_WIN(win).sep_tx_idx,
-                          MPL_MEM_RMA);
+        utarray_push_back(MPIDI_OFI_global.ctx[ctx_idx].rma_sep_idx_array,
+                          &MPIDI_OFI_WIN(win).sep_tx_idx, MPL_MEM_RMA);
         MPIDI_OFI_WIN(win).sep_tx_idx = -1;
     }
     if (MPIDI_OFI_WIN(win).ep != NULL) {
@@ -403,13 +425,40 @@ static int win_init_stx(MPIR_Win * win)
     int ret, mpi_errno = MPI_SUCCESS;
     struct fi_info *finfo;
     bool have_per_win_cntr = false;
+    int nic = 0;
+    int vni = MPIDI_OFI_WIN(win).vni;
+    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_WIN_INIT_STX);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_WIN_INIT_STX);
+    MPIR_FUNC_ENTER;
 
-    finfo = fi_dupinfo(MPIDI_OFI_global.prov_use);
+    /* WB TODO - Only setting up nic[0] for RMA right now. */
+    finfo = fi_dupinfo(MPIDI_OFI_global.prov_use[0]);
     MPIR_Assert(finfo);
 
+
+    /* Initialize rma shared context when first window is created. */
+    if (MPIDI_OFI_global.ctx[ctx_idx].rma_stx_ctx == NULL) {
+        /* NOTE: if MPIDI_OFI_VNI_USE_SEPCTX, we could share rma_stx_ctx across vnis */
+
+        struct fi_tx_attr tx_attr;
+        memset(&tx_attr, 0, sizeof(tx_attr));
+        /* A shared transmit context’s attributes must be a union of all associated
+         * endpoints' transmit capabilities. */
+        tx_attr.caps = FI_RMA | FI_WRITE | FI_READ | FI_ATOMIC;
+        tx_attr.msg_order = FI_ORDER_RAR | FI_ORDER_RAW | FI_ORDER_WAR | FI_ORDER_WAW;
+        tx_attr.op_flags = FI_DELIVERY_COMPLETE | FI_COMPLETION;
+        MPIDI_OFI_CALL_RETURN(fi_stx_context(MPIDI_OFI_global.ctx[ctx_idx].domain, &tx_attr,
+                                             &MPIDI_OFI_global.ctx[ctx_idx].rma_stx_ctx, NULL),
+                              ret);
+        if (ret < 0) {
+            MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
+                        "Failed to create RMA shared TX context.\n");
+            MPIDI_OFI_global.ctx[ctx_idx].rma_stx_ctx = NULL;
+            mpi_errno = MPIDI_OFI_ENAVAIL;
+            goto fn_fail;
+        }
+    }
+
     /* Set per window transmit attributes. */
     set_rma_fi_info(win, finfo);
     /* Still need to take out rx capabilities for shared context. */
@@ -417,7 +466,7 @@ static int win_init_stx(MPIR_Win * win)
 
     finfo->ep_attr->tx_ctx_cnt = FI_SHARED_CONTEXT;     /* Request a shared context */
     finfo->ep_attr->rx_ctx_cnt = 0;     /* We don't need RX contexts */
-    MPIDI_OFI_CALL_RETURN(fi_endpoint(MPIDI_OFI_global.ctx[0].domain,
+    MPIDI_OFI_CALL_RETURN(fi_endpoint(MPIDI_OFI_global.ctx[ctx_idx].domain,
                                       finfo, &MPIDI_OFI_WIN(win).ep, NULL), ret);
     if (ret < 0) {
         MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
@@ -429,8 +478,8 @@ static int win_init_stx(MPIR_Win * win)
 
     if (win_set_per_win_sync(win) == MPI_SUCCESS) {
         have_per_win_cntr = true;
-        MPIDI_OFI_CALL_RETURN(fi_ep_bind
-                              (MPIDI_OFI_WIN(win).ep, &MPIDI_OFI_global.rma_stx_ctx->fid, 0), ret);
+        MPIDI_OFI_CALL_RETURN(fi_ep_bind(MPIDI_OFI_WIN(win).ep,
+                                         &MPIDI_OFI_global.ctx[ctx_idx].rma_stx_ctx->fid, 0), ret);
         if (ret < 0) {
             MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
                         "Failed to bind endpoint to shared transmit contxt.\n");
@@ -439,7 +488,7 @@ static int win_init_stx(MPIR_Win * win)
         }
 
         MPIDI_OFI_CALL_RETURN(fi_ep_bind(MPIDI_OFI_WIN(win).ep,
-                                         &MPIDI_OFI_global.ctx[0].cq->fid,
+                                         &MPIDI_OFI_global.ctx[ctx_idx].cq->fid,
                                          FI_TRANSMIT | FI_SELECTIVE_COMPLETION), ret);
         if (ret < 0) {
             MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
@@ -449,7 +498,8 @@ static int win_init_stx(MPIR_Win * win)
         }
 
         MPIDI_OFI_CALL_RETURN(fi_ep_bind
-                              (MPIDI_OFI_WIN(win).ep, &MPIDI_OFI_global.ctx[0].av->fid, 0), ret);
+                              (MPIDI_OFI_WIN(win).ep, &MPIDI_OFI_global.ctx[ctx_idx].av->fid, 0),
+                              ret);
         if (ret < 0) {
             MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE,
                         "Failed to bind endpoint to address vector.\n");
@@ -471,7 +521,7 @@ static int win_init_stx(MPIR_Win * win)
 
   fn_exit:
     fi_freeinfo(finfo);
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_WIN_INIT_STX);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     if (MPIDI_OFI_WIN(win).ep != NULL) {
@@ -488,14 +538,24 @@ static int win_init_stx(MPIR_Win * win)
 
 static int win_init_global(MPIR_Win * win)
 {
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_WIN_INIT_GLOBAL);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_WIN_INIT_GLOBAL);
+    MPIR_FUNC_ENTER;
 
-    MPIDI_OFI_WIN(win).ep = MPIDI_OFI_global.ctx[0].tx;
-    MPIDI_OFI_WIN(win).cmpl_cntr = MPIDI_OFI_global.ctx[0].rma_cmpl_cntr;
-    MPIDI_OFI_WIN(win).issued_cntr = &MPIDI_OFI_global.rma_issued_cntr;
+    int vni = MPIDI_OFI_WIN(win).vni;
+    int nic = 0;
+    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_WIN_INIT_GLOBAL);
+    MPIDI_OFI_WIN(win).ep = MPIDI_OFI_global.ctx[ctx_idx].tx;
+#ifdef MPIDI_OFI_VNI_USE_DOMAIN
+    MPIDI_OFI_WIN(win).cmpl_cntr = MPIDI_OFI_global.ctx[ctx_idx].rma_cmpl_cntr;
+    MPIDI_OFI_WIN(win).issued_cntr = &MPIDI_OFI_global.ctx[ctx_idx].rma_issued_cntr;
+#else
+    ctx_idx = MPIDI_OFI_get_ctx_index(NULL, 0, nic);
+    /* NOTE: shared with ctx[0] */
+    MPIDI_OFI_WIN(win).cmpl_cntr = MPIDI_OFI_global.ctx[ctx_idx].rma_cmpl_cntr;
+    MPIDI_OFI_WIN(win).issued_cntr = &MPIDI_OFI_global.ctx[ctx_idx].rma_issued_cntr;
+#endif
+
+    MPIR_FUNC_EXIT;
 
     return 0;
 }
@@ -504,20 +564,31 @@ static int win_init(MPIR_Win * win)
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_WIN_INIT);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_WIN_INIT);
+    MPIR_FUNC_ENTER;
 
     MPL_COMPILE_TIME_ASSERT(sizeof(MPIDI_Devwin_t) >= sizeof(MPIDI_OFI_win_t));
 
     memset(&MPIDI_OFI_WIN(win), 0, sizeof(MPIDI_OFI_win_t));
 
-    MPIDI_OFI_WIN(win).win_id = MPIDI_OFI_mr_key_alloc();
+    MPIDI_OFI_WIN(win).win_id =
+        MPIDI_OFI_mr_key_alloc(MPIDI_OFI_COLL_MR_KEY, win->comm_ptr->context_id);
+    if (MPIDI_OFI_WIN(win).win_id == -1ULL) {
+        MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE, "Failed to get global mr key.\n");
+        mpi_errno = MPIDI_OFI_ENAVAIL;
+        goto fn_exit;
+    }
 
     MPIDIU_map_set(MPIDI_OFI_global.win_map, MPIDI_OFI_WIN(win).win_id, win, MPL_MEM_RMA);
 
     MPIDI_OFI_WIN(win).sep_tx_idx = -1; /* By default, -1 means not using scalable EP. */
     MPIDI_OFI_WIN(win).progress_counter = 1;
 
+    /* Assign vni to window.
+     * NOTE: we could assign vni per epoch, then we need run `win_init_{sep,stx,global}`
+     * at start of every epoch.
+     */
+    MPIDI_OFI_WIN(win).vni = MPIDI_OFI_get_win_vni(win);
+
     /* First, try to enable scalable EP. */
     if (MPIR_CVAR_CH4_OFI_ENABLE_SCALABLE_ENDPOINTS && MPIR_CVAR_CH4_OFI_MAX_RMA_SEP_CTX > 0) {
         /* Create tx based on scalable EP. */
@@ -536,74 +607,73 @@ static int win_init(MPIR_Win * win)
     win_init_global(win);
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_WIN_INIT);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 /* Callback of MPL_gavl_tree_create to delete local registered MR for dynamic window */
 static void dwin_close_mr(void *obj)
 {
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_DWIN_CLOSE_MR);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_DWIN_CLOSE_MR);
+    MPIR_FUNC_ENTER;
 
     /* Close local MR */
     struct fid_mr *mr = (struct fid_mr *) obj;
     if (mr) {
-        int ret;
+        int ret ATTRIBUTE((unused));
+        if (!MPIDI_OFI_ENABLE_MR_PROV_KEY) {
+            uint64_t requested_key = fi_mr_key(mr);
+            MPIDI_OFI_mr_key_free(MPIDI_OFI_LOCAL_MR_KEY, requested_key);
+        }
         MPIDI_OFI_CALL_RETURN(fi_close(&mr->fid), ret);
         MPIR_Assert(ret >= 0);
     }
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_DWIN_CLOSE_MR);
+    MPIR_FUNC_EXIT;
     return;
 }
 
 /* Callback of MPL_gavl_tree_create to delete remote registered MR for dynamic window */
 static void dwin_free_target_mr(void *obj)
 {
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_DWIN_FREE_TARGET_MEM);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_DWIN_FREE_TARGET_MEM);
+    MPIR_FUNC_ENTER;
 
     /* Free cached buffer for remote MR */
     MPL_free(obj);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_DWIN_FREE_TARGET_MEM);
+    MPIR_FUNC_EXIT;
     return;
 }
 
 int MPIDI_OFI_mpi_win_set_info(MPIR_Win * win, MPIR_Info * info)
 {
     int mpi_errno;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_SET_INFO);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_SET_INFO);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_set_info(win, info);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_SET_INFO);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 int MPIDI_OFI_mpi_win_get_info(MPIR_Win * win, MPIR_Info ** info_p_p)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_GET_INFO);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_GET_INFO);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_get_info(win, info_p_p);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_GET_INFO);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 int MPIDI_OFI_mpi_win_free(MPIR_Win ** win_ptr)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_FREE);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_FREE);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_free(win_ptr);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_FREE);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
@@ -611,24 +681,22 @@ int MPIDI_OFI_mpi_win_create(void *base, MPI_Aint length, int disp_unit, MPIR_In
                              MPIR_Comm * comm_ptr, MPIR_Win ** win_ptr)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_create(base, length, disp_unit, info, comm_ptr, win_ptr);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 int MPIDI_OFI_mpi_win_attach(MPIR_Win * win, void *base, MPI_Aint size)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_attach(win, base, size);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
@@ -637,25 +705,23 @@ int MPIDI_OFI_mpi_win_allocate_shared(MPI_Aint size, int disp_unit, MPIR_Info *
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_SHARED);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_SHARED);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_allocate_shared(size, disp_unit, info_ptr, comm_ptr, base_ptr,
                                                win_ptr);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_SHARED);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 int MPIDI_OFI_mpi_win_detach(MPIR_Win * win, const void *base)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_DETACH);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_DETACH);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_detach(win, base);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_DETACH);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
@@ -663,24 +729,22 @@ int MPIDI_OFI_mpi_win_allocate(MPI_Aint size, int disp_unit, MPIR_Info * info, M
                                void *baseptr, MPIR_Win ** win_ptr)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_allocate(size, disp_unit, info, comm, baseptr, win_ptr);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
 int MPIDI_OFI_mpi_win_create_dynamic(MPIR_Info * info, MPIR_Comm * comm, MPIR_Win ** win_ptr)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_DYNAMIC);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_DYNAMIC);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDIG_mpi_win_create_dynamic(info, comm, win_ptr);
 
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_DYNAMIC);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
@@ -688,8 +752,8 @@ int MPIDI_OFI_mpi_win_create_hook(MPIR_Win * win)
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_HOOK);
+    MPIR_FUNC_ENTER;
+    win_init_am(win);
 
     /* This hook is called by CH4 generic call after CH4 initialization */
     if (MPIDI_OFI_ENABLE_RMA) {
@@ -703,7 +767,7 @@ int MPIDI_OFI_mpi_win_create_hook(MPIR_Win * win)
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_HOOK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -713,8 +777,8 @@ int MPIDI_OFI_mpi_win_allocate_hook(MPIR_Win * win)
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_HOOK);
+    MPIR_FUNC_ENTER;
+    win_init_am(win);
 
     /* This hook is called by CH4 generic call after CH4 initialization */
     if (MPIDI_OFI_ENABLE_RMA) {
@@ -726,7 +790,7 @@ int MPIDI_OFI_mpi_win_allocate_hook(MPIR_Win * win)
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_HOOK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -736,8 +800,8 @@ int MPIDI_OFI_mpi_win_allocate_shared_hook(MPIR_Win * win)
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_SHARED_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_SHARED_HOOK);
+    MPIR_FUNC_ENTER;
+    win_init_am(win);
 
     /* This hook is called by CH4 generic call after CH4 initialization */
     if (MPIDI_OFI_ENABLE_RMA) {
@@ -749,7 +813,7 @@ int MPIDI_OFI_mpi_win_allocate_shared_hook(MPIR_Win * win)
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_ALLOCATE_SHARED_HOOK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
@@ -759,18 +823,13 @@ int MPIDI_OFI_mpi_win_create_dynamic_hook(MPIR_Win * win)
 {
     int mpi_errno = MPI_SUCCESS;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_DYNAMIC_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_DYNAMIC_HOOK);
+    MPIR_FUNC_ENTER;
+    win_init_am(win);
 
     MPIR_CHKPMEM_DECL(1);
 
     /* This hook is called by CH4 generic call after CH4 initialization */
     if (MPIDI_OFI_ENABLE_RMA) {
-        /* FIXME: should the OFI specific size be stored inside OFI rather
-         * than overwriting CH4 info ? Now we simply followed original
-         * create_dynamic routine.*/
-        win->size = (uintptr_t) UINTPTR_MAX - (uintptr_t) MPI_BOTTOM;
-
         mpi_errno = win_init(win);
         MPIR_ERR_CHECK(mpi_errno);
 
@@ -807,7 +866,7 @@ int MPIDI_OFI_mpi_win_create_dynamic_hook(MPIR_Win * win)
     MPIR_CHKPMEM_COMMIT();
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_CREATE_DYNAMIC_HOOK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     MPIR_CHKPMEM_REAP();
@@ -823,16 +882,16 @@ typedef struct dwin_target_mr {
 int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH_HOOK);
+    MPIR_FUNC_ENTER;
 
     MPIR_Errflag_t errflag = MPIR_ERR_NONE;
     MPIR_Comm *comm_ptr = win->comm_ptr;
     dwin_target_mr_t *target_mrs;
     int mpl_err = MPL_SUCCESS, i;
+    int nic = 0;
+    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, MPIDI_OFI_WIN(win).vni, nic);
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH_HOOK);
+    MPIR_FUNC_ENTER;
 
     MPIR_CHKLMEM_DECL(1);
 
@@ -841,7 +900,7 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
 
     uint64_t requested_key = 0ULL;
     if (!MPIDI_OFI_ENABLE_MR_PROV_KEY)
-        requested_key = MPIDI_OFI_mr_key_alloc();
+        requested_key = MPIDI_OFI_mr_key_alloc(MPIDI_OFI_LOCAL_MR_KEY, MPIDI_OFI_INVALID_MR_KEY);
 
     int rc = 0, allrc = 0;
     struct fid_mr *mr = NULL;
@@ -849,7 +908,7 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
     if (MPIR_GPU_query_pointer_is_dev(base))
         rc = -1;
     else
-        MPIDI_OFI_CALL_RETURN(fi_mr_reg(MPIDI_OFI_global.ctx[0].domain, /* In:  Domain Object */
+        MPIDI_OFI_CALL_RETURN(fi_mr_reg(MPIDI_OFI_global.ctx[ctx_idx].domain,   /* In:  Domain Object */
                                         base,   /* In:  Lower memory address */
                                         size,   /* In:  Length              */
                                         FI_REMOTE_READ | FI_REMOTE_WRITE,       /* In:  Expose MR for read  */
@@ -895,7 +954,7 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
             (MPIDI_OFI_target_mr_t *) MPL_malloc(sizeof(MPIDI_OFI_target_mr_t), MPL_MEM_RMA);
         MPIR_Assert(target_mr);
         /* Store addr only for calculating offset when !MPIDI_OFI_ENABLE_MR_VIRT_ADDRESS */
-        target_mr->addr = (uint64_t) target_mrs[i].base;
+        target_mr->addr = (uintptr_t) target_mrs[i].base;
         target_mr->mr_key = target_mrs[i].mr_key;
 
         mpl_err = MPL_gavl_tree_insert(MPIDI_OFI_WIN(win).dwin_target_mrs[i],
@@ -905,7 +964,7 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_ATTACH_HOOK);
+    MPIR_FUNC_EXIT;
     MPIR_CHKLMEM_FREEALL();
     return mpi_errno;
   fn_fail:
@@ -915,8 +974,7 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
 int MPIDI_OFI_mpi_win_detach_hook(MPIR_Win * win, const void *base)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_DETACH_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_DETACH_HOOK);
+    MPIR_FUNC_ENTER;
 
     MPIR_Errflag_t errflag = MPIR_ERR_NONE;
     MPIR_Comm *comm_ptr = win->comm_ptr;
@@ -956,7 +1014,7 @@ int MPIDI_OFI_mpi_win_detach_hook(MPIR_Win * win, const void *base)
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_DETACH_HOOK);
+    MPIR_FUNC_EXIT;
     MPIR_CHKLMEM_FREEALL();
     return mpi_errno;
   fn_fail:
@@ -966,20 +1024,24 @@ int MPIDI_OFI_mpi_win_detach_hook(MPIR_Win * win, const void *base)
 int MPIDI_OFI_mpi_win_free_hook(MPIR_Win * win)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_OFI_MPI_WIN_FREE_HOOK);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_OFI_MPI_WIN_FREE_HOOK);
+    int nic = 0;
+
+    MPIR_FUNC_ENTER;
 
     if (MPIDI_OFI_ENABLE_RMA) {
-        MPIDI_OFI_mr_key_free(MPIDI_OFI_WIN(win).win_id);
+        int vni = MPIDI_OFI_WIN(win).vni;
+        int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+        MPIDI_OFI_mr_key_free(MPIDI_OFI_COLL_MR_KEY, MPIDI_OFI_WIN(win).win_id);
         MPIDIU_map_erase(MPIDI_OFI_global.win_map, MPIDI_OFI_WIN(win).win_id);
         /* For scalable EP: push transmit context index back into available pool. */
         if (MPIDI_OFI_WIN(win).sep_tx_idx != -1) {
-            utarray_push_back(MPIDI_OFI_global.rma_sep_idx_array, &(MPIDI_OFI_WIN(win).sep_tx_idx),
-                              MPL_MEM_RMA);
+            utarray_push_back(MPIDI_OFI_global.ctx[ctx_idx].rma_sep_idx_array,
+                              &(MPIDI_OFI_WIN(win).sep_tx_idx), MPL_MEM_RMA);
         }
-        if (MPIDI_OFI_WIN(win).ep != MPIDI_OFI_global.ctx[0].tx)
+        if (MPIDI_OFI_WIN(win).ep != MPIDI_OFI_global.ctx[ctx_idx].tx)
             MPIDI_OFI_CALL(fi_close(&MPIDI_OFI_WIN(win).ep->fid), epclose);
-        if (MPIDI_OFI_WIN(win).cmpl_cntr != MPIDI_OFI_global.ctx[0].rma_cmpl_cntr)
+        /* FIXME: comparing pointer is fragile, ctx[ctx_idx].rma_cmpl_cntr may be a dummy */
+        if (MPIDI_OFI_WIN(win).cmpl_cntr != MPIDI_OFI_global.ctx[ctx_idx].rma_cmpl_cntr)
             MPIDI_OFI_CALL(fi_close(&MPIDI_OFI_WIN(win).cmpl_cntr->fid), cntrclose);
         if (MPIDI_OFI_WIN(win).mr)
             MPIDI_OFI_CALL(fi_close(&MPIDI_OFI_WIN(win).mr->fid), mr_unreg);
@@ -1002,7 +1064,7 @@ int MPIDI_OFI_mpi_win_free_hook(MPIR_Win * win)
     /* This hook is called by CH4 generic call before CH4 finalization */
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_OFI_MPI_WIN_FREE_HOOK);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
