@@ -16,39 +16,22 @@
 
 #include "posix_impl.h"
 
-MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_mpi_send(const void *buf, MPI_Aint count,
-                                                  MPI_Datatype datatype, int rank, int tag,
-                                                  MPIR_Comm * comm, int context_offset,
-                                                  MPIDI_av_entry_t * addr, MPIR_Request ** request)
-{
-    return MPIDIG_mpi_send(buf, count, datatype, rank, tag, comm, context_offset, addr, request);
-}
-
-MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_send_coll(const void *buf, MPI_Aint count,
-                                                   MPI_Datatype datatype, int rank, int tag,
-                                                   MPIR_Comm * comm, int context_offset,
-                                                   MPIDI_av_entry_t * addr,
-                                                   MPIR_Request ** request,
-                                                   MPIR_Errflag_t * errflag)
-{
-    return MPIDIG_send_coll(buf, count, datatype, rank, tag, comm, context_offset, addr, request,
-                            errflag);
-}
-
-MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_mpi_ssend(const void *buf, MPI_Aint count,
-                                                   MPI_Datatype datatype, int rank, int tag,
-                                                   MPIR_Comm * comm, int context_offset,
-                                                   MPIDI_av_entry_t * addr, MPIR_Request ** request)
-{
-    return MPIDIG_mpi_ssend(buf, count, datatype, rank, tag, comm, context_offset, addr, request);
-}
+#define MPIDI_POSIX_SEND_VSIS(vsi_src_, vsi_dst_) \
+    do { \
+        vsi_src_ = MPIDI_POSIX_get_vsi(SRC_VCI_FROM_SENDER, comm, comm->rank, rank, tag); \
+        vsi_dst_ = MPIDI_POSIX_get_vsi(DST_VCI_FROM_SENDER, comm, comm->rank, rank, tag); \
+    } while (0)
 
 MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_mpi_isend(const void *buf, MPI_Aint count,
                                                    MPI_Datatype datatype, int rank, int tag,
                                                    MPIR_Comm * comm, int context_offset,
                                                    MPIDI_av_entry_t * addr, MPIR_Request ** request)
 {
-    return MPIDIG_mpi_isend(buf, count, datatype, rank, tag, comm, context_offset, addr, request);
+    int vsi_src, vsi_dst;
+    MPIDI_POSIX_SEND_VSIS(vsi_src, vsi_dst);
+
+    return MPIDIG_mpi_isend(buf, count, datatype, rank, tag, comm, context_offset, addr,
+                            vsi_src, vsi_dst, request);
 }
 
 MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_isend_coll(const void *buf, MPI_Aint count,
@@ -58,8 +41,11 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_isend_coll(const void *buf, MPI_Aint co
                                                     MPIR_Request ** request,
                                                     MPIR_Errflag_t * errflag)
 {
-    return MPIDIG_isend_coll(buf, count, datatype, rank, tag, comm, context_offset, addr, request,
-                             errflag);
+    int vsi_src, vsi_dst;
+    MPIDI_POSIX_SEND_VSIS(vsi_src, vsi_dst);
+
+    return MPIDIG_isend_coll(buf, count, datatype, rank, tag, comm, context_offset, addr,
+                             vsi_src, vsi_dst, request, errflag);
 }
 
 MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_mpi_issend(const void *buf, MPI_Aint count,
@@ -68,7 +54,11 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_mpi_issend(const void *buf, MPI_Aint co
                                                     MPIDI_av_entry_t * addr,
                                                     MPIR_Request ** request)
 {
-    return MPIDIG_mpi_issend(buf, count, datatype, rank, tag, comm, context_offset, addr, request);
+    int vsi_src, vsi_dst;
+    MPIDI_POSIX_SEND_VSIS(vsi_src, vsi_dst);
+
+    return MPIDIG_mpi_issend(buf, count, datatype, rank, tag, comm, context_offset, addr,
+                             vsi_src, vsi_dst, request);
 }
 
 MPL_STATIC_INLINE_PREFIX int MPIDI_POSIX_mpi_cancel_send(MPIR_Request * sreq)
