@@ -1,7 +1,6 @@
-/* -*- Mode: c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2011 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "mpidimpl.h"
@@ -19,7 +18,7 @@ cvars:
       category    : COLLECTIVE
       type        : boolean
       default     : false
-      class       : device
+      class       : none
       verbosity   : MPI_T_VERBOSITY_USER_BASIC
       scope       : MPI_T_SCOPE_ALL_EQ
       description : >-
@@ -48,12 +47,63 @@ static const char *entry_to_str(enum MPIDU_Sched_entry_type type)
     }
 }
 
+static void entry_dump(FILE * fh, struct MPIDU_Sched_entry *e)
+{
+    switch (e->type) {
+        case MPIDU_SCHED_ENTRY_SEND:
+            {
+                struct MPIDU_Sched_send *s = &(e->u.send);
+                fprintf(fh, "\t\tSend: " MPI_AINT_FMT_DEC_SPEC " of type %x from %d\n", s->count,
+                        s->datatype, s->dest);
+                fprintf(fh, "\t\t from buff: %p\n", s->buf);
+            }
+            break;
+        case MPIDU_SCHED_ENTRY_RECV:
+            {
+                struct MPIDU_Sched_recv *r = &(e->u.recv);
+                fprintf(fh, "\t\tRecv: " MPI_AINT_FMT_DEC_SPEC " of type %x from %d\n", r->count,
+                        r->datatype, r->src);
+                fprintf(fh, "\t\t Into buff: %p\n", r->buf);
+            }
+            break;
+        case MPIDU_SCHED_ENTRY_REDUCE:
+            {
+                struct MPIDU_Sched_reduce *rd = &(e->u.reduce);
+                fprintf(fh, "\t\tReduce: %p -> %p\n", rd->inbuf, rd->inoutbuf);
+                fprintf(fh, "\t\t  " MPI_AINT_FMT_DEC_SPEC " elements of type %x\n", rd->count,
+                        rd->datatype);
+                fprintf(fh, "\t\t Op: %x\n", rd->op);
+            }
+            break;
+        case MPIDU_SCHED_ENTRY_COPY:
+            {
+                struct MPIDU_Sched_copy *cp = &(e->u.copy);
+                fprintf(fh, "\t\tFrom: %p " MPI_AINT_FMT_DEC_SPEC " of type %x\n", cp->inbuf,
+                        cp->incount, cp->intype);
+                fprintf(fh, "\t\tTo:   %p " MPI_AINT_FMT_DEC_SPEC " of type %x\n", cp->outbuf,
+                        cp->outcount, cp->outtype);
+            }
+            break;
+        case MPIDU_SCHED_ENTRY_NOP:
+            break;
+        case MPIDU_SCHED_ENTRY_CB:
+            {
+                struct MPIDU_Sched_cb *cb = &(e->u.cb);
+                fprintf(fh, "\t\tcb_type=%d\n", cb->cb_type);
+                fprintf(fh, "\t\tcb_addr: %p\n", cb->u.cb_p);
+            }
+            break;
+        default:
+            break;
+    }
+}
+
 /* utility function for debugging, dumps the given schedule object to fh */
 static void sched_dump(struct MPIDU_Sched *s, FILE * fh)
 {
     int i;
 
-    fprintf(fh, "--------------------------------\n");
+    fprintf(fh, "================================\n");
     fprintf(fh, "s=%p\n", s);
     if (s) {
         fprintf(fh, "s->size=%zd\n", s->size);
@@ -63,14 +113,16 @@ static void sched_dump(struct MPIDU_Sched *s, FILE * fh)
         fprintf(fh, "s->req=%p\n", s->req);
         fprintf(fh, "s->entries=%p\n", s->entries);
         for (i = 0; i < s->num_entries; ++i) {
+            fprintf(fh, "--------------------------------\n");
             fprintf(fh, "&s->entries[%d]=%p\n", i, &s->entries[i]);
-            fprintf(fh, "s->entries[%d].type=%s\n", i, entry_to_str(s->entries[i].type));
-            fprintf(fh, "s->entries[%d].status=%d\n", i, s->entries[i].status);
-            fprintf(fh, "s->entries[%d].is_barrier=%s\n", i,
+            fprintf(fh, "\ts->entries[%d].type=%s\n", i, entry_to_str(s->entries[i].type));
+            fprintf(fh, "\ts->entries[%d].status=%d\n", i, s->entries[i].status);
+            fprintf(fh, "\ts->entries[%d].is_barrier=%s\n", i,
                     (s->entries[i].is_barrier ? "TRUE" : "FALSE"));
+            entry_dump(fh, &(s->entries[i]));
         }
     }
-    fprintf(fh, "--------------------------------\n");
+    fprintf(fh, "================================\n");
     /*
      * fprintf(fh, "s->next=%p\n", s->next);
      * fprintf(fh, "s->prev=%p\n", s->prev);
@@ -83,23 +135,19 @@ struct MPIDU_Sched_state {
 };
 
 /* holds on to all incomplete schedules on which progress should be made */
-struct MPIDU_Sched_state all_schedules = { NULL };
+static struct MPIDU_Sched_state all_schedules = { NULL };
 
 /* returns TRUE if any schedules are currently pending completion by the
  * progress engine, FALSE otherwise */
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_are_pending
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_are_pending(void)
 {
+    /* this function is only called within a critical section to decide whether
+     * yield is necessary. (ref: .../ch3/.../mpid_nem_inline.h)
+     * therefore, there is no need for additional lock protection.
+     */
     return (all_schedules.head != NULL);
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_next_tag
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_next_tag(MPIR_Comm * comm_ptr, int *tag)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -111,8 +159,7 @@ int MPIDU_Sched_next_tag(MPIR_Comm * comm_ptr, int *tag)
     int end = MPI_UNDEFINED;
     struct MPIDU_Sched *elt = NULL;
 #endif
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDU_SCHED_NEXT_TAG);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDU_SCHED_NEXT_TAG);
+    MPIR_FUNC_ENTER;
 
     *tag = comm_ptr->next_sched_tag;
     ++comm_ptr->next_sched_tag;
@@ -129,11 +176,13 @@ int MPIDU_Sched_next_tag(MPIR_Comm * comm_ptr, int *tag)
         end = tag_ub / 2;
     }
     if (start != MPI_UNDEFINED) {
+        MPID_THREAD_CS_ENTER(VCI, MPIDIU_THREAD_SCHED_LIST_MUTEX);
         DL_FOREACH(all_schedules.head, elt) {
             if (elt->tag >= start && elt->tag < end) {
                 MPIR_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, "**toomanynbc");
             }
         }
+        MPID_THREAD_CS_EXIT(VCI, MPIDIU_THREAD_SCHED_LIST_MUTEX);
     }
 #endif
 
@@ -145,14 +194,15 @@ int MPIDU_Sched_next_tag(MPIR_Comm * comm_ptr, int *tag)
 #if defined(HAVE_ERROR_CHECKING)
   fn_fail:
 #endif
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDU_SCHED_NEXT_TAG);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_start_entry
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
+void MPIDU_Sched_set_tag(struct MPIDU_Sched *s, int tag)
+{
+    s->tag = tag;
+}
+
 /* initiates the schedule entry "e" in the NBC described by "s", where
  * "e" is at "idx" in "s".  This means posting nonblocking sends/recvs,
  * performing reductions, calling callbacks, etc. */
@@ -162,8 +212,7 @@ static int MPIDU_Sched_start_entry(struct MPIDU_Sched *s, size_t idx, struct MPI
     MPIR_Request *r = s->req;
     MPIR_Comm *comm;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDU_SCHED_START_ENTRY);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDU_SCHED_START_ENTRY);
+    MPIR_FUNC_ENTER;
 
     MPIR_Assert(e->status == MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED);
 
@@ -229,25 +278,39 @@ static int MPIDU_Sched_start_entry(struct MPIDU_Sched *s, size_t idx, struct MPI
                 e->status = MPIDU_SCHED_ENTRY_STATUS_STARTED;
             }
             break;
+        case MPIDU_SCHED_ENTRY_PT2PT_SEND:
+            ret_errno = MPID_Isend(e->u.send.buf, e->u.send.count, e->u.send.datatype,
+                                   e->u.send.dest, e->u.send.tag, e->u.send.comm,
+                                   MPIR_CONTEXT_INTRA_PT2PT, &e->u.send.sreq);
+            if (unlikely(ret_errno)) {
+                e->status = MPIDU_SCHED_ENTRY_STATUS_FAILED;
+            } else {
+                e->status = MPIDU_SCHED_ENTRY_STATUS_STARTED;
+            }
+            break;
+        case MPIDU_SCHED_ENTRY_PT2PT_RECV:
+            ret_errno = MPID_Irecv(e->u.recv.buf, e->u.recv.count, e->u.recv.datatype,
+                                   e->u.recv.src, e->u.recv.tag, e->u.recv.comm,
+                                   MPIR_CONTEXT_INTRA_PT2PT, &e->u.recv.rreq);
+            if (unlikely(ret_errno)) {
+                e->status = MPIDU_SCHED_ENTRY_STATUS_FAILED;
+            } else {
+                e->status = MPIDU_SCHED_ENTRY_STATUS_STARTED;
+            }
+            break;
         case MPIDU_SCHED_ENTRY_REDUCE:
             MPL_DBG_MSG_D(MPIR_DBG_COMM, VERBOSE, "starting REDUCE entry %d\n", (int) idx);
             mpi_errno =
                 MPIR_Reduce_local(e->u.reduce.inbuf, e->u.reduce.inoutbuf, e->u.reduce.count,
                                   e->u.reduce.datatype, e->u.reduce.op);
-            if (mpi_errno)
-                MPIR_ERR_POP(mpi_errno);
-            MPIR_Op_release_if_not_builtin(e->u.reduce.op);
-            MPIR_Datatype_release_if_not_builtin(e->u.reduce.datatype);
+            MPIR_ERR_CHECK(mpi_errno);
             e->status = MPIDU_SCHED_ENTRY_STATUS_COMPLETE;
             break;
         case MPIDU_SCHED_ENTRY_COPY:
             MPL_DBG_MSG_D(MPIR_DBG_COMM, VERBOSE, "starting COPY entry %d\n", (int) idx);
             mpi_errno = MPIR_Localcopy(e->u.copy.inbuf, e->u.copy.incount, e->u.copy.intype,
                                        e->u.copy.outbuf, e->u.copy.outcount, e->u.copy.outtype);
-            if (mpi_errno)
-                MPIR_ERR_POP(mpi_errno);
-            MPIR_Datatype_release_if_not_builtin(e->u.copy.intype);
-            MPIR_Datatype_release_if_not_builtin(e->u.copy.outtype);
+            MPIR_ERR_CHECK(mpi_errno);
             e->status = MPIDU_SCHED_ENTRY_STATUS_COMPLETE;
             break;
         case MPIDU_SCHED_ENTRY_NOP:
@@ -258,8 +321,12 @@ static int MPIDU_Sched_start_entry(struct MPIDU_Sched *s, size_t idx, struct MPI
             MPL_DBG_MSG_D(MPIR_DBG_COMM, VERBOSE, "starting CB entry %d\n", (int) idx);
             if (e->u.cb.cb_type == MPIDU_SCHED_CB_TYPE_1) {
                 ret_errno = e->u.cb.u.cb_p(r->comm, s->tag, e->u.cb.cb_state);
-                /* Sched entries list can be reallocated inside callback */
-                e = &s->entries[idx];
+                if (s->kind == MPIR_SCHED_KIND_GENERALIZED) {
+                    /* Sched entries list can be reallocated inside callback */
+                    e = &s->entries[idx];
+                } else {
+                    MPIR_Assert(e == &s->entries[idx]);
+                }
                 if (unlikely(ret_errno)) {
                     if (MPIR_ERR_NONE == r->u.nbc.errflag) {
                         if (MPIX_ERR_PROC_FAILED == MPIR_ERR_GET_CLASS(ret_errno)) {
@@ -274,8 +341,12 @@ static int MPIDU_Sched_start_entry(struct MPIDU_Sched *s, size_t idx, struct MPI
                 }
             } else if (e->u.cb.cb_type == MPIDU_SCHED_CB_TYPE_2) {
                 ret_errno = e->u.cb.u.cb2_p(r->comm, s->tag, e->u.cb.cb_state, e->u.cb.cb_state2);
-                /* Sched entries list can be reallocated inside callback */
-                e = &s->entries[idx];
+                if (s->kind == MPIR_SCHED_KIND_GENERALIZED) {
+                    /* Sched entries list can be reallocated inside callback */
+                    e = &s->entries[idx];
+                } else {
+                    MPIR_Assert(e == &s->entries[idx]);
+                }
                 if (unlikely(ret_errno)) {
                     if (MPIR_ERR_NONE == r->u.nbc.errflag) {
                         if (MPIX_ERR_PROC_FAILED == MPIR_ERR_GET_CLASS(ret_errno)) {
@@ -301,7 +372,7 @@ static int MPIDU_Sched_start_entry(struct MPIDU_Sched *s, size_t idx, struct MPI
     }
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDU_SCHED_START_ENTRY);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     e->status = MPIDU_SCHED_ENTRY_STATUS_FAILED;
@@ -313,17 +384,12 @@ static int MPIDU_Sched_start_entry(struct MPIDU_Sched *s, size_t idx, struct MPI
 /* Posts or performs any NOT_STARTED operations in the given schedule that are
  * permitted to be started.  That is, this routine will respect schedule
  * barriers appropriately. */
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_continue
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 static int MPIDU_Sched_continue(struct MPIDU_Sched *s)
 {
     int mpi_errno = MPI_SUCCESS;
     size_t i;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDU_SCHED_CONTINUE);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDU_SCHED_CONTINUE);
+    MPIR_FUNC_ENTER;
 
     for (i = s->idx; i < s->num_entries; ++i) {
         struct MPIDU_Sched_entry *e = &s->entries[i];
@@ -332,8 +398,7 @@ static int MPIDU_Sched_continue(struct MPIDU_Sched *s)
             mpi_errno = MPIDU_Sched_start_entry(s, i, e);
             /* Sched entries list can be reallocated inside callback */
             e = &s->entries[i];
-            if (mpi_errno)
-                MPIR_ERR_POP(mpi_errno);
+            MPIR_ERR_CHECK(mpi_errno);
         }
 
         /* _start_entry may have completed the operation, but won't update s->idx */
@@ -350,25 +415,20 @@ static int MPIDU_Sched_continue(struct MPIDU_Sched *s)
         }
     }
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDU_SCHED_CONTINUE);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_create
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* creates a new opaque schedule object and returns a handle to it in (*sp) */
-int MPIDU_Sched_create(MPIR_Sched_t * sp)
+int MPIDU_Sched_create(MPIR_Sched_t * sp, enum MPIR_Sched_kind kind)
 {
     int mpi_errno = MPI_SUCCESS;
     struct MPIDU_Sched *s;
     MPIR_CHKPMEM_DECL(2);
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDU_SCHED_CREATE);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDU_SCHED_CREATE);
+    MPIR_FUNC_ENTER;
 
     *sp = NULL;
 
@@ -382,6 +442,9 @@ int MPIDU_Sched_create(MPIR_Sched_t * sp)
     s->tag = -1;
     s->req = NULL;
     s->entries = NULL;
+    s->kind = kind;
+    s->buffers = NULL;
+    s->handles = NULL;
     s->next = NULL;     /* only needed for sanity checks */
     s->prev = NULL;     /* only needed for sanity checks */
 
@@ -395,17 +458,13 @@ int MPIDU_Sched_create(MPIR_Sched_t * sp)
     MPIR_CHKPMEM_COMMIT();
     *sp = s;
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDU_SCHED_CREATE);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     MPIR_CHKPMEM_REAP();
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_clone
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* clones orig and returns a handle to the new schedule in (*cloned) */
 int MPIDU_Sched_clone(MPIR_Sched_t orig, MPIR_Sched_t * cloned)
 {
@@ -416,23 +475,82 @@ int MPIDU_Sched_clone(MPIR_Sched_t orig, MPIR_Sched_t * cloned)
     return mpi_errno;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_start
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
-/* sets (*sp) to MPIR_SCHED_NULL and gives you back a request pointer in (*req).
- * The caller is giving up ownership of the opaque schedule object. */
-int MPIDU_Sched_start(MPIR_Sched_t * sp, MPIR_Comm * comm, int tag, MPIR_Request ** req)
+int MPIDU_Sched_free(struct MPIDU_Sched *s)
+{
+    MPL_free(s->entries);
+    if (s->buffers) {
+        for (void **p = (void **)utarray_front(s->buffers); p;
+             p = (void **) utarray_next(s->buffers, p)) {
+            MPL_free(*p);
+        }
+        utarray_free(s->buffers);
+    }
+    if (s->handles) {
+        for (int *p = (int *)utarray_front(s->handles); p; p = (int *) utarray_next(s->handles, p)) {
+            if (HANDLE_GET_MPI_KIND(*p) == MPIR_COMM) {
+                MPIR_Comm *comm;
+                MPIR_Comm_get_ptr(*p, comm);
+                MPIR_Comm_release(comm);
+            } else if (HANDLE_GET_MPI_KIND(*p) == MPIR_DATATYPE) {
+                MPIR_Datatype_release_if_not_builtin(*p);
+            } else if (HANDLE_GET_MPI_KIND(*p) == MPIR_OP) {
+                MPIR_Op_release_if_not_builtin(*p);
+            } else {
+                MPIR_Assert(0);
+            }
+        }
+        utarray_free(s->handles);
+    }
+    MPL_free(s);
+    return MPI_SUCCESS;
+}
+
+int MPIDU_Sched_reset(struct MPIDU_Sched *s)
+{
+    MPIR_Assert(s->kind == MPIR_SCHED_KIND_PERSISTENT);
+
+    for (int i = 0; i < s->num_entries; ++i) {
+        s->entries[i].status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
+    }
+    s->idx = 0;
+    /* do not reset tag */
+    s->req = NULL;
+    s->next = NULL;     /* only needed for sanity checks */
+    s->prev = NULL;     /* only needed for sanity checks */
+    return MPI_SUCCESS;
+}
+
+void *MPIDU_Sched_alloc_state(struct MPIDU_Sched *s, MPI_Aint size)
+{
+    void *p = MPL_malloc(size, MPL_MEM_OTHER);
+    if (p == NULL) {
+        /* Caller should process error */
+        return p;
+    }
+
+    if (s->buffers == NULL) {
+        utarray_new(s->buffers, &ut_ptr_icd, MPL_MEM_OTHER);
+    }
+    utarray_push_back(s->buffers, &p, MPL_MEM_OTHER);
+    return p;
+}
+
+static void sched_add_ref(struct MPIDU_Sched *s, int handle)
+{
+    if (s->handles == NULL) {
+        utarray_new(s->handles, &ut_int_icd, MPL_MEM_OTHER);
+    }
+    utarray_push_back(s->handles, &handle, MPL_MEM_OTHER);
+}
+
+int MPIDU_Sched_start(struct MPIDU_Sched *s, MPIR_Comm * comm, MPIR_Request ** req)
 {
     int mpi_errno = MPI_SUCCESS;
     MPIR_Request *r;
-    struct MPIDU_Sched *s = *sp;
 
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDU_SCHED_START);
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDU_SCHED_START);
+    MPIR_FUNC_ENTER;
 
     *req = NULL;
-    *sp = MPIR_SCHED_NULL;
 
     /* sanity check the schedule */
     MPIR_Assert(s->num_entries <= s->size);
@@ -456,28 +574,28 @@ int MPIDU_Sched_start(MPIR_Sched_t * sp, MPIR_Comm * comm, int tag, MPIR_Request
     *req = r;
     /* cc is 1, which is fine b/c we only use it as a signal, rather than
      * incr/decr on every constituent operation */
-    s->tag = tag;
 
     /* Now kick off any initial operations.  Do this before we tell the progress
      * engine about this req+sched, otherwise we have more MT issues to worry
      * about.  Skipping this step will increase latency. */
     mpi_errno = MPIDU_Sched_continue(s);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     /* finally, enqueue in the list of all pending schedules so that the
      * progress engine can make progress on it */
-    if (all_schedules.head == NULL)
-        MPID_Progress_activate_hook(MPIR_Nbc_progress_hook_id);
 
+    MPID_THREAD_CS_ENTER(VCI, MPIDIU_THREAD_SCHED_LIST_MUTEX);
     DL_APPEND(all_schedules.head, s);
+    MPID_THREAD_CS_EXIT(VCI, MPIDIU_THREAD_SCHED_LIST_MUTEX);
+
+    MPIR_Progress_hook_activate(MPIR_Nbc_progress_hook_id);
 
     MPL_DBG_MSG_P(MPIR_DBG_COMM, TYPICAL, "started schedule s=%p\n", s);
     if (MPIR_CVAR_COLL_SCHED_DUMP)
         sched_dump(s, stderr);
 
   fn_exit:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDU_SCHED_START);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
   fn_fail:
     if (*req)
@@ -491,10 +609,6 @@ int MPIDU_Sched_start(MPIR_Sched_t * sp, MPIR_Comm * comm, int tag, MPIR_Request
 }
 
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_add_entry
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* idx and e are permitted to be NULL */
 static int MPIDU_Sched_add_entry(struct MPIDU_Sched *s, int *idx, struct MPIDU_Sched_entry **e)
 {
@@ -527,10 +641,6 @@ static int MPIDU_Sched_add_entry(struct MPIDU_Sched *s, int *idx, struct MPIDU_S
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_send
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* do these ops need an entry handle returned? */
 int MPIDU_Sched_send(const void *buf, MPI_Aint count, MPI_Datatype datatype, int dest,
                      MPIR_Comm * comm, MPIR_Sched_t s)
@@ -539,8 +649,7 @@ int MPIDU_Sched_send(const void *buf, MPI_Aint count, MPI_Datatype datatype, int
     struct MPIDU_Sched_entry *e = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_SEND;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -560,6 +669,10 @@ int MPIDU_Sched_send(const void *buf, MPI_Aint count, MPI_Datatype datatype, int
      * release it at entry completion time */
     MPIR_Comm_add_ref(comm);
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
 
   fn_exit:
     return mpi_errno;
@@ -567,11 +680,45 @@ int MPIDU_Sched_send(const void *buf, MPI_Aint count, MPI_Datatype datatype, int
     goto fn_exit;
 }
 
+int MPIDU_Sched_pt2pt_send(const void *buf, MPI_Aint count, MPI_Datatype datatype,
+                           int tag, int dest, MPIR_Comm * comm, MPIR_Sched_t s)
+{
+    int mpi_errno = MPI_SUCCESS;
+    struct MPIDU_Sched_entry *e = NULL;
+
+    mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
+    MPIR_ERR_CHECK(mpi_errno);
+
+    e->type = MPIDU_SCHED_ENTRY_PT2PT_SEND;
+    e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
+    e->is_barrier = FALSE;
+
+    e->u.send.buf = buf;
+    e->u.send.count = count;
+    e->u.send.count_p = NULL;
+    e->u.send.datatype = datatype;
+    e->u.send.dest = dest;
+    e->u.send.sreq = NULL;      /* will be populated by _start_entry */
+    e->u.send.comm = comm;
+    e->u.send.is_sync = FALSE;
+    e->u.send.tag = tag;
+
+    /* the user may free the comm & type after initiating but before the
+     * underlying send is actually posted, so we must add a reference here and
+     * release it at entry completion time */
+    MPIR_Comm_add_ref(comm);
+    MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_ssend
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_ssend(const void *buf, MPI_Aint count, MPI_Datatype datatype, int dest,
                       MPIR_Comm * comm, MPIR_Sched_t s)
 {
@@ -579,8 +726,7 @@ int MPIDU_Sched_ssend(const void *buf, MPI_Aint count, MPI_Datatype datatype, in
     struct MPIDU_Sched_entry *e = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_SEND;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -600,6 +746,10 @@ int MPIDU_Sched_ssend(const void *buf, MPI_Aint count, MPI_Datatype datatype, in
      * release it at entry completion time */
     MPIR_Comm_add_ref(comm);
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
 
   fn_exit:
     return mpi_errno;
@@ -608,10 +758,6 @@ int MPIDU_Sched_ssend(const void *buf, MPI_Aint count, MPI_Datatype datatype, in
 }
 
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_send_defer
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_send_defer(const void *buf, const MPI_Aint * count, MPI_Datatype datatype, int dest,
                            MPIR_Comm * comm, MPIR_Sched_t s)
 {
@@ -619,8 +765,7 @@ int MPIDU_Sched_send_defer(const void *buf, const MPI_Aint * count, MPI_Datatype
     struct MPIDU_Sched_entry *e = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_SEND;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -640,6 +785,10 @@ int MPIDU_Sched_send_defer(const void *buf, const MPI_Aint * count, MPI_Datatype
      * release it at entry completion time */
     MPIR_Comm_add_ref(comm);
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
 
   fn_exit:
     return mpi_errno;
@@ -647,10 +796,6 @@ int MPIDU_Sched_send_defer(const void *buf, const MPI_Aint * count, MPI_Datatype
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_recv_status
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_recv_status(void *buf, MPI_Aint count, MPI_Datatype datatype, int src,
                             MPIR_Comm * comm, MPI_Status * status, MPIR_Sched_t s)
 {
@@ -658,8 +803,7 @@ int MPIDU_Sched_recv_status(void *buf, MPI_Aint count, MPI_Datatype datatype, in
     struct MPIDU_Sched_entry *e = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_RECV;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -675,6 +819,10 @@ int MPIDU_Sched_recv_status(void *buf, MPI_Aint count, MPI_Datatype datatype, in
     status->MPI_ERROR = MPI_SUCCESS;
     MPIR_Comm_add_ref(comm);
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
 
   fn_exit:
     return mpi_errno;
@@ -682,10 +830,6 @@ int MPIDU_Sched_recv_status(void *buf, MPI_Aint count, MPI_Datatype datatype, in
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_recv
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_recv(void *buf, MPI_Aint count, MPI_Datatype datatype, int src, MPIR_Comm * comm,
                      MPIR_Sched_t s)
 {
@@ -693,8 +837,7 @@ int MPIDU_Sched_recv(void *buf, MPI_Aint count, MPI_Datatype datatype, int src,
     struct MPIDU_Sched_entry *e = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_RECV;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -710,6 +853,45 @@ int MPIDU_Sched_recv(void *buf, MPI_Aint count, MPI_Datatype datatype, int src,
 
     MPIR_Comm_add_ref(comm);
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
+
+  fn_exit:
+    return mpi_errno;
+  fn_fail:
+    goto fn_exit;
+}
+
+int MPIDU_Sched_pt2pt_recv(void *buf, MPI_Aint count, MPI_Datatype datatype,
+                           int tag, int src, MPIR_Comm * comm, MPIR_Sched_t s)
+{
+    int mpi_errno = MPI_SUCCESS;
+    struct MPIDU_Sched_entry *e = NULL;
+
+    mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
+    MPIR_ERR_CHECK(mpi_errno);
+
+    e->type = MPIDU_SCHED_ENTRY_PT2PT_RECV;
+    e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
+    e->is_barrier = FALSE;
+
+    e->u.recv.buf = buf;
+    e->u.recv.count = count;
+    e->u.recv.datatype = datatype;
+    e->u.recv.src = src;
+    e->u.recv.rreq = NULL;      /* will be populated by _start_entry */
+    e->u.recv.comm = comm;
+    e->u.recv.status = MPI_STATUS_IGNORE;
+    e->u.recv.tag = tag;
+
+    MPIR_Comm_add_ref(comm);
+    MPIR_Datatype_add_ref_if_not_builtin(datatype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, comm->handle);
+        sched_add_ref(s, datatype);
+    }
 
   fn_exit:
     return mpi_errno;
@@ -717,10 +899,6 @@ int MPIDU_Sched_recv(void *buf, MPI_Aint count, MPI_Datatype datatype, int src,
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_reduce
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_reduce(const void *inbuf, void *inoutbuf, MPI_Aint count, MPI_Datatype datatype,
                        MPI_Op op, MPIR_Sched_t s)
 {
@@ -729,8 +907,7 @@ int MPIDU_Sched_reduce(const void *inbuf, void *inoutbuf, MPI_Aint count, MPI_Da
     struct MPIDU_Sched_reduce *reduce = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_REDUCE;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -745,6 +922,10 @@ int MPIDU_Sched_reduce(const void *inbuf, void *inoutbuf, MPI_Aint count, MPI_Da
 
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
     MPIR_Op_add_ref_if_not_builtin(op);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, datatype);
+        sched_add_ref(s, op);
+    }
 
   fn_exit:
     return mpi_errno;
@@ -752,10 +933,6 @@ int MPIDU_Sched_reduce(const void *inbuf, void *inoutbuf, MPI_Aint count, MPI_Da
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_copy
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* Schedules a copy of "incount" copies of "intype" from "inbuf" to "outbuf" as
  * specified by "outcount" and "outtype".  It is erroneous to attempt to copy
  * more data than will fit into the (outbuf,outcount,outtype)-triple.  This
@@ -771,8 +948,7 @@ int MPIDU_Sched_copy(const void *inbuf, MPI_Aint incount, MPI_Datatype intype,
     struct MPIDU_Sched_copy *copy = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_COPY;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -788,6 +964,10 @@ int MPIDU_Sched_copy(const void *inbuf, MPI_Aint incount, MPI_Datatype intype,
 
     MPIR_Datatype_add_ref_if_not_builtin(intype);
     MPIR_Datatype_add_ref_if_not_builtin(outtype);
+    if (s->kind != MPIR_SCHED_KIND_GENERALIZED) {
+        sched_add_ref(s, intype);
+        sched_add_ref(s, outtype);
+    }
 
     /* some sanity checking up front */
 #if defined(HAVE_ERROR_CHECKING) && !defined(NDEBUG)
@@ -810,10 +990,6 @@ int MPIDU_Sched_copy(const void *inbuf, MPI_Aint incount, MPI_Datatype intype,
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_barrier
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* require that all previously added ops are complete before subsequent ops
  * may begin to execute */
 int MPIDU_Sched_barrier(MPIR_Sched_t s)
@@ -829,10 +1005,6 @@ int MPIDU_Sched_barrier(MPIR_Sched_t s)
     return mpi_errno;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_cb
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* buffer management, fancy reductions, etc */
 int MPIDU_Sched_cb(MPIR_Sched_cb_t * cb_p, void *cb_state, MPIR_Sched_t s)
 {
@@ -841,8 +1013,7 @@ int MPIDU_Sched_cb(MPIR_Sched_cb_t * cb_p, void *cb_state, MPIR_Sched_t s)
     struct MPIDU_Sched_cb *cb = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_CB;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -860,10 +1031,6 @@ int MPIDU_Sched_cb(MPIR_Sched_cb_t * cb_p, void *cb_state, MPIR_Sched_t s)
     goto fn_exit;
 }
 
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_cb2
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 /* buffer management, fancy reductions, etc */
 int MPIDU_Sched_cb2(MPIR_Sched_cb2_t * cb_p, void *cb_state, void *cb_state2, MPIR_Sched_t s)
 {
@@ -872,8 +1039,7 @@ int MPIDU_Sched_cb2(MPIR_Sched_cb2_t * cb_p, void *cb_state, void *cb_state2, MP
     struct MPIDU_Sched_cb *cb = NULL;
 
     mpi_errno = MPIDU_Sched_add_entry(s, NULL, &e);
-    if (mpi_errno)
-        MPIR_ERR_POP(mpi_errno);
+    MPIR_ERR_CHECK(mpi_errno);
 
     e->type = MPIDU_SCHED_ENTRY_CB;
     e->status = MPIDU_SCHED_ENTRY_STATUS_NOT_STARTED;
@@ -891,12 +1057,7 @@ int MPIDU_Sched_cb2(MPIR_Sched_cb2_t * cb_p, void *cb_state, void *cb_state2, MP
     goto fn_exit;
 }
 
-
 /* returns TRUE in (*made_progress) if any of the outstanding schedules in state completed */
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_progress_state
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 static int MPIDU_Sched_progress_state(struct MPIDU_Sched_state *state, int *made_progress)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -925,8 +1086,10 @@ static int MPIDU_Sched_progress_state(struct MPIDU_Sched_state *state, int *made
                             e->status = MPIDU_SCHED_ENTRY_STATUS_COMPLETE;
                         MPIR_Request_free(e->u.send.sreq);
                         e->u.send.sreq = NULL;
-                        MPIR_Comm_release(e->u.send.comm);
-                        MPIR_Datatype_release_if_not_builtin(e->u.send.datatype);
+                        if (s->kind == MPIR_SCHED_KIND_GENERALIZED) {
+                            MPIR_Comm_release(e->u.send.comm);
+                            MPIR_Datatype_release_if_not_builtin(e->u.send.datatype);
+                        }
                     }
                     break;
                 case MPIDU_SCHED_ENTRY_RECV:
@@ -947,8 +1110,41 @@ static int MPIDU_Sched_progress_state(struct MPIDU_Sched_state *state, int *made
                             e->status = MPIDU_SCHED_ENTRY_STATUS_COMPLETE;
                         MPIR_Request_free(e->u.recv.rreq);
                         e->u.recv.rreq = NULL;
-                        MPIR_Comm_release(e->u.recv.comm);
-                        MPIR_Datatype_release_if_not_builtin(e->u.recv.datatype);
+                        if (s->kind == MPIR_SCHED_KIND_GENERALIZED) {
+                            MPIR_Comm_release(e->u.recv.comm);
+                            MPIR_Datatype_release_if_not_builtin(e->u.recv.datatype);
+                        }
+                    }
+                    break;
+                case MPIDU_SCHED_ENTRY_PT2PT_SEND:
+                    if (e->u.send.sreq != NULL && MPIR_Request_is_complete(e->u.send.sreq)) {
+                        if (s->req->status.MPI_ERROR) {
+                            e->status = MPIDU_SCHED_ENTRY_STATUS_FAILED;
+                        } else {
+                            e->status = MPIDU_SCHED_ENTRY_STATUS_COMPLETE;
+                        }
+                        MPIR_Request_free(e->u.send.sreq);
+                        e->u.send.sreq = NULL;
+                        if (s->kind == MPIR_SCHED_KIND_GENERALIZED) {
+                            MPIR_Comm_release(e->u.send.comm);
+                            MPIR_Comm_release(e->u.send.comm);
+                        }
+                        MPIR_Datatype_release_if_not_builtin(e->u.send.datatype);
+                    }
+                    break;
+                case MPIDU_SCHED_ENTRY_PT2PT_RECV:
+                    if (e->u.recv.rreq != NULL && MPIR_Request_is_complete(e->u.recv.rreq)) {
+                        if (s->req->status.MPI_ERROR) {
+                            e->status = MPIDU_SCHED_ENTRY_STATUS_FAILED;
+                        } else {
+                            e->status = MPIDU_SCHED_ENTRY_STATUS_COMPLETE;
+                        }
+                        MPIR_Request_free(e->u.recv.rreq);
+                        e->u.recv.rreq = NULL;
+                        if (s->kind == MPIR_SCHED_KIND_GENERALIZED) {
+                            MPIR_Comm_release(e->u.recv.comm);
+                            MPIR_Datatype_release_if_not_builtin(e->u.recv.datatype);
+                        }
                     }
                     break;
                 default:
@@ -963,8 +1159,7 @@ static int MPIDU_Sched_progress_state(struct MPIDU_Sched_state *state, int *made
                 if (e->is_barrier) {
                     /* post/perform the next round of operations */
                     mpi_errno = MPIDU_Sched_continue(s);
-                    if (mpi_errno)
-                        MPIR_ERR_POP(mpi_errno);
+                    MPIR_ERR_CHECK(mpi_errno);
                 }
             } else if (e->is_barrier && e->status < MPIDU_SCHED_ENTRY_STATUS_COMPLETE) {
                 /* don't process anything after this barrier entry */
@@ -992,14 +1187,11 @@ static int MPIDU_Sched_progress_state(struct MPIDU_Sched_state *state, int *made
                     break;
             }
 
-            mpi_errno = MPID_Request_complete(s->req);
-            if (mpi_errno != MPI_SUCCESS) {
-                MPIR_ERR_POP(mpi_errno);
-            }
+            MPIR_Request_complete(s->req);
 
-            s->req = NULL;
-            MPL_free(s->entries);
-            MPL_free(s);
+            if (s->kind != MPIR_SCHED_KIND_PERSISTENT) {
+                MPIDU_Sched_free(s);
+            }
 
             if (made_progress)
                 *made_progress = TRUE;
@@ -1013,21 +1205,27 @@ static int MPIDU_Sched_progress_state(struct MPIDU_Sched_state *state, int *made
 }
 
 /* returns TRUE in (*made_progress) if any of the outstanding schedules completed */
-#undef FUNCNAME
-#define FUNCNAME MPIDU_Sched_progress
-#undef FCNAME
-#define FCNAME MPL_QUOTE(FUNCNAME)
 int MPIDU_Sched_progress(int *made_progress)
 {
-    int mpi_errno;
+    /* Sched progress may call callback functions that will call into progress again.
+     * For example, with MPI_Comm_idup, sched_cb_gcn_allocate_cid will call MPIR_Allreduce.
+     * This inner progress should skip Sched progress to avoid recursive situation.
+     */
+    static int in_sched_progress = 0;
 
-    MPID_THREAD_CS_ENTER(VNI, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
+    if (in_sched_progress) {
+        return MPI_SUCCESS;
+    } else {
+        int mpi_errno;
 
-    mpi_errno = MPIDU_Sched_progress_state(&all_schedules, made_progress);
-    if (!mpi_errno && all_schedules.head == NULL)
-        MPID_Progress_deactivate_hook(MPIR_Nbc_progress_hook_id);
+        MPID_THREAD_CS_ENTER(VCI, MPIDIU_THREAD_SCHED_LIST_MUTEX);
+        in_sched_progress = 1;
+        mpi_errno = MPIDU_Sched_progress_state(&all_schedules, made_progress);
+        if (!mpi_errno && all_schedules.head == NULL)
+            MPIR_Progress_hook_deactivate(MPIR_Nbc_progress_hook_id);
+        in_sched_progress = 0;
+        MPID_THREAD_CS_EXIT(VCI, MPIDIU_THREAD_SCHED_LIST_MUTEX);
 
-    MPID_THREAD_CS_EXIT(VNI, MPIR_THREAD_GLOBAL_ALLFUNC_MUTEX);
-
-    return mpi_errno;
+        return mpi_errno;
+    }
 }
