@@ -6,11 +6,15 @@
 
 autoreconf=${AUTORECONF:-autoreconf}
 
-echo "=== running autoreconf in 'mpl' ==="
-(cd mpl && $autoreconf ${autoreconf_args:-"-vif"}) || exit 1
+if test -d "mpl" ; then
+    echo "=== running autoreconf in 'mpl' ==="
+    (cd mpl && $autoreconf ${autoreconf_args:-"-vif"}) || exit 1
+fi
 
-echo "=== running autoreconf in 'tools/topo/hwloc/hwloc' ==="
-(cd tools/topo/hwloc/hwloc && ./autogen.sh) || exit 1
+if test -d "tools/topo/hwloc/hwloc" ; then
+    echo "=== running autoreconf in 'tools/topo/hwloc/hwloc' ==="
+    (cd tools/topo/hwloc/hwloc && ./autogen.sh) || exit 1
+fi
 
 echo "=== running autoreconf in '.' ==="
 $autoreconf ${autoreconf_args:-"-vif"} || exit 1
