@@ -15,6 +15,14 @@ static char **config_argv = NULL;
 static int reading_config_file = 0;
 static struct HYD_arg_match_table match_table[];
 
+#define ASSERT_ARGV \
+    do { \
+        if (!**argv) { \
+            status = HYD_FAILURE; \
+            HYDU_ERR_POP(status, "missing command line argument, add -h for help\n"); \
+        }\
+    } while (0)
+
 static void init_ui_mpich_info(void)
 {
     HYD_ui_mpich_info.ppn = -1;
@@ -135,7 +143,7 @@ static void help_help_fn(void)
     printf("    -gpus-per-proc                   number of GPUs per process (default: auto)\n");
 
     printf("\n");
-    printf("Please see the intructions provided at\n");
+    printf("Please see the instructions provided at\n");
     printf("http://wiki.mpich.org/mpich/index.php/Using_the_Hydra_Process_Manager\n");
     printf("for further details\n\n");
 }
@@ -440,6 +448,32 @@ static HYD_status profile_fn(char *arg, char ***argv)
     goto fn_exit;
 }
 
+static void output_from_help_fn(void)
+{
+    printf("\n");
+    printf("-output-from: only show the output from this rank\n\n");
+}
+
+static HYD_status output_from_fn(char *arg, char ***argv)
+{
+    HYD_status status = HYD_SUCCESS;
+
+    if (reading_config_file && HYD_ui_info.output_from != -1) {
+        /* global variable already set; ignore */
+        goto fn_exit;
+    }
+
+    status = HYDU_set_int(arg, &HYD_ui_info.output_from, atoi(**argv));
+    HYDU_ERR_POP(status, "error setting output_from\n");
+
+  fn_exit:
+    (*argv)++;
+    return status;
+
+  fn_fail:
+    goto fn_exit;
+}
+
 static void prepend_rank_help_fn(void)
 {
     printf("\n");
@@ -467,11 +501,12 @@ static HYD_status prepend_rank_fn(char *arg, char ***argv)
 
 static void pattern_info(void)
 {
-    printf("   Regular expressions can include:\n");
-    printf("       %%r: Process rank\n");
+    printf("   Supported substitutions include:\n");
+    printf("       %%r: Process rank (in MPI_COMM_WORLD)\n");
     printf("       %%g: Process group ID\n");
     printf("       %%p: Proxy ID\n");
     printf("       %%h: Hostname\n");
+    printf("       %%t: Timestamp in seconds\n");
 }
 
 static void prepend_pattern_help_fn(void)
@@ -750,6 +785,7 @@ static HYD_status np_fn(char *arg, char ***argv)
     status = get_current_exec(&exec);
     HYDU_ERR_POP(status, "get_current_exec returned error\n");
 
+    ASSERT_ARGV;
     status = HYDU_set_int(arg, &exec->proc_count, atoi(**argv));
     HYDU_ERR_POP(status, "error getting executable process count\n");
 
@@ -1779,6 +1815,7 @@ static struct HYD_arg_match_table match_table[] = {
     {"hostlist", hostlist_fn, hostlist_help_fn},
     {"ppn", ppn_fn, ppn_help_fn},
     {"profile", profile_fn, profile_help_fn},
+    {"output-from", output_from_fn, output_from_help_fn},
     {"prepend-rank", prepend_rank_fn, prepend_rank_help_fn},
     {"l", prepend_rank_fn, prepend_rank_help_fn},
     {"prepend-pattern", prepend_pattern_fn, prepend_pattern_help_fn},
@@ -1844,5 +1881,5 @@ static struct HYD_arg_match_table match_table[] = {
     {"gpus-per-proc", gpus_per_proc_fn, gpus_per_proc_help_fn},
     {"g", gpus_per_proc_fn, gpus_per_proc_help_fn},
 
-    {"\0", NULL}
+    {"\0", NULL, NULL}
 };
