@@ -1,7 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2017 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "hydra_rmk_sge.h"
@@ -10,19 +9,13 @@
 
 static HYD_status process_mfile_tokens(char **tokens, struct HYD_node *node)
 {
-    HYD_status status = HYD_SUCCESS;
-
     MPL_strncpy(node->hostname, tokens[0], HYD_MAX_HOSTNAME_LEN);
     node->core_count = atoi(tokens[1]);
 
-  fn_exit:
-    return status;
-
-  fn_fail:
-    goto fn_exit;
+    return HYD_SUCCESS;
 }
 
-HYD_status HYDI_rmk_sge_query_node_list(int *node_count, struct HYD_node **nodes)
+HYD_status HYDI_rmk_sge_query_node_list(int *node_count, struct HYD_node ** nodes)
 {
     char *hostfile;
     HYD_status status = HYD_SUCCESS;
