@@ -3,16 +3,17 @@
  *     See COPYRIGHT in top-level directory
  */
 
+#include "mpitest.h"
 #include "mpi.h"
 #include <stdio.h>
 #include <stdlib.h>
-#include "mpitest.h"
+#include "mtest_dtp.h"
 
 /*
 static char MTEST_Descrip[] = "A simple test of Reduce with all choices of root process";
 */
 
-void set_send_buf(int count, int *buf)
+static void set_send_buf(int count, int *buf)
 {
     int i;
     for (i = 0; i < count; i++) {
@@ -20,7 +21,7 @@ void set_send_buf(int count, int *buf)
     }
 }
 
-void set_recv_buf(int count, int *buf)
+static void set_recv_buf(int count, int *buf)
 {
     int i;
     for (i = 0; i < count; i++) {
@@ -28,7 +29,7 @@ void set_recv_buf(int count, int *buf)
     }
 }
 
-void check_buf(int rank, int size, int count, int *errs, int *buf)
+static void check_buf(int rank, int size, int count, int *errs, int *buf)
 {
     int i;
     for (i = 0; i < count; i++) {
@@ -42,23 +43,25 @@ void check_buf(int rank, int size, int count, int *errs, int *buf)
     }
 }
 
-int main(int argc, char *argv[])
+static int test_reduce(mtest_mem_type_e oddmem, mtest_mem_type_e evenmem)
 {
     int errs = 0;
-    int rank, size, root, i;
+    int rank, size, root;
     void *sendbuf, *recvbuf, *sendbuf_h, *recvbuf_h;
     int minsize = 2, count;
     MPI_Comm comm;
     mtest_mem_type_e memtype;
 
-    MTest_Init(&argc, &argv);
-    MTestArgList *head = MTestArgListCreate(argc, argv);
-
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     if (rank % 2 == 0)
-        memtype = MTestArgListGetMemType(head, "evenmemtype");
+        memtype = evenmem;
     else
-        memtype = MTestArgListGetMemType(head, "oddmemtype");
+        memtype = oddmem;
+
+    if (rank == 0) {
+        MTestPrintfMsg(1, "./reduce -evenmemtype=%s -oddmemtype=%s\n",
+                       MTest_memtype_name(evenmem), MTest_memtype_name(oddmem));
+    }
 
     while (MTestGetIntracommGeneral(&comm, minsize, 1)) {
         if (comm == MPI_COMM_NULL)
@@ -68,8 +71,8 @@ int main(int argc, char *argv[])
         MPI_Comm_size(comm, &size);
 
         for (count = 1; count < 130000; count = count * 2) {
-            MTestAlloc(count * sizeof(int), memtype, &recvbuf_h, &recvbuf, 0);
-            MTestAlloc(count * sizeof(int), memtype, &sendbuf_h, &sendbuf, 0);
+            MTestMalloc(count * sizeof(int), memtype, &recvbuf_h, &recvbuf, rank);
+            MTestMalloc(count * sizeof(int), memtype, &sendbuf_h, &sendbuf, rank);
 
             for (root = 0; root < size; root++) {
                 set_send_buf(count, sendbuf_h);
@@ -89,6 +92,20 @@ int main(int argc, char *argv[])
         }
         MTestFreeComm(&comm);
     }
+    return errs;
+}
+
+int main(int argc, char *argv[])
+{
+    int errs = 0;
+    MTest_Init(&argc, &argv);
+
+    struct dtp_args dtp_args;
+    dtp_args_init(&dtp_args, MTEST_COLL_NOCOUNT, argc, argv);
+    while (dtp_args_get_next(&dtp_args)) {
+        errs += test_reduce(dtp_args.u.coll.evenmem, dtp_args.u.coll.oddmem);
+    }
+    dtp_args_finalize(&dtp_args);
 
     MTest_Finalize(errs);
     return MTestReturnValue(errs);
