@@ -1,9 +1,8 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *
- *  (C) 2003 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include "mpi.h"
 #include <stdio.h>
 #include "mpitest.h"
@@ -25,15 +24,14 @@ int main(int argc, char **argv)
     if (rank == 0) {
         blockcnt[0] = 1;
         offsets[0] = 3;
-        MPI_Type_hindexed(1, blockcnt, offsets, MPI_BYTE, &tmp_type);
+        MPI_Type_create_hindexed(1, blockcnt, offsets, MPI_BYTE, &tmp_type);
         blockcnt[0] = 1;
         offsets[0] = 1;
-        MPI_Type_hindexed(1, blockcnt, offsets, tmp_type, &newtype);
+        MPI_Type_create_hindexed(1, blockcnt, offsets, tmp_type, &newtype);
         MPI_Type_commit(&newtype);
 
-        MPI_Type_lb(newtype, &lb);
-        MPI_Type_extent(newtype, &extent);
-        MPI_Type_ub(newtype, &ub);
+        MPI_Type_get_extent(newtype, &lb, &extent);
+        ub = lb + extent;
 
         /* Check that the results are correct */
 #ifdef DEBUG
