@@ -11,14 +11,13 @@
 
 int main(int argc, char *argv[])
 {
-    int rank, nproc, i;
+    int nproc, i;
     int errors = 0, errs = 0;
     int buf = 0, *my_buf;
     MPI_Win win;
     MPI_Group world_group;
 
     MTest_Init(&argc, &argv);
-    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &nproc);
 
     MPI_Win_create(&buf, sizeof(int), sizeof(int), MPI_INFO_NULL, MPI_COMM_WORLD, &win);
