@@ -19,9 +19,8 @@ int main(int argc, char *argv[])
     MPI_Datatype dtype;
     MPI_T_enum enumtype;
     MPI_T_cvar_handle bcastHandle, bcastLongHandle;
-    int bcastCount, bcastScope, bcastCvar = -1;
-    int bcastLongCount, bcastLongScope, bcastLongCvar = -1;
-    int gatherScope, gatherCvar = -1;
+    int bcastCount, bcastScope = -1, bcastCvar = -1;
+    int bcastLongCount, bcastLongScope = -1, bcastLongCvar = -1;
     int newval;
     int errs = 0;
 
@@ -41,10 +40,7 @@ int main(int argc, char *argv[])
         cnameLen = sizeof(cname);
         MPI_T_cvar_get_info(i, cname, &cnameLen, &verbosity, &dtype,
                             &enumtype, NULL, NULL, &binding, &scope);
-        if (strcmp(cname, "MPIR_CVAR_GATHER_VSMALL_MSG_SIZE") == 0) {
-            gatherCvar = i;
-            gatherScope = scope;
-        } else if (strcmp(cname, "MPIR_CVAR_BCAST_SHORT_MSG_SIZE") == 0) {
+        if (strcmp(cname, "MPIR_CVAR_BCAST_SHORT_MSG_SIZE") == 0) {
             bcastCvar = i;
             bcastScope = scope;
             if (binding != MPI_T_BIND_NO_OBJECT && binding != MPI_T_BIND_MPI_COMM) {
@@ -58,7 +54,6 @@ int main(int argc, char *argv[])
                 fprintf(stderr, "Unexpected binding for MPIR_CVAR_BCAST_LONG_MSG\n");
                 errs++;
             }
-        } else if (strcmp(cname, "MPIR_CVAR_BCAST_MIN_PROCS") == 0) {
         }
     }
 
