@@ -1,8 +1,8 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2006 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include "mpi.h"
 #include <stdio.h>
 #include <stdlib.h>
@@ -100,16 +100,13 @@ int main(int argc, char *argv[])
                 fprintf(stderr, "message %d (%d) should be %d\n", i, rmsg[i], 10 + i);
             }
         }
-        /* The MPI standard says that there is no way to use MPI_Request_free
-         * safely with receive requests.  A strict MPI implementation may
-         * choose to consider these erroreous (an IBM MPI implementation
-         * does so)  */
-#ifdef USE_STRICT_MPI
-        MPI_Wait(&r[4], MPI_STATUS_IGNORE);
-#else
+        /* Previously, this test was thought to be non-standard behavior. However,
+         * recent MPI Forum discussions have concluded that freeing a request without
+         * explicit completion via MPI_TEST/MPI_WAIT is allowed. The implementation is
+         * required to complete the request during MPI_FINALIZE. (2021-08-24)
+         */
         MTestPrintfMsg(10, "About  free Irecv request\n");
         MPI_Request_free(&r[4]);
-#endif
     }
 
     if (rank != dest && rank != src) {
